
/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * alist.c
 *
 * list of agents or mns mobile node is talking to.
 * The list head is maintained by the caller.
 *
 * routines include:
 *
 * Module should hide how list is constructed from callers.
 *
 * XXX. sort entries according to delta lifetimes
 * ala callout list.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "mnstate.h"
#include "route.h"
#include "alist.h"
#include "mip_log.h"

extern int	debug;
static int	noticeFunction = 3;



/*
 * search alist specified to see if ip addr is in list.
 *
 * note dependencies in alist_create().
 *
 * returns:
 *	struct alist * - pointer to alist entry if found
 *	NULL - no such entry.
 */
static struct alist *prev;

struct alist *
alist_search(struct alist *head, struct in_addr bip)
{
	prev = head;
	while(head) {
		if (((unsigned long)head->bip.s_addr) == ((unsigned long)bip.s_addr))
			return(head);
		prev = head;
		head = head->next;
	}
	return(NULL);
}


struct alist *
alist_findFirst( struct alist *head, int daemonType, int match )
/*
 *  INPUT:	pointer to the head of a list of agents and (possibly)
 *		mobile nodes; identifier of a given type of agent or node
 *		(OURHA, OURFA, THEIRFA...)
 *  OUTPUT:	pointer to the first instance of daemonType if match is TRUE
 *		or pointer to first anything that isn't daemonType if
 *		match is FALSE.  the matching agent must also have its
 *		retryWait field set to zero and if it is not the first agent
 *		in the list, its sigaverage must be >= MIN_GOODSIG.
 *		NULL is returned if no match is found.
 *
 *  -- sets prev in an identical manner to alist_search() in case this
 *     is useful sometime...
 *
 *  TBD?: might be simpler if the type fields were bits instead of
 *  numbers...  the loops should be collapsed into one.
 */
{
	int	firstone = TRUE;

entryMark( 1, "alist_findFirst" );

	prev = head;
	if (match) {
	    while(head) {
		    if ((unsigned long)head->identity == daemonType 
			&& head->retryWait == 0
		    	&& (firstone ||
			  (!firstone && head->sigaverage >= MIN_GOODSIG))
		       )
		    {
			    return(head);
		    }
		    firstone = 0;
		    prev = head;
		    head = head->next;
	    }
	}
	else {
	    while(head) {
		    if ((unsigned long)head->identity != daemonType
			&& head->retryWait == 0
		    	&& (firstone ||
			  (!firstone && head->sigaverage >= MIN_GOODSIG))
		       )
		    {
			    return(head);
		    }
		    firstone = 0;
		    prev = head;
		    head = head->next;
	    }
	}

	return(NULL);

}  /* end alist_findFirst() */



/* create an entry and insert the node with the key bip.  make insertion
 * at the end of the list.
 *
 * because prev is used for insertion, this must be called after 
 * alist_search() and before alist_delete(), where the alist_search() 
 * failed to find an entry.
 *
 * returns:
 *	struct *alist - pointer to alist entry if found
 *	NULL - out of buffer space
*/
struct alist *
alist_create(struct alist **head, struct in_addr bip)
{
	struct alist *n = NULL;
	struct alist *a = malloc(sizeof(struct alist));
	if (!a) {
		return(NULL);
	}
	/* zero it out */
	memset(a, 0, sizeof(struct alist));
	
	a->bip = bip;	/* store key */

	/*
	 *  insert at tail
	 */
	if ( ! prev )
	    *head = a;
	else
	    prev->next = a;


	return(a);
}



void
alist_sort( struct alist **head, struct alist *a )
/*
 *  --	change position of a in list pointed to by head.  sort values
 *	by sigaverage.  assume that all elements are ordered except for
 *	a, which may be in the proper place already, so search before
 *	making any changes.
 *
 *	if more than one entry has the same average, insert after
 *	all current entries with that value.
 */
{
    struct alist	*beforea = NULL,
			*p = NULL,	/* p's before q's */
			*q,
			*ptarget = NULL,
			*qtarget = NULL;
    int			gotnext = 0,
			gottarget = 0;

entryMark( 1, "alist_sort" );


    /* a is no worse than the next element in the list --
     * don't move it, as it may lead to an unnecessary reRegistration
     *
     * once we have a pointer to the a (below), we can also check
     * that the item before has a same/different value.
     */
    if ( a->next && a->sigaverage == a->next->sigaverage )
	return;

    /* find new location */
    for ( p = *head, q = NULL; p; q = p, p = p->next ) {
	if ( p == a ) {			/* keep "next pointer" to a */
	    beforea = q;
	    gotnext = 1;
	}
	/* remember all elements (and "pointers to") up to the 
	 * one that logically follows a
	 */
	if ( !gottarget ) {
	    ptarget = p;
	    qtarget = q;
	    if ( p->sigaverage < a->sigaverage )
		gottarget = 1;
	}

	if ( gotnext && ptarget )	/* quit when all pointers are found */
	    break;
    }

    /* a is sorted properly */
    if ( a == ptarget || a->next == ptarget )
	return;

    /* if a is equal to its successor in the list, it is (again)
     * sorted properly
     */
    if ( beforea && beforea->sigaverage == a->sigaverage )
	return;

    /* remove a */
    if ( a == *head )
	*head = (*head)->next;
    else
	beforea->next = a->next;

    /* put a back */
    a->next = ptarget;

    if ( qtarget == a )		/* beware of previous list state */
	qtarget = beforea;
    if ( !qtarget ) {
	a->next = *head;
	*head = a;
    }
    else
	qtarget->next = a;

}  /* end alist_sort() */



/*
 * alist_delete
 *	delete a list entry.  don't free it if it represents the connected
 *	agent.  NOTE! if the calling code intends to delete the entire
 *	contents of an agent/node list, be sure that the current agent is
 *	zero'd out in the MN state before traversing the list(s).
 *
 * returns:
 *	1 if list entry could not be deleted because it is the current agent
 *	0 otherwise (success)
 *
 * returned indirectly:
 *	**head - reset if it was pointing to deleted item
 */
int
alist_delete(struct alist **head, struct in_addr bip)
{
	struct alist *a = alist_search(*head, bip);
	int rval = 0;

	if (a == NULL)
		return rval;
	if (*head == a) {
		*head = (*head)->next;
	}
	else {
		prev->next = a->next;
	}

	a->next = NULL;
	a->notinlist = TRUE;
	if ( mns.agent != a ) {
		free(a);
	}
	else {
		rval = 1;
	}

	return rval;
}

static struct alist *whead;

void
alist_initwalk(struct alist *head)
{
	whead = head;
}

struct alist *
alist_nextwalk()
{
	struct alist *cur = whead;
	if (whead == NULL)
		return(NULL);
	whead = whead->next;
	return(cur);
}


/*
 *  clears retry field (sets to FALSE) for ALL entries in the agentlist,
 *  whether agents or mobile nodes.
 */
void
alist_clrRetry( struct alist *head )
{
	struct alist *a;

entryMark( 2, "alist_clrRetry" );

	alist_initwalk(head);
	while (a = alist_nextwalk()) {
	    a->retryWait = 0;
	}

}  /* end alist_clrRetry() */


void
alist_dump(struct alist *head)
{
	int i = 0;
	alist_initwalk(head);
	alist_header(1);
	while(1) {
		head = alist_nextwalk();
		if (!head)
			return;
		printf("---------------------\n");
		alist_dumpone(head, 1);
		printf("---------------------\n");
	}
}


void
alist_header(int longform) {
	char	*s = "netauth";

	if (!longform) {
		printf("%-33s %s %s\n",
			"beaconIP / coa", "sig/ave", "location");
		return;
	}

	printf("%-27s %s  %s  %s %s %s\n", "beaconIP / coa", s,
		"sig/ave", "cnt/tick", "wait", "lifetime");
}


int
alist_dumpone(struct alist *a, int longform)
{
	char buf[1024];
	char buf2[1024];
	char *id;
	char *s;
	char *secstate = "";

#ifdef SUPERLONGFORM
	/* for debug testing
	*/
	printf("bip %s\n", inet_ntoa(a->bip));
	printf("coa %s\n", inet_ntoa(a->coa));
	printf("ticks %d\n", a->ticks);
	printf("identity %d\n", a->identity);
	printf("secpolicy %d\n", a->secpolicy);
	printf("signal %d\n", a->sigstrength);
	printf("seqno %d\n", a->seqno);
	printf("lifetime %d\n", a->lifetime);
	printf("sawAuth %d\n", a->sawAuth);
	printf("auth_ok %d\n", a->auth_ok);
	printf("location <%s>\n", a->location);
	return;
#endif

	/* one or two line listing per entry */
	strcpy(buf, inet_ntoa(a->bip));
	sprintf(buf2, "%s / %s", buf, inet_ntoa(a->coa));

	s = "%-33s %d  %3d/%-3d  %3d/%-3d  %4d %-d\n";

	if (!longform) {
		printf("%-33s %3d/%-3d %s\n",
			buf2, a->sigstrength, a->sigaverage,
			a->location?a->location:"");
		return;
	}

	printf(s, buf2, a->auth_ok,
		a->sigstrength, a->sigaverage,
		a->sigcount, a->ticks, 
		a->retryWait, a->lifetime
		);

	switch(a->identity) {
	case OURHA:
		id = "ha";
		break;
	case OURFA:
		id = "our fa";
		break;
	case THEIRFA:
		id = "their fa";
		break;
	case MOBILENODE:
		id = "mn";
		break;
	}
	if (a->location[0])
		printf("\t%s @ <%s>\n", id, a->location);
	else
		printf("\t%s\n", id);

}  /* end alist_dumpone() */


#ifdef TESTMAIN
/*
 * test all primitives above
 */
main()
{
	struct alist *myhead = NULL;
	struct alist *myhead2 = NULL;
	struct in_addr bip;
	struct alist *v;

	printf("sizeof alist member %d\n", sizeof (struct alist));
	/* test case 1.  put 5 entries in list and dump
	*/
	bip.s_addr = 0x01010101;
	alist_create(&myhead, bip);
	bip.s_addr = 0x02020202;
	alist_create(&myhead, bip);
	bip.s_addr = 0x03030303;
	alist_create(&myhead, bip);

	bip.s_addr = 0x04040404;
	alist_create(&myhead, bip);
	bip.s_addr = 0x05050505;
	alist_create(&myhead, bip);

	alist_dump(myhead); 

	/* now find all the list entries
	 * but first search for one that is not there.
	 */
	printf("search for entry not in list\n");
	bip.s_addr = 0x01020304;
	v = alist_search(myhead, bip);
	if (v == NULL) 
		printf("not found, correct!\n");
	else {
		printf("found %x\n", v->bip.s_addr);
		printf("test failed\n");
		exit(1);
	}

	bip.s_addr = 0x01010101;
	v = alist_search(myhead, bip);
	if (v == NULL)  {
		printf("not found, incorrect!\n");
		exit(1);
	}
	else {
		printf("found %x\n", v->bip.s_addr);
	}

	bip.s_addr = 0x05050505;
	v = alist_search(myhead, bip);
	if (v == NULL) {
		printf("not found, incorrect!\n");
		exit(1);
	}
	else
		printf("found %x\n", v->bip.s_addr);

	bip.s_addr = 0x03030303;
	v = alist_search(myhead, bip);
	if (v == NULL)  {
		printf("not found, incorrect!\n");
		exit(1);
	}
	else
		printf("found %x\n", v->bip.s_addr);

	/* test series 2.  put 1 entry in list and delete it
	*/

	bip.s_addr = 0x01010101;
	alist_create(&myhead2, bip);
	alist_delete(&myhead2, bip);
	alist_dump(myhead2); 
	v = alist_search(myhead2, bip);
	if (v == NULL) 
		printf("correct - didn't find it\n");
		
	/* test series 3.  put 2 entries in list and delete first
	*/
	bip.s_addr = 0x02020202;
	alist_create(&myhead2, bip);
	bip.s_addr = 0x01010101;
	alist_create(&myhead2, bip);
	alist_delete(&myhead2, bip);
	printf("list should only have 02020202 in it\n");
	alist_dump(myhead2); 
	bip.s_addr = 0x02020202;
	v = alist_search(myhead2, bip);
	v->ticks = 47;
	if (v == NULL) 
		printf("incorrect - didn't find it\n");
	else
		printf("found %x\n", v->bip.s_addr);

	/* test series 4. delete 3rd entry in original list
	*/
	bip.s_addr = 0x03030303;
	alist_delete(&myhead, bip);
	printf("following list should have 5/4/2/1 in that order\n");
	alist_dump(myhead); 

	/* test series 5. delete 1st and last entries in list
	*/
	bip.s_addr = 0x01010101;
	alist_delete(&myhead, bip);
	bip.s_addr = 0x05050505;
	alist_delete(&myhead, bip);
	printf("following list should only have 4/2 in it\n");
	alist_dump(myhead); 

	/* final test. put entry back in and print out.
	 * list should have 3/4/2 in that order
	 */
	bip.s_addr = 0x03030303;
	alist_create(&myhead, bip);
	printf("following list should have 3/4/2 in that order\n");
	alist_dump(myhead); 
}
#endif

