/*
 * Copyright (c) 1996,1997,2001 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* DHCP client program 
 *
 * functions:
 *
 * initDhcpSocket - setup dhcp socket for i/o
 * dhcpStateMachine - drive dhcp discovery process, dealing with
 *	two kinds of state machine events,  dhcp input, and time ticks.
 * dhcpEnterInitState - make sure dhcp internal state machine is
 *	ready to start
 * myifconfig - set an interface IP address
 * getifaddr - get an interface IP address
 * getifmask - get an interface netmask
 * getInnerIP - get mvif inner ip address
 * getOuterIP - get mvif outer ip address
 * setInnerIP - set inner mvif ip
 * setOuterIP - set outer mvif ip
 * enable_dhcp - setup routing mechanism for dhcp/MIP combined 
 * disable_dhcp - turn off dhcp function and restore normal MIP
 * get_broadcast_address - get broadcast address of interface
 * send_discover - send DHCP DISCOVER message
 * send_request - send DHCP REQUEST message
 * parse_packet - parse a DHCP input message
 * dhcp_msg_type - determine dhcp message type
 * error - syslog error.  FATAL in all cases.
 * dhcpdump - dump a dhcp input message.
 * 
*/

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/sockio.h>
#include <net/if.h>
#include <netinet/in.h>
#define	KERNEL	/* suppress a warning message */
#include <netkey/key.h>
#undef	KERNEL
#include <netinet/if_mvif.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <syslog.h>
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include "mnconfig.h"
#include "mnstate.h"
#include "dhcp.h"


#define IN              1
#define OUT             2

#define	BROADCAST	1
#define	UNICAST		0
#define	TRUE		1
#define	FALSE		0

/* dhcp UDP ports
*/
#define DHCPS_PORT	67	/* we send to this */
#define DHCPC_PORT	68	/* client port - we use this */

/* dhcp state machine
*/
#define DHCP_STATE_INIT		0
#define DHCP_STATE_SELECTING	1
#define DHCP_STATE_REQUESTING	2
#define DHCP_STATE_BOUND	3
#define DHCP_STATE_RENEWING	4
#define DHCP_STATE_REBINDING	5

#define IN              1
#define OUT             2
/* time interval lease be renewed. a short renewing time allows for
 * effective movement detection. client includes lease suggestions
 * in DHCPDISCOVER and DHCPREQUEST to request a short lease 
 */
#define DHCP_DEFAULT_RENEWING_TIME	60

/* time to wait for and select OFFERs. the best to select servers
 * according to signal strength. select first server in this impl. 
 */
#define DHCP_DEFAULT_SELECTING_TIME	5

#define DHCP_DEFAULT_RETRY_LIMIT	6
int backoff[] = { 2, 3, 4, 5, 8, 16 };

char * dhcp_message_types[] = { "BOOTP", "DHCPDISCOVER", "DHCPOFFER", 
				"DHCPREQUEST", "DHCPDECLINE", "DHCPACK", 
				"DHCPNAK", "DHCPRELEASE", "DHCPINFORM" };
struct option {
	u_char type;
	u_char length;
	u_char value[255];
};

extern struct mn_config mn_config;
extern int pubOrpriv;
extern int udpSocket;

struct dhcp_packet packet, pkt;
int dhcpSocket;
int dhcpState;
static int previous_policy = NOSTATE;
int lastxid, serverxid;
static struct in_addr client;		/* DHCP COA */ 
static struct in_addr subnet_mask; 
static struct in_addr router; 
static struct in_addr name_server;
static struct in_addr server;
/* client must maintain two timers, T1 and T2
 * general purpose: lease renewal
 * T1: time at which client enters RENEWING state
 *	and attempts to contact the dhcp server
 *	client sends DHCPREQUEST msg to server
 * T2: time at which client enters REBINDING state
 *	and attempts to contact ANY server
 * T1 must be less than T2
 * T2 must be less than renewal time.
 * T1 by default must be .5 * the lease time.
 * T2 by default must be .875 * the least time.
 * T1 and T2 MAY be told to the client by the server (may not).
*/
static int T1;  /* dhcp renewal time from dhcp message */ 
static int T2;  /* dhcp rebinding time from dhcp message */
static int lease_time; /* the dhcp lease time */
int msgtype;
int selecting_timer, retry_timer, retry_count, T1_timer, T2_timer, lease_timer;

static void send_discover();
void send_request(int, int);
void error(char * fmt, ...);
extern void mnStateMachine(int, struct alist *, struct in_addr);

/* 
 * setup socket for dhcp to use
 */
initDhcpSocket() 
{
	struct sockaddr_in name;
	struct timeval tv;
	int flag;

	/* get time of day and set random seed.  Note: this
	 * does this for ALL of mnd
	 */
	gettimeofday(&tv, NULL);
	srandom(tv.tv_sec);

	/* udp socket
	*/
	dhcpSocket = socket(AF_INET, SOCK_DGRAM, 0);
	if (dhcpSocket < 0) {
		error("socket");
	}

	flag = 1;

	/* we can broadcast
	*/
	if (setsockopt(dhcpSocket, SOL_SOCKET, SO_BROADCAST, 
					(char *)&flag, sizeof(flag)) < 0) {
		error("setsockopt");
	}

	name.sin_family = AF_INET;
	name.sin_addr.s_addr = INADDR_ANY;
	name.sin_port = htons(DHCPC_PORT);
	name.sin_len = sizeof(name);
	bzero(name.sin_zero, sizeof(name.sin_zero));
	
	/*  bind client port
	*/
	if (bind(dhcpSocket, (struct sockaddr *)&name, sizeof(name)) < 0) {
		error("dhcp socket bind failed - dhcp running?");
		exit(1);
	}
}

/*
 * 
 *  The dhcp_tunnel_policy enables the dhcp setup 
 *  based on the policy, that is given to it as 
 *  input. The policies includes IPIP/NAT/NOTUNNEL/NOSTATE
 *
 *  It disables any previous policy and installs the new 
 *  policy.
 *
 */  
void 
dhcp_tunnel_policy(int new_policy)
{
	extern int debug;
	
	if (debug){
		printf("dhcp_tunnel_policy -- previous/new --> %d/%d \n",previous_policy,new_policy);
	}
	
	if (new_policy == previous_policy){
		return;
	}

	/* undo the previous policy */ 
	switch(previous_policy){
		
		case NOSTATE  :
				disable_dhcp(NOSTATE);
				break;
		case IPIP     :
				disable_dhcp(IPIP);
				break;
		case NAT      :
				disable_dhcp(NAT);
				break;
		case NOTUNNEL :
				disable_dhcp(NOTUNNEL);
				break;
		
	}

	/* install the new policy */
	switch(new_policy){

		case NOSTATE  :
				enable_dhcp(NOSTATE);
				break;
		case IPIP     :
				enable_dhcp(IPIP);
				break;
		case NAT      :
				enable_dhcp(NAT);
				break;			
		case NOTUNNEL :
				enable_dhcp(NOTUNNEL);
				break;
	}

	/* dhcp_present_policy is for the sake of mnstat
	*/
	mns.dhcp_present_policy = previous_policy = new_policy;
}

/* if set, we have backed up the existing /etc/mnd.conf file
*/
static int dhcp_wipes_resolver = 0;

void 
saveResolver() 
{
	FILE *fp; 

	/* backup /etc/resolv.conf so as to not obliterate it
	 * only do this once.
	*/
	if (dhcp_wipes_resolver == 0) {
		dhcp_wipes_resolver = 1;
		system ("cp /etc/resolv.conf /etc/resolv.conf.mnd.bak");
	}

	fp = fopen("/etc/resolv.conf","w");
	if ( fp == NULL ) {
		error("Could not open /etc/resolv.conf");
	}
	/*
	fprintf(fp, "search \n");
	*/
	fprintf(fp, "nameserver %s\n",inet_ntoa(name_server));
	fclose(fp);
}

/*
 * put /etc/resolv.conf back at HUP time 
 */
restoreResolver()
{
	if (dhcp_wipes_resolver == 0)
		return;

	system("cp /etc/resolv.conf.mnd.bak /etc/resolv.conf");
}

/* DHCP State Machine
 *  if event, then the state machine is driven by an incoming message,
 *  otherwise, the state machine is driven by mnStateTick 
 *
 * event values:
 *	1:  dhcp message has been received and socket is ready to read
 *	0:  state machine time tick (1 second has passed)
 *
 * mns.mode state variable.  This is a sub-state used by the dhcp
 * popup mechanism and is NOT an official mn state.
 *	if DHCP_MODE, then we have done the dhcp thing.
 *	if AGENT_MODE, then we have NOT done the dhcp thing.
 *		This would be the normal MIP state; i.e., agents
 *		get higher priority over the COA function, although
 *		the COA function should be manually forceable.
 * the official mn state for "dhcp in use" is CO_LOCATED.
 * That means we are using DHCP and MIP and in point of fact
 * are acting as our own foreign agent.
*/

dhcpStateMachine(int event)
{
	struct sockaddr_in from;
	int fromlen;
	extern int debug;

	if (debug) {
		printf("dhcp: dhcpStateMachine: event %d\n", event);
		printf("dhcp: mnstate %d, mns.mode %d\n", mns.state, mns.mode);
		printf("dhcp: dhcpstate %d\n", dhcpState);
	}
	/* paranoid code: make sure dhcp is off
	*/
	if (mns.state != NOWHERE && mns.state != CO_LOCATED) {
		if (mns.mode > 0) {
		    	mns.mode = AGENT_MODE;
			/* go to no dhcp tunnel state
			*/
			dhcp_tunnel_policy(NOSTATE);
		}
		return;
	}
	/* if we get to here, then get ready to do the dhcp thing
	*/
	if (mns.mode == AGENT_MODE) {
		mns.mode = mn_config.dhcp_policy;
		dhcpState = DHCP_STATE_INIT;
	}

	/* if message pending, read a dhcp message
	*/
	if (event) {
		if (debug) {
			printf("dhcp: message read event\n");
		}
		bzero(&pkt, sizeof(pkt));
		fromlen = sizeof(struct sockaddr);
		if (recvfrom(dhcpSocket, &pkt, sizeof(pkt), 0,
				(struct sockaddr *)&from, &fromlen) < 0) {
			error("recvfrom");
		}
		
		/* figure out what type of message
		*/
	        msgtype = dhcp_msg_type(&pkt);

		/* if not a DHCP message (BOOTP message) or Transaction_ID
		 *   mismatch, ignore the message 
		 */
	        if (msgtype == 0 || ntohl(pkt.xid) != lastxid) 
			return;
	}

	/* now figure out what to do based on internal DHCP state
	 * machine. in theory, there could be two external
	 * events for each state here.  sometimes they do NOT matter.
	 */
	switch (dhcpState) {
	    /* initial state
	    */
	    case DHCP_STATE_INIT :
		/* note: we ignore any inputs in this state
		*/
		/* send discovery message to get DHCP started
		*/
		send_discover();
		if (debug) {
			printf("dhcp: leaving INIT state, broadcasted DISCOVER message\n");
		}
		/* next state
		*/
		dhcpState = DHCP_STATE_SELECTING;
		selecting_timer = DHCP_DEFAULT_SELECTING_TIME;
		break;

	    /* hopefully we get an offer and go to the next state
	    */
	    case DHCP_STATE_SELECTING :
		if (event) {
		    if (msgtype == DHCPOFFER) {
			if (debug)
				printf("dhcp: got OFFER from %s\n", 
					inet_ntoa(from.sin_addr));
			T1 = T2 = lease_time = 0;
		        parse_packet(&pkt);
			if (T1 == 0) {
				T1 = lease_time * .5;
			}
			if (T2 == 0) {
				T2 = lease_time * .875;
			}
			/* constrain T1 to be less than or equal
			 * configured time, if set
			*/
			if (mn_config.dhcp_renewing_period > 0) {
				if (T1 > mn_config.dhcp_renewing_period) {
					T1 = mn_config.dhcp_renewing_period;
				}
			}
			if (debug) {
				printf("IP = %s\n", inet_ntoa(client));
				printf("Subnet_mask = %s\n", inet_ntoa(subnet_mask));
				printf("First router = %s\n", inet_ntoa(router));
				if (mn_config.doDns) {
					printf("Name Server = %s\n", inet_ntoa(name_server));
				}
				printf("Lease time = %d\n", lease_time);
				printf("renewal time T1 = %d\n", T1);
				printf("rebind time T2 = %d\n", T2);
			}
			send_request(BROADCAST, TRUE);
			if (debug)
				printf("broadcasted REQUEST, entering "
							"REQUESTING state\n");
			dhcpState = DHCP_STATE_REQUESTING;

			retry_count = 0;
			retry_timer = backoff[retry_count];
		    }
		}
	        /* else time elapsed and we may get to start over again
	        */
		else {
		    selecting_timer--;
		    if (selecting_timer <= 0) 
			dhcpState = DHCP_STATE_INIT;
		}
		break;
	    /* hopefully we get an ACK for our request
	    */
	    case DHCP_STATE_REQUESTING :
		if (event) {
		    if (msgtype == DHCPACK) {
			if (debug)
				printf("got ACK from %s, entering BOUND state\n", 
						inet_ntoa(from.sin_addr));
			parse_packet(&pkt);
			/* 
			 * if auto mode
			 *	policy comes from config
			 * else
			 *	policy is manual and forced
			 *		get from mns.mode
			 * OLD: enable_dhcp();
			 */
			if (mns.auto_mode == TRUE) {
				dhcp_tunnel_policy(mn_config.dhcp_policy);
			}
			else {
				dhcp_tunnel_policy(mns.mode);
			}
			dhcpState = DHCP_STATE_BOUND;
			T1_timer = T1;
			T2_timer = T2;
			lease_timer = lease_time;

			/* Let's rewrite the /etc/resolv.conf, if mnd is 
			 * configured to do so.  we only do this the first time.
			 */
			if (mn_config.doDns) {
				saveResolver();
			}
			/* change state to CO_LOCATED and send out a MIP reg. 
			*/
			mnStateMachine(udpSocket, NULL, client);
		    }
		    /* but we could get a NAK, start over
		    */
		    else if (msgtype == DHCPNAK) {
			if (debug)
				printf("dhcp: got NAK from %s, go back to INIT state\n", 
						inet_ntoa(from.sin_addr));
			dhcpState = DHCP_STATE_INIT;
		    }
		}
		/* time elapsed, we have have to retry the request
		 * or worse, start over
		*/
		else {
		    retry_timer--;
		    if (retry_timer <= 0) {
			if (retry_count < DHCP_DEFAULT_RETRY_LIMIT-1) {
			    send_request(BROADCAST, TRUE);
			    if (debug)
			    	printf("dhcp: time out, retrying...\n");
			    retry_count++;
			    retry_timer = backoff[retry_count];
			}
			else {
			    dhcpState = DHCP_STATE_INIT;
			}
		    }
		}
		break;
	    /* we have sent MIP registration, we may have to renew
	     * note: we don't deal with Mobile-IP, we deal with
	     * DHCP renewals.  ticks drive this state.
	    */
	    case DHCP_STATE_BOUND :
		/* ignore any messages in this state
		*/
		T1_timer--;
		T2_timer--;
		lease_timer--;
		if (debug) {
			printf("dhcp: state bound, t1 %d t2 %d, lease %d\n",
				T1_timer, T2_timer, lease_timer);
		}
		/* renewal timer */
		if (T1_timer <= 0) {
			send_request(UNICAST, FALSE);
			if (debug)
		 		printf("T1 expired, sent REQUEST, "
						"entering RENEWING state\n");
			dhcpState = DHCP_STATE_RENEWING;
			retry_count = 0;
			retry_timer = (T2-T1)/2;
			/* don't retry if retry_timer is too small */
			if (retry_timer < 3) retry_count++;
		}
		break;
	    /* we have to renew
	    */
	    case DHCP_STATE_RENEWING :
		/* hopefully we got an ACK
		*/
		if (event) {
		    if (msgtype == DHCPACK) {
			if (debug)
				printf("got ACK from %s, entering BOUND state\n", 
						inet_ntoa(from.sin_addr));
			parse_packet(&pkt);
			/* network parameters should be the same */
			/* enable_dhcp(); */
			dhcpState = DHCP_STATE_BOUND;
			T1_timer = T1;
			T2_timer = T2;
			lease_timer = lease_time;
		    }
		    else if (msgtype == DHCPNAK) {
			if (debug)
				printf("got NAK from %s, go back to INIT state\n", 
						inet_ntoa(from.sin_addr));
			/*
			disable_dhcp();
			dhcpState = DHCP_STATE_INIT; */

			enterNowhereState(mns.state, FALSE);
		    }
		}
		else {
		    T2_timer--;
		    lease_timer--;
		    if (!retry_count) {
		        retry_timer--;
		        if (retry_timer <= 0) {
			    send_request(UNICAST, FALSE);
			    if (debug) {
			    	printf("time out, retrying...\n");
			    }
			    retry_count++;
			}
		    }
		    if (T2_timer <= 0) {
			/* enter REBINDING state only if plenty of time left */
			if (lease_timer >= 4) {
			    send_request(BROADCAST, FALSE);
			    if (debug) 
			    	printf("T2 expired, broadcasted REQUEST, entering"
							" REBINDING state\n");
			    dhcpState = DHCP_STATE_REBINDING;
			    retry_count = 0;
			    retry_timer = (lease_time - T2) / 2;
			    /* don't retry if retry_timer is too small */
			    if (retry_timer < 3) retry_count++;
			}
			else {
			    if (debug)
			    	printf("dhcp: T2 expired, go back to INIT state "
								"directly\n");
			/*
			    disable_dhcp();
			    dhcpState = DHCP_STATE_INIT; */
			    enterNowhereState(mns.state, FALSE);
			}
		    }
		}
		break;
	    case DHCP_STATE_REBINDING :
		if (event) {
		    if (msgtype == DHCPACK) {
			if (debug) {
				printf("got ACK from %s, entering BOUND state\n", 
						inet_ntoa(from.sin_addr));
			}

			parse_packet(&pkt);
			/* network parameters should be the same */
			/* enable_dhcp(); */
			dhcpState = DHCP_STATE_BOUND;
			T1_timer = T1;
			T2_timer = T2;
			lease_timer = lease_time;
		    }
		    else if (msgtype == DHCPNAK) {
			if (debug)
				printf("got NAK from %s, go back to INIT state\n",
						inet_ntoa(from.sin_addr));
			/*
			disable_dhcp(FALSE);
			dhcpState = DHCP_STATE_INIT; */
			enterNowhereState(mns.state, FALSE);
		    }
		}
		else {
		    lease_timer--;
		    if (!retry_count) {
		        retry_timer--;
		        if (retry_timer <= 0) {
			    send_request(BROADCAST, FALSE);
			    if (debug)
			    	printf("time out, retrying\n");
			    retry_count++;
			}
		    }
		    if (lease_timer <= 0) {
			if (debug)
				printf("lease expired, go back to INIT state\n");
			/*
			enable_dhcp();
			dhcpState = DHCP_STATE_INIT; */
			enterNowhereState(mns.state, FALSE);
		    }
		}
		break;
	    default :
		error("this should not happen, DHCP entered an unknown state");
		break;
	}
}

/* DHCP state machine enters INIT state */
dhcpEnterInitState() {
        extern int debug;
	if (dhcpState == DHCP_STATE_BOUND || dhcpState == DHCP_STATE_RENEWING
	    || dhcpState == DHCP_STATE_REBINDING){
		dhcp_tunnel_policy(NOSTATE);
	}
	dhcpState = DHCP_STATE_INIT;
}

/* ifconfig: configure the interface with an address 
 * if address is zero, delete the address 
 */
      
myifconfig(char * ifname, struct in_addr ip, struct in_addr mask)
{       
        struct ifreq ifr;
        struct ifaliasreq ifra;
        struct sockaddr_in * ifa; 
 
	/* get address of the interface */
        strcpy(ifr.ifr_name, ifname);
        if (ioctl(dhcpSocket, SIOCGIFADDR, &ifr) < 0) {
		if (errno != EADDRNOTAVAIL) 
                	error("ioctl: SIOCGIFADDR: %s", strerror(errno));
        }
	else errno = 0;

        /* delete old address first */
        if (!errno && ioctl(dhcpSocket, SIOCDIFADDR, &ifr) < 0) {
                error("ioctl: SIOCDIFADDR");
        }

        /* add new address */

	if (ip.s_addr) {
        	bzero(&ifra, sizeof(ifra));
        	strcpy(ifra.ifra_name, ifname);
        	ifa = (struct sockaddr_in *)(&ifra.ifra_addr);
        	ifa->sin_family = AF_INET;
        	ifa->sin_addr.s_addr = ip.s_addr;
        	ifa->sin_len = sizeof(struct sockaddr_in);
        	ifa = (struct sockaddr_in *)(&ifra.ifra_mask);
        	ifa->sin_family = AF_INET;
        	ifa->sin_addr.s_addr = mask.s_addr;
        	ifa->sin_len = sizeof(struct sockaddr_in);
        	ifa = (struct sockaddr_in *)(&ifra.ifra_broadaddr);
        	ifa->sin_family = AF_INET;
        	ifa->sin_addr.s_addr = ip.s_addr | ~mask.s_addr;
        	ifa->sin_len = sizeof(struct sockaddr_in);
        	if (ioctl(dhcpSocket, SIOCAIFADDR, &ifra) < 0) {
                	error("ioctl: SIOCAIFADDR");
        	}
	}

}

/* get the IP address of interface ifname */

void getifaddr(char * ifname, struct in_addr * addr)
{
	struct ifreq ifr;
        struct sockaddr_in * ifa; 

	strcpy(ifr.ifr_name, ifname);
	if (ioctl(dhcpSocket, SIOCGIFADDR, &ifr) < 0) {
		if (errno != EADDRNOTAVAIL) 
			error("ioctl: SIOCGIFADDR: %s", strerror(errno));
	}
	else errno = 0;
	ifa = (struct sockaddr_in *)(&ifr.ifr_addr);
	if (!errno)
		addr->s_addr = ifa->sin_addr.s_addr;
	else
		addr->s_addr = 0;
}

/* get the subnet mask of interface ifname */

void getifmask(char * ifname, struct in_addr * mask)
{
	struct ifreq ifr;
        struct sockaddr_in * ifa; 

	strcpy(ifr.ifr_name, ifname);
	if (ioctl(dhcpSocket, SIOCGIFNETMASK, &ifr) < 0) {
		if (errno != EADDRNOTAVAIL)
			error("ioctl: SIOCGIFNETMASK");
	}
	else errno = 0;
	ifa = (struct sockaddr_in *)(&ifr.ifr_addr);
	if (!errno)
		mask->s_addr = ifa->sin_addr.s_addr;
	else
		mask->s_addr = 0;
}

getInnerIP(char * ifname)
{
	struct ifreq ifr;

	strcpy(ifr.ifr_name, ifname);
	if (ioctl(dhcpSocket, SIOCGMVINIP, &ifr) < 0) {
		error("ioctl: SIOCGMVINIP");
	}
	return((int)ifr.ifr_data);
}

getOuterIP(char * ifname)
{
	struct ifreq ifr;

	strcpy(ifr.ifr_name, ifname);
	if (ioctl(dhcpSocket, SIOCSMVOUTIP, &ifr) < 0) {
		error("ioctl: SIOCSMVOUTIP");
	}
	return((int)ifr.ifr_data);
}

/*
 * set the "inner" mvif ip address to some value.
 * The inner ip address is not set by default, but should
 * be set to the MN's home address so that encapsulated
 * packets sent "thru" the mvif device will have a sane
 * inner IP datagram.
 */
void 
setInnerIP(char * ifname, struct in_addr inip)
{
	struct ifreq ifr;

	strcpy(ifr.ifr_name, ifname);
	ifr.ifr_data = (caddr_t) inip.s_addr;
	if (ioctl(dhcpSocket, SIOCSMVINIP, &ifr) < 0) {
		if (errno != EINVAL)
			error("ioctl: SIOCSMVINIP");
		else
			errno = 0;
	}
}

void setOuterIP(char * ifname, struct in_addr outip)
{
	struct ifreq ifr;

	strcpy(ifr.ifr_name, ifname);
	ifr.ifr_data = (caddr_t) outip.s_addr;
	if (ioctl(dhcpSocket, SIOCSMVOUTIP, &ifr) < 0) {
		error("ioctl: SIOCSMVOUTIP");
	}
}


/* variables used to save the old address of mvif interface */

static struct in_addr mvif_old_addr, mvif_old_mask;
static struct in_addr mvif_old_inner_ip, mvif_old_outer_ip;

/* 
 * enable_dhcp and disable_dhcp.  These two functions
 * turn the dhcp routing infrastructure on the host on or off
 */

/*
 * enable dhcp based on the policy given to it as input
 * do NOSTATE          
 *       nothing to do
 *     
 * do IPIP
 *       1. set mobile device (wi0) to dhcp coa
 *       2. set mvif to MN ip, save any previous addr
 *       3. set inner/outer mvif to be MN
 *       4. set HA to talk directly via host route to local dhcp router
 *       5. set default route thru mvif device for IPIP tunnel
 * do NAT
 *       1. set mobile device (wi0) to dhcp coa
 *       2. set mvif to MN ip, save any previous addr
 *       3. set inner/outer mvif to be MN
 *       4. set HA to talk directly via host route to local dhcp router
 *       5. set default route thru mvif device for IPIP tunnel
 *		note: this is mostly useful for making sure the
 *		innermost ip_src address is set to the MN MIP ip.
 *		Packets sent thru this route are caught by divert.
 *		They are not as a result IPIP packets, but IP/UDP/MIP/IP.
 *       6. setup ipfw firewall rules for divert  
 * do NOTUNNEL
 *	note: NOTUNNEL for the most part, just makes us an
 *	ordinary non-MIP node, although we still can receive
 *	packets sent to our MIP ip_src.  Packets sent from us
 *	will have the local COA as their ip_src.  
 *       1. set mobile device (wi0) to dhcp coa
 *       2. set mvif to MN ip, save any previous addr
 *       3. set inner/outer mvif to be MN
 *       4. set default route thru local dhcp router
 *
 */
enable_dhcp(int policy)
{
	char command[256];
	struct in_addr zero;
	struct in_addr mask;
	extern int debug; 

	zero.s_addr = 0;

	if ( debug ){
		printf("Enabledhcp() -- policy --> %d \n",policy); 
	}
	
	if (policy == NOSTATE)
		return;
	
	if ( (policy == NAT) || (policy == IPIP)){

		/* configure mobile device with DHCP COA
		 */
		myifconfig(mn_config.ifname, client, subnet_mask);
		
		/* configure some virtual device (mvif) with fixed
		 * ip address (mn home address). note: we save the
		 * previous address if any first.
		 */
		getifaddr(mn_config.mvifname, &mvif_old_addr);
		
		getifmask(mn_config.mvifname, &mvif_old_mask);
		myifconfig(mn_config.mvifname, mn_config.mn, mn_config.netmask);
		
		/* set the inner source address of mvif to the local MN's
		 * home address.  Thus any encapsulated packets will
		 * have an inner IP datagram that for sure originates from us.
		 * note: we save any original setting, and then restore it when done.
		 */
		mvif_old_inner_ip.s_addr = getInnerIP(mn_config.mvifname);
		setInnerIP(mn_config.mvifname, mn_config.mn);
		
		/* set the outer source address of mvif to our home MN ip too.
		 * note: we save any original value again.
		 */
		mvif_old_outer_ip.s_addr = getOuterIP(mn_config.mvifname);
		setOuterIP(mn_config.mvifname, client);
		
		/* now decide routing strategy:
		 * if default IPIP tunnel home 
		 * we set a host route to the HA, and do IPIP
		 * tunnel for the default route thru the HA
		 */
		
		/* host route to the Home Agent BEFORE we route all other packets
		 * to the home agent via a tunnel.  This makes sure
		 * that we can route MIP registration packets to the HA
		 */
		route_add_host(mn_config.ha, router, client);

		mn_route_add_tunnel(mn_config.ha, zero, NULL);
	}	

	if ( policy == NAT ){
		/*
		setupDivertTunnel();
		*/
	}

	if ( policy == NOTUNNEL){
	
		/* Delete the MN Home IP bound to mvif0 */
		del_ifaddr(mn_config.mn, mn_config.netmask, mn_config.mvifname);

		/* configure mobile device with DHCP COA
		 */
		myifconfig(mn_config.ifname, client, subnet_mask);


		/*  set default route thru local dhcp router */
	  	route_add_default(router);
	}

}

/*
 * disable dhcp based on the policy given to it as input
 *
 * undo NOSTATE  
 *            nothing to do
 *
 * undo IPIP
 *	put saved address back on mvif       
 *      put MN ip back on mobile device 
 *      set client address back to 0 (dhcp coa internal state)
 *      set inner src of mvif with old saved address
 *      do same with outer src
 *      delete default tunnel route thru mvif
 *      delete HA host route
 * undo NAT
 *      put saved address back on mvif
 *      put MN ip back on mobile device
 *      set client address back to 0 (dhcp coa internal state)
 *      set inner src of mvif with old saved address
 *      do same with outer src
 *      delete default tunnel route thru mvif
 *      delete HA host route
 * undo NOTUNNEL
 *      put saved address back on mvif
 *      put MN ip back on mobile device
 *      set client address back to 0 (dhcp coa internal state)
 *      set inner src of mvif with old saved address
 *      do same with outer src
 *      delete default route
 */
disable_dhcp(int policy)
{
	char command[256]; 
	extern int debug;
	struct in_addr zero;
	struct in_addr mask;

	if ( debug ){
		printf("Disabledhcp() -- policy --> %d \n",policy); 
	}

	if (policy == NOSTATE)
		return;
	
	if ( (policy == IPIP) || (policy == NAT)){ 
		/* configure mvif with old mvif saved address 	  
		 */	
		myifconfig(mn_config.mvifname, mvif_old_addr, mvif_old_mask);
	
		/* configure mobile device with our home MIP address
		 */	
		myifconfig(mn_config.ifname, mn_config.mn, mn_config.netmask);
		
		/* set the inner source address of mvif with old saved address
		 */
		setInnerIP(mn_config.mvifname, mvif_old_inner_ip);
		
		/* set the outer source address of mvif 
		 */
		setOuterIP(mn_config.mvifname, mvif_old_outer_ip);
		
		/* delete default tunnel route 
		*/
		mn_route_delete_default(mn_config.ha, FALSE);	  	  
		
		/* delete the auxilliary host route for HA to temporary local 
		 * router 
		 */
		route_delete_host(mn_config.ha, router);
	}
	
	if ( (policy == NAT)){ 
		/*
		if (mns.fr_number > 0){	  
			delFireWallRule();
		}
		*/
	}

	if ( policy == NOTUNNEL){
	  
		/* configure mvif with old mvif saved address 	  
		 */	
		myifconfig(mn_config.mvifname, mvif_old_addr, mvif_old_mask);

		/* configure mobile device with our home MIP address
		 */	
		myifconfig(mn_config.ifname, mn_config.mn, mn_config.netmask);

		/* delete default route */
		route_del_default(router);	  	
	}
}

/* get the broadcast address of interface */

u_long 
get_broadcast_address()
{
	struct ifreq ifr;
	struct sockaddr_in * ifa;
	struct in_addr addr;	
	extern int debug;

	strcpy(ifr.ifr_name, mn_config.ifname);
	bzero(&ifr.ifr_addr, sizeof(struct sockaddr));
	if (ioctl(dhcpSocket, SIOCGIFBRDADDR, &ifr) < 0) {
		error("ioctl");
	}
	ifa = (struct sockaddr_in *)(&ifr.ifr_addr);
	return(ifa->sin_addr.s_addr);
}

/* broadcast a DHCP discover message. ip src for discover is always
 * 0,  dst is always limited broadcast.
*/
static
void 
send_discover()
{
	struct sockaddr_in to;
	struct in_addr ip_src, ip_dst;
	struct option * opt;
	int optlen, packetlen, rc;

	/* setup address 
	*/

#ifdef OLDWAY
	to.sin_family = AF_INET;
	to.sin_addr.s_addr = get_broadcast_address();
	to.sin_port = htons(DHCPS_PORT);
	to.sin_len = sizeof(to);
	bzero(to.sin_zero, sizeof(to.sin_zero));
#endif

	/* init/setup a DHCP discovery packet for sending
	*/
	bzero(&packet, sizeof(packet));
	packet.op = BOOTREQUEST;
	packet.htype = HTYPE_ETHER;
	packet.hlen = 6;
	packet.hops = 0;
	lastxid = random();	/* generate a random transaction id */
	packet.xid = htonl(lastxid);
	packet.secs = 0;
	packet.flags = htons(BOOTP_BROADCAST);
	packet.ciaddr.s_addr = 0;
	packet.yiaddr.s_addr = 0;
	packet.siaddr.s_addr = 0;
	packet.giaddr.s_addr = 0;
	bcopy(mn_config.mac, packet.chaddr, 6);

	/* put in COOKIE */

	bcopy(DHCP_OPTIONS_COOKIE, packet.options, 4);
	optlen = 4;
	opt = (struct option *)(packet.options + optlen);

	/* put in message type 
	 * DISCOVER as we are client needing info
	*/

	opt->type = DHO_DHCP_MESSAGE_TYPE;
	opt->length = 1;
	opt->value[0] = DHCPDISCOVER;
	optlen += 3;
	opt = (struct option *)(packet.options + optlen);

	/* tell the server we are interested in these parameters: 
	 * subnet mask
	 * routers
	*/

	opt->type = DHO_DHCP_PARAMETER_REQUEST_LIST;
	opt->length = 3;
	opt->value[0] = DHO_SUBNET_MASK;
	opt->value[1] = DHO_ROUTERS;
	opt->value[2] = DHO_DOMAIN_NAME_SERVERS;
	optlen += 5;
	opt = (struct option *)(packet.options + optlen);

	/* tell server lease renewal time if configured to do so
	 */

	if (mn_config.dhcp_renewing_period > 0) {
		opt->type = DHO_DHCP_LEASE_TIME;
		opt->length = 4;
		*((u_long *)opt->value) = 
				htonl(mn_config.dhcp_renewing_period * 2);
		optlen += 6;
		opt = (struct option *)(packet.options + optlen);
	}

	/* mark the end of options 
	*/
	opt->type = DHO_END;
	optlen += 1;

	packetlen = DHCP_FIXED_LEN + optlen;

	/* send the message */

#ifdef OLDWAY
	rc = sendto(dhcpSocket, &packet, packetlen, 0, 
					(struct sockaddr *)&to, sizeof(to));
#else
	/* these addresses do not need to be byte-swapped
	*/
	ip_src.s_addr = 0L;
	ip_dst.s_addr = 0xffffffff;

	rc = writebpf_udp(&packet, packetlen,
		     (char *) mn_config.mac,  NULL, /* NULL means bcast */
		     ip_src, ip_dst,
		     DHCPC_PORT, DHCPS_PORT);
	if (rc < 0) {
		syslog(LOG_WARNING, "bpf write failure %m");
	}
#endif
}

/* send a DHCPREQUEST message to server
 * if broadcast, the message will be broadcasted
 * if requesting, the message is sent in REQUESTING state 
*/
void 
send_request(int broadcast, int requesting)
{
	struct in_addr ip_src, ip_dst;
	struct sockaddr_in to;
	struct option * opt;
	int optlen, packetlen, rc;

	/* setup possible send ip address
	*/

	to.sin_family = AF_INET;
	if (broadcast) {
		/* we use bpf */
	}
	else {
		to.sin_addr.s_addr = server.s_addr;
	}
	to.sin_port = htons(DHCPS_PORT);
	to.sin_len = sizeof(to);
	bzero(to.sin_zero, sizeof(to.sin_zero));

	/* fillin BOOTP parameters 
	*/

	bzero(&packet, sizeof(packet));
	packet.op = BOOTREQUEST;
	packet.htype = HTYPE_ETHER;
	packet.hlen = 6;
	packet.hops = 0;
	lastxid = serverxid;	/* according to RFC, use 'xid' from server
				   DHCPOFFER message */
	packet.xid = htonl(serverxid);
	packet.secs = 0;
	if (broadcast) {
		packet.flags = htons(BOOTP_BROADCAST);
	}
	else {
		packet.flags = 0;
	}
	if (requesting)		/* use client IP if we are configured */
		packet.ciaddr.s_addr = 0;
	else
		bcopy(&client, &packet.ciaddr, sizeof(client));
	packet.yiaddr.s_addr = 0;
	packet.siaddr.s_addr = 0;
	packet.giaddr.s_addr = 0;
	bcopy(mn_config.mac, packet.chaddr, 6);

	/* put in COOKIE */

	bcopy(DHCP_OPTIONS_COOKIE, packet.options, 4);
	optlen = 4;
	opt = (struct option *)(packet.options + optlen);

	/* put in message type */

	opt->type = DHO_DHCP_MESSAGE_TYPE;
	opt->length = 1;
	opt->value[0] = DHCPREQUEST;
	optlen += 3;
	opt = (struct option *)(packet.options + optlen);

	/* include requested address only if we are requesting an 
	   initial lease, no requested address if extending lease */

	if (requesting) {
		opt->type = DHO_DHCP_REQUESTED_ADDRESS;
		opt->length = 4;
		bcopy(&client, opt->value, 4);
		optlen += 6;
		opt = (struct option *)(packet.options + optlen);
	}

	/* include server ID to indicate the server being selected */

	if (requesting) {
		opt->type = DHO_DHCP_SERVER_IDENTIFIER;
		opt->length = 4;
		bcopy(&server, opt->value, 4);
		optlen += 6;
		opt = (struct option *)(packet.options + optlen);
	}

	/* tell the server we are interested in these parameters */

	opt->type = DHO_DHCP_PARAMETER_REQUEST_LIST;
	opt->length = 3;
	opt->value[0] = DHO_SUBNET_MASK;
	opt->value[1] = DHO_ROUTERS;
	opt->value[2] = DHO_DOMAIN_NAME_SERVERS;
	optlen += 5;
	opt = (struct option *)(packet.options + optlen);

	/* suggest the server with a lease time suitable for 
	   movement detection */

	if (mn_config.dhcp_renewing_period > 0) {
		opt->type = DHO_DHCP_LEASE_TIME;
		opt->length = 4;
		*((u_long *)opt->value) = 
				htonl(mn_config.dhcp_renewing_period * 2);
		optlen += 6;
		opt = (struct option *)(packet.options + optlen);
	}

	/* mark the end of options */

	opt->type = DHO_END;
	optlen += 1;

	packetlen = DHCP_FIXED_LEN + optlen;

	/* send the message */

	if (broadcast) {
		ip_dst.s_addr = 0xffffffff;
		rc = writebpf_udp(&packet, packetlen,
		        (char *) mn_config.mac,  NULL, /* NULL means bcast */
		     	client, ip_dst,
		     	DHCPC_PORT, DHCPS_PORT);
	}
	else {
		rc = sendto(dhcpSocket, &packet, packetlen, 0, 
					(struct sockaddr *)&to, sizeof(to));
	}
	if (rc < 0) {
		syslog(LOG_WARNING, "bpf write failure %m");
	}
}

/* parse an incoming DHCP message
   TBD: return a value indicating result of parsing (could have problems)
	initialize those values to be overwritten by parser so that upon
	return, calling function can examine what options are available,
	avoiding erronously using old values not setup in this call.
	Should also deal with the situation when "sname" and "file"
	fields are overloaded */

parse_packet(struct dhcp_packet * pkt)
{
	struct option * opt;

	serverxid = ntohl(pkt->xid);

	bcopy(&(pkt->yiaddr), &client, sizeof(client));

	opt = (struct option *)(pkt->options + 4);

	while(opt->type != DHO_END) {
	    switch (opt->type) {
		case DHO_PAD : break;
		case DHO_SUBNET_MASK :
			bcopy(opt->value, &subnet_mask, sizeof(subnet_mask));
			break;
		case DHO_ROUTERS :
			/* use the first router */
			bcopy(opt->value, &router, sizeof(router));
			break;
	    case DHO_DOMAIN_NAME_SERVERS :
			/* use the first Name Server */
			bcopy(opt->value, &name_server, sizeof(name_server));
			break;
		case DHO_DHCP_MESSAGE_TYPE :
			msgtype = opt->value[0];
			break;
		case DHO_DHCP_SERVER_IDENTIFIER :
			bcopy(opt->value, &server, sizeof(server));
			break;
		case DHO_DHCP_LEASE_TIME :
			lease_time = ntohl(*((u_long *)opt->value));
			break;
		case DHO_DHCP_RENEWAL_TIME :
			T1 = ntohl(*((u_long *)opt->value));
			break;
		case DHO_DHCP_REBINDING_TIME :
			T2 = ntohl(*((u_long *)opt->value));
			break;
		default :
			/* skip other options */
			break;
	    }
	    if (opt->type != DHO_PAD)
		opt = (struct option *)((char *)opt + opt->length + 2);
	    else
		opt = (struct option *)((char *)opt + 1);
	}
}

/* determine the message type of current incoming message */

dhcp_msg_type(struct dhcp_packet * pkt)
{
	struct option * opt;

	opt = (struct option *)(pkt->options + 4);
	while (opt->type != DHO_END) {
		if (opt->type == DHO_DHCP_MESSAGE_TYPE)
			return(opt->value[0]);
		if (opt->type != DHO_PAD)
			opt = (struct option *)((char *)opt + opt->length + 2);
		else
			opt = (struct option *)((char *)opt + 1);
	}
	return 0;
}

void error(char * fmt, ...)
{
	va_list list;
	char msg[256];

	va_start(list, fmt);
	vsprintf(msg, fmt, list);
	va_end(list);
	syslog(LOG_ERR, "%s", msg);
	exit(1);
}


static
dhcpdump(struct dhcp_packet * pkt)
{
	struct option * opt;
	int i;

	printf("BOOTP fields:\n");
	printf("\tOpcode = %d\n", pkt->op);
	printf("\tHardware type = %d\n", pkt->htype);
	printf("\tHardware addr length = %d\n", pkt->hlen);
	printf("\tHops = %d\n", pkt->hops);
	printf("\tTransaction id = %u\n", ntohl(pkt->xid));
	printf("\tSeconds = %d\n", ntohs(pkt->secs));
	printf("\tFlags = 0x%x\n", ntohs(pkt->flags));
	printf("\tClient IP = %s\n", inet_ntoa(pkt->ciaddr));
	printf("\tYour IP = %s\n", inet_ntoa(pkt->yiaddr));
	printf("\tServer IP = %s\n", inet_ntoa(pkt->siaddr));
	printf("\tGateway IP = %s\n", inet_ntoa(pkt->giaddr));
	printf("\tClient hardware address = %02x:%02x:%02x:%02x:%02x:%02x\n",
		pkt->chaddr[0], pkt->chaddr[1], pkt->chaddr[2],
		pkt->chaddr[3], pkt->chaddr[4], pkt->chaddr[5]);
	printf("\tServer name = %s\n", pkt->sname);
	printf("\tBoot file = %s\n", pkt->file);

	printf("DHCP options:\n");
	opt = (struct option *)(pkt->options + 4);
	while (opt->type != DHO_END) {
		switch (opt->type) {
		    case DHO_PAD : break;
		    case DHO_SUBNET_MASK :
			printf("\tSubnet mask = %s\n", 
				inet_ntoa(*((struct in_addr *)opt->value)));
			break;
		    case DHO_ROUTERS :
			printf("\tRouters = %s",
				inet_ntoa(*((struct in_addr *)opt->value)));
			for (i=4; i<opt->length; i+=4)
			    printf(",%s", 
				inet_ntoa(*((struct in_addr *)(opt->value+i))));
			printf("\n");
			break;
		    case DHO_DHCP_MESSAGE_TYPE :
			printf("\tDHCP message type = %s\n", 
				dhcp_message_types[opt->value[0]]);
			break;
		    case DHO_DHCP_SERVER_IDENTIFIER :
			printf("\tServer identifier = %s\n",
				inet_ntoa(*((struct in_addr *)opt->value)));
			break;
		    case DHO_DHCP_LEASE_TIME :
			printf("\tLease time = %u\n", 
				ntohl(*((u_long *)opt->value)));
			break;
		    default :
			printf("\tOption type = %d, length = %d, value = 0x",
				opt->type, opt->length);
			for (i=0; i<opt->length; i++)
				printf("%02x", opt->value[i]);
			printf("\n");
			break;
		}
		if (opt->type != DHO_PAD)
			opt = (struct option *)((char *)opt + opt->length + 2);
		else
			opt = (struct option *)((char *)opt + 1);
	}
}
