
/*
 * Copyright (c) 2003 Portland State University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * divertHa.c
 * provides divert in and divert out encapsulation/decapsulation tunneling
 * at Home Agent.
 *
 * functions:
 * handleDivertIn(int fd) - handle CH to MN pkts, encapsulate for UDP tunnel
 * handleDivertOut(int fd) - handle MN to CH pkts, decapsulate and forward
 * setupFirewallRule(...) - insert divert firewall rules 
 * delFireWallRule(int number) - delete divert firewall rules
 */

#include <stdio.h>
#include <errno.h>
#include <limits.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <netdb.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/ip_icmp.h>
#include <netinet/ip_fw.h>
#include <math.h>
#include <sys/timeb.h>

#include "mipconfig.h"
#include "vlist.h"
#include "divertInterfaces.h"
#include "mip_log.h"

/*
 *  This function handles IP datagrams that have to be routed to the 
 *  Mobile Nodes.  Packets appear here from the CH headed to the MN.
 * 
 *  To the IP datagram prepend an IP-UDP-MIP_Message header and write it
 *  back into the divert socket to the MN.
 *
 * called from: mipmain/select function
 */ 

void 
handleDivertIn(int fd)
{
	struct sockaddr_in sin;
	int ii,n,sinlen;
	unsigned char packet[BUFSIZE];
	unsigned char *pack;
	extern int debug;
	struct vlist *hhead;
	struct ip *hdr; 
	unsigned short sport,dport;
	struct udphdr *udphdr;
	struct tcphdr *tcphdr;
	unsigned long seq,ack;
	unsigned short th_win;
	unsigned int th_off;
	unsigned int th_x2;
	unsigned char th_flags;
	struct in_addr saddr,daddr;
	int dataSize ;
	struct vlist *visitor;
	static int rawSock = -1;
	
	sinlen = sizeof(struct sockaddr);	
	pack = (packet + sizeof(struct ip) + sizeof(struct udphdr) + sizeof(struct mip_MsgHdr) );

	n=recvfrom(fd, pack, BUFSIZE, 0
		   ,(struct sockaddr *) &sin, &sinlen);

	if (n < 0){
		syslog( LOG_ERR, "handleDivertIn(): recvfrom (%m)");
		exit(1);
	}

	if (debug) {
		printf("handleDivertIn \n");
		/*dumpme(packet, n);*/
	}

	hdr=(struct ip *)pack;

	hhead   =  (struct vlist *) vlist_gethahead();

	/* 
	 * Search the visitor in the visitor list.
	 * we need the udp_src and udp_port fields
	 * of the corresponding visitor's structure
	 * for encapsulation.
 	 *
	 * TBD: we really should hash the visitor list lookup.
	 *
	 */
	visitor =  vlist_search(hhead,hdr->ip_dst);

	/* if this box truly is a visitor,
	 * note that we get info out of the visitor list
	 */
 	if (visitor){

		if (debug) {	  
			printf("It is  a visitor \n");
		}
	  
		/*  prepend an IP-UDP-MIP_Message Header */
		encapsulate(packet, n, mip_config.ha_sin.sin_addr,
			visitor->udp_src.sin_addr, visitor->udp_port);			    
		
		dataSize = n + sizeof(struct ip) + sizeof(struct udphdr)
		  + sizeof(struct mip_MsgHdr) ; 
		
		if (rawSock == -1)
			rawSock = getRawSock();
		
		/* Write it back out on the raw socket
		 * note the divert socket is not used here because we
		 * need to make sure ip_id is set. divert does not do that.
		*/
		sin.sin_addr.s_addr = INADDR_ANY;

		while(1){
			n=sendto(rawSock,(void *) packet, dataSize, 0,		  
				 (struct sockaddr *) &sin, sizeof(sin));
		  
			if (n < 0){

				if (errno == ENOBUFS)			  
					continue;
		
				if (errno == EHOSTUNREACH)
					return;

				syslog( LOG_ERR,
					"handleDivertIn(): sendto (%m), ip dst %s",
						inet_ntoa(visitor->udp_src.sin_addr.s_addr));
				exit(1);
			}
			break;
		}
	}
}

/*     This function handles data OR registration messages coming from MNs' 
 *     situated behind NAT machines. (both control/data on same stream).
 *	MNs are trying to send packets to CH thru tunnel to HA.
 *     
 *     If Data Packets, remove the IP-UDP-MIP_Message header from the
 *     IP-UDP-MIP_Message-IP datagram and write it back into the divert
 *     socket. Check the values of the MIP_MSGHDR_TYPE and Next_header
 *     before decapsulation as specified by the RFC.
 * 
 *     If Registration Packet, write it back into the divert socket 
 *     without any changes.
 */ 

void 
handleDivertOut(int fd)
{

	struct sockaddr_in sin,mysocket;
	int ii,n,n1,sinlen;
        unsigned char packet[BUFSIZE];
	unsigned char *decap_packet; 
	extern int debug;
	struct vlist *hhead;
	struct ip *hdr;
	struct vlist *visitor;
	struct mip_MsgHdr *msgHdr;
	struct udphdr *udpHdr;
	int raw; 
	int dataSize;

	sinlen = sizeof(struct sockaddr);
	n=recvfrom(fd, packet, BUFSIZE, 0
		   ,(struct sockaddr *) &sin, &sinlen);

	if (n < 0){
		syslog( LOG_ERR, "handleDivertOut(): recvfrom (%m)");
		exit(1);
	}

	hdr=(struct ip *)packet; 
	
	if (debug) {
		printf("handleDivertOut \n");
		/*dumpme(packet, n);*/
	}


	msgHdr = (struct mip_MsgHdr *) (packet + sizeof(struct ip) 
					+ sizeof(struct udphdr) );

	udpHdr = (struct udphdr *) (packet + sizeof(struct ip) );

	/* if type is MIP/NAT data packet (not MIP registration)
	*/
	if (msgHdr->type == MIP_MSGHDR_TYPE && 
		msgHdr->nextHeader == IPVERSION ){

		if (debug){	  
			printf("handleDivertOut - It is a Data Packet \n");
		}	  
	  
		/* move the ptr ahead to point to the actual packet */
		decap_packet = packet + (sizeof(struct ip) 
				 + sizeof(struct udphdr)
				 + sizeof(struct mip_MsgHdr));

		/* Length of the decapsulated packet */
		dataSize = n - (sizeof(struct ip) 
				 + sizeof(struct udphdr)
				 + sizeof(struct mip_MsgHdr));


		/* Write it back into the Divert Socket */
		sin.sin_addr.s_addr = INADDR_ANY;
		while(1){
			n1=sendto(fd,decap_packet, dataSize ,0		  
				  ,(struct sockaddr *) &sin, sizeof(sin));
		 
			if (n1 < 0){
				if (errno == ENOBUFS)			  
					continue;				  
				
				syslog( LOG_ERR, 
					"handleDivertOut(): sendto (%m)");
				exit(1);
			}
			break;
		}
		
	}else{

		/* Make sure that the ip destination of this packet is HA
		 * and the destination port is 434. 
		 */
		if (hdr->ip_dst.s_addr == mip_config.ha_sin.sin_addr.s_addr && 
			ntohs(udpHdr->uh_dport) == UDP_PORT){		    

			if (debug) {		  
				printf("handleDivertOut - It is a registration Packet \n");			  
			}	  
			
			/* Write it back into the Divert Socket, untouched */ 	  		
/*
			sin.sin_addr.s_addr = INADDR_ANY;	  		  
*/
			sin.sin_addr.s_addr = mip_config.ha_sin.sin_addr.s_addr;

			while(1){
				n=sendto(fd,(void *) packet, n ,0			  
					 ,(struct sockaddr *) &sin, sizeof(sin));
			  
				if (n < 0){			  

					if (errno == ENOBUFS)			  				  
						continue;
					syslog( LOG_ERR, "handleDivertOut(): sendto (self/HA) (%m)");				  
					exit(1);
				}
				break;
			}

		}else{

			if (debug) {		  
				printf("handleDivertOut - Dropping An Unknown Packet \n");
			}	  
		}
	}
}

/*
 * This Function sets-up a Firewall rule(ipfw divert rule). 
 * This Rules should be set-up after the  first three default rules dealing with 
 * loopback interface packets. Hence the rule number 301.
 * 
 * srcAddr --- Source Address 
 * dstAddr --- Destination Address
 * inOrOut --- 1 for incoming and 2 for outgoing
 * port    --- DIVERTPORTIN or DIVERTPORTOUT
 * proto   --- protocol ( For E.g. UDP -- 17, icmp -- 1 , TCP -- 6 )
 */
void setupFirewallRule(struct in_addr srcAddr,struct in_addr dstAddr
		       ,int inOrOut, int port,int proto,int num)
{

	struct ip_fw fw;         
	int fw_sock = -1;
	struct in_addr saddr;
	struct in_addr daddr;
	
	daddr = dstAddr;
	saddr = srcAddr;

        /* Specify the firewall rule and write                 	
	 *  it via a special raw socket 
	 *  at the appropriate section of the kernel 
	 */	   
	bzero(&fw, sizeof (struct ip_fw));
	
	fw.fw_dst.s_addr      = daddr.s_addr;
	if (daddr.s_addr == 0)
		fw.fw_dmsk.s_addr     = daddr.s_addr;	  	
	else
		fw.fw_dmsk.s_addr     = inet_addr("255.255.255.255"); 	        


	fw.fw_src.s_addr      = saddr.s_addr;
	if (saddr.s_addr == 0)
		fw.fw_smsk.s_addr     = saddr.s_addr;	         
	else
		fw.fw_smsk.s_addr     = inet_addr("255.255.255.255"); 	         

	
	fw.fw_flg |= IP_FW_F_DIVERT;
	
	/* Protocol number: 0=all, 1=ICMP, 6=TCP, 17=UDP, see RFC 1700  */
	fw.fw_prot=proto;
	
	fw.fw_un.fu_divert_port = port;
	
	/* open a socket */
	if ((fw_sock=socket(AF_INET, SOCK_RAW, IPPROTO_RAW))==-1) {
		syslog( LOG_ERR, "setupFireWallRule() : socket (%m)");
		exit(1);
	}
	
	fw.fw_number = num;

	/* debug only 
	mip_log(LOG_ERR, "setupFireWallRule(): number is %d", num);
	*/
	
	if (inOrOut == IN)
		fw.fw_flg |=  IP_FW_F_IN;	  
	else
		fw.fw_flg |=  IP_FW_F_OUT;	  
	
	if ((setsockopt(fw_sock, IPPROTO_IP, IP_FW_ADD, &fw
			, sizeof(struct ip_fw))) < 0 ){

		if (inOrOut == IN){	  	  
			fw.fw_flg &=  ~IP_FW_F_IN;		  
			syslog( LOG_ERR, "setupFireWallRule() : setsockopt - IN (%m), rule: %d", num);
			mip_log( LOG_ERR, "setupFireWallRule() : setsockopt - IN (%m), rule: %d", num);
			exit(1);
		}
		else{
			fw.fw_flg &= ~IP_FW_F_OUT;		  
			syslog( LOG_ERR, "setupFireWallRule() : setsockopt - OUT (%m), rule: %d", num);
			mip_log( LOG_ERR, "setupFireWallRule() : setsockopt - OUT (%m), rule: %d", num);
			exit(1);
		}
	}
	close(fw_sock);
}


/*
 * This Function deletes a Firewall rule(ipfw divert rule). 
 * 
 * All rules matching the firewall rule number are deleted.
 * 
 */
void delFireWallRule(int number)
{
	struct ip_fw fw; 
	int fw_sock;
	int inOrOut = IN;
	  
	bzero(&fw, sizeof (struct ip_fw));
	fw.fw_number = number;

	fw.fw_prot=1;

	/* open a socket */
	if ((fw_sock=socket(AF_INET, SOCK_RAW, IPPROTO_RAW))==-1) {
		syslog( LOG_ERR, " delFireWallRule() : socket (%m)");
		exit(1);
	}

	if (inOrOut == IN)
		fw.fw_flg |=  IP_FW_F_IN;	  
	else
		fw.fw_flg |=  IP_FW_F_OUT;	  

	if ((setsockopt(fw_sock, IPPROTO_IP, IP_FW_DEL, &fw
			, sizeof(struct ip_fw))) < 0 ) {

		if (inOrOut == IN) {	  	  
			fw.fw_flg &=  ~IP_FW_F_IN;		  
			syslog( LOG_ERR, " delFireWallRule() : setsocketopt - OUT (%m), rule %d",
				number);
			mip_log( MLOG_ERR, " delFireWallRule() : setsocketopt - OUT (%m), rule %d",
				number);
			exit(1);
		}
		else {
			fw.fw_flg &= ~IP_FW_F_OUT;		  
			syslog( LOG_ERR, " delFireWallRule() : setsocketopt - OUT (%m)");
			exit(1);
		}
	}
	close(fw_sock);
}
