/*
 * Copyright (c) 2003 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * divertMn.c
 *	module does reverse-tunneling needed at MN to
 *	deal with HA to MN NAT/UDP tunnels.
 *
 * functions:
 */

#include <stdio.h>
#include <errno.h>
#include <limits.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <netdb.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <net/if.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/ip_icmp.h>
#include <netinet/ip_fw.h>
#include <netinet/in_systm.h>
#include <math.h>

#include "mnstate.h"
#include "mnconfig.h"
#include "udp_reg.h"
#include "mip_log.h"
#include "divertInterfaces.h"

/*   Packets that show-up at this socket are either 1. MIP registration replys 
 *   or 2. Data packets that MN is getting from HA (in NAT/UDP tunnel from HA).
 *   
 *   If Data Packets and MIP Message Header is found, remove  
 *   outer IP-UDP-MIP_Message header and write packet back 
 *   to the divert socket to deliver to self.  
 *   Check the values of the MIP_MSGHDR_TYPE 
 *   and Next_header before decapsulation as specified by the RFC.
 *
 *   If Registration Reply Packet, write it back into the divert socket    
 *   without any changes.  
 *
 *   TBD:
 *	registration reply should be delivered to self, not fed
 *	back to kernel.
 */

void 
handleDivertIn(int fd)
{

	struct sockaddr_in sin;  
	int ii,n,sinlen;
	unsigned char packet[BUFSIZE];        
	unsigned char *decap_packet;
	extern int debug;
	struct vlist *hhead;
	struct ip *hdr;
	struct vlist *visitor;
	struct mip_MsgHdr *msgHdr;
	struct udphdr *udpHdr;
	int dataSize;
	
	sinlen = sizeof(struct sockaddr);
	
	n=recvfrom(fd, packet, BUFSIZE, 0
		   ,(struct sockaddr *) &sin, &sinlen);
	
	if (n < 0){
		syslog( LOG_ERR, "handleDivertIn(): recvfrom (%m)");		
		exit(1);
	}


	if (debug) {
		//printf("handleDivertIn \n");
		/*dumpme(packet, n);*/
	}
	
	hdr=(struct ip *)packet; 
	
	msgHdr = (struct mip_MsgHdr *) (packet + sizeof(struct ip) 
					+ sizeof(struct udphdr) );
	
	udpHdr = (struct udphdr *) (packet + sizeof(struct ip));
	
	/* if packet is data
	 */
	if (msgHdr->type == MIP_MSGHDR_TYPE && 
		msgHdr->nextHeader == IPVERSION ){

		if (debug){	  
			//printf("handleDivertIn - It is a Data Packet \n");
		}	  

		/* move the pointer ahead to point to the actual packet */
		decap_packet = packet + (sizeof(struct ip) 
					+ sizeof(struct udphdr)
					+ sizeof(struct mip_MsgHdr));

		/* Length of the decapsulated packet */
		dataSize  = n - ( sizeof(struct ip) 
					+ sizeof(struct udphdr)
					+ sizeof(struct mip_MsgHdr));
		
		/* Write it back into the Divert Socket */
		sin.sin_addr.s_addr = mn_config.mn.s_addr;
		while(1){
			n=sendto(fd,decap_packet, dataSize, 0
				 ,(struct sockaddr *) &sin, sizeof(sin));
		
			if (n < 0){
				if (errno == ENOBUFS)
					continue; 
				syslog( LOG_ERR, "handleDivertIn(): sendto (%m)");		  
				exit(1);
			}
			break;
		}
	}
	else {
		/* Make Sure that the IPSrc is HA and the destination port is UDP_PORT */ 
		if(hdr->ip_src.s_addr == mn_config.ha.s_addr && 
			ntohs(udpHdr->uh_dport) == UDP_PORT ){		   

			if (debug){	  		  
				//printf("handleDivertIn - this a Registration Reply \n");
			}
			
			/*  If the MIP Message Header is Not Found, then the packet
			 *  could be a registration reply. We write it back to the
			 *  divert socket,untouched.
			 */
			sin.sin_addr.s_addr = mns.coa.s_addr;
			while(1){
				n=sendto(fd,(void *) packet, n ,0
					 ,(struct sockaddr *) &sin, sizeof(sin));
			
				if (n < 0){
					if (errno == ENOBUFS)
						continue; 
					syslog( LOG_ERR, "handleDivertIn(): sendto (%m)");		  			  
					exit(1);
				}
				break;
			}
		}
	}
}

/*   This function handles the reverse IP-UDP-IP tunnel. 
 *   Packets are sent here for encapsulation from MN outbound to HA
 *   for COA. The MN's interface is bound to 2 IP addresses.
 *   
 *                1. co_located care-of-address
 *                2. Mobile Node's Home Address. 
 *   Steps.
 *                1. Read the packets leaving the interface.
 *                2. set the inner IP_src to Mobile Node's fixed MN ip addr.
 *                3. If inner IP datagram protocol field is tcp or udp 
 *                   change their checksums accordingly as their      
 *                   checksums depend on the ip_src and ip_dst and
 *                   we just changed the ip_dst. If protocol field
 *                   is ICMP, no problem with the checksum.
 *                4. Encapsulate the packet i.e attach
 *                   IP-UDP-MIP_Message Header and write it back 
 *                   into the divert socket.      
 *
 *  If Registration Request Packet, write it back into the divert socket    
 *  without any changes.   
 *
 * Bug:
 *	initial check on input ip datagram should be based on ip_proto field.
 *		if udp and mip/registration ...
 */
 
void 
handleDivertOut(int fd)
{

	struct sockaddr_in sin,mysocket;
	int ii,n,n1,sinlen;
        unsigned char packet[BUFSIZE];
	unsigned char *pack;
	extern int debug;
	struct ip *hdr ;
	struct udphdr *udphdr;
	struct tcphdr *tcphdr;
	struct udp_register *reg;
	unsigned short sport,dport;
	tcp_seq seq,ack;
	unsigned int th_off;
	unsigned int th_x2; 
	unsigned short th_win;
	unsigned short th_urp;
	unsigned short tcp_sum;
	unsigned char th_flags;
	struct in_addr saddr,daddr;
	int protocol;
	int dataSize;
	static int rawSock = -1;
	
	sinlen = sizeof(struct sockaddr);

	pack = (packet + sizeof(struct ip) + sizeof(struct udphdr) + sizeof(struct mip_MsgHdr) );

	n=recvfrom(fd, pack, BUFSIZE, 0
		   ,(struct sockaddr *) &sin, &sinlen);
	
	if (n < 0){
		syslog( LOG_ERR, "handleDivertOut(): recvfrom (%m)");	  
		exit(1);
	}

	if (debug) {
		//printf("handleDivertOut \n");
		/*dumpme(packet, n);*/
	}
	
	hdr=(struct ip *)packet; 

	((struct udphdr *)reg) = ((struct udp_reg *)(packet 
						     + sizeof(struct ip) 
						     + sizeof(struct udphdr)));
	
	
	if ( (  ((struct ip *)packet)->ip_p == IPPROTO_UDP) && (reg->type == UDP_REGISTER_TYPE) 
		){  	     	     

		if (debug) {		  	  
			//printf("handleDivertOut - It is a registration Packet \n");
		}	  

		/*  so it is a registration packet. Write it back to the   
		 *  into the divert socket , untouched.   
		 */
		sin.sin_addr.s_addr = INADDR_ANY;	  	  
		while(1){
			n1=sendto(fd,&packet, n,0
				  ,(struct sockaddr *) &sin, sizeof(sin));
		
			if (n1 < 0){
				if (errno == ENOBUFS)
					continue; 
				syslog( LOG_ERR, "handleDivertOut(): sendto (%m)");		  
				exit(1);
			}
			break;
		}
	} else {

		if (debug){	  
			//printf("handleDivertOut - It is a Data Packet \n");
		}

		/*  Attach an IP-UDP-MIP_Message Header */  
		encapsulate(packet,n,mns.coa,mn_config.ha,htons(UDP_PORT));		      

	        if (rawSock == -1) {
			rawSock = getRawSock();
		}
		
		dataSize = n + sizeof(struct ip) + sizeof(struct udphdr) 
		  + sizeof(struct mip_MsgHdr) ;
		
		/* Write it back out on the raw socket
		 * note divert does not work because we
		 * need to make sure ip_id is set.
		*/
		sin.sin_addr.s_addr = INADDR_ANY;

		while(1) {
			n1=sendto(rawSock,&packet, dataSize, 0,
				(struct sockaddr *) &sin, sizeof(sin));
		
			if (n1 < 0){
				if (errno == ENOBUFS)
					continue; 
				syslog( LOG_ERR, 
					"handleDivertOut(): sendto (%m)");
				exit(1);
			}
			break;
		}
	}
}

/*
 * This Function sets-up a Firewall rule(ipfw divert rule). 
 * This Rules should be set-up after the  first three default rules dealing with 
 * loopback interface packets. Hence the rule number 301.
 * 
 * srcAddr --- Source Address 
 * dstAddr --- Destination Address
 * inOrOut --- 1 for incoming and 2 for outgoing
 * port    --- DIVERTPORTIN or DIVERTPORTOUT
 * proto   --- protocol ( For E.g. UDP -- 17, icmp -- 1 , TCP -- 6 )
 */
void 
setupFirewallRule(struct in_addr srcAddr,struct in_addr dstAddr
		       ,int inOrOut, int port,int proto)
{
	struct ip_fw fw;         
	int fw_sock = -1;
	extern int debug;
	struct in_addr saddr;
	struct in_addr daddr;
	
	daddr = dstAddr;
	saddr = srcAddr;
	
        /* Specify the firewall rule and write                 	
	 *  it via a special raw socket 
	 *  at the appropriate section of the kernel 
	 */	   
	bzero(&fw, sizeof (struct ip_fw));
	
	fw.fw_dst.s_addr    =  daddr.s_addr; 
	if (daddr.s_addr == 0)
	  	fw.fw_dmsk.s_addr     = daddr.s_addr;
	else
	        fw.fw_dmsk.s_addr     = inet_addr("255.255.255.255"); 


	fw.fw_src.s_addr    =  saddr.s_addr;
	if (saddr.s_addr == 0)
	         fw.fw_smsk.s_addr     = saddr.s_addr;
	else
	         fw.fw_smsk.s_addr     = inet_addr("255.255.255.255"); 
	
	fw.fw_flg                |= IP_FW_F_DIVERT;
	fw.fw_prot               = proto;
	fw.fw_un.fu_divert_port  = port;	
	fw.fw_number = mn_config.fw_rule;

	if (inOrOut == IN)
		fw.fw_flg |=  IP_FW_F_IN;	  
	else
		fw.fw_flg |=  IP_FW_F_OUT;	  
	
	/* open a socket */
	if ((fw_sock=socket(AF_INET, SOCK_RAW, IPPROTO_RAW))==-1) {
		syslog( LOG_ERR, "setupFireWallRule() : socket (%m)");
		exit(1);
	}
	
	if ((setsockopt(fw_sock, IPPROTO_IP, IP_FW_ADD, &fw
			, sizeof(struct ip_fw))) < 0 ){

		if (inOrOut == IN){	  	  
			fw.fw_flg &=  ~IP_FW_F_IN;		  
			syslog( LOG_ERR, "setupFireWallRule() : setsocketopt - IN (%m)");
			exit(1);
		}
		else{
			fw.fw_flg &= ~IP_FW_F_OUT;		  
			syslog( LOG_ERR, "setupFireWallRule() : setsocketopt - OUT (%m)");
			exit(1);
		}

	}
	close(fw_sock);

}

/*
 *   setup a IP-UDP tunnel.
 *
 */
void 
setupDivertTunnel()
{

	struct in_addr dst_addr;			    
	struct in_addr src_addr;
	
	dst_addr.s_addr = 0;
	

	/* send UDP packets from the home agent to DIVERTPORTIN */   
	setupFirewallRule(mn_config.ha,dst_addr
			  ,IN,DIVERTPORTIN,17);
  
	/* send IP packets, from MN and Havin Ipsrc == MN's Home Address , to DIVERTPORTOUT */
	setupFirewallRule(mn_config.mn,dst_addr
			  ,OUT,DIVERTPORTOUT,0);

	mns.fr_number = mn_config.fw_rule;

	mip_log(MLOG_INFO, "NAT mode: ipfw rules installed\n");
}

/*
 * This Function deletes a Firewall rule(ipfw divert rule). 
 *
 * This Rules should be the number after the  first three default 
 * rules dealing with loopback interface packets. Hence the rule 
 * number 301.
 * 
 * All the rules matching the firewall number are deleted.
 * 
 */
void 
delFireWallRule()
{
	struct ip_fw fw;           
	int fw_sock;
	int inOrOut = IN;
	
	bzero(&fw, sizeof (struct ip_fw));
	fw.fw_number = mn_config.fw_rule;
	
	fw.fw_prot=1;
	
	/* open a socket */
	if ((fw_sock=socket(AF_INET, SOCK_RAW, IPPROTO_RAW))==-1) {
		syslog( LOG_ERR, " delFireWallRule() : socket (%m)");
		exit(1);
	}
	
	if (inOrOut == IN)
		fw.fw_flg |=  IP_FW_F_IN;	  
	else
		fw.fw_flg |=  IP_FW_F_OUT;	  
	
	
	if ((setsockopt(fw_sock, IPPROTO_IP, IP_FW_DEL, &fw
			, sizeof(struct ip_fw))) < 0 ){

		if (inOrOut == IN){	  	  
			fw.fw_flg &=  ~IP_FW_F_IN;		  
			syslog( LOG_ERR, " delFireWallRule() : setsocketopt - OUT (%m)");
			exit(1);
		}
		else{
			fw.fw_flg &= ~IP_FW_F_OUT;		  
			syslog( LOG_ERR, " delFireWallRule() : setsocketopt - OUT (%m)");
			exit(1);
		}

	}
	close(fw_sock);
	mns.fr_number = 0;	/* reference count */
}

