
/*
 * bpf input to icmp and general write capability which
 *	may be used by any entity interested in low-level pkt emissions
 *
 * Two functions at this point:
 *	1.  In ARP mode, we read icmp router discovery packets (in mnd),
 *	this allows us to learn FA's mac addresses, as we may not arp for them.
 *	In ad hoc mode, they tell us their mac address in their ICMP enhanced
 *	beacon, so we don't use the bpf there.
 *	2.  we send dhcp "broadcast" packets to 255.255.255.255, using
 *	packets written on the bpf.  This allows us to avoid the
 *	"writing to all 1's" problem with BSD sockets.  It is also kinky.
 */

/*
 * Copyright (c) 2001 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/time.h>
#include <sys/filio.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <net/bpf.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <netinet/ip_icmp.h>
#include <netinet/ip_var.h>
#include <netinet/udp_var.h>
#include <netinet/if_ether.h>
#include <net/if_llc.h>

static int bpf_bufsize;
#define BPFBUFSIZE 32768
static char bpf_buf[BPFBUFSIZE];

/* type of interface */
static int if_op;

/* Link-level header length */
#ifndef ETHER_HDRLEN
#define ETHER_HDRLEN    sizeof(struct ether_header)
#endif

static int bpf_open(char *errbuf);

/* cheerfully borrowed from dynamix linux code
 * bpf filter to get rid of everything except icmp type 9
 * packets
*/
/* ethhdr - 14 octets, iphdr - 20 octets */
static
struct bpf_insn icmp_filter[] = {
        /* A <- P[23:1] - load byte - IP protocol */
        BPF_STMT(BPF_LD+BPF_B+BPF_ABS, 14 + 9),
        /* pc += (A == IPPROTO_ICMP) ? 0 : 3 */
        BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_K, IPPROTO_ICMP, 0, 3),
        /* A <- P[34:1] - load byte - ICMP type */
        BPF_STMT(BPF_LD+BPF_B+BPF_ABS, 14 + 20),
        /* pc += (A == 9) ? 0 : 1 (agentadv) */
        BPF_JUMP(BPF_JMP+BPF_JEQ+BPF_K, 9, 0, 1),
        BPF_STMT(BPF_RET+BPF_K, (u_int)-1), /* accept the packet */
        BPF_STMT(BPF_RET+BPF_K, 0) /* filter packet */
};

static
int icmp_filter_len = sizeof icmp_filter / sizeof (struct bpf_insn);

static int bpf_fd=-1;

/*
 * init_bpf     
 *	char *device_name, the name of the ifconfig mobile device
 *	char *ebuf - store any errors as strings here please.
 *		(syslog calls)
 *
 * filter on "icmp" type 9 only.
 * returns: fd or -1 on error
 */
int
init_bpf(char *device, char *ebuf, int need_read)
{
	int rc;
	int fd;
	int v = BPFBUFSIZE;
 	struct ifreq ifr;
	struct bpf_program p;

	p.bf_len = icmp_filter_len;
	p.bf_insns = icmp_filter;

	fd = bpf_open(ebuf);
	if (fd < 0)
		return(-1);

	bpf_fd = fd;

	/* set the buffer size, must match read size
	*/
        (void) ioctl(fd, BIOCSBLEN, (caddr_t)&v);

	/* set the network device
	*/
        (void)strncpy(ifr.ifr_name, device, sizeof(ifr.ifr_name));
        if (ioctl(fd, BIOCSETIF, (caddr_t)&ifr) < 0) {
                sprintf(ebuf, "%s: %s", device, strerror(errno));
		close(fd);
		return(-1);
        }

	if (!need_read) {
		shutdown(fd, 0);	/* shut down read */
		return(fd);
	}

        /* Get the data link layer type. */
        if (ioctl(fd, BIOCGDLT, (caddr_t)&v) < 0) {
                sprintf(ebuf, "BIOCGDLT: %s", strerror(errno));
		close(fd);
		return(-1);
        }
	if_op = v;

	/* turn on immediate mode so we don't hang in reads
	 * and get 1 packet at a time. overkill on that there buffer,
	 * but oh well (size is too big).
	*/
	v = 1;
	rc = ioctl(fd, BIOCIMMEDIATE, &v);
	if (rc < 0) {
		sprintf(ebuf, "biocimmediate: %s", strerror(errno));
		close(fd);
		return(-1);
	}
	rc = ioctl(fd, BIOCGBLEN, &bpf_bufsize);
	if (rc < 0) {
		sprintf(ebuf, "biocgblen: %s", strerror(errno));
		close(fd);
		return(-1);
	}
	/* install the local packet filter for reads
	*/
	rc = ioctl(fd, BIOCSETF, &p);
	if (rc < 0) {
		sprintf(ebuf, "biocsetf: %s", strerror(errno));
		close(fd);
		return(-1);
	}
	return(fd);
}


/*
 * set the device for bpf output.  This is only used by
 * agents, which do not use the bpf to read packets.
 */
aim_bpf(char *device, char *ebuf)
{
	int rc;
 	struct ifreq ifr;
	int fd;
	struct bpf_program p;

	fd = bpf_fd;

	/* set the network device
	*/
        (void)strncpy(ifr.ifr_name, device, sizeof(ifr.ifr_name));
        if (ioctl(fd, BIOCSETIF, (caddr_t)&ifr) < 0) {
                sprintf(ebuf, "%s: %s", device, strerror(errno));
		close(fd);
		return(-1);
        }
	return(0);
} 

/*
 * bpf_open
 *	find a bpf device
 *	open rd/wr
 *
 * note: opened read/write !!!
 *
 * returns:
 *	valid fd or -1, plus str errno in errbuf
 */ 

static 
int
bpf_open(char *errbuf)
{
        int fd;
        int n = 0;
	extern int errno;
        char device[sizeof "/dev/bpf000"];
 
        /*
         * Go through all the minors and find one that isn't in use.
         */
        do {
                (void)sprintf(device, "/dev/bpf%d", n++);
                fd = open(device, O_RDWR);
        } while (fd < 0 && errno == EBUSY);

        /*
         * XXX better message for all minors used
         */
        if (fd < 0)
                sprintf(errbuf, "%s: %s", device, strerror(errno));

        return (fd);
}


bpf_close()
{
	close(bpf_fd);
}

/*
 * return TRUE if ethernet device, else
 * return FALSE
 */
bpf_needed()
{
	if (if_op == DLT_EN10MB || if_op == DLT_IEEE802)  
		return(1);
	return(0);
}

#ifdef TESTMAIN
int
bpf_select(char *buf, int size)
{
	int fd1;
	int rc;
	int notdone = 1;
        struct timeval timeout = {0, 0};
        fd_set ready;
	int on = 0;	/* on is byte count to read */

	fd1 = bpf_fd;

#ifdef USEFIONREAD
	ioctl(fd1, FIONREAD, &on);
	if (on) {
		rc=bpf_read(fd1, buf, size);
		return(rc);
	}
	else {
		return(0);
	}
#else /* select, but not working -- donnah why */

again:
	FD_ZERO(&ready);
       	FD_SET(fd1, &ready);
	rc = select(100, &ready, NULL, NULL, NULL);

	/* alarm will cause this to spin
	*/
	if (rc < 0) {
		/*
		perror("select");
		notdone = 0;
		*/
		goto again;
	}
	if (FD_ISSET (fd1, &ready)) {
		rc=bpf_read(fd1, buf, size);
	}
	return(rc);
#endif
}
#endif

/*
 * note:
 *	only call this with bpf_buf (static global above)
 *	and size set to bpf_bufsize (static global above)
 */
int
bpf_read(int fd, char *buf, int size)
{
	int cc;
	extern int errno;

 	cc = read(fd, buf, size);
        if (cc < 0) {
        /* Don't choke when we get ptraced */
        	switch (errno) {
                case EINTR:
			return (-1);

                case EWOULDBLOCK:
                        return (-1);
		}
	}
	return(cc);
}

/*
 * return directly: pointer to remainder of buffer (ICMP portion)
 * inputs:
 *	buf buffer read from bpf
 *	insize - size of that buffer/packet
 * outputs:
 *	outsize - size LEFT after we skip numerous things
 *	macflag - 1 if valid MAC seen
 *	mac - 6 byte mac address for MAC source
 *	iplen - length of this ip packet
 */
char *
bpf_process_read(char *buf, int insize, int *outsize, int *macflag, char *mac, int *iplen)
{
	char *bp = buf;
	int hdrlen;
	int phdrlen = 0;
	struct ether_header *eh;
	unsigned short type;
	struct llc *l;
	struct ip *ipp;

	*macflag = 0;		/* assume no */

	/* 1.  skip bpf header
	*/
#define bhp ((struct bpf_hdr *)bp)
	/*
	printf("bh_caplen bh_datalen %d %d\n", bhp->bh_caplen, bhp->bh_datalen);
	*/
	hdrlen = bhp->bh_hdrlen;
	bp += hdrlen;
	insize -= hdrlen;

	if (if_op != DLT_EN10MB)
		return(bp);

	eh = (struct ether_header *) bp;

	type = ntohs(eh->ether_type);
	/* TBD. best do something about this
	 * LLC madness.
	*/
	insize -= ETHER_HDRLEN;
	bp += ETHER_HDRLEN;

	/* sanity check 802.3 header MORE
	*/
	if (type <= 1500) {
		l = (struct llc *) bp;
		insize -= sizeof (struct llc);
		bp += sizeof(struct llc);
	}

	memcpy(mac, eh->ether_shost, 6);
	*macflag = 1;

	/* get length of packet
	*/ 
	ipp = (struct ip *) bp;
	*iplen =  (int) ntohs(ipp->ip_len);
	insize -= *iplen;

	*outsize = insize;	
	
	return(bp);
}

#ifndef TESTMAIN
handleBpfInput(int sd)
{
	int rc;
	char *bp;
	int macflag;
	char mac[6];
	int outsize;
	int iplen;
	char *p;		/* pointer to pkt in trimmed bpf buffer */

	rc=bpf_read(sd, bpf_buf, bpf_bufsize);
	if (rc < 0) {
		/* tbd. log this
		*/
		return;
	}
	/* may have multiple input packets
	*/
	/* p is pointer to ip packet
	 * p + iplen is pointer to next bpf packet if it exists
	 * outsize is remainder in bytes
	 * iplen is length of ip packet
	 */
	bp = bpf_buf;
	while (1) {
		p = bpf_process_read(bp, rc, &outsize,
			&macflag, &mac[0], &iplen);
		handleIcmpInput(p, iplen, macflag, mac);
		if (outsize <= 0)
			break;
		bp = bp + (rc - outsize);
		rc = outsize;
	}
}
#endif

/*
 * append mac/ip/udp headers and issue packet on current bpf interface
 * note: bpf_fd found internally for reasons of convenience.
 *
 * DHCP broadcast is the current consumer of this function.
 *
 * note: if macdst is NULL, we use limited broadcast (all ffs).
 * note: ipsrc, ipdst assumed in network byte order on input.
 * 	 srcport/dstport assumed to NOT be in network byte order.
 *
 * function currently uses writev, but doesn't need it.
 * Some code ideas borrowed from ISC dhcp client, but in the main
 * based on BSD kernel code.
 * 
 */
writebpf_udp(char *packet, int length, 
		char *macsrc, char *macdst,
		struct in_addr ipsrc, struct in_addr ipdst, 
		int srcport, int dstport)
{
	int bufoffset=0;
	unsigned char buf[1514];
	struct iovec iov[2];
	int rc; 
	extern int debug;
	struct ether_header *eh;
	struct udpiphdr *ui;

	/* copy payload into same packet buffer, skipping layers 2/3/4
	*/
	char *pkt = (char *) &buf[0];
	pkt = pkt + sizeof(struct ether_header) + sizeof(struct udpiphdr);	
	bcopy(packet, pkt, length);	/* copy data part into pkt part */

	/* build the ethernet header first.
	*/
	eh = (struct ether_header *)&buf[0];
	memcpy(eh->ether_shost, macsrc, 6);
	if (macdst == NULL) {
		memset(eh->ether_dhost, 0xff, 6);
	}
	else {
		memcpy(eh->ether_dhost, macsrc, 6);
	}
	eh->ether_type = htons(ETHERTYPE_IP);
	bufoffset = ETHER_HDRLEN;	/* length so far */

        /*
         * Fill in mbuf with extended UDP header
         * and addresses and length put into network format.
         */
        ui = (struct udpiphdr *) &buf[bufoffset];
        bzero(ui->ui_x1, sizeof(ui->ui_x1));	/* zero out-non csum 1st 9 ip bytes */
        ui->ui_pr = IPPROTO_UDP;
        ui->ui_len = htons((u_short)length + sizeof (struct udphdr));
        ui->ui_src = ipsrc;	/* already byte-swapped */
        ui->ui_dst = ipdst;	/* already byte-swapped */

        ui->ui_sport = htons(srcport);
        ui->ui_dport = htons(dstport);
        ui->ui_ulen = ui->ui_len;

        /*
         * do checksum
         */
        ui->ui_sum = 0;
        if ((ui->ui_sum = in_cksum(ui, sizeof (struct udpiphdr) + length)) == 0)
                ui->ui_sum = 0xffff;

	/* udp header is finished, ip header needs work
	*/

        ((struct ip *)ui)->ip_v = 4;
        ((struct ip *)ui)->ip_hl = 5;
        ((struct ip *)ui)->ip_tos = IPTOS_LOWDELAY;
        ((struct ip *)ui)->ip_len = htons(sizeof (struct udpiphdr) + length);
        ((struct ip *)ui)->ip_id = 0;
        ((struct ip *)ui)->ip_off = 0;
        ((struct ip *)ui)->ip_ttl = 16;
	/* one more cksum calculation on ip hdr this time
	*/
        ((struct ip *)ui)->ip_sum = 0;
        ((struct ip *)ui)->ip_sum = in_cksum ((struct ip *)ui, sizeof(struct ip));

	bufoffset += sizeof(struct udpiphdr);
	bufoffset += length;	/* now we have total length */
/*
	if (debug) {
		printf("bpf write, total length %d\n", bufoffset);
	}
*/

	iov[0].iov_base = buf;
	iov[0].iov_len = bufoffset;
	rc = writev(bpf_fd, iov, 1);
/*
	if (debug) {
		printf("bpf write returns %d\n", rc);
	}
*/
	return(rc);
}

/*
 * append mac/ip/udp headers and issue packet on current bpf interface
 * note: bpf_fd found internally for reasons of convenience.
 *
 * writing to ICMP 255.255.255.255 is the reason for this routine.
 *
 * note: if macdst is NULL, we use limited broadcast (all ffs).
 * note: ipsrc, ipdst assumed in network byte order on input.
 * 	 srcport/dstport assumed to NOT be in network byte order.
 *
 * function currently uses writev, but doesn't need it.
 * Some code ideas borrowed from ISC dhcp client, but in the main
 * based on BSD kernel code.
 * 
 */
writebpf_icmp(char *packet, int length, 
		char *macsrc, char *macdst,
		struct in_addr ipsrc, struct in_addr ipdst)  
{
	int bufoffset=0;
	unsigned char buf[1514];
	struct iovec iov[2];
	int rc; 
	extern int debug;
	struct ether_header *eh;
	struct ip *ip;

	/* build the ethernet header first.
	*/
	eh = (struct ether_header *)&buf[0];
	memcpy(eh->ether_shost, macsrc, 6);
	if (macdst == NULL) {
		memset(eh->ether_dhost, 0xff, 6);
	}
	else {
		memcpy(eh->ether_dhost, macsrc, 6);
	}
	eh->ether_type = htons(ETHERTYPE_IP);
	bufoffset = ETHER_HDRLEN;	/* length so far */

	ip = (struct ip *) &buf[bufoffset];

        ip->ip_v = 4;
        ip->ip_hl = 5;
        ip->ip_tos = 0;
        ip->ip_len = htons(sizeof (struct ip) + length);
        ip->ip_id = 0;
        ip->ip_off = 0;
        ip->ip_ttl = 1;
	ip->ip_p = IPPROTO_ICMP;
        ip->ip_src = ipsrc;
        ip->ip_dst = ipdst;

	/* one more cksum calculation on ip hdr this time
	*/
        ip->ip_sum = 0;
        ip->ip_sum = in_cksum ((struct ip *)ip, sizeof(struct ip));

	bufoffset += sizeof(struct ip);

/*
	if (debug) {
		printf("bpf write, total length %d\n", bufoffset);
	}
*/

	iov[0].iov_base = buf;
	iov[0].iov_len = bufoffset;
	iov[1].iov_base = packet;
	iov[1].iov_len = length;

	rc = writev(bpf_fd, iov, 2);
/*
	if (debug) {
		printf("bpf write returns %d\n", rc);
	}
*/
	return(rc);
}

#ifdef TESTMAIN
main(int argc, char **argv)
{

	char *dev = argv[1];
	char *expr = "icmp";
	char errbuf[2048];
	int fd;
	char buf[BPFBUFSIZE];
	int rc;
	char *bp;
	int macflag;
	char mac[6];
	int outsize;
	int iplen;
	char *p;		/* pointer to pkt in trimmed bpf buffer */

	fd = init_bpf(dev, "icmp[0] == 9", errbuf);
	if (fd < 0) {
		printf("errbuf %s\n", errbuf);
	}

	/* just to test ... ignore rc
	*/
	rc = bpf_needed();

	/* select and read a packet, we hope
	*/
	while(1) {
		rc = bpf_select(bpf_buf, bpf_bufsize);
		/* p is pointer to ip packet
	         * p + iplen is pointer to next bpf packet if it exists
		 * outsize is remainder in bytes
		 * iplen is length of ip packet
		 */
		bp = bpf_buf;
		while (1) {
			p = bpf_process_read(bp, rc, &outsize,
				&macflag, &mac[0], &iplen);
			if (outsize <= 0)
				break;
			bp = bp + (rc - outsize);
			rc = outsize;
		}
	}
}
#endif

