/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * icmpinit.c
 *
 * initialization for handling icmp beacon
 */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/signal.h>
#include <termios.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/ip_var.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>

#include "mip_icmp.h"	/* icmp router advert + appropriate mse extension/s */
#include "mnstate.h"
#include "mnstats.h"
#include "unixsock.h"
#include "mip_log.h"
#include "mndversion.h"

#define TRUE	1
#define FALSE	0

struct mn_stats mn_stats;

extern int debug;

static void sendMulticast(int);
static void recvMulticast(int, struct in_addr, struct in_addr);

int beaconSd;

/*
 * initIcmp
 *
 * open raw icmp socket so that we can send and recv router adverts
 * on it.  should be able to send solicitations too.
 *
 * Note: if and when we can send multicast (224.0.0.1) as well as
 * limited broadcast, we can't do both on the same socket descriptor.
 * The code in netinet/ip_output.c won't support that.
 * 
 * returns:
 *	int beaconSd - socket descriptor
 */ 
int
initIcmp(sendBcast, sendMcast, maddr, mif)
int sendBcast;	/* want to send broadcast only */
int sendMcast;   /* want to send multicast only */
struct in_addr maddr;  /* multicast address for output */
struct in_addr mif;  /* if non-zero, a default multicast ip interface */
{
	struct protoent *proto;
	int on=1;
	int rc;

	if (!(proto = getprotobyname("icmp"))) {
 		syslog(LOG_ERR, "initIcmp(): unknown protocol icmp (%m).");
		exit(1);
	}
	/* icmp raw socket for sending out ICMP adverts
	*/
	if ((beaconSd = socket(AF_INET, SOCK_RAW, proto->p_proto)) < 0) {
	  	syslog(LOG_ERR, "initIcmp(): socket (%m)");
		exit(1);
	}
	
	/* Set close-on-exec flag to prevent fd leaks. */
	if (fcntl(beaconSd, F_SETFD, 1) < 0) {
		syslog(LOG_INFO, "initIcmp(): F_SETFD (%m)");
		/* not catastrophic, don't exit */
	}

	/* if no multicast addr set, we still want to enable
	 * default reception of multicast
	 */
	if (sendMcast == FALSE) {
		maddr.s_addr = inet_addr("224.0.0.1");
	}
	recvMulticast(beaconSd, maddr, mif);

	/* configure send side of socket, either broadcast or multicast
	*/
	if (sendBcast)  {
		/*
		sendBroadcast(beaconSd);
		*/
	}
	else if (sendMcast) {
		sendMulticast(beaconSd);
	}
	else {
		/* has to be one or the other */
		syslog(LOG_ERR, "internal icmp configuration error");
		exit(1);
	}

	/* configure socket to receive incoming i/f index
	 * which allows us to determine which i/f a packet came in on
	 */
	rc = setsockopt(beaconSd, IPPROTO_IP, IP_RECVIF, (char *)&on, sizeof(on));
	if (debug) {
		printf("setsockopt on beaconSd %d, IP_RECVF flag set\n", 
			beaconSd);
	}
	if ( rc < 0 ) {
		syslog( LOG_ERR, "initIcmp(): setsockopt(IP_RECVIF) (%m)");
		exit(1);
	}
	return(beaconSd);
}

#ifdef NOTUSED
/*
 * sendBroadcast - old routine.  we now use the bpf to send
 * 255.255.255.255 broadcast packets on arbitrary interfaces.
 *
 * done when we want to send 255.255.255.255 icmp router messages
 * on any interface.  We set ROUTETOIF so that the destination
 * ip address used with sendto will determine the interface.
 * We set BROADCAST because we want to broadcast.  We set
 * LIMITBCAST, so that our ip dst will be munged into 255.255.255.255
 * after ROUTETOIF has determined the interface.
 * ROUTETOIF hardwires the ttl to 1.  LIMITBCAST was a PSU
 * hack, and is now no longer in the os.
 */
static void
sendBroadcast(int beaconSd)
{
	int on;
	int rc;

	/* set ROUTETOIF for this socket.  packets sent to i/f ip
	 * address are not routed but delivered directly
	 */
  	on = 1;
        rc = setsockopt(beaconSd, SOL_SOCKET, SO_DONTROUTE, (char *)&on,
                    sizeof(on));
	if ( rc < 0 ) {
 		syslog( LOG_ERR, "initIcmp(): setsockopt(SO_DONTROUTE) (%m)");
		exit(1);
	}
	/* so we can broadcast on this socket
	*/
        rc = setsockopt(beaconSd, SOL_SOCKET, SO_BROADCAST, (char *)&on,
                    sizeof(on));
	if ( rc < 0 ) {
		syslog( LOG_ERR, "initIcmp(): setsockopt(SO_BROADCAST) (%m)");
		exit(1);
	}

	/* so our broadcasts will be limited broadcast 255.255.255.255
	 * as specified by mobile-ip (there is a kernel mod associated 
	 * with this, it only affects "route to interface" code)
	 */
        rc = setsockopt(beaconSd, SOL_SOCKET, SO_LIMITBCAST, (char *)&on,
                    sizeof(on));
	if ( rc < 0 ) {
		syslog( LOG_ERR, "initIcmp(): setsockopt(SO_LIMITBCAST) (%m)");
		exit(1);
	}
}
#endif

/*
 * recvMulticast 
 *	1. merely enables the reception of
 * 	multicast addresses on the MIP default multicast port.
 * 	This allows us to recv icmp advert/solicitations from
 * 	hosts using multicast.
 *
 *	2. disables loopback of multicast packets
 *
 * Note: this code does NOT cause a fatal exit if the attempt
 * to add a multicast group on a particular interface fails.
 * It is possible that a given i/f might not support multicast.  
 * A message will be logged however.
 */
static void
recvMulticast(int beaconSd, struct in_addr maddr, struct in_addr mif)
{
	struct ip_mreq mreq;
	unsigned char loop = 0;

 	mreq.imr_multiaddr.s_addr = maddr.s_addr;
	if (mif.s_addr)
		mreq.imr_interface.s_addr = mif.s_addr;
	else
		mreq.imr_interface.s_addr = htonl(INADDR_ANY);

        if (setsockopt(beaconSd, IPPROTO_IP, 
		IP_ADD_MEMBERSHIP, &mreq, sizeof(mreq)) < 0){
 		syslog(LOG_ERR, "initIcmp(): IP_ADD_MEMBERSHIP failed (%m).");
		return;
	}
        if (setsockopt(beaconSd, IPPROTO_IP, 
		IP_MULTICAST_LOOP, &loop, sizeof(loop))) {
 		syslog(LOG_ERR, "initIcmp(): IP_MULTICAST_LOOP failed (%m).");
		exit(1);
	}
}

/*
 * sendMulticast
 *
 * setup send side.  recvMulticast already joined the mcast group.
 * We set the TTL to 1.  Setting the i/f either is only done
 * in the recv routine BECAUSE we only support one interface
 * for mns (mnd).  mipd has to explicitly set the output i/f
 * for each interface a multicast beaocn is sent out on.
 */
static void
sendMulticast(int beaconSd)
{
	unsigned char ttl = 1;

 	if (setsockopt(beaconSd, IPPROTO_IP, 
		IP_MULTICAST_TTL, &ttl, sizeof(ttl)) < 0) {
 		syslog(LOG_ERR, "initIcmp(): IP_MULTICAST_TTL failed (%m).");
		exit(1);
	}
}

/*
 * aim multicast socket at a given interface
 */
setMcastIf(struct in_addr addr)
{
	int rc;
     	rc = setsockopt(beaconSd, IPPROTO_IP, IP_MULTICAST_IF, &addr, sizeof(addr));
	if (rc < 0) {
 		syslog(LOG_ERR, "setMcastIF(): IP_MULTICAST_IF failed (%m).");
		exit(1);
	}
}
