/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * icmprecv.c
 *
 * handle icmp input
 */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/signal.h>
#include <termios.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/ip_var.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>

#include "mip_icmp.h"	/* icmp router advert + appropriate mse extension/s */
#include "mip_log.h"

#include "mnstate.h"
#include "unixsock.h"
#include "mndversion.h"
#include "alist.h"
#ifdef MND
#include "mnconfig.h"
#endif
#include "icmpstats.h"
#ifdef MADRP
#include "sr_mad.h"
struct SR_hlist * srhlist_search( struct in_addr mn);
#ifdef MIPD
struct vlist * vlist_search(struct vlist *, struct in_addr);
struct vlist * vlist_getfahead();
struct vlist * vlist_gethahead();
#else 
struct alist * alist_search(struct alist *, struct in_addr);
struct alist * alist_getmnhead();
extern struct alist * sr_hagentlist;
extern int srhagentcount;
#endif
#endif

extern int debug;

/* prototypes
*/

static int
icmpSecurityCheck(char *, struct mip_mse_icmp_auth *);

void stateIcmp( struct in_addr, unsigned short, unsigned short, int,
	char *, int, char *, int, int, char *, struct in_addr, unsigned long,
	unsigned long);

int 	getNetauthSecurity();

int 	getAdhocSecurity();

char    *getNetauthKey();

char    *getAdhocKey();

int     getNetauthSpi();

int     getAdhocSpi();

int     getConfigArp(struct in_addr  *ifaddr);

int	indextoif(int);

int recvMsg(int , char *, int , struct sockaddr *, int *, int *);

static int beaconSd;

/*
 * handleIcmpAdvert
 *
 * function reads, parses, and sets up icmp router advert info
 * for both mnd/mipd state machines. 
 *
 * roughly:
 *
 *	read the packet - note that we can get an icmp packet
 *		from any interface (including any we send)
 *		and we will get ALL of them.  raw sockets are not
 *		terribly discriminating.
 *
 *	we get ip header, form pointer to it
 *	check header len for min icmp, if too short, toss
 *	
 *	we know we appended ip address of incoming i/f at end of
 *	packet, else kernel tosses packet.  store it.
 *
 *	various sanity checks:
 *		if packet from SELF, toss it.
 *		form pointer to ICMP part (router advert we hope)
 *
 *		if not router advert, toss it.
 *
 *		query our config structure to see two things:
 *			1. if packet not sent to one of our mobile i/fs
 *			(in config structure), toss it.
 *			2. if packet IS to i/f, determine if arp on/off
 *			on per interface basis.
 *		if we are MIPD
 *			and arp is on, toss beacon, agents don't care
 *			about beacons.
 *
 *		make size checks on advert part
 *
 *	form pointer to MSE
 *
 *	if not MIPAGENTAD
 *		toss
 *
 * 	figure out if HA/FA or our MN bit   
 *
 *	if FA
 *		check advert size again since bigger
 *		if MND 
 *			check to see if FA address matches our HA,
 *			if so, treat this as HA beacon
 *			store COA
 *			set sawFa
 *			set rbit if on in flags
 *		else if MIPD
 *			toss as agents don't care about FA beacons
 *	if HA
 *		if MND
 *			set sawHa
 *		else if MIPD
 *			toss as agents don't care about HA beacons either
 *	if MN
 *		if arp is on, ignore this beacon 
 *		set sawMn
 *
 *	if both fa/ha bits set
 *		check if it's actually our HA in disguise, if not
 *		interpret the beacon as from a FA.
 *
 *	if no ha/fa/mn bits set
 *		really wierd error
 *
 *	store/byte-swap sequence and lifetime
 *
 * 	now walk any remaining TLVs including PSU extra TLVs
 *
 *	loop until end of bytes
 *		case 0: pad
 *		case MAC:
 *			store MAC
 *		case MIPAGENTAUTH:
 *			size check.
 *			zero out ICMP checksum because that is how
 *				this auth was done.
 *			form pointer to this tlv structure.
 *			if MIPD
 *				if agent field is 0 or this interface,
 *					we must process this,
 *
 *				else we can ignore it as not for us
 * 
 *			if NETAUTH type of authentication
 *				do auth hash, note if OK or not
 *				if OK
 *					set bit in auth_type for NETAUTH
 *					store timestamp
 *					no need to continue, we are done
 *				else
 *					increment auth failure counters
 *			if ADHOCAUTH (optionally follows NETAUTH 
 *				processing same as NETAUTH
 *		case MIPAUTHINFO:
 *			store location string
 *		case PREFIX:
 *			ignore
 *		advance acc. to TLV size
 *
 *	DONE with TLVs
 *
 *	if beacon lacks authentication and arp is not being used, error
 *		increment icmp counters acc to type ha/fa/mn
 *
 *	if we are authenticated (auth_type set) and no MAC address
 *		somebody goofed, log and ignore
 *
 *	complete debug info if -d set
 *
 *	if auth ok
 *		increment valid auth counter
 *	else
 *		increment valid arp/beacon counter
 *
 *	if MN
 *		checkAlist - beacon goes to alist mechanism which
 *			will either put in agent list or mn list
 *
 *	if MIPD
 *		if arp in use, beacons should be tossed by here.
 *		if authentication in use, pass to 
 *		stateIcmp
 *
 * whew...
 */

/*
 * mnd only.  we already have the packet, and likely a mac to
 * go with it (using the bpf as input in arp mode)
 */
handleIcmpInput(char *p, int iplen, int macflag, char *mac)
{
	int icmpSd = -1;		/* not used */
	handleIcmpAdvert(icmpSd, p, iplen, macflag, mac);
}

/*
 * icmpSd is either -1, meaning packet is passed to this function,
 *	else >=0, meaning packet is read by this function
 */
handleIcmpAdvert(int icmpSd, char *p, int iplen, int macflag, char *mac)
{
	register struct icmp_advert *icp;
        register struct mip_mseicmp *mse;
	struct mip_mse_icmp_auth *mseauth;
	struct mip_mse_icmp_mac *msemac;
	char location[257];
	char url[257];
	int noaddrs = 0;
	struct ip *ipp;
        int rc = -1;
	int auth_ok = 0;	/* we got auth key, and auth passed */
	int auth_type = 0;	
	unsigned long auth_tssecs = 0;
	unsigned long auth_tsfract = 0;
        int cc;
	int total;
        extern int debug;
        u_short flags = 0;
	struct sockaddr_in from;
	int fromlen;
	u_char packet[MAXICMPPACKET];
	char *ppt;
	char *dumpppt;
	/* info extracted from MSE
	*/
	int sawha = 0;
	int sawfa = 0;
	int sawmn = 0;
	int useha = 0;	/* TRUE if recv'd beacon is used for HA data */
	int rbit = 0;
	int faisha = 0;   /* TRUE in case FA is actually our HA */
	u_short mse_sequence;
	u_short mse_lifetime;
	char mac_addr[MACMSIZ];
	int sawMac = 0;
	int advert_length;
	int hlen;	/* ip header length */
	int i;
	struct in_addr mse_coa;
	struct in_addr ip_src;
	int size, lsize;
	int usize;
	char *l = NULL;
	char *u = NULL;
	int doArp;
	struct in_addr ifaddr;	
	int ifindex;
	int netauthbad = 0;
	int adhocauthbad = 0;
#ifdef MADRP            
        struct icmp_unreachable * icu;
        struct SR_hlist * ss;
        struct in_addr nexthop;
        struct vlist * vlisthead=NULL, *thisvlist=NULL;
#endif

	lsize = 0;
	usize = 0;
	mse_coa.s_addr = 0;
	ip_src.s_addr = 0;
	fromlen = sizeof(from);

	ppt = (char *)packet;
	/* if non-bpf, read the icmp packet. note ip header is in there
	*/
	if (icmpSd >= 0) {
		cc = recvMsg(icmpSd, (char *)packet, MAXICMPPACKET,
			(struct sockaddr *) &from, &fromlen, &ifindex);
		if (cc < 0) {
			if (errno == EINTR)
				return;
			mip_log( MLOG_ERR,
				"handleIcmpAdvert(): recvmsg failed (errno=%d)", errno);
			if (debug)
				perror("mnd: icmp recvmsg error");
			return;
		}
	}
	else {
		ppt = p;	/* pointer to passed in packet buffer */
		cc = iplen;	/* length of input at this stage */
#ifdef MND
		/* trifle hardwired, but we can only have one if
		*/ 
		ifaddr.s_addr = mn_config.mn.s_addr;
#endif
		sawMac = macflag;
		for (i=0;i<MACMSIZ;i++)
			mac_addr[i]=*mac++;
	}

	dumpppt = ppt;

	/* count all beacons recv */
	icmp_stats.all_beacons_recv++;

	total = cc;
	if (debug) {
		printf("icmprecv: input pkt from %s, len=%d\n", 
                                inet_ntoa(from.sin_addr),
				total);
		dumpme(dumpppt, total);	
	}
	
	/* we have ip header in incoming ICMP packet,
	 * we setup a pointer to it
	 */
	ipp = (struct ip*) ppt;
	hlen = (int) (ipp->ip_hl << 2);
        if (cc < hlen + ICMP_MINLEN) {
		icmp_stats.tooshort++;
		if (debug) {
			printf("error: icmp packet less than min icmp length\n");
		}
                return;
        }

	/* determine interface
	 */
	if (icmpSd >= 0) {
		ifaddr.s_addr = indextoif(ifindex);
		if (debug) {
			printf("ifaddr %x, ipp src %x\n", 
				ifaddr.s_addr, ipp->ip_src.s_addr);
		}
		/* if icmp packet came from us or 0, bail asap
		*/
		if ((ifaddr.s_addr == 0) || (ifaddr.s_addr == ipp->ip_src.s_addr)) {
			icmp_stats.all_beacons_recv--;
			if (debug) {
				printf("icmprecv error, packet ignored\n");
			}
			return;
		}
	}
	ip_src = ipp->ip_src;

	cc -= hlen;	/* throw away ip header */
	/* pointer to router advert body */
	icp = (struct icmp_advert *) ((int)ipp + hlen);

	/* we care about only router adverts. 
	 * not worth logging otherwise
	 */
	if (icp->icmp_type != ICMP_ROUTERADVERT)  {
		if (debug) {
			printf("icmprecv: packet NOT routeradvert\n");
		}
#ifdef MADRP
              if(icp->icmp_type==ICMP_UNREACHABLE){
                       icu=(struct icmp_unreachable *) ((int)ipp + hlen);
		       if(debug)
				printf("!!!!recvive icmp unreachable of %s\n",
                               		inet_ntoa(icu->_ip.ip_dst));
#ifdef MND
                       ss=srhlist_search(icu->_ip.ip_dst);
                       if (ss) {
#ifdef DEBUG
		       		if(debug)
                                printf("delete host route %s, hlist too\n",
                                        inet_ntoa(ss->destHost));
#endif
                                route_delete_host(ss->destHost, ss->nextHop);
                                srhlist_delete(icu->_ip.ip_dst);
				alist_delete(&sr_hagentlist,icu->_ip.ip_dst);
				srhagentcount--;
                        } else {
				struct alist * alisthead=NULL, *a=NULL;
				alisthead=alist_getmnhead();
				a=alist_search(alisthead, icu->_ip.ip_dst);
				if( a){
#ifdef DEBUG
		       			if(debug)
                                	printf("delete link route %s, alist too\n",
                                        inet_ntoa(a->bip));
#endif
					alist_delete(&alisthead, a->bip);
                                	route_delete_link(a->bip, a->mac);
				}
                        }

                       Send_SR_request( icu->_ip.ip_dst, TRUE, ifaddr);

#else /* MIPD */
                       ss= srhlist_search(icu->_ip.ip_dst);
                       if (ss) {
#ifdef DEBUG
		       		if(debug)
                                printf("delete %s in hlist\n",
                                        inet_ntoa(ss->destHost));
#endif
                                srhlist_delete(ss->destHost);
		       }
		       if(is_host_route(icu->_ip.ip_dst, &nexthop)) {
#ifdef DEBUG
		       		if(debug)
                                printf("delete host %s in routing table\n",
                                        inet_ntoa(icu->_ip.ip_dst));
#endif
                                route_delete_host(icu->_ip.ip_dst, nexthop);
                       } else if(is_link_route(icu->_ip.ip_dst)) {
#ifdef DEBUG
		       		if(debug)
                                printf("delete link %s in routing table\n",
                                        inet_ntoa(icu->_ip.ip_dst));
#endif
                               route_delete_link(icu->_ip.ip_dst,NULL);
		       }
                 	vlisthead=vlist_getfahead();
		 	thisvlist=vlist_search(vlisthead,icu->_ip.ip_dst);
		  	if(! thisvlist) {
				  vlisthead=vlist_gethahead();
				  thisvlist=vlist_search(vlisthead,icu->_ip.ip_dst);
			}
			if( thisvlist){
#ifdef DEBUG
		       		if(debug)
                               	printf("delete %s in vlist\n", 
					inet_ntoa(icu->_ip.ip_dst));
#endif
				vlist_delete(&vlisthead, icu->_ip.ip_dst);
                        }
#endif  /* MND */
               	}
#endif  /* MADRP */
		return;
	}

	/* find out whether arp is on or off of the incoming i/f */
	doArp=getConfigArp(&ifaddr);

	/* we had no matching interface, toss packet
	 * as it was a beacon recv. on an interface we don't care
	 * about
	 */
	if (doArp < 0) {
		if(debug)
			printf("icmprecv: doArp failure - pkt ignored\n");
		return;
	}
#ifdef MIPD
	/* if mipd and arp is ON on this if, ignore icmp.
	 * TBD. modify to handle icmp router solicitations.
	 */
	if (doArp) {
		if(debug)
			printf("icmprecv: we are mipd, doArp - pkt ignored\n");
		return;
	}
#endif

	/* perform sanity checks per RFC 1256, section 5.2 ("Message
	 * Validation by Hosts")
	 */
	/* U. of Helsinki linux MIP sets naddrs to 0 ...
	*/
	noaddrs = icp->icmp_naddrs;
	if (icp->icmp_naddrs < 1) {
		noaddrs = 0;
#ifdef NOPE
		icmp_stats.tooshort++;
		if (debug) {
			printf("handleIcmpAdvert(): icmp \"Num Addrs\" is %d.\n",
				icp->icmp_naddrs);
		}
		return;
#endif
	}
	else {
		ip_src = *((struct in_addr *) &icp->icmp_router);
	}
	/* at this point ip_src is the ip address of the beacon entity
	 * in question
	 */

	/* 2 long words for IPaddr and preference */
	if (icp->icmp_asize != 2) {
		icmp_stats.malformed++;
		if (debug) {
			printf("handleIcmpAdvert(): icmp \"Addr Entry Size\" is %d.\n",
				icp->icmp_asize);
		}
		return;
	}

	/* size of router advert, basic header + variable length
	 * ip address part at end
	 */
	advert_length = sizeof(struct icmp_advert) + (8 * (icp->icmp_naddrs-1));

	/* check size, invalid mobile ip packet if too small 
	 */
	if (cc <= advert_length ) {
		icmp_stats.tooshort++;
		if (debug) {
			printf("icmprecv: message too short\n");
		}
		return;
	}

	/* check size of remaining packet. unfortunate part here
	 * is that this only checks it for HA, not for FA.
	 */
	if (cc < (advert_length + (sizeof(struct mip_mseicmp) - 4))) { 
		icmp_stats.msetooshort++;
		if (debug) {
			printf("icmprecv: icmp+mse message too short\n");
		}
		return;
	}

	/* form pointer to point to mse, still can't trust it though
	 */
	mse = (struct mip_mseicmp *) ((int)icp + advert_length);
	if (mse->mse_type != MIPAGENTAD) {
		icmp_stats.invalidicmpmse++;
		if (debug) {
			printf("icmprecv: invalid mse type in icmp advert\n");
		}
		return;
	}
	
	/* check flags */
	flags = ntohs(mse->mse_flags);

	size = mse->mse_len + 2;

	/* came from FA */
	if (TST(flags, FABIT)) {
		struct in_addr router;
		/* sigh. must do fa check twice
		*/
		if (cc < (advert_length + (sizeof(struct mip_mseicmp)))) { 
			icmp_stats.msetooshort++;
			return;
		}
#ifdef MND
		/* HACK!... if mnd gets beacon from fa that is same as ha,
		 * treat as ha
		 */
 		router = ip_src;
                if (router.s_addr == mn_config.ha.s_addr) {
                        faisha = useha = 1;
			icmp_stats.faisha_recv++;
                }
                else {
                        sawfa = 1;
                        if (mse->mse_len < 10) {
                                /* invalid mse++ */
				icmp_stats.msetooshort++;
                                return;
                        }
                        if (TST(flags, RBIT)) {
                                rbit = 1;
                        }
                        mse_coa.s_addr = mse->mse_coa;

			/* treat a mistaken coa of zero as truncated packet */
			if (mse_coa.s_addr == INADDR_ANY) {
				icmp_stats.msetooshort++;
				return;
			}
		}
#else	/* ifdef MIPD */
		/* agent doesn't care about beacons from FAs */
		return;
#endif
	}
	/* came from HA */
	if (TST(flags, HABIT) || faisha) {
#ifdef MND
 		struct in_addr router = ip_src;
		sawha = useha = 1;
		if (mse->mse_len < 6) {
			/* invalid mse++ */
			icmp_stats.msetooshort++;
			return;
		}

		/* If the src IPaddr of the packet sent by the "router"
		 * matches our HA IPaddr then we recognize it as our HA.
		 * Otherwise the advertisement is not important to us
		 * unless it also contains a set FA bit.
		 */
                if (router.s_addr != mn_config.ha.s_addr) {
			useha = 0;
                        if (debug) {
                                char buf[1024];
                                strcpy(buf, inet_ntoa(router));

				if (sawfa) {
					printf("HA beacon is not ours, but FA bit is set.\n");
				}
				else {
                                	printf("ignoring HA beacon, my ha is %s, beacon is from %s\n",
						inet_ntoa(mn_config.ha), buf);
				}
                        }
			if (!sawfa) {
				return;
			}
                }
#else
		/* agent doesn't care about HA beacons either */
		return;	
#endif
	}
	if (TST(flags, MNBIT)) {
		/* if arp is on, ignore icmpbeacon from other mn */
		if (doArp) {
			if (debug) 
				printf("Arp is on, Mn ignores icmp beacon from Mn\n");
			icmp_stats.arp_on_mn_beacons_ignored++;
			return;
		}
		sawmn = 1;
		if (mse->mse_len < 6) {
			/* invalid mse++ */
			icmp_stats.msetooshort++;
			return;
		}
	}

	if (sawfa && sawha) {
		icmp_stats.bothhafa++;
	}
	/* ignore packet if no bit flag set */
	else if (sawfa == 0 && sawha == 0 && sawmn == 0) {
		return;
	}

	mse_sequence = ntohs(mse->mse_sequence);
	mse_lifetime = ntohs(mse->mse_lifetime);

	/* now deal with optional parts if any
	*/
	cc = cc - (advert_length + size);
	mse = (struct mip_mseicmp *) (((char *)mse)  + size);

	if (debug) {
		printf("icmprecv: now to do optional TLV parts\n");
		printf("cc %d\n", cc);
	}

	/* unlike the draft says, we don't discard messages
	 * with types in range 0..127, we just ignore the extension
	 */
	while(cc >= 2) {
		int len;
		if (debug) {
			printf("cc %d, mse type (hex) %x\n", cc, 
				mse->mse_type);
		}
		switch(mse->mse_type) {
		case 0:  /* one byte pad, should be end */
			cc = 0;
			break;
		case MIPMAC:
			msemac = (struct mip_mse_icmp_mac *) mse;
			if (cc < sizeof(struct mip_mse_icmp_mac) 
				|| msemac->mse_len < 6) {
				/* invalid mse++ */
				icmp_stats.msetooshort++;
				if (debug) {
					printf("MIPMAC too short %d\n",
						msemac->mse_len);
				}
				return;
			}
			for (i=0;i<MACMSIZ;i++)
				mac_addr[i]=msemac->mse_mac[i];
			sawMac = 1;
			if (debug) {
				printf("MIPMAC option: %x %x %x %x %x %x\n",
					mac_addr[0], mac_addr[1], mac_addr[2],
					mac_addr[3], mac_addr[4], mac_addr[5]);
			}
			break;
		case MIPAGENTAUTH:
			if (cc < sizeof(struct mip_mse_icmp_auth)) {
				/* invalid mse++ */
				icmp_stats.msetooshort++;
				if (debug) {
					printf("MIPAGENTAUTH too short %d\n",
						msemac->mse_len);
				}
				return;
			}
			if (debug) {
				printf("MIPGAGENTAUTH option: \n");
			}
			icp->icmp_cksum = 0;
			icmp_stats.all_netauth_recv++;
			mseauth = (struct mip_mse_icmp_auth *) mse;
#ifdef MIPD
			/* if agent is 0 or this interface, we accept
			 * else message is ignored.  Note we do this
			 * BEFORE we run auth. csum, which can
			 * save the agent time spent in otherwise
			 * useless processing.
			 */
			if (mseauth->agent == 0 || mseauth->agent == ifaddr.s_addr ) {
				;
			}
			/* else we can ignore it
			*/
			else {
			  	if(debug) {
			    		printf("MSEAUTH: ICMP ignored, ifaddr %x, agent %x\n", 
						ifaddr.s_addr, mseauth->agent);
				}
				return;
			}
#endif
			if (TST(mseauth->flags, NETAUTH)) {
				/* configured to reject netauth */
				if (checkNetauthReject()) {
					if (debug) {
						printf("NETAUTH rejected -- acc. to config policy\n");
					}
					icmp_stats.netauth_rejected++;
					goto nexttlv;
				}
				/* discard beacons if ARP is true
				 * and beacon has authentication 
				 */
				if (doArp == TRUE) {
					if (debug) {
						printf("pkt with NETAUTH ignored as arp is on \n");
					}
					return;
				}

				/* auth_ok is a boolean, 1 if auth worked */
				auth_ok = icmpSecurityCheck(packet + hlen, mseauth);
				if (debug) {
					printf("icmpseccheck auth_ok %d\n",
						auth_ok);
				}
				if (auth_ok > 0) {
					/* set auth_type bit to indicate 
					 * which type
					 */
					SET(auth_type, NETAUTH);
					auth_tssecs = ntohl(mseauth->tssecs);
					auth_tsfract = ntohl(mseauth->tsfract);
					/* if pass net auth, don't bother to continue, 
					 * no matter there is adhoc auth or not 
					 */
					cc=0;
					break;
				}
				/* count authentication failures
				*/
				else {
					if (sawha)  {
						icmp_stats.ha_netauthbad++;
					}
					if (sawfa) {
						icmp_stats.fa_netauthbad++;
					}
					if (sawmn) {
						icmp_stats.mn_netauthbad++;
					}
					netauthbad = 1;
				}
			}
			else if (TST(mseauth->flags, ADHOCAUTH)) {
				cc = 0;		/* last by definition */
				/* configured to reject adhoc */ 
				if (checkAdhocReject()) {
					if (debug) {
						printf("ADHOC rejected -- acc. to policy\n");
					}
					icmp_stats.adhoc_rejected++;
					goto nexttlv;
				}
				if (doArp == TRUE) {
					return;
				}

				auth_ok = icmpSecurityCheck(packet + hlen, mseauth);
				if (debug) {
					printf("icmpseccheck auth_ok %d\n",
						auth_ok);
				}
				if (auth_ok > 0) {
					SET(auth_type, ADHOCAUTH);
					auth_tssecs = ntohl(mseauth->tssecs);
					auth_tsfract = ntohl(mseauth->tsfract);
				}
				else {
					if (sawha)  {
						icmp_stats.ha_adhoc_authbad++;
					}
					if (sawfa) {
						icmp_stats.fa_adhoc_authbad++;
					}
					if (sawmn) {
						icmp_stats.mn_adhoc_authbad++;
					}
					adhocauthbad = 1;
				}
			}
			break;
		case MIPAGENTINFO:
			if (cc < 4) {
				return;
			}
			lsize = (int) mse->mse_len & 0xff;
			/* put in location array and make sure it is
			 * null terminated
			 */
			strncpy(location, ((char *)mse) + 2, lsize);
			location[lsize] = 0;
			if (debug) {
				printf("MIPAGENTINFO, lsize %d, location %s\n", 					lsize, location);
			}
			l = (char *)location;
			break;
		case 19: /* prefix message, ignore it */
		default:
			if (debug) {
				printf("mse_type %d 0x%x\n",
					mse->mse_type, mse->mse_type);
			}
			break;
		}
		/* advance pointer */
nexttlv:
		if (cc) {
			cc -= (mse->mse_len + 2);
			mse = (struct mip_mseicmp *) (((char *)mse) + 
				(((int) mse->mse_len + 2) & 0xff));
		}
	}

	/* if beacon lacks authentication, and arp is not used, we toss it
	 * but count it.  This may indicate agents (or mns?)
	 * that are using arp or it may indicate simply bad authentication.
	 */
	if (doArp == FALSE && auth_type == 0) {
		/* toss if had authentication but it failed
		*/
		if (netauthbad || adhocauthbad) {
			if (debug)
			    printf("icmp auth invalid, netauthbad %d, adhocauthbad %d\n", netauthbad, adhocauthbad);
			return;
		}		
		/* beacon not authenticated and discarded
		 * - probably arp mechanism 
	   	 * primary use here is to tell us we are getting
		 * beacons, BUT they are from agents lacking
		 * authentication.  
		 */
		if (sawha)  {
			icmp_stats.ha_maybe_arp_lacks_auth++;
		}
		if (sawfa) {
			icmp_stats.fa_maybe_arp_lacks_auth++;
		}
		if (sawmn) {
			icmp_stats.mn_maybe_arp_lacks_auth++;
		}
		return;
	}
	/* if we are authenticated and we don't have a mac address,
	 * then we have a foobar!
	 */
	if (auth_type && sawMac == FALSE) {
		icmp_stats.auth_nomac++;
		return;
	}

#ifdef MAYBENOTUSEFUL
	/* count type of auth if that is what we are doing and
	 * it worked out
	 */
	if (doArp == FALSE && auth_ok) {
		if (sawha)  {
			icmp_stats.ha_netauthok++;
		}
		if (sawfa) {
			icmp_stats.fa_netauthok++;
		}
		if (sawmn) {
			icmp_stats.mn_netauthok++;
		}
	}
#endif

	/* dump mse information
	*/
	if (debug) {
		printf("ICMPRECV: icmp packet:\n");
		
		dumpme(dumpppt, total);	
		if (sawha && sawfa) {
			char buf[1024];
			strcpy(buf, inet_ntoa(mse_coa));
 			printf("ICMPRECV: HA/FA(%s): icmp_router_advert/mse, sequence %u, lifetime %u, coa %s\n",
                                inet_ntoa(from.sin_addr),
                                mse_sequence, mse_lifetime,
				buf);
		}
		else if (sawha) {
 			printf("ICMPRECV: HA(%s): icmp_router_advert/mse, sequence %u, lifetime %u%s\n",          
                                inet_ntoa(from.sin_addr),
                                mse_sequence, mse_lifetime,
                                faisha?" (HA advertising as FA)":""
                                );

		}
		else if (sawfa) {
			char buf[1024];
			strcpy(buf, inet_ntoa(mse_coa));
			printf("ICMPRECV: FA(%s): icmp_router_advert/mse, sequence %u, lifetime %u, coa %s\n", 
				inet_ntoa(from.sin_addr),
				mse_sequence, mse_lifetime, 
				buf
				);
		}
		else {
			printf("ICMPRECV: MN(%s): icmp_router_advert/mse, sequence %u, lifetime %u\n", 
				inet_ntoa(from.sin_addr), 
				mse_sequence, 
				mse_lifetime);
		}
		if (sawMac) {
			printf("ICMPRECV: MAC address: ");
			for (i=0; i<6; i++)
				printf("%02x%c",
					mac_addr[i]&0xff, (i<5)?':':' ');
			printf("\n");
		}
		if (l) {
			printf("location <%s>\n", l);
		}
		if (u) {
			printf("url <%s>\n", u);
		}
		if (auth_ok) {
			printf("ICMPRECV: auth was valid\n");
			if (TST(auth_type,NETAUTH)) 
				printf("ICMPRECV: netauth_key authentication passed:  timestamp %x.%x\n", auth_tssecs, auth_tsfract);
			if (TST(auth_type,ADHOCAUTH))
				printf("ICMPRECV: adhoc_key authentication passed: timestamp %x.%x\n", auth_tssecs, auth_tsfract);
		}
		else {
			printf("ICMPRECV: icmp not authenticated!\n");
		}
	}

	/* if beacon has valid auth, count it, else throw it
	 * in the arp/valid pile
	 */
	if (auth_ok) {
		icmp_stats.valid_auth_beacons++;
	}
	else {
		icmp_stats.valid_arp_beacons++;
	}

	if (debug) {
		printf("valid beacon, ip src is %s\n", 
                               inet_ntoa(ip_src));
	}

	/* at this point:
	 * sawha/sawfa - kind of agent
	 * router - target ip address to use for route
	 * mse_sequence - current sequence number
	 * mse_lifetime - lifetime before reregistration
	 * mse_coa - coa from fa
 	 * auth_ok - boolean, true if either authentication verified
 	 * auth_type - either NETAUTH or ADHOCAUTH (bit flag)
	 * auth_tssecs, auth_tsfract - value for any auth timestamp in secs
	 * location - pointer to location, NULL if none.
	 * lsize - size of location 
	 * sawMac - boolean, true if we have MAC address
	 * mac_addr[6] - MAC address of the sender
	 *	only valid if sawMac is TRUE.
	 * doArp - configuration variable, arp on/off on this interface
	 * struct in_addr ifaddr - if ip address
	 */
	icmp_stats.valid_beacons_recv++;
#ifdef MND
        /* take this beacon and see if it is in the alist
         * if it is not, we will store it in there
         */
        checkAlist(	useha,
			ip_src,
                	mse_sequence, 
			mse_lifetime, 
			mse_coa, 
			auth_type, 
			auth_ok, 
			auth_tssecs, 
			auth_tsfract, 
			l, 
			lsize,
			sawMac,	
			mac_addr,	/* mac address */
			sawmn,		/* if TRUE, then mn */
			doArp,		/* are we doing arp or not on this interface */
			ifaddr		/* if ip addr */
		);
	return(0);
#endif
#ifdef MIPD
	if (!sawmn)
		return(0);

	/* note: if arp is in use, this function is not called
	 * authentication can be taken for granted too.
	 */
	
	
        stateIcmp(	
		*((struct in_addr *) &icp->icmp_router),  /* ip addr */
                mse_sequence, 
                mse_lifetime, 
		auth_type, 
		l, 
		lsize,
		u, 
		usize,
		sawMac,	
		mac_addr,	/* mac address */
		ifaddr,		/* if ip addr */
		auth_tssecs,
		auth_tsfract
		);
	
	return(0);
#endif
	syslog(LOG_ERR, "icmpHandleAdvert - should not get here\n");
	exit(1);
}

/*
 * check icmp security if it shows up.  note packet should have
 * had ip header removed from it.
 *
 * we don't log anything here.  we do however keep
 * totals in stats 
 *
 * return:
 *	1: auth ok
 *	0: failed authentication.  That means we got an authentication
 *	but the hash recompute failed
 *     -1: no key (mode switch).
 */
static
int
icmpSecurityCheck(char *packet, struct mip_mse_icmp_auth *mseauth)
{
	int spi = mseauth->spi;
	unsigned char hash[16];
	extern int debug;
	int size;
	char *key;

	/* sanity check */
	if (mseauth == NULL)
		return (0);

	/* do security hash check
	 */
	if (((getNetauthSecurity())&&(TST(mseauth->flags,NETAUTH)))
	  ||((getAdhocSecurity())&&(TST(mseauth->flags,ADHOCAUTH)))){
		size = (((char *)mseauth) - packet) +
			MIPAGENTAUTH_OFFSET;
		if (debug) {
			printf("icmp auth size is %d\n", size);
		}
		/* check hash first */
                /* recompute MD5 hash, result in local hash[]
                */ 
		if (TST(mseauth->flags,NETAUTH)) {
			if (spi != htonl(getNetauthSpi()))
				return(-1);
			key = getNetauthKey();
			if (key == NULL)
				return(-1);
	                computeMd5(packet, size, key, hash);
		}
		else {
			if (spi != htonl(getAdhocSpi()))
				return(-1);
			key = getAdhocKey();
			if (key == NULL)
				return(-1);
	                computeMd5(packet, size, key, hash);
		}

		if (bcmp(mseauth->hash, hash, 16) != 0) {
                        if (debug) {
                                printf("mn agent authentication failed\n");
                        }
                        /* authentication failed
                        */
                        return(0);
                }
		if (debug) {
			printf("mn agent authentication ok\n");
		}
		return(1);
	}

        if (debug) {
        	printf("mn agent authentication TEST failed\n");
        }
	return(0);
}










