
/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * ifs.c
 *
 * module for walking interfaces and extracting ifconfig-style
 * information.
 */
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/ioctl.h>

#include <net/if.h>
#include <net/if_var.h>
#include <net/if_dl.h>
#include <netinet/in.h>
#include <netinet/in_var.h>
#include <arpa/inet.h>

#define	NSIP
#include <netdb.h>

#include <sys/protosw.h>

#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "mip_log.h"


getIfs()
{
	int             s;
	struct ifconf   ifc;
#define MAX_INTERFACES 50	/* Yeah right. */
	char            buffer[MAX_INTERFACES * sizeof(struct ifreq)];
	struct ifreq   *ifptr, *end;
	int             ifflags;
	char 		name[1024]; 
	int		rc;
	struct 		sockaddr_in if_sin;
	struct 		sockaddr_in netmask;
	struct 		sockaddr_dl *sdl;
	int 		mcast;
	int 		found_index = -1;

	struct ifreq    ifr;

	s = socket(AF_INET, SOCK_DGRAM, 0);
	if (s < 0) {
		syslog(LOG_ERR, "getIfs(): UDP ifconfig socket (%m).");
		exit(1);
	}
	ifc.ifc_len = sizeof(buffer);
	ifc.ifc_buf = buffer;

	/*
	 * use SIOCGIFCONF socket ioctl to suck in all the interface
	 * structures in one swell foop
	 */
	if (ioctl(s, SIOCGIFCONF, (char *) &ifc) < 0) {
		syslog(LOG_ERR, "getIfs(): ifconfig (SIOCGIFCONF) (%m)");
		exit(1);
	}
	/*
	 * ifc_req points to first if structure
	 */
	ifflags = ifc.ifc_req->ifr_flags;
	/*
	 * setup end pointer, ifc.ifc_len was returned with true length
	 */
	end = (struct ifreq *) (ifc.ifc_buf + ifc.ifc_len);
	ifptr = ifc.ifc_req;
	while (ifptr < end) {
		sprintf(ifr.ifr_name, "%s", ifptr->ifr_name);
		sprintf(name, "%s", ifptr->ifr_name);
 		if (ifptr->ifr_addr.sa_family == AF_LINK) { 
                        sdl = (struct sockaddr_dl *) &ifptr->ifr_addr;
#ifdef TESTMAIN
			{
				char *c = (char *) LLADDR(sdl);
                        	printf("LINK: %s: [%d]\n", name, sdl->sdl_index);
				printf("link addr %2x:%2x:%2x:%2x:%2x:%2x\n",
					c[0], c[1], c[2], c[3], c[4], c[5], c[6]);
			}
#endif
                        found_index = sdl->sdl_index;
                }

		close(s);
		if ((s = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
			syslog(LOG_ERR, "getIfs(): ifconfig socket (%m)");
			exit(1);
		}
		if (ifptr->ifr_flags == ifflags) {
			if (ioctl(s, SIOCGIFFLAGS, (caddr_t) & ifr) < 0) {
				syslog( LOG_ERR, "getIfs(): ioctl (SIOCGIFFLAGS) (%m)");
				exit(1);
			}
			/* if interface is not up, ignore it
			*/
			if (!(ifr.ifr_flags & IFF_UP)) {
				goto advance;
			}

			mcast = 0;
			if (ifr.ifr_flags & IFF_MULTICAST) {
				mcast = 1;  /* if can multicast */
			}
			/* get ip address
			*/
			rc = getInfo(s, name, &if_sin, &netmask);
			if (rc) {
				/* pass name and ip address to config
				 * sanity check routine so that it can
				 * verify existing interface against
				 * desired i/f and store ip address
				 */
#ifdef TESTMAIN
				printf("ifname: %s, ip=%s, mcast %d \n", 
					ifptr->ifr_name,
					inet_ntoa(if_sin.sin_addr),
					mcast);
#else
				configInterface(ifptr->ifr_name,
					&if_sin, &netmask, mcast, found_index);
#endif
				      
			}
		}

advance:
		/* advance ifptr
		*/
		if (ifptr->ifr_addr.sa_len)	/* Dohw! */
			ifptr = (struct ifreq *) ((caddr_t) ifptr +
						  ifptr->ifr_addr.sa_len -
						  sizeof(struct sockaddr));
		ifptr++;
	}
	close(s);
}

/* pry the IP interface address out of the interface
 *
 * returns:
 *	0: failed
 *	1: got it
 *
 * outputs:
 * 	the address is returned via the sin pointer passed in.
 * inputs:
 *	s - socket
 *	name - name of interface
 */
int
getInfo(int s, char *name, struct sockaddr_in *sin, struct sockaddr_in *netmask)
{
	struct ifreq    ifr;
	char *inet_ntoa();

        strncpy(ifr.ifr_name, name, sizeof (ifr.ifr_name));
        if (ioctl(s, SIOCGIFADDR, (caddr_t)&ifr) < 0) {
                if (errno == EADDRNOTAVAIL || errno == EAFNOSUPPORT) {
                        return(0);
                } else {
                        syslog( LOG_ERR, "getInfo(): ioctl SIOCGIFADDR (%m)");
			exit(1);
		}
        }
	/* paranoid check, only want IP
	*/
	if (ifr.ifr_addr.sa_family != AF_INET) 
		return(0);

	bcopy(&ifr.ifr_addr, sin, sizeof (struct sockaddr));

	/* get the netmask too
	*/

        if (ioctl(s, SIOCGIFNETMASK, (caddr_t)&ifr) < 0) {
                if (errno == EADDRNOTAVAIL || errno == EAFNOSUPPORT) {
                        return(0);
                } else {
                        syslog( LOG_ERR, "getInfo(): ioctl SIOCGIFADDR (%m)");
			exit(1);
		}
        }
	bcopy(&ifr.ifr_addr, netmask, sizeof (struct sockaddr));
#ifdef TESTMAIN
	printf("netmask %s, anded %x\n", inet_ntoa(netmask->sin_addr),
		ntohl(netmask->sin_addr.s_addr) &
		ntohl(sin->sin_addr.s_addr));
#endif

	return(1);
}

#ifdef TESTMAIN
main()
{
	getIfs();
}
#endif
