/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * keylist.c
 *
 * a module for the support of a keylist data type.
 *
 * a keylist is a list of keys indexed by ip address
 * The keylist search mechanism is based on a simple HASH
 * bucket and should support the quick search of MANY ip/key bindings.
 *
 * This module does NOT limit the size of keys to any fixed number of
 * butes.  The keysize must be passed in.  
 * More than one keylist is supported.   A pointer is passed
 * back from init_keylist that should be used with the enter/get
 * routines.
 *
 * functional interface:
 *	struct keylist *init_keylist() - returns a keylist (array)
 *	int enter_keylist() - enter a key into a list
 *	search_keylist() - routine for searching a list
 * 	char * get_key() - external routine for searching for entry
 *		returns key directly and other values indirectly 
 *	void dump_keylist() - given a keylist, print it out (debug).
 *
 * init_keylist should be called to allocate and init a keylist.
 * The pointer returned by it is used as a handle for the other routines.
 * enter_keylist()  puts a key into the keylist.
 * get_key is called to recover the key based on a ipaddr.
 */
#include <stdio.h>
#include "keylist.h"

/*
 * returns pointer to array of keybuckets
 *
 * returns:
 *	NULL: failed due to no space
 *	k: array of keybuckets
 */
struct keylist *
init_keylist()
{
	struct keylist *k;

	k = (struct keylist *) malloc(sizeof(struct keylist *) * KEYBUCKETS);
	if (k == NULL)
		return(NULL);
	/* don't need to do this, but we are paranoid ...
	*/
	bzero(k, sizeof(struct keylist *) * KEYBUCKETS);
	return(k);
}

/*
 * enter_keylist
 *
 * algorithm:
 *
 *	search key list on index ipaddr
 *	if found, fatal error
 *	malloc space
 *	store key in key structure
 *	glue key into hash list
 *
 *	spi is an array of outgoing & incoming AH and ESP spi values
 *	ordered as follows:
 *		spi[0] = ah outgoing
 *		spi[1] = ah incoming
 *		spi[2] = esp outgoing
 *		spi[3] = esp incoming
 *
 * returns:
 *	1 - ok
 *	0 - failed (addr in keylist)
 *     -1 - malloc out of room
 */

int
enter_keylist(struct keylist *kb, unsigned long ipaddr, 
	char* key, int keysize, unsigned long spi[], int transform,
	unsigned long ipaddr2, int inuse, int timeflag)
{
	struct keylist *k;
	struct keylist *kbl;
	struct keylist **kba;
	int index;

	/* search the key list */
	k = search_keylist(kb, ipaddr);
	if (k)
		return(0);
	/* malloc the space for the keylist structure */
	k = (struct keylist *) malloc(sizeof(struct keylist));
	if (k == NULL) {
		return(-1);
	}
	bzero(k, sizeof(struct keylist));

	/* malloc the space for the key itself
	*/
	k->key = (char *)malloc(keysize);
	if (k->key == NULL) {
		return(-1);
	}
	bzero(k->key, keysize);

	/* store the key info */
	k->ipaddr = ipaddr;
	k->spi_ah[SPI_OUTB]	= spi[0];
	k->spi_ah[SPI_INB]	= spi[1];
	k->spi_esp[SPI_OUTB]	= spi[2];
	k->spi_esp[SPI_INB]	= spi[3];
	k->transform = transform;
	k->inuse = inuse;
	k->timeflag = timeflag;
	k->next = NULL;
	bcopy(key, k->key, keysize);
	k->keysize = keysize;
	k->ipaddr2 = ipaddr2;

	index = KEYHASH(ipaddr);
	kba = (struct keylist **) kb; 
	kbl = kba[index];
	if ( kbl == NULL) {
		kba[index] = k;
	}
	else {
		while(kbl->next) {
			kbl = kbl->next;
		}
		kbl->next = k;
	}
	return(1);
}

struct keylist *
search_keylist(struct keylist *kb, unsigned long ipaddr)
{
	int index;
	struct keylist *k;
	struct keylist **kba;

	/* hash the key */
	/* linear search on hash bucket */
	index = KEYHASH(ipaddr);
	kba = (struct keylist **)kb;
	k = kba[index];
	if ( k == NULL)
		return(NULL);
	while(1) {
		if (k->ipaddr == ipaddr)
			return(k);
		if (k->next == NULL)
			return(NULL);
		k = k->next;
	}
}

/*
 * given a keylist pointer and an ip address as index, return
 * the key (and other values through indirect pointers) for a given
 * entry in a keylist.
 *
 * returns:
 *	NULL - if no such ipaddr
 *	key string otherwise
 *
 * indirectly:
 *	size of key string
 *	spi value
 *	transform value
 *
 *	NOTE: only the outbound spi is returned as this is all that
 *	MobileIP uses and this function answers to it, not to ipsec.
 */
char *
get_key(struct keylist *kb,
	unsigned long ipaddr, int *keysize, unsigned long *spi, int *transform, 
	unsigned long *ipaddr2, int *timeflag)
/* ipaddr is input and index for search */
/* keysize/spi/transform are outputs */
{
	struct keylist *k;

	k = search_keylist(kb, ipaddr);
	if (k == NULL)
		return(NULL);
	*keysize = k->keysize;
	*spi = k->spi_ah[SPI_OUTB];
	*transform = k->transform;
	*timeflag = k->timeflag;
	*ipaddr2 = k->ipaddr2;
	return(k->key);
}

/*
 * given a keylist and an ip index, change the keylist entry transform value
 *
 * returns:
 *	1: no such entry
 *	0: entry found and value stored ok.
 */
int
modify_key_transform(struct keylist *kb, unsigned long ipaddr, int transform)
/* ipaddr is input and index for search */
{
	struct keylist *k;

	k = search_keylist(kb, ipaddr);
	if (k == NULL)
		return(1);
	k->transform = transform;
	return(0);
}


static struct keylist **warray;
static struct keylist *whead;
static windex;

void
klist_initwalk(struct keylist *list)
{
	windex = 0;
        warray = (struct keylist **) list;
	whead = warray[windex];
}

struct keylist *
klist_nextwalk()
{
        struct keylist *cur = whead;
        while (cur == NULL && windex < KEYBUCKETS) {
		cur = whead = warray[++windex];
	}
	if (cur)
		whead = whead->next;
        return(cur);
}


static void
print_key(char *key, int keysize)
{
	int j;
	char *kp;

	kp = key;
	printf("key: ");
	for (j = 0; j < keysize; j++, kp++) {
		printf("%2.2x", *kp & 0xff);
	} 
	printf("\n");
}

void
dump_keylist(struct keylist *kb)
{
	int i;
	int j;
	struct keylist *k;
	char *kp;
	struct keylist **kba = (struct keylist **) kb;

	for ( i = 0; i < KEYBUCKETS; i++) {
		k = kba[i];
		if ( k == NULL)
			continue;
		printf("keybucket index %d\n", i);
		while(1) {
			
			printf("\tipaddr %ld(0x%lx)\n", k->ipaddr, k->ipaddr);
			printf("\tipaddr2 %ld(0x%lx)\n", k->ipaddr2, 
				k->ipaddr2);
			printf("\tspis: AH_out 0x%x, AH_in 0x%x, ESP_out 0x%x, ESP_in 0x%x\n", 
				k->spi_ah[SPI_OUTB], k->spi_ah[SPI_INB],
				k->spi_esp[SPI_OUTB], k->spi_esp[SPI_INB]);
			printf("\ttransform %d keysize %d inuse %d\n\t", 
				k->transform, k->keysize, k->inuse);
			print_key(k->key, k->keysize);
			if (k->next == NULL)
				break;
			k = k->next;
		}
	}
}

#ifdef TESTMAIN
main()
{
	char testkey[16];
	int i;
	unsigned long ipaddr;
	char *keystring;
	unsigned long	spi,
		spiset[4];
	int keysize, transform;
	unsigned long ipaddr2;
	struct keylist *kl;
	int timestamp; 

	bzero(spiset, sizeof(unsigned long)*4);
	spiset[0] = 256;

	kl = init_keylist();
	if (kl == NULL) {
		printf("init_keylist failed\n");
		exit(1);
	}
	for ( i = 0; i < 16;  i++) {
		testkey[i] = i;
	}

	ipaddr = 0x0a0b0c0d;
	enter_keylist(kl, ipaddr, testkey, 16, spiset, 1, 0L, 1);
	dump_keylist(kl);

	/* simple lookup test */
	keystring = get_key(kl, ipaddr, &keysize, &spi, &transform, &ipaddr2, 
		&timestamp);
	printf("get_key got keysize %d, spi %d, transform %d\n",
			keysize, spi, transform);
	print_key(keystring, keysize);

	/* now insert 15 and make sure hash buckets overflow */
	for ( i = 0; i < 2000; i++) {
		ipaddr = i;
		enter_keylist(kl, ipaddr, testkey, 16, spiset, ipaddr, 0L, 1);
	}
	dump_keylist(kl);
}
#endif
