/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * sr_hagentlist.c
 *
 * list of agents or mns mobile node is talking to.
 * The list head is maintained by the caller.
 *
 * routines include:
 *
 * Module should hide how list is constructed from callers.
 *
 * XXX. sort entries according to delta lifetimes
 * ala callout list.
 */

#ifdef MADRP

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "mnstate.h"
#ifdef IPSEC
#include "keylist.h"
#include "ipsec.h"
#endif
#include "route.h"
#include "sr_mad.h"
#include "mip_log.h"

extern int	debug;
static int	noticeFunction = 3;
int 		srhagentcount=0;



/*
 * search SR_hlist specified to see if ip addr is in list.
 *
 * note dependencies in SR_hlist_create().
 *
 * returns:
 *	struct SR_hlist * - pointer to SR_hlist entry if found
 *	NULL - no such entry.
 */
static struct SR_hlist *prev;

struct SR_hlist *
sr_hagentlist_search(struct SR_hlist *head, struct in_addr bip)
{
	prev = head;
	while(head) {
		if (head->destHost.s_addr == bip.s_addr)
			return(head);
		prev = head;
		head = head->next;
	}
	return(NULL);
}



/* create an entry and insert the node with the key bip.  make insertion
 * at the end of the list.
 *
 * because prev is used for insertion, this must be called after 
 * sr_hagentlist_search() and before sr_hagentlist_delete(), where the sr_hagentlist_search() 
 * failed to find an entry.
 *
 * returns:
 *	struct *sr_hagentlist - pointer to sr_hagentlist entry if found
 *	NULL - out of buffer space
*/
struct SR_hlist *
sr_hagentlist_create(struct SR_hlist **head, struct in_addr bip)
{
	struct SR_hlist *n = NULL;
	struct SR_hlist *a = malloc(sizeof(struct SR_hlist));
	if (!a) {
		return(NULL);
	}
	/* zero it out */
	memset(a, 0, sizeof(struct SR_hlist));
	
	a->destHost.s_addr = bip.s_addr;	/* store key */
	srhagentcount++;
	/*
	 *  insert at tail
	 */
	if ( ! prev )
	    *head = a;
	else
	    prev->next = a;


	return(a);
}



void
sr_hagentlist_sort( struct SR_hlist **head, struct SR_hlist *a )
/*
 *  --	change position of a in list pointed to by head.  sort values
 *	by metric.  assume that all elements are ordered except for
 *	a, which may be in the proper place already, so search before
 *	making any changes.
 *
 *	if more than one entry has the same average, insert after
 *	all current entries with that value.
 */
{
    struct SR_hlist	*beforea = NULL,
			*p = NULL,	/* p's before q's */
			*q,
			*ptarget = NULL,
			*qtarget = NULL;
    int			gotnext = 0,
			gottarget = 0;

entryMark( 1, "sr_hagentlist_sort" );


    /* a is no worse than the next element in the list --
     * don't move it, as it may lead to an unnecessary reRegistration
     *
     * once we have a pointer to the a (below), we can also check
     * that the item before has a same/different value.
     */
    if ( a->next && a->metric == a->next->metric )
	return;

    /* find new location */
    for ( p = *head, q = NULL; p; q = p, p = p->next ) {
	if ( p == a ) {			/* keep "next pointer" to a */
	    beforea = q;
	    gotnext = 1;
	}
	/* remember all elements (and "pointers to") up to the 
	 * one that logically follows a
	 */
	if ( !gottarget ) {
	    ptarget = p;
	    qtarget = q;
	    if ( p->metric < a->metric )
		gottarget = 1;
	}

	if ( gotnext && ptarget )	/* quit when all pointers are found */
	    break;
    }

    /* a is sorted properly */
    if ( a == ptarget || a->next == ptarget )
	return;

    /* if a is equal to its successor in the list, it is (again)
     * sorted properly
     */
    if ( beforea && beforea->metric == a->metric )
	return;

    /* remove a */
    if ( a == *head )
	*head = (*head)->next;
    else
	beforea->next = a->next;

    /* put a back */
    a->next = ptarget;

    if ( qtarget == a )		/* beware of previous list state */
	qtarget = beforea;
    if ( !qtarget ) {
	a->next = *head;
	*head = a;
    }
    else
	qtarget->next = a;

}  /* end sr_hagentlist_sort() */

 
/*
 * sr_hagentlist_delete
 *	delete a list entry.  don't free it if it represents the connected
 *	agent.  NOTE! if the calling code intends to delete the entire
 *	contents of an agent/node list, be sure that the current agent is
 *	zero'd out in the MN state before traversing the list(s).
 *
 * returns:
 *	1 if list entry could not be deleted because it is the current agent
 *	0 otherwise (success)
 *
 * returned indirectly:
 *	**head - reset if it was pointing to deleted item
 */
int
sr_hagentlist_delete(struct SR_hlist **head, struct in_addr bip)
{
	struct SR_hlist *a = sr_hagentlist_search(*head, bip);
	int rval = 0;

	if (a == NULL)
		return rval;
	if (*head == a) {
		*head = (*head)->next;
	}
	else {
		prev->next = a->next;
	}
	srhagentcount--;

	a->next = NULL;
	free(a);
	return rval;
}

static struct SR_hlist *whead;

void
sr_hagentlist_initwalk(struct SR_hlist *head)
{
	whead = head;
}

struct SR_hlist *
sr_hagentlist_nextwalk()
{
	struct SR_hlist *cur = whead;
	if (whead == NULL)
		return(NULL);
	whead = whead->next;
	return(cur);
}

void
sr_hagentlist_dump(struct SR_hlist *head)
{
	int i = 0;
	sr_hagentlist_initwalk(head);
	printf("\nSR h agent list\n");
	printf("============================\n");
	printf("lifetime\tdestination\tnext hop\tmatric\tisAgent\n");  

	while(1) {
		head = sr_hagentlist_nextwalk();
		if (!head)
			return;
		printf("---------------------\n");
		srhlist_dumpone(head);
		printf("---------------------\n");
	}
}

#endif
