
/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifdef MADRP

/*
 * srrecvlist.c
 *
 * srrecv list module.  Implement a linked list of srrecv
 * entries to be used by hosts.  The list head
 * is maintained by the caller.
 *
 * Module should hide how list is constructed from callers.
 * XXX. Current search mechanism is simple linear search.  Should
 * eventually be replaced with hash search on ip address.
 *
 * XXX. more important, sort entries according to delta lifetimes
 * ala callout list.
 */

#include "sr_mad.h"
struct SR_recv  *srrecvlist;

/* create an entry and insert the node
 *
 * returns:
 *	struct *srrecvlist - pointer to srrecvlist entry if found
 *	NULL - out of buffer space
*/
struct SR_recv *
srrecvlist_create()
{
	struct SR_recv *n = NULL;
	struct SR_recv *v = malloc(sizeof(struct SR_recv));
	if (!v) {
		return(NULL);
	}
	bzero(v, sizeof(struct SR_recv));

	/* for now just put at head of list
	*/
	if (srrecvlist)
		n = srrecvlist;
	srrecvlist = v;
	v->next = n;
		
	return(v);
}
/* 
 * search srrecvlist specified to see if mn is in list.
 *    
 * returns: 
 *      struct vlist * - pointer to vlist entry if found
 *      NULL - no such entry.
 */
       
struct SR_recv *
srrecvlist_search( struct in_addr mn){
	struct SR_recv * head=srrecvlist;
	while(head) {
		if (((unsigned long)head->ip_target.s_addr)
		       == ((unsigned long)mn.s_addr))
		       return(head);
	  	head = head->next;
  	}
	return NULL;
}

void tick_clear_recvlist(){

	struct SR_recv *ss, * prev=NULL;
	ss=srrecvlist;
	while (ss !=NULL){
	        if (--ss->lifetime <=0) {
			if(prev==NULL) {
				srrecvlist=ss->next;
				free(ss);
				ss=srrecvlist;
			} else {
				prev->next=ss->next;
				free(ss);
				ss=prev->next;
			}
		} else {
			prev=ss;
			ss=ss->next;
		}
	}
}

static struct SR_recv *whead;

void
srrecvlist_initwalk()
{
	whead = srrecvlist;
}

struct SR_recv *
srrecvlist_nextwalk()
{
	struct SR_recv *cur = whead;
	if (whead == NULL)
		return(NULL);
	whead = whead->next;
	return(cur);
}

void
srrecvlist_dump()
{
	struct SR_recv *head;
	int i = 0;
        printf("\nSR recv list\n");
	printf("============================\n");
	srrecvlist_initwalk();
	while(1) {
		head = srrecvlist_nextwalk();
		if (!head)
			return;
		printf("---------------------\n");
		srrecvlist_dumpone(head, 1);
		printf("---------------------\n");
	}
}


/*
 * print one srrecvlist entry out
 */
int
srrecvlist_dumpone(struct SR_recv *v)
{
	int j;
        printf("request ID : %u+%u\n",v->requestId_secs,v->requestId_fract);
     	printf("initiator: %s\t", inet_ntoa(v->ip_initiator)); 
     	printf("target: %s \t",inet_ntoa(v->ip_target));
     	printf("replier:%s \n",inet_ntoa(v->ip_replier));
     	printf("origial lifetme: %d\t\t",v->orig_lifetime); 
        printf("lifetime:%d\t\t",v->lifetime);
     	printf("tuplecount: %d\n",v->tuplecount);
     	printf("===ip addr\t\tsignal strength\t\tpowerstrength====\n");
	for (j=0;j<v->tuplecount;j++){
	    printf("   %s\t%d\t\t%d\n",
		    inet_ntoa(v->tuple[j].ipaddr),
		    v->tuple[j].signalstrength,
		    v->tuple[j].powerstrength);
	}
}

#ifdef TESTSRSENT
/*
 * test all primitives above
 */
main()
{
	struct SR_recv *myhead = NULL;
	struct SR_recv *myhead2 = NULL;
	struct in_addr mn;
	struct SR_recv *v;

	/* test case 1.  put 5 entries in list and dump
	*/
	mn.s_addr = 0x01010101;
	srrecvlist_create(&myhead, mn);
	mn.s_addr = 0x02020202;
	srrecvlist_create(&myhead, mn);
	mn.s_addr = 0x03030303;
	srrecvlist_create(&myhead, mn);
	mn.s_addr = 0x04040404;
	srrecvlist_create(&myhead, mn);
	mn.s_addr = 0x05050505;
	srrecvlist_create(&myhead, mn);

	srrecvlist_dump(myhead); 

	/* now find all the list entries
	 * but first search for one that is not there.
	 */
	mn.s_addr = 0x01020304;
	v = srrecvlist_search(myhead, mn);
	if (v == NULL) 
		printf("not found, correct!\n");
	else
		printf("found %x\n", v->mn.s_addr);

	mn.s_addr = 0x01010101;
	v = srrecvlist_search(myhead, mn);
	if (v == NULL) 
		printf("not found, incorrect!\n");
	else
		printf("found %x\n", v->mn.s_addr);

	mn.s_addr = 0x05050505;
	v = srrecvlist_search(myhead, mn);
	if (v == NULL) 
		printf("not found, incorrect!\n");
	else
		printf("found %x\n", v->mn.s_addr);

	mn.s_addr = 0x03030303;
	v = srrecvlist_search(myhead, mn);
	if (v == NULL) 
		printf("not found, incorrect!\n");
	else
		printf("found %x\n", v->mn.s_addr);

	/* test series 2.  put 1 entry in list and delete it
	*/

	mn.s_addr = 0x01010101;
	srrecvlist_create(&myhead2, mn);
	srrecvlist_delete(&myhead2, mn);
	srrecvlist_dump(myhead2); 
	v = srrecvlist_search(myhead2, mn);
	if (v == NULL) 
		printf("correct - didn't find it\n");
		
	/* test series 3.  put 2 entries in list and delete first
	*/
	mn.s_addr = 0x02020202;
	srrecvlist_create(&myhead2, mn);
	mn.s_addr = 0x01010101;
	srrecvlist_create(&myhead2, mn);
	srrecvlist_delete(&myhead2, mn);
	printf("list should only have 02020202 in it\n");
	srrecvlist_dump(myhead2); 
	mn.s_addr = 0x02020202;
	v = srrecvlist_search(myhead2, mn);
	if (v == NULL) 
		printf("incorrect - didn't find it\n");
	else
		printf("found %x\n", v->mn.s_addr);

	/* test series 4. delete 3rd entry in original list
	*/
	mn.s_addr = 0x03030303;
	srrecvlist_delete(&myhead, mn);
	printf("following lisit should have 5/4/2/1 in that order\n");
	srrecvlist_dump(myhead); 

	/* test series 5. delete 1st and last entries in list
	*/
	mn.s_addr = 0x01010101;
	srrecvlist_delete(&myhead, mn);
	mn.s_addr = 0x05050505;
	srrecvlist_delete(&myhead, mn);
	printf("following list should only have 4/2 in it\n");
	srrecvlist_dump(myhead); 

	/* final test. put entry back in and print out.
	 * list should have 3/4/2 in that order
	 */
	mn.s_addr = 0x03030303;
	srrecvlist_create(&myhead, mn);
	printf("following list should have 3/4/2 in that order\n");
	srrecvlist_dump(myhead); 
}
#endif

#ifdef TESTDUMP
main()
{
	struct SR_recv *head = NULL;
	struct SR_recv *v;
	struct in_addr mn;
	struct in_addr coa;

	mn.s_addr = 0xc1d1e1f1;
	v = srrecvlist_create(&head, mn);
	v->vstate = 0;
	strncpy(v->location, "my pc", 5);
	v->lsize = 5;
	v->lifetime = 1;
	v->coa.s_addr = 0x01020304;
	v->auth_type = 0;

	mn.s_addr = 0x02020202;
	v = srrecvlist_create(&head, mn);
	v->vstate = 1;
	strncpy(v->location, "your pc", 7);
	v->lsize = 7;
	v->coa.s_addr = 0xcccd02ce;
	v->auth_type = 0x1;
	v->lifetime = 2;

	mn.s_addr = 0x03030303;
	v = srrecvlist_create(&head, mn);
	v->vstate = 2;
	strncpy(v->location, "bobs pc", 7);
	v->lsize = 7;
	v->coa.s_addr = 0xacbd02cd;
	v->auth_type = 0x2;
	v->lifetime = 3;

	mn.s_addr = 0x04040404;
	v = srrecvlist_create(&head, mn);
	v->vstate = 3;
	strncpy(v->location, "tena pc", 7);
	v->lsize = 7;
	v->coa.s_addr = 0xcccccc12;
	v->auth_type = 0x3;
	v->lifetime = 4;

	mn.s_addr = 0x05050505;
	v = srrecvlist_create(&head, mn);
	v->vstate = 4;
	strncpy(v->location, "xyxz pc", 7);
	v->lsize = 7;
	v->coa.s_addr = 0xcccccc11;
	v->auth_type = 0x0;
	v->lifetime = 5;

	srrecvlist_shortHeader();
	srrecvlist_initwalk(head);
	while(1) {
		head = srrecvlist_nextwalk();
		if (!head)
			return;
		srrecvlist_dumpone(head, 0);
	}
}
#endif
#endif
