
/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifdef MADRP

/*
 * srsentlist.c
 *
 * srsent list module.  Implement a linked list of srsent
 * entries to be used by hosts.  The list head
 * is maintained by the caller.
 *
 * Module should hide how list is constructed from callers.
 * XXX. Current search mechanism is simple linear search.  Should
 * eventually be replaced with hash search on ip address.
 *
 * XXX. more important, sort entries according to delta lifetimes
 * ala callout list.
 */

#include "sr_mad.h"
struct SR_sent  *srsentlist;

/* create an entry and insert the node
 *
 * returns:
 *	struct *srsentlist - pointer to srsentlist entry if found
 *	NULL - out of buffer space
*/
struct SR_sent *
srsentlist_create()
{
	struct SR_sent *n = NULL;
	struct SR_sent *v = malloc(sizeof(struct SR_sent));
	if (!v) {
		return(NULL);
	}
	bzero(v, sizeof(struct SR_sent));

	/* for now just put at head of list
	*/
	if (srsentlist)
		n = srsentlist;
	srsentlist = v;
	v->next = n;
		
	return(v);
}
/* 
 * search vlist specified to see if mn is in list.
 *    
 * returns: 
 *      struct vlist * - pointer to vlist entry if found
 *      NULL - no such entry.
 */
       
struct SR_sent *
srsentlist_search( struct in_addr mn){
	struct SR_sent * head=srsentlist;
	while(head) {
		if (((unsigned long)head->ip_target.s_addr)
		       == ((unsigned long)mn.s_addr))
		       return(head);
	  	head = head->next;
  	}
	return NULL;
}

int tick_clear_sentlist(struct in_addr * tosend){

	struct SR_sent *ss, * prev=NULL;
	int k=0, delete;
	ss=srsentlist;
	while (ss !=NULL){
		delete=0;
		if( ss->replied_flag ) {
			delete=1;
		} else {
			ss->backoff_timer--;
			if (ss->backoff_timer <=0) {
				if (ss->backoff_count==(MAX_INDEX-1)) {
					delete=1;
				} else {
					if(is_link_route(ss->ip_target.s_addr)){
						delete=1;
					}
					else{
						tosend[k].s_addr=
							ss->ip_target.s_addr;
						k++;
					}
				}
			} 
		}
		if (delete) {
			if(prev==NULL) {
				srsentlist=ss->next;
				free(ss);
				ss=srsentlist;
			} else {
				prev->next=ss->next;
				free(ss);
				ss=prev->next;
			}
		} else {	
			prev=ss;
			ss=ss->next;
		}
	}
	return k;
}

static struct SR_sent *whead;

void
srsentlist_initwalk()
{
	whead = srsentlist;
}

struct SR_sent *
srsentlist_nextwalk()
{
	struct SR_sent *cur = whead;
	if (whead == NULL)
		return(NULL);
	whead = whead->next;
	return(cur);
}

void
srsentlist_dump()
{
	struct SR_sent *head;
	int i = 0;
        printf("\nSR sent list\n");
	printf("============================\n");
	printf("request-ID\t\tip-target\t\treplied\tcount\ttimer\n");
	srsentlist_initwalk();
	while(1) {
		head = srsentlist_nextwalk();
		if (!head)
			return;
		printf("---------------------\n");
		srsentlist_dumpone(head, 1);
		printf("---------------------\n");
	}
}


/*
 * print one srsentlist entry out
 */
int
srsentlist_dumpone(struct SR_sent *v)
{
	int i;
        printf("%u+%u\t",v->requestId_secs,
	     v->requestId_fract);
	printf("%s\t\t",inet_ntoa(v->ip_target));
     	printf("%d\t",v->replied_flag);
    	printf("%d\t",v->backoff_count);
     	printf("%d\n",v->backoff_timer);
}

#ifdef TESTSRSENT
/*
 * test all primitives above
 */
main()
{
	struct SR_sent *myhead = NULL;
	struct SR_sent *myhead2 = NULL;
	struct in_addr mn;
	struct SR_sent *v;

	/* test case 1.  put 5 entries in list and dump
	*/
	mn.s_addr = 0x01010101;
	srsentlist_create(&myhead, mn);
	mn.s_addr = 0x02020202;
	srsentlist_create(&myhead, mn);
	mn.s_addr = 0x03030303;
	srsentlist_create(&myhead, mn);
	mn.s_addr = 0x04040404;
	srsentlist_create(&myhead, mn);
	mn.s_addr = 0x05050505;
	srsentlist_create(&myhead, mn);

	srsentlist_dump(myhead); 

	/* now find all the list entries
	 * but first search for one that is not there.
	 */
	mn.s_addr = 0x01020304;
	v = srsentlist_search(myhead, mn);
	if (v == NULL) 
		printf("not found, correct!\n");
	else
		printf("found %x\n", v->mn.s_addr);

	mn.s_addr = 0x01010101;
	v = srsentlist_search(myhead, mn);
	if (v == NULL) 
		printf("not found, incorrect!\n");
	else
		printf("found %x\n", v->mn.s_addr);

	mn.s_addr = 0x05050505;
	v = srsentlist_search(myhead, mn);
	if (v == NULL) 
		printf("not found, incorrect!\n");
	else
		printf("found %x\n", v->mn.s_addr);

	mn.s_addr = 0x03030303;
	v = srsentlist_search(myhead, mn);
	if (v == NULL) 
		printf("not found, incorrect!\n");
	else
		printf("found %x\n", v->mn.s_addr);

	/* test series 2.  put 1 entry in list and delete it
	*/

	mn.s_addr = 0x01010101;
	srsentlist_create(&myhead2, mn);
	srsentlist_delete(&myhead2, mn);
	srsentlist_dump(myhead2); 
	v = srsentlist_search(myhead2, mn);
	if (v == NULL) 
		printf("correct - didn't find it\n");
		
	/* test series 3.  put 2 entries in list and delete first
	*/
	mn.s_addr = 0x02020202;
	srsentlist_create(&myhead2, mn);
	mn.s_addr = 0x01010101;
	srsentlist_create(&myhead2, mn);
	srsentlist_delete(&myhead2, mn);
	printf("list should only have 02020202 in it\n");
	srsentlist_dump(myhead2); 
	mn.s_addr = 0x02020202;
	v = srsentlist_search(myhead2, mn);
	if (v == NULL) 
		printf("incorrect - didn't find it\n");
	else
		printf("found %x\n", v->mn.s_addr);

	/* test series 4. delete 3rd entry in original list
	*/
	mn.s_addr = 0x03030303;
	srsentlist_delete(&myhead, mn);
	printf("following lisit should have 5/4/2/1 in that order\n");
	srsentlist_dump(myhead); 

	/* test series 5. delete 1st and last entries in list
	*/
	mn.s_addr = 0x01010101;
	srsentlist_delete(&myhead, mn);
	mn.s_addr = 0x05050505;
	srsentlist_delete(&myhead, mn);
	printf("following list should only have 4/2 in it\n");
	srsentlist_dump(myhead); 

	/* final test. put entry back in and print out.
	 * list should have 3/4/2 in that order
	 */
	mn.s_addr = 0x03030303;
	srsentlist_create(&myhead, mn);
	printf("following list should have 3/4/2 in that order\n");
	srsentlist_dump(myhead); 
}
#endif

#ifdef TESTDUMP
main()
{
	struct SR_sent *head = NULL;
	struct SR_sent *v;
	struct in_addr mn;
	struct in_addr coa;

	mn.s_addr = 0xc1d1e1f1;
	v = srsentlist_create(&head, mn);
	v->vstate = 0;
	strncpy(v->location, "my pc", 5);
	v->lsize = 5;
	v->lifetime = 1;
	v->coa.s_addr = 0x01020304;
	v->auth_type = 0;

	mn.s_addr = 0x02020202;
	v = srsentlist_create(&head, mn);
	v->vstate = 1;
	strncpy(v->location, "your pc", 7);
	v->lsize = 7;
	v->coa.s_addr = 0xcccd02ce;
	v->auth_type = 0x1;
	v->lifetime = 2;

	mn.s_addr = 0x03030303;
	v = srsentlist_create(&head, mn);
	v->vstate = 2;
	strncpy(v->location, "bobs pc", 7);
	v->lsize = 7;
	v->coa.s_addr = 0xacbd02cd;
	v->auth_type = 0x2;
	v->lifetime = 3;

	mn.s_addr = 0x04040404;
	v = srsentlist_create(&head, mn);
	v->vstate = 3;
	strncpy(v->location, "tena pc", 7);
	v->lsize = 7;
	v->coa.s_addr = 0xcccccc12;
	v->auth_type = 0x3;
	v->lifetime = 4;

	mn.s_addr = 0x05050505;
	v = srsentlist_create(&head, mn);
	v->vstate = 4;
	strncpy(v->location, "xyxz pc", 7);
	v->lsize = 7;
	v->coa.s_addr = 0xcccccc11;
	v->auth_type = 0x0;
	v->lifetime = 5;

	srsentlist_shortHeader();
	srsentlist_initwalk(head);
	while(1) {
		head = srsentlist_nextwalk();
		if (!head)
			return;
		srsentlist_dumpone(head, 0);
	}
}
#endif
#endif
