
/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * mip_fastate.c 
 *
 * state machine routines for fa 
 * 
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/times.h>
#include <errno.h>
#include <signal.h>
#include <string.h>

#include "mipconfig.h"
#include "mipstate.h"
#include "udp_reg.h"
#include "faha_pinfo.h"
#include "vlist.h"
#include "mipstats.h"
#include "unixsock.h"
#include "mip_log.h"

extern struct fa_stats fa_stats;
extern int debug; 
void
faUdp(struct mip_ifs *, u_long *, struct sockaddr_in *, struct packet_info *);

void
faUdp_pending(struct mip_ifs *, struct vlist *, u_long *, 
	struct sockaddr_in *, struct packet_info *);
void
faUdp_present(struct mip_ifs *, struct vlist *, u_long *, 
	struct sockaddr_in *, struct packet_info *);
void
faUdp_forward(struct mip_ifs *, struct vlist *, u_long *, 
	struct sockaddr_in *, struct packet_info *);

void
storeMd5(char *, int , unsigned char *, unsigned char *);

void
computeMd5(char *, int, char *, char *);

void faTick();

void haTick();

void
sendOneUdp(int, char *, int , struct in_addr, u_short);

static void
sendUdpErrorReply (struct sockaddr_in *, struct packet_info *, int);

char *reverseAddress();

/* 
 * visitor list for fa
 */

struct vlist *fahead = NULL;

extern int udpsock;

/* udp registration from MN to HA or udp reply from HA to MN
 * with FA in the middle.
 */
void
faUdp(struct mip_ifs *mifs, u_long *if_addr, struct sockaddr_in *from, struct packet_info *pinfo)
{
	struct vlist *v;

	/* security check.  if on this interface we are not doing
	 * arp and we don't already have a cached mac address (we have
	 * to get a beacon before we get a udp registration msg),
	 * then this packet is toast.  We will only have a cached
	 * mac address if network authentication is being used
	 * and the cache entry has been made already.
	 * This prevents the FA from forwarding UDP registrations
	 * when net. authentication is used and an MN lacking
	 * such authentication presents the FA with a regisration packet.
	 */
	if (pinfo->reg && mifs->doArp == FALSE) {
		v = vlist_search(fahead, pinfo->reg->home_address);
		if (!v) {
#ifdef MADRP
			if (!srhlist_search(pinfo->reg->home_address))
#endif
			{
#ifdef DEBUG
				if(debug)
				printf("can't find %s in hlist/vlist\n",
					inet_ntoa(pinfo->reg->home_address));
#endif
				fa_stats.nonetauth++;
				return;
			}
		}
		/* sanity invariant check. make sure auth. indicated
		*/
		if (v && v->auth_type == 0 
#ifdef MADRP
		      && !srhlist_search(pinfo->reg->home_address)
#endif
		   ) {
#ifdef DEBUG
			if(debug)
			printf("v->auth_type=0 so return \n");
#endif
			fa_stats.nonetauth++;
			return;
		}
	}
	
	/* it's a registration packet, presumably from MN to HA
	*/
	if (pinfo->reg) {
		u_short lifetime;
		fa_stats.requests++;
		/* make various checks
		 * 1. flags - say no to things that this FA doesn't do
		 * 2. lifetime
		 * 3. reserved field non-zero
		 */
		if (pinfo->reg->flags) {
			if (pinfo->reg->flags & M_BIT) {
				sendUdpErrorReply (from, pinfo, 
					REG_FAIL_FA_ENCAPS);
				return;
			}
			if (pinfo->reg->flags & G_BIT) {
				sendUdpErrorReply (from, pinfo, 
					REG_FAIL_FA_ENCAPS);
				return;
			}
			if (pinfo->reg->flags & V_BIT) {
				sendUdpErrorReply (from, pinfo, 
					REG_FAIL_FA_VJ);
				return;
			}
		}
		lifetime = ntohs(pinfo->reg->lifetime);
		if (lifetime > mip_config.mobileLifetime) {
			fa_stats.badlifetime++;  
			mip_log( MLOG_SEC, "FA: Bad lifetime value %d from MN %s(%s)",
				lifetime,
				reverseAddress(from->sin_addr),
				inet_ntoa(from->sin_addr));
			sendUdpErrorReply (from, pinfo, REG_FAIL_FA_LIFETIME);
			return;
		}
		/* search the vlist to see if we already have one
		*/
		v = vlist_search(fahead, pinfo->reg->home_address);
		/* no such beast, make one
		*/
		if (v == NULL ){
			v = vlist_create(&fahead, pinfo->reg->home_address);
			if (v == NULL) {
				fatal("faUdp(): out of memory in vlist");
			}
#ifdef MADRP
			if (srhlist_search(pinfo->reg->home_address)){
				v->routeset = ROUTEHOST;
			}
#endif
			fa_stats.visitors++;
			mip_log( MLOG_INFO, "faUdp(): new MN %s(%s) in visitor list",
				reverseAddress(from->sin_addr),
				inet_ntoa(from->sin_addr));
			/* pending means new; i.e., we have not seen
			 * a HA roundtrip message yet
			 */
			faUdp_pending(mifs, v, if_addr, from, pinfo);
		}
		/* we have an entry already */
		else {
			/* BEACON_NOSTATE goes to PENDING
			*/
			if (v->vstate == BEACON_NOSTATE) {
				fa_stats.visitors++;
				mip_log( MLOG_INFO, "faUdp(): new MN %s(%s) in visitor list",
					reverseAddress(from->sin_addr),
					inet_ntoa(from->sin_addr));
				faUdp_pending(mifs, v, if_addr, from, pinfo);
			}
			else {
				faUdp_present(mifs, v, if_addr, from, pinfo);
			}
		}
	}
	/* its a reply packet, presumably from HA to MN
	*/
	else if (pinfo->rep) {
		fa_stats.replies++;
		/*
		mip_log( MLOG_INFO, "faUdp(): received reply packet from ip=%s",
				inet_ntoa(from->sin_addr));	
		*/

		/* search the vlist to see if we already have one
		*/
		v = vlist_search(fahead, pinfo->rep->home_address);
		/* we don't have an entry. presume the MN has split.
		 * make a stat entry and drop the packet
		*/
		if (v == NULL) {
			fa_stats.replydropped++;
			/*
			mip_log( MLOG_INFO, "faUdp(): dropped reply as no visitor list entry");
			*/
			return;
		}
		/* see if the lower 32 bits of the incoming ident
		 * matches the visitor list ident, if not, drop it
		 */
		if (pinfo->rep->i.l[1] != v->i.l[1] ) {
			fa_stats.identmismatch++;
			/*
			mip_log( MLOG_SEC, "faUdp(): vlist exists, but identity mismatch for some MN");
			*/
			return;
		}
		/* 
		 * forward the packet back to the mn
		 */
		faUdp_forward(mifs, v, if_addr, from, pinfo);
	}
}





/*
 * STATE: NONE->PENDING
 *	  BEACON_ZERO->PENDING
 *
 * packet from MN sent to HA. 
 *
 * assumption is that we can store all info and forward packet
 * to HA without doing routing table updates for MN.
 * those are not done until we hear back from the HA.
 *
 * pending state indicates we have never heard back from the HA yet
 * even once.
 */
void
faUdp_pending(struct mip_ifs *mifs, struct vlist *v, u_long *if_addr, 
	struct sockaddr_in *from, struct packet_info *pinfo)
{
	int psize;
	int rc;
	struct mse_authext *mse = NULL;
	int spi;

	if (mifs == NULL) {
		mip_log( MLOG_ERR, "faUdp_pending(): mifs was NULL");
		exit(1);
	}
	/* store info in new visiting list entry
	*/
	v->vstate = FA_PENDING;
	v->i_f.s_addr = *if_addr;	/* which interface */
	v->vmifs = mifs;		/* pointer to that interface */
	v->udp_src = *from;
	v->udp_port = from->sin_port;  /* store port explicitly */
	v->ha = pinfo->reg->home_agent;
	v->mn = pinfo->reg->home_address;
	/* TBD. check lifetime validity. for now
	 * just be uncritical.
	 */
	v->orig_lifetime = ntohs(pinfo->reg->lifetime);
	v->lifetime = mip_config.mobileLifetime;
	v->i.l[0] = pinfo->reg->i.l[0];
	v->i.l[1] = pinfo->reg->i.l[1];

	/* forward packet to HA
	 * packet dst is pinfo->reg->home_agent
	 * packet contents consists of:
 	 *	. registration packet
	 *	. any pre non-auth mse bits
	 *	. mn-ha mse
	 * if we have no mse relationship with ha, we
	 * can simply forward the above, else we must copy
	 * the bits after the packet and send it on.
	 */
	psize = sizeof(struct udp_register) + pinfo->premnha_size + 
			pinfo->mn_ha_auth_size;
	fa_stats.pending++;
	/*
	mip_log( MLOG_INFO, "faUdp_pending(): new MN ip=%s in pending list",
		inet_ntoa(v->mn));
	*/
	if (mip_config.faha_security && pinfo->mn_ha_auth) {
		u_long	ha_keysize, ha_spi, ha_transform, dummy; 
		int dummy2;
		char	*ha_key;

		/* construct, append, and secure the fa mse
		 * to append what is needed AFTER the mn_ha mse
		 */
                mse = (struct mse_authext *)
                        (((char *)pinfo->packet) +  psize);
                mse->type = FA_HA_AUTHENTICATOR_TYPE;
                mse->length = 4 + 16;

		ha_key = get_key(
			mip_config.sa[FAHASECURITY].kl,
			(unsigned long) pinfo->reg->home_agent.s_addr,  
			(int *) &ha_keysize,   
			&ha_spi,       
			(int *) &ha_transform,
			&dummy, &dummy2); 

		if (ha_key == NULL) {
			mip_log(MLOG_SEC, "FA does not have FAHA key for HA (%s).", 
				inet_ntoa(pinfo->reg->home_agent));
			fa_stats.visitors--;		/* incremented before this fn called */
			/* vlist_delete(&fahead, v->mn); 
				-- bad idea; unlikely case, unless MN is under attack */
			return;
		}

		spi = htonl(ha_spi);
                bcopy(&spi, &mse->spi, sizeof(int));
                storeMd5((char *)pinfo->packet, psize + AUTHEXT_HASHEDBYTES,
			mse->hash, ha_key);
                psize += mse->length + 2;
	}

	/* pkt to HA
	*/
	sendOneUdp(udpsock, pinfo->packet, 
			psize , v->ha, htons(UDP_PORT));
}





/*
 * we already had a list entry and we got a registration
 * packet from the MN.  update vlist entry and send packet to HA.
 *
 * We do NOT change the state.  if it was pending, it still is.
 * Hence the name of this function is a bit of a misnomer.
 * The state is changed only when we get an ack from the HA.
 * Note that we do NOT update the vlist timer either.  
 *
 * A tricky mobile host might have switched interfaces on us by
 * the way.  If so, we send it back to PENDING state.
 * TBD: there is a potential denial of service attack here.
 * the only way it can be dealt with is by either assuming
 * MN/FA authentication or by waiting for the vlist entry
 * on an old i/f to timeout.  This problem has not been dealt with
 * as we hoping for divine (or even good) inspiration.
 *
 */
void
faUdp_present(struct mip_ifs *mifs, struct vlist *v, u_long *if_addr, 
	struct sockaddr_in *from, struct packet_info *pinfo)
{
	int psize;
	int rc;
	struct mse_authext *mse = NULL;
	int spi;

	/* TBD. check lifetime validity. for now
	 * just be uncritical.  
	if (ntohs(pinfo->reg->lifetime) > mip_config.mobileLifetime) {
	}
	 */

	/* mobile host changed interface.  has to wait for ack to come
	 * back but we must note that change here.  we return it
	 * to pending state, if it wasn't in pending state.
	 */ 
	if (*if_addr != v->i_f.s_addr) {
		char buf[1024];
		char buf2[1024];
		struct in_addr t;
		t.s_addr = *if_addr;

		sprintf(buf, "if_addr %s ", inet_ntoa(t));
		sprintf(buf2, "vif addr %s ", inet_ntoa(v->i_f));

		/* if not pending state, mn has a route, which
		 * is now wrong.  we zap the route and put
		 * mn back in pending state.
		 *
		 * This is slightly wrong, however, even if we're authenticated,
		 * as we are throwing away an existing valid registration before
		 * the new one has been completed.  Alternative:  Add the new
		 * link layer route, don't go into PENDING.
		 */
		if (v->vstate != FA_PENDING) {
			char *mac = NULL;

			if (v->sawMac)
				mac = v->mac;
			if (v->routeset  == ROUTELINK)
			{
				route_delete_link(v->mn, v->mac);
				v->routeset = 0;
			}
		}
		fa_stats.pending--;
		mip_log( MLOG_INFO, 
			"faUdp_present(): returned MN (%s) to pending state as interface changed (%s -> %s)",
			inet_ntoa(v->mn), buf, buf2);
		faUdp_pending(mifs, v, if_addr, from, pinfo);
		return;
	}
	 
	/* update fields that FA might care about
	*/
	v->ha = pinfo->reg->home_agent;
	v->mn = pinfo->reg->home_address;
	v->i.l[0] = pinfo->reg->i.l[0];
	v->i.l[1] = pinfo->reg->i.l[1];
	v->udp_src = *from;
	v->udp_port = from->sin_port;  /* in case mn changes port */

	/* forward packet to HA
	 * packet dst is pinfo->reg->home_agent
	 * packet contents consists of:
 	 *	. registration packet
	 *	. any pre non-auth mse bits
	 *	. mn-ha mse
	 * if we have no mse relationship with ha, we
	 * can simply forward the above, else we must copy
	 * the bits after the packet and send it on.
	 */
	psize = sizeof(struct udp_register) + pinfo->premnha_size + 
			pinfo->mn_ha_auth_size;
	if (mip_config.faha_security && pinfo->mn_ha_auth) {
		u_long	ha_keysize, ha_spi, ha_transform, dummy;
		int dummy2;
		char	*ha_key;

		/* construct, append, and secure the fa mse
		 * to append what is needed AFTER the mn_ha mse
		 */
                mse = (struct mse_authext *)
                        (((char *)pinfo->packet) +  psize);
                mse->type = FA_HA_AUTHENTICATOR_TYPE;
                mse->length = 4 + 16;

		/* lookup key.  index is home agent (MN) out of
		 * keylist. note that the ip addr came from the udp packet.
		 */
		ha_key = get_key(
			mip_config.sa[FAHASECURITY].kl,
			(unsigned long) pinfo->reg->home_agent.s_addr,  
			(int *) &ha_keysize,   
			&ha_spi,       
			(int *) &ha_transform,
			&dummy, &dummy2); 

		if (ha_key == NULL) {
			mip_log(MLOG_SEC, "FA does not have FAHA key for HA (%s).", 
				inet_ntoa(pinfo->reg->home_agent));
			/* XXX statistics to adjust? */
			return;
		}

		spi = htonl(ha_spi);
                bcopy(&spi, &mse->spi, sizeof(int));
		storeMd5((char *)pinfo->packet, psize + AUTHEXT_HASHEDBYTES,
			mse->hash, ha_key);
                psize += mse->length + 2;
	}

	/* pkt to HA */
	sendOneUdp(udpsock, pinfo->packet, psize , 
		v->ha, htons(UDP_PORT));
}





/*
 * fa got reply packet from ha and wishes to forward it
 * back to mn that it knows is present.
 *
 * STATE: -> FA_PRESENT
 *
 * note: mifs pointer here is probably not useful.
 */
void
faUdp_forward(struct mip_ifs *mifs, struct vlist *v, u_long *if_addr, 
	struct sockaddr_in *from, struct packet_info *pinfo)
{
	int psize;
	int rc;
	extern int debug;
	int ltime;
	struct mse_authext *mse = NULL;
	int spi;
	
	/* reset lifetime.
	 */
	v->lifetime = mip_config.mobileLifetime;

	/* forward packet to MN
	 * packet dst is v->mn
	 * packet contents consists of:
 	 *	. reply packet
	 *	. any pre non-auth mse bits
	 *	. mn-ha mse
	 * plus what fa glues on if desired.
	 * if we have no mse relationship with mn, we
	 * can simply forward the above, else we must copy
	 * the bits after the packet and send it on.
	 */
	psize = sizeof(struct udp_reply) + pinfo->premnha_size + 
			pinfo->mn_ha_auth_size;
	
	if (debug) {
		dumpUdpReply((struct udp_reply *) pinfo->packet);
	}

	if (mip_config.famn_security && pinfo->mn_ha_auth) {
		u_long	mn_keysize, mn_spi, mn_transform, dummy;
		int dummy2;
		char	*mn_key;

		/* construct, append, and secure the fa mse
		 * to append what is needed AFTER the mn_ha mse
		 */
                mse = (struct mse_authext *)
                        (((char *)pinfo->packet) +  psize);
                mse->type = MN_FA_AUTHENTICATOR_TYPE;
                mse->length = 4 + 16;

		mn_key = get_key(
			mip_config.sa[FAMNSECURITY].kl,
			(unsigned long) pinfo->rep->home_address.s_addr,
			(int *) &mn_keysize,   
			&mn_spi,       
			(int *) &mn_transform,
			&dummy, &dummy2); 

		if (mn_key == NULL) {
			/* How can this be?  We forwarded the request from the MN,
			 * but now we don't have a key to forward the response?
			 * Strange.... */
			mip_log(MLOG_SEC, "FA does not have MNFA key for MN (%s).", 
				inet_ntoa(v->mn));
			/* vlist_delete(&fahead, v->mn); */
			return;
		}

		spi = htonl(mn_spi);
                bcopy(&spi, &mse->spi, sizeof(int));
                storeMd5((char *)pinfo->packet, psize + AUTHEXT_HASHEDBYTES,
			mse->hash, mn_key);
		if (debug) {
			printf("famn mse follows: size %d\n", psize);
			dumpMse(mse);	
		}
                psize += mse->length + 2;
	}

	/* it was pending, then we must install route table entry
	 * BEFORE we try and send packet to mn, else talking to
	 * the poor devil will be tough.  However, we add the route as late
	 * as possible just in case there is something else wrong that prevents
	 * us from forwarding the reply to the mobile node.
	 */
	if (v->vstate == FA_PENDING)   {
		char *mac = NULL;

		if (v->sawMac) {
			mac = v->mac;	
		}
		route_add_link(v->mn, mac,
				v->vmifs->ifname );
		v->routeset = ROUTELINK;
		fa_stats.pending--;
	}
	v->vstate = FA_PRESENT;  /* change state */

	sendOneUdp(udpsock, pinfo->packet, psize , v->mn, v->udp_port);

/* this is out for now, because it seems that it would be too easy
 * for somebody to spoof the HA and set lifetime == 0 and simply
 * produce a very simple denial of service attack.
 */
#ifdef RETHINK
        /* if we lack a security relationship between the fa
         * and ha, we dare not make judgements about the packet
         * coming back as it would be too easy to spoof it
         */
        if (!mip_config.faha_security) {
                return;
        }

	/* if HA denied access to MN, we drop it from our visitor list
	 * note: we can only do this after we send on the packet,
	 * else we wouldn't have routing.
	 */
	ltime = ntohs(pinfo->rep->lifetime);
	if (ltime == 0 || (pinfo->rep->code >= 128 &&
		pinfo->rep->code <= 136)) {
		char habuf[1024];
		char mnbuf[1024];
		char mnname[1024];

		strncpy(habuf, inet_ntoa(from->sin_addr),  1024);
		strncpy(mnbuf, inet_ntoa(v->mn), 1024);
		strncpy(mnname, reverseAddress(v->mn), 1024);
		/* TBD. error codes should be mapped into strings
		*/
		mip_log( MLOG_SEC, "FA: HA %s(%s) denied access to MN %s(%s), deleting from FA visitor list, error code was: %d, lifetime %d", 
			reverseAddress(from->sin_addr), 
			habuf,
			mnname,
			mnbuf,
			pinfo->rep->code,
			ltime);

		fa_stats.hadenied++;
		vlist_delete(&fahead,v->mn);
		fa_stats.visitors--;
		fa_route_delete(v->mn, v->i_f);
	}
#endif
}





/* fa state machine tick
 *
 * TODO:
 *	implement fa_route_delete
 */
void
faTick()
{
	struct vlist *v;
	extern int debug;

	/* walk the vlist and decrement timer state,
	 * deleting list entries as appropriate
	 */
	vlist_initwalk(fahead);
	while (v = vlist_nextwalk()) {
		/* ignore infinity
		*/
		if (v->lifetime == 0xffff) {
			continue;
		}
		v->lifetime--;	
		/* if list timeout
		*/
		if (v->lifetime <= 0) {
			if (debug) {
				printf("faTick: timeout on %s\n",
					inet_ntoa(v->mn));
			}
			fa_stats.timeouts++;
			/* PENDING->NOSTATE
			 * all we have to do here is delete the
			 * list entry.  
			 */
			if (v->vstate == FA_PENDING) {
				vlist_delete(&fahead,v->mn);
				fa_stats.visitors--;
				mip_log( MLOG_INFO, "FA timeout: MN %s PENDING->NOSTATE, dropped from visitor list", inet_ntoa(v->mn) );
			}
			/* PRESENT->NOSTATE
			*/
			else if (v->vstate == FA_PRESENT) {
				/* TBD. delete host route entry
				 * including any arp entries grat/otherwise
				 */
				char *mac = NULL;

				vlist_delete(&fahead,v->mn);
				fa_stats.visitors--;
				if (v->sawMac){
					mac = v->mac;
					route_delete_link(v->mn, mac);
				}
				mip_log( MLOG_INFO, "FA: timeout: MN %s PRESENT->NOSTATE, dropped from visitor list", inet_ntoa(v->mn) );
			}
                        else if (v->vstate == BEACON_NOSTATE ) {
                                vlist_delete(&fahead,v->mn);
                                fa_stats.visitors--;
                                {
                                        char *mac = NULL;
                                        if (v->sawMac){
                                                mac = v->mac;
                                        	route_delete_link(v->mn, mac);
					}
                                }
                                mip_log( MLOG_INFO, "FA: timeout: MN %s in BEACON_NOSTATE, dropped from visitor list", inet_ntoa(v->mn) );
                        }
		}
	}
}

/* wipe out all routes in fa "visitor list"
 */
void
faRouteWipe()
{
	struct vlist *v;
	extern int debug;

	/* not a fa, therefore get out of here
	*/
	if (mip_config.sawFa == 0) {
		return;
	}

	/* walk the vlist and wipe out route entripes
	 */
	vlist_initwalk(fahead);
	while (v = vlist_nextwalk()) {
		if (v->vstate == FA_PRESENT) {
			char *mac = NULL;
			if (v->sawMac){
				mac = v->mac;
				route_delete_link(v->mn, mac);
			}
		}
	}
}

/*
 * walk the vlist and
 *	1. send the count of vlist nodes AND then
 *	2. each vlist node
 *
 * to the process that requested the info via a unix socket
 */
fa_dumpVlist()
{
	struct vlist *v;
	int count = 0;

	vlist_initwalk(fahead);
	while (v = vlist_nextwalk()) {
		count++;
	}

	/* write the number of vlist entries 
	*/
	sendUnixMessage(MOBYSOCK_CLIENT, (char *)&count, sizeof(int));	

	/* write the vlist nodes
	*/
	vlist_initwalk(fahead);
	while (v = vlist_nextwalk()) {
		sendUnixMessage(MOBYSOCK_CLIENT, (char *)v,
						sizeof(struct vlist));
	}
}


/*
 * fa uses this function to send reply on error
 */
static
void
sendUdpErrorReply (struct sockaddr_in *from, struct packet_info *pinfo, 
	int code)
{
	char packet[1024];
	struct mse_authext *mse = NULL;
	int spi;
	extern int debug;
	struct udp_reply *rep = (struct udp_reply *) packet;
	int psize = sizeof(struct udp_reply);

	bzero(packet, 1024);
	rep->type = UDP_REPLY_TYPE;
	rep->code = (u_char) code;
	/* if failure due to lifetime, return acceptable lifetime
	*/
	if (code == REG_FAIL_FA_LIFETIME)
		rep->lifetime = htons(mip_config.mobileLifetime);

	rep->home_address = pinfo->reg->home_address;
	rep->home_agent = pinfo->reg->home_agent;
	rep->i.l[0] = pinfo->reg->i.l[0];
	rep->i.l[1] = pinfo->reg->i.l[1];

	/* append famn security if desired
	*/
	if (mip_config.famn_security && pinfo->mn_ha_auth) {
		mse = (struct mse_authext *)
			((char *)packet +  sizeof (struct udp_reply));
		mse->type = MN_FA_AUTHENTICATOR_TYPE;
		mse->length = 4 + 16;
		spi = htonl(mip_config.sa[FAMNSECURITY].spi); /* Zhong Chen */
		bcopy(&spi, &mse->spi, sizeof(int));
		storeMd5((char *)rep, psize + AUTHEXT_HASHEDBYTES,
			mse->hash, mip_config.sa[FAMNSECURITY].key);
		psize += mse->length + 2;
	}

	/* ship it
	*/
	if (debug) {
		dumpUdpReply((struct udp_reply *) packet);
		dumpMse(mse);
	}
	sendOneUdp(udpsock, packet, psize , from->sin_addr, from->sin_port);
}


struct vlist * vlist_getfahead(){ return fahead;}
struct vlist ** vlist_getfahead_pointer(){ return &fahead;}

