/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * mip_icmp.h
 *
 * logical structure if icmp/router advertisement for mobile ip is:
 *
 * ip | icmp_advert | mip_mse_icmp | (optional) mip_mse_icmp_info |
 *	(optional) mip_mse_icmp_auth
 *
 * The icmp_advert and mip_mse_icmp fields are required by mobile-ip.
 * The latter two are *features* and are optional.
 */

/*
 * Structure of an icmp router advertisement message .
 *
 * current size of structure: 16 bytes
 */
struct icmp_advert {
	u_char	icmp_type;		/* type of message, see below */
	u_char	icmp_code;		/* type sub code */
	u_short	icmp_cksum;		/* ones complement cksum of struct */
	u_char  icmp_naddrs;            /* number of addresses in message */
	u_char  icmp_asize;             /* address size, always 2 */
	/* mn can use this as one way to detect movement.
	 * if it does not recv. advert in this lifetime, then
	 * may no longer be talking to agent
	 */
	u_short icmp_adlifetime;        /* lifetime of advert in seconds */
/* 2 long words pairs of (router address, preference)
 * preference = 0 is default
 * This part can be variable length.
 */
	u_long  icmp_router;
        u_long  icmp_preference;
};

/*     
 * structure of mobile ip icmp mobile service extension
 */
struct mip_mseicmp {
/* mseicmp types
 *
 * current size of structure: 12 bytes
 */ 
#define MIPAGENTAD 16
	u_char  mse_type;	     /* type of mobile service extension */
	u_char  mse_len;	     /* length in bytes */
	u_short mse_sequence;        /* sequence number, count of adverts */
	/* longest time (0xffff is infinity) agent is willing to
	 * accept in UDP registration request.  Not the same as
	 * the icmp lifetime above.
	 */
	u_short mse_lifetime;	     /* list timeout */ 
	u_short mse_flags;           /* bit-field flags */
	u_long  mse_coa;             /* Care-Of Address, not sent for ha and mns */
};

/* psu has defined three extensions of our own
 *
 * MIPAGENTINFO - basically include a string that describes where
 * the agent is located. A mn that removes > 1 beacon from > 1 source
 * can tell the user where the beacons are coming from.
 *
 * MIPAGENTAUTH - last extension in packet.  Authenticates using
 * TBD (keyed md5 to start with) entire icmp packet (icmp + data payload).
 * icmp_key is used.  Used by both mnd and mipd.
 *
 * MIPMAC - used to put link layer mac address in beacon
 */

#define MIPAGENTINFO	128
struct mip_mse_icmp_info {
	u_char mse_type;
	u_char mse_len;
#define MAX_INFO_LEN   30
	u_char info[MAX_INFO_LEN];
};

/* note that this field must come before the
 * authentication TLV
 */
#define MIPMAC  129
struct mip_mse_icmp_mac {
	u_char  mse_type;     /* type of mobile service extension */
	u_char  mse_len;      /* length in bytes */
	u_char  mse_mac[6];   /* mac address */
};

/* The authentication tlv must come last.  There may logically
 * be any number of authentication TLVs, one per logical link.
 * We currently only implement two logical links distinquished
 * by the network_auth and adhoc flag bits.
 *
 * The icmp_router field is assumed to carry the ip address
 * of the agent or mn.  The mac address is included in the authentication.
 * All fields up to and including the spi are included in the authentication.
 *
 * In general, replay attacks or simply passive use of a MAC/IP address
 * pair may not be defeated.  However active replay attacks may optionally
 * be defeated by use of a challenge response protocol.  The protocol
 * would work as follows:
 *	node sends beacon with MAC, IP address and authentication.
 *	recv. takes icmp_advert portion and fills it in, sends
 *		UNICAST router advert with auth field filled in
 *		with fresh nonce field and CHALLENGE bit set.
 *		Key type is indicated as well.
 *	sender must echo packet material enclosed; i.e., recv. nonce
 *	value to prove existance of shared key.  echo is
 *	UNICAST to recv. 
 *
 * The agent field is filled in by a Mobile Node only and should be
 * 0 for agents themselves or 0 for MNs that lack an agent.  
 * An agent may choose to ignore an MN that is not using it as 
 * an agent.  This field indicates the current HA or FA as a
 * point of attachment.  This defeats a simple form of replay attack
 * and is a minor routing optimization simply because a recv. agent
 * that is not indicated by this field can choose to ignore the beacon.
 */
#define MIPAGENTAUTH   130
struct mip_mse_icmp_auth {
        u_char mse_type;    
        u_char mse_length;                  /* length of data + spi */
	u_char flags;			    /* flags field */
/* bits in flags - id which kind of auth is used */
#define   NETAUTH	0
#define   ADHOCAUTH	1
	u_char pad;	     
	u_long  tssecs;	        /* ntp secs timestamp */
	u_long  tsfract;	/* ts fractional part */
	u_long  agent;		/* ip addr for mn of current agent */
/* how far checksum is calculated
*/
#define MIPAGENTAUTH_OFFSET	20
	unsigned long spi;
        u_char hash[16];                /* enough room for the hash */
};      

/* support for flags bitfields
*/

#define	B(bit)		(1 << ((bit) & 0x0ff))	
#define	SET(word,bit)	(word |= B(bit))
#define	CLR(word,bit)	(word &= (~B(bit)))
#define	TST(word,bit)	(word & B(bit))

/* bit number in mse_flags
*/
#define MNBIT    0	/* PSU extension: mn, not an agent */
#define VBIT     9	/* Van Jacobson bit */
#define GBIT    10	/* GRE cisco encapsulation available bit */
#define MBIT	11	/* Minimal encapsulation offered */
#define FABIT   12	/* agent is FA */
#define HABIT	13	/* agent is HA */
#define BBIT	14	/* agent is BUSY */
/* no coa allowed */
#define RBIT	15	/* agent requires you to register with it */

#define MAXICMPPACKET	1024
#define MACMSIZ		6
#ifndef MAXIFNAME
#define MAXIFNAME	128
#endif

/* identity of system
*/
#define SELF_MN		-1
#define SELF_HA		0
#define SELF_FA 	1

/* authentication  type */
#define       SAWNETAUTH      0
#define       SAWADHOCAUTH    1

#ifdef MADRP  
#define ICMP_UNREACHABLE 3
struct icmp_unreachable{
       u_char type;
       u_char code;
       u_short chksum;
       u_long unuse;
       struct ip _ip;
};
#endif

