
/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 *  FILE:	mip_log.c
 *  SYNOPSIS:	
 *	Implements a separate log for each mipd/mnd daemon.  This is parallel
 *	functionality to syslog; i.e., *NOT* the same log.  We basically put
 *	daemon startup and exit/failure info in syslog and mobile-ip specifics
 *	(security information ...) in the mipd and mnd logs in /var/log.
 *
 *  FUNCTIONS:	mip_openlog, mip_closelog, mip_log, onusr2, mip_chlogname
 *
 *  EXIT VALUES:	
 *
 *  --	NOTE:	failure of local logging functions cannot send errors via
 *		syslogd because of incompatibility between older and newer
 *		versions of the variable argument definitions in varargs.h
 *		and stdarg.h.
 *
 *	Only one logfile may be open at a time.  Opening a second logfile
 *	will redirect all entries from the old file to the new.
 *
 */


#define MLOGROUTINES	/* uses stdarg.h over varargs.h */
#include "mip_log.h"

static int	noticeFunction = 2;

static int	llFd = NOLFILE;			/* local logfile fd */

/* names of current, previous and default logfiles */
static char	*curlog = NULL,
		*prevlog = NULL;

#ifdef MIPDVERSION
char	*deflog = MIPD_LOGFILE;
#elif defined( MNDVERSION )
char	*deflog = MND_LOGFILE;
#else
char	*deflog = NULL;
#endif

extern int debug;

int	rotate = TRUE,		/* TRUE to rotate curlog to prevlog */
	changed = FALSE;	/* TRUE between calls to mip_chlogname and
				     mip_openlog */




int
mip_openlog( char *logfile )
/*
 *  INPUT:	pathname to local logfile
 *  OUTPUT:
 *	0	logfile is open
 *	-1	open failed on logfile
 *
 *  --	failure of open() is not fatal.  global file descriptor is set so
 *	sibling functions do nothing if the open fails.
 *
 *	if a logfile is already open, this call simply returns.
 */
{

entryMark( 1, "mip_openlog" );



    if ( llFd != NOLFILE )
	return 0;


    if ( (llFd = open( logfile, O_WRONLY|O_APPEND|O_CREAT, MLOG_MODE )) == -1 )
    {
	llFd = NOLFILE;
	curlog = NULL;
	perror("open");
	return(-1);
    }
    curlog = logfile;
    changed = FALSE;


    return(0);

}  /* end mip_openlog() */





void
mip_closelog( void )
{

entryMark( 0, "mip_closelog" );


    if ( llFd == NOLFILE )    /* do nothing -- original mip_openlog() failed */
	return;

    close( llFd );
    llFd = NOLFILE;

    if ( rotate ) {
	prevlog = curlog;		/* rotate remembered logfile name */
	curlog = NULL;
    }
    rotate = TRUE;

}  /* end mip_closelog() */





int
mip_log( int pri, char *fmt, ... )
/*
 *  INPUT:	priority of log, format string and variable list of
 *		arguments appropriate for printf(3).
 *  RETURN VALUES:	
 *	0	successful write to local logfile
 *	-1	no action because mip_openlog failed or has not been called
 *	-2	write failed for local log
 *	-3	time() failed.
 *
 *  --	a string is generated using vsnprintf() and is headed by a string
 *	representing the priority value.  this is then written to the local
 *	logfile.  failure is not fatal.
 *
 *	the log entry is written even if the current time cannot be obtained.
 *
 *	log entries have the following form:
 *
 *	    <MM dd hh:mm:ss> <priority>: <message>
 *	
 *	where each line is appended with a newline if one is not provided
 *	in the format string or its arguments.
 */
{
    int		nchar = 0,	/* number of characters */
		nbytes;
    char	buf[MAXBUF+2],
		*s;
    time_t	rval = 0;	/* return value */
    va_list	ap;


entryMark( 0, "mip_log" );


    if ( llFd == NOLFILE )    /* do nothing -- original mip_openlog() failed */
	return(-1);


    /* 
     * get a date for the log message
     */
    if ( (rval = time( NULL )) == -1 ) {
	perror("time");
	rval = -3;
    }
    else		/* strips off weekday and year */
	nchar += sprintf(buf, "%.15s ", ctime(&rval)+4 );


    /* 
     * translate priority into a priority type string
     */
    if ( pri < 1 || pri > MLOG_PRI_MAX )  pri = MLOG_UNKNOWN;
    nchar += sprintf(buf+nchar, "%s: ", mtypestr[pri]);


    /* 
     * complete the string with printf variant and write to local logfile
     */
    va_start(ap, fmt);
    nchar += vsnprintf(buf+nchar, MAXBUF-nchar, fmt, ap);
    if ( buf[nchar-1] != '\n' ) {
	buf[nchar] = '\n';
	buf[nchar+1] = '\0';
	nchar += 1;
    }
    for ( s = buf; nchar > 0; s += nbytes, nchar -= nbytes )
	if ( (nbytes = write( llFd, s, nchar )) == -1 ) {
	    perror("write");
	    return(-2);
	}


    return( 0 );

}  /* end mip_log() */





void
onusr2( void )
/*
 *  --	closes the current logfile (if one is open) and opens a new logfile.
 *	if no logfile can be remembered, use the default logfile name
 *	for this daemon.
 *
 *	this may be invoked prior to mip_openlog().  to intentionally change
 *	the name of the logfile use mip_closelog() and mip_openlog().
 */
{
    char	*templog;

entryMark( 1, "onusr2" );


    if ( curlog )
	templog = curlog;
    else
	templog = deflog;

    if ( llFd != NOLFILE ) {
	mip_log( MLOG_INFO,
	    "Closing logfile (SIGUSR2).  Log continues in \"%s\".",
	    curlog);
    }
    if ( changed )	rotate = FALSE;
    mip_closelog();

    if ( mip_openlog( templog ) < 0 )
	return;

    mip_log( MLOG_INFO,
	"New logfile (SIGUSR2).  Previous logfile was \"%s\".",
	prevlog?prevlog:"<undefined>" );

}  /* end onusr2() */





void
mip_chlogname( char *logname )
/*
 *  --	give the logfile a new static name.  this is only useful if
 *	the next log function call is onusr2.
 */
{

entryMark( 1, "mip_chlogname" );

    changed = TRUE;

    prevlog = curlog;
    curlog = logname;

}  /* end mip_chlogname() */





#ifdef TESTLOG

void
main( int argc, char *argv[] )
{

    extern void	onusr2( void );

    signal( SIGUSR2, onusr2 );

    kill( getpid(), SIGUSR2 );
    /* mip_openlog( "1yuletime" ); 	*/
    mip_openlog( "2enterprise" );
    kill( getpid(), SIGUSR2 );
    mip_closelog();
    mip_openlog( "2enterprise" );
    mip_chlogname( "1yuletime" );
    kill( getpid(), SIGUSR2 );

    return;

    /* ================================================================= */


#ifdef LOGTEST1
{
    int	rval;

    mip_openlog(MIPD_LOGFILE);

    rval = mip_log( MLOG_INFO, "%s %d.%d", "foobar daemon:", 1,0);
    rval = mip_log( -1, "%s%d", "string", 42 );
    rval = mip_log( MLOG_INFO, "%s %x", "sample: something informative has occured", 42);
    rval = mip_log( MLOG_SEC, "%s\n", "sample: something of a secure nature with newline");
    rval = mip_log( 0, "fmt only");

    /* printf("rval == %d\n");	/* */

    mip_closelog();
}
#endif /* LOGTEST1 */


}  /* end main() */

#endif /* TESTLOG */
