
/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * mipinit
 *
 * module for initialization of mobile-ip functionality in kernel.
 * Both mipd and mn use this.
 */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/signal.h>
#include <termios.h>

#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/ip_var.h>
#include <netdb.h>
#include <unistd.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>

#include "mipconfig.h"
#include "mip_log.h"


/* open a raw socket to set various mobile-ip parameters
 * in the kernel
 *
 * params:
 *	am_mn: 1 to say "i am a mobile node (agent and/or host).
 *		all mobile nodes will set this.  agents don't set it.
 *	mnaddr: am Mobile Node and want to set the foreign default
 *		address; i.e., the i/f that the system is mobile on.
 *		Normally FAs don't want to do that.  MNs do.
 *		A FA that wanted to be a MN would do it however.
 */
mip_init(int am_mn, struct in_addr mnaddr)
{
	int rc;
        struct sockaddr_in ifaddr;
	int s;
	int on = 1;

	/* get a raw socket
	*/ 
	if ((s = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP)) < 0) {
		syslog( LOG_ERR, "mip_init(): socket (%m).");
                exit(1); 
        }

	/* init mobile-ip
	*/
        rc = setsockopt(s, IPPROTO_IP, IP_MIP_MN_ON, (char *)&on, sizeof(on));
        if ( rc < 0 ) {
		syslog( LOG_ERR, "mip_init(): setsockopt(IP_MIP_MN_ON) (%m).");
                exit(1);
        }

	/* if mobile node, set mobile ip address (address of i/f
	 * to use for default route)
	 */
	if (am_mn) {
		ifaddr.sin_family = AF_INET;    
		ifaddr.sin_addr = mnaddr;
		if (setsockopt(s, IPPROTO_IP, IP_MIP_MN_IF,
			       (char *)&ifaddr.sin_addr,
			       sizeof(ifaddr.sin_addr)) < 0) {
			syslog( LOG_ERR,
			    "mip_init(): setsockopt(IP_MIP_MN_IF) (%m).");
			exit(1);
		}
	}
	close(s);
}

mip_off()
{
	int rc;
        struct sockaddr_in ifaddr;
	int s;
	int on = 0;

	/* get a raw socket
	*/ 
	if ((s = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP)) < 0) {
                logperror("socket");
                exit(1); 
        }

	/* init mobile-ip
	*/
        rc = setsockopt(s, IPPROTO_IP, IP_MIP_MN_ON, (char *)&on, sizeof(on));
        if ( rc < 0 ) {
                logperror("setsockopt(IP_MIP_MN_ON)");
                exit(1);
        }

	close(s);
}

#ifdef TESTMAIN
logperror(s)
char *s;
{
	printf("%s\n", s);
}

/*
 * syntax: mipinit <mobile_if_ip>
 * 	   mipinit -h,  give help
 */
main(int argc, char **argv)
{
	struct in_addr dummy;

	if (argc == 2) { /* mn */
		/* turn it off
		*/
		if (strcmp(argv[1], "-h") == 0) {
			printf("mipinit - init for FA or HA\n");
			printf("mipinit <if ip addr> - init for MN\n");
			printf("mipinit off - disable mobile ip config\n");
			return(0);
		}
		if (strcmp(argv[1], "off") == 0) {
			mip_off();
			return(0);
		}
		/* turn it on and set mn default address
		*/
		dummy.s_addr = inet_addr(argv[1]);
		mip_init(1, dummy);
	}
	else /* not mn */
		/* turn it on and don't set default address
		*/
		mip_init(0, dummy);
}
#endif
