
/*
 * Copyright (c) 1996, 2000 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * mipmain.c
 *
 * mobile ip agent daemon for FreeBSD.
 *
 * runs on both Home Agents and Foreign Agents
 *
 * syntax:
 *	mipd [-p] [-d] [-c]
 *
 * -d: debug mode, print to printf.  will not go background as daemon.
 * -p: parse only, parse config file and exit.  As a result,
 *	the config file can be checked for errors.
 *	You can get more information with -d here too.
 * -c: config file.  default is /etc/mipd.conf, with this switch
 *	you can provide a config file elsewhere.
 *
 * Any invalid switch will cause a usage message to be printed.
 *
 * SIGNALS:
 *	kill -HUP <mipd.pid>		stop daemon & reset routing facilities
 *	kill -USR1 <mipd.pid>		reset routing && restart daemon
 *	kill -USR2 <mipd.pid>		open a new logfile
 *	
 *	routes in visitor list are removed from route table and
 *	mipd will exit on HUP.
 *
 *      on USR1, mipd will exit, exec and reread mipd.conf.
 */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/signal.h>
#include <termios.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/ip_var.h>
#ifdef KAMEIPSEC
#include <netinet6/ipsec.h>
#endif
#include <netdb.h>
#include <unistd.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>

#include "mipconfig.h"

#ifdef MADRP
#include "sr_mad.h"
struct sr_config  sr_config;
#endif

#include "mipstate.h"
#include "mipstats.h"
#include "unixsock.h"
#include "mip_log.h"
#include "mipdversion.h"
#include "icmpstats.h"
#include "route.h"
#include "vlist.h"

extern struct ha_stats ha_stats;
extern struct fa_stats fa_stats;
extern struct udp_errors udp_errors;
struct icmp_stats icmp_stats;
#ifdef MADRP
struct sr_metric sr_metric;
#endif

static char configFile[1024];
int debug = 0;
int parseOnly = 0;
char *progname;
char **progargs;

#ifdef HARP
#define MIPSEC
#include "harp.h"
#endif

void parseArgs(int, char **);
unsigned short getSequence(u_short *);
void sendBeacon( struct sockaddr_in *, struct sockaddr_in *, 
	struct sockaddr_in *, int, int, int, u_short, char *,
	int, struct sockaddr_in *, char *);
void tickClock(void);
static void unblockSignals();
static void doit();
#ifdef KAMEIPSEC
void enableKameIpsec();
#endif

main(argc, argv)
int argc;
char **argv;
{
	struct in_addr dummy;
	extern void onhup();
	extern void onusr1();
	extern void onusr2();
	extern void mipTick();

	/* check status - root only
	*/
	if (getuid()) {
		fprintf(stderr, "mipd must be invoked by root\n");
		exit(1);
	}
	progname = argv[0];
	progargs = argv;

	/* open log files: logging via syslogd for fatal errors, startup
	 * and HUP shutdown, logging to MIPD_LOGFILE for security specific
	 * errors and events, and for packet parsing errors.
	 */
	openlog("mipd", LOG_CONS | LOG_PID, LOG_DAEMON);
	if ( mip_openlog( MIPD_LOGFILE ) < 0 ) {
	    syslog( LOG_ERR,
		"Failed to open local logfile \"%s\".", MIPD_LOGFILE );
	    exit(1);
	}


	/* parse input arguments.  
	 * configuration file read in here, but linkage to
	 * ifconfig not checked yet.
	 */
	parseArgs(argc, argv);
#ifdef HARP
#ifdef MIPSEC
	if ( mip_config.harp && (mip_config.haha_security == FALSE )){
		mip_log( MLOG_ERR, "haha security required.\n" );
		exit(1); 
	}
#endif
#endif

	/* Route socket must be available to initialize interfaces. */
        initSockets();

	/* get kernel ifs and make sure they match config
	 * request
	 */
	initIfs();

	/* become a daemon
	*/
	if (!debug) {
		if (daemon(0,0) < 0) {
			syslog(LOG_ERR, "main(): daemon (%m)");
			exit(1);
		}
	}

	syslog(LOG_INFO, "Mobile agent daemon started [%d].  (%s)",
		getpid(), MIPDVERSION);
	mip_log(MLOG_INFO, "Mobile agent daemon started [%d].  (%s)",
		getpid(), MIPDVERSION);

	signal(SIGHUP, onhup);
	signal(SIGINT, onhup);	/* allows cleanup even when debugging */
	signal(SIGUSR1, onusr1);
	unblockSignals();
	signal(SIGALRM, mipTick);
	
	storepid();

	initRouting();

	doit();
}

void
parseArgs(argc, argv)
	int             argc;
	char          **argv;
{
	extern int      errno;
	extern char    *optarg;
	extern int      optind;
	int             ch, fd;
	int 		rc;
	extern void configInit();

	configInit();
	strncpy(configFile, CONFIGDEFAULT, 1024);
	while ((ch = getopt(argc, argv, "pdc:")) != EOF) {
		switch (ch) {
		case 'p':
			parseOnly = 1;
		case 'd':
			debug = 1;
			break;
		case 'c':
			strncpy(configFile, optarg, 1024);
			break;
		case '?':
		default:
			usage();
		}
	}

	/* read in config file
	*/
	rc = configRead(configFile);
	if (rc < 0) {
		exit(1);    /* NOTE: all errors logged within configRead() */
	}
#ifdef MADRP
        sr_initconfig(&sr_metric);
#endif

	if (parseOnly) {
		debug = 1;
	}

	if (debug) {
		printf("configFile %s\n", configFile);
		configDump();
	}

	if (parseOnly) {
		printf("parse mode -- successfully completed, exiting\n");
		exit(0);
	}
}

usage()
{
	fprintf(stderr,"mipd: -d -p -c <config file>\n");
	exit(1);
}

/*
 * check config interfaces against ifconfig/kernel configuration
 * and verify that all requested interfaces exist.  We also
 * get and store their ip address.
 *
 * This routine's subroutines may fail with a message logged via syslog
 * if the config file doesn't match the real configuration
 */
initIfs()
{
	int dontcare;
	unsigned long ipaddr;

	getIfs();
	verifyIfConfig();
	/* if we have a ha interface (we can only have one),
	 * determine its ip address for later use
	 */
	if (mip_config.sawHa && mip_config.ha_ifname != NULL) {
		mapIfname(mip_config.ha_ifname, &dontcare,  	
			&ipaddr);
		mip_config.ha_sin.sin_addr.s_addr = ipaddr;
		if (debug) {
			printf("ha if info follows %s %x:\n",
				mip_config.ha_ifname,
				ntohl(mip_config.ha_sin.sin_addr.s_addr));
		}
	}
	
#ifdef WRONG
	zapClones();	/* zap clone routes on non-arp interfaces */
#endif
}

/*
 * do any boot time routing setup
 * 1. setup LLINFO clone routes acc. to per i/f ad hoc or arp setup
 *      if arp, do nothing
 *
 * note: cannot call this routine until initSockets() called.
 */     
initRouting()
{
        zapClones();                   /* zap clone routes on non-arp interfaces */
}

extern void setMcastIf(struct in_addr);

/* send beacons on desired interfaces
*/
void beaconLoop()
{
	struct mip_ifs *mips;  
	extern struct mip_ifs mip_ifs[MAX_INTERFACES];
	extern int mip_ifcount;
	extern struct sockaddr_in fa_coa;
	int doubledBeacon = 0;
	int haorfa = 0;
	struct sockaddr_in *dst_addr = NULL;
	int i;

	mips = &mip_config.mip_ifs[0];
	for ( i = 0; i < mip_config.mip_ifcount; i++, mips++ ) {
		/* if this interface defines both a ha and a fa, send
		 * a combined beacon for the ha form, and skip the fa form
		 */
		if (mips->both == 2)
			continue;


		/* check whether to send zero, one, or a combined beacon */
		haorfa = mips->amfa;
		dst_addr = &mips->r_sin;

		if (mips->nextAgent) {
			/* beacon for both agents if both have non-NULL routers
			 */
			if (mips->router && mips->nextAgent->router) {
				doubledBeacon = 1;
			}

			/* beacon only for the FA if just it's router is set */
			else if (mips->nextAgent->router) {
				dst_addr = &mips->nextAgent->r_sin;
				haorfa = mips->nextAgent->amfa;
			}

			/* no beacons as no routers are set */
			else if (!mips->router && !mips->nextAgent->router) {
				continue;
			}

			/* NOTE: if all cases fail then beacon only for
			 * the HA which is the default setting
			 */
		}
		else if (mips->both) {
			mip_log(MLOG_ERR, "beaconLoop(): (internal) nextAgent==NULL but i/f defined for 2 agents.\n");
			exit(1);
		}


		/* sawMcast means at least 1 interface wants a multicast
		 * beacon.
		 */
		if (mip_config.sawMcast) {
			setMcastIf(mips->if_sin.sin_addr);
		}

		/* parameters
		 * 	router in advert packet, same as if address at moment
		 *	dest - the ip dst of the ip header
		 *		note: if not mcast, doesn't matter
		 *	registrationLifetime
		 *	fa or ha
		 *	whether fa AND ha
		 *	current sequence number for this stream
		 *		of advertisements
		 */
		sendBeacon(
			&mips->if_sin, 	/* icmp advert. router value */
			dst_addr,		/* ip dst address */
			&mip_config.fa_coa,		/* fa coa address */
			mip_config.mobileLifetime, 	/* list timer */
			haorfa, 		/* for fa or ha */
			doubledBeacon,		/* TRUE if both ha and fa */
			getSequence(&mips->seq),  /* sequence number */
 			mip_config.mip_ifs[i].mac,   /* mac_addr */
			mips->mcast,	/* mcast or unicast flag */
			&mips->if_sin,  /* ip src address */
			mips->ifname 	/* string name of interface */
			);  
	}
}

/*
 * 1. initialize code to send adverts
 * 2. open a udp socket at port 434, the mobile-ip port.
 * We recv. authentication requests on this port
 */
#ifdef HARP
int harpSocket;               /* harp socket descriptor must be externally */
				 /* available for harpForward() */
static int harpRegSocket;	/* socket for requesting and dump harp tables */
#endif
static int udpSocket;
static int unixSocket;
static int icmpSocket;
#ifdef DIVERT
static int divertSocketIn;
static int divertSocketOut;
#endif
#ifdef MADRP
int madSocket;
#endif          
static int routeSocket;

initSockets()
{
	struct mip_ifs *mips;  
	int rc;
	char errbuf[2048];

	icmpSocket = initIcmp(mip_config.sawBroadcast, mip_config.sawMcast,
		inet_addr("224.0.0.1"), 0);
	udpSocket = initUdp(UDP_PORT);
	routeSocket = routeInit();

#ifdef DIVERT
	/* Get a Divert Socket For Incoming Packets */
	divertSocketIn = getDivertSock();

	/* Bind The Socket To a DIVERTPORTIN */
	divertBind(divertSocketIn,DIVERTPORTIN);

	/* Get a Divert Socket For Outgoing Packets */
	divertSocketOut = getDivertSock();

	/* Bind The Socket To a DIVERTPORTOUT */
	divertBind(divertSocketOut,DIVERTPORTOUT);
#endif


#ifdef KAMEIPSEC
	/* if ipsec bypass switch, then udpSocket and icmpSocket
	 * bypass any IPSEC policy setup
	 */
	if ( mip_config.kameipsecEnable ) {
		enableKameIpsec();
	}
#endif

#ifdef HARP
	if ( mip_config.harp ){
		if ( debug ){
			printf("initSockets(): initializing HARP sockets\n");
		}
		harpSocket = initUdp( HARP_PORT );
		harpRegSocket = initHarpSocket( HARP_PORT );
	}
#endif
	unixSocket = initUnixSocket(MOBYSOCK_SERVER, FALSE);
#ifdef MADRP
        madSocket= init_sr_Udp();
#endif
        routeSocket = routeInit();

#ifdef HARP
	/* initialize ripv2 advertisement if desired
	*/
	if (mip_config.dorip) {
		(void) initRIP( 
			inet_addr("224.0.0.9"),
			mip_config.rip_if.sin_addr, 
			mip_config.rip_ha.sin_addr, 
			mip_config.rip_mask.sin_addr);
	}
#endif

	/* setup bpf for writing. note we aim it at the 1st
	 * interface in the mips table
	 */
	mips = &mip_config.mip_ifs[0];
	rc = init_bpf(mips->ifname, errbuf, 0 /* don't need read */ );
	if (rc < 0) {
		syslog( LOG_ERR, "init_bpf: failed (%s).", errbuf);
		exit(1);
	}
}

static int pending_tick;

#ifdef HARP
static ripClock = 30; 
#endif

static void
doit()
{
	fd_set readmask;
	int rc;
	extern int errno;

	startAlarm();
	while(1) {

 		/* alarm only sets this pending_tick flag 
		 * tickClock
		 *	sends beacons
		 *	drives state machine/list timeouts
		 */
                if (pending_tick) {
                        tickClock();
                        pending_tick--;
                }

		/* select on icmp, udp, unix sockets
		 * select on 1 second interval.  This
		 * is heartbeat clock for daemon and timer/s
		 * are driven from it. This allows mnd to 
		 * react to lack of agent, respond to registration timer, etc.
		 */
		FD_ZERO(&readmask);
		FD_SET(icmpSocket, &readmask);
		FD_SET(udpSocket, &readmask);
#ifdef DIVERT
		FD_SET(divertSocketIn, &readmask);
		FD_SET(divertSocketOut, &readmask);
#endif
#ifdef HARP
		if ( mip_config.harp ){
			FD_SET( harpSocket, &readmask );
			FD_SET( harpRegSocket, &readmask );
		}
#endif
		FD_SET(routeSocket, &readmask);
#ifdef MADRP
                FD_SET(madSocket, &readmask);
#endif
		FD_SET(unixSocket, &readmask);
		rc = select(unixSocket + 1, /* most recently created socket */
				(fd_set *) &readmask, NULL, NULL, NULL);
		if (rc < 0) {
			if (errno == EINTR)
				continue;
			syslog( LOG_ERR, "doit(): select (%m).");
			exit(1);
		}
		if (FD_ISSET(icmpSocket, &readmask)) {
			handleIcmpAdvert(icmpSocket, NULL, 0, 0, NULL);
		}
		if (FD_ISSET(udpSocket, &readmask)) {
			handleUdpInput(udpSocket
#ifdef HARP
			, UDP_SOC
#endif
			);
		}
#ifdef HARP
	 	if ( mip_config.harp ){ 
			if ( FD_ISSET( harpSocket, &readmask )) {
				 if ( debug ){
					printf("doit(): received message on HARP socket\n");
				 }
				 handleUdpInput( harpSocket, HARP_UDP_SOC );
			}
			if ( FD_ISSET( harpRegSocket, &readmask )){
				 if ( debug ){
					printf("doit(): received message on HARP socket\n");
				 }
				 harpRegDump( harpRegSocket );
			}
		} 
#endif
		if (FD_ISSET(unixSocket, &readmask)) {
			handleUnixInput(unixSocket);
		}
#ifdef DIVERT
		if (FD_ISSET(divertSocketIn, &readmask)) {
                        handleDivertIn(divertSocketIn);
                }
		if (FD_ISSET(divertSocketOut, &readmask)) {
                        handleDivertOut(divertSocketOut);
                }
#endif
		if (FD_ISSET(routeSocket, &readmask)) {
                        routeRead();
                }
#ifdef MADRP
                if (FD_ISSET(madSocket, &readmask)) {
                        handleMadInput(madSocket, &sr_metric);
                }
#endif
	}
}

/* clock tick.  needs to do two things
 *
 * 1. beacon sending mechanism
 * 2. send tick to state machine
 *
 * assume tickClock is called one time per second.
 */
static int beaconClock=0;

void 
tickClock()
{

	/* if virtual beacon clock times out, send beacons
	 * and reload counter
	 */
	if (beaconClock <= 0) {
		beaconClock = mip_config.beaconRate;
		beaconLoop();	/* send beacons out ifs */
	}
	beaconClock--;

#ifdef HARP
	if (mip_config.dorip) {
		ripClock--;
		if (ripClock <= 0) {
			ripClock = 30;
			sendRIP();
		}
	}
#endif

	/* send tick event to state machine
	*/
	stateTick();
#ifdef MADRP
        mip_tickBackoffTimer();
#endif
}


/*
 * read message from unix socket
 */
int
handleUnixInput(int s)
{
	int rc;
	char buf[MAXBUF];

	rc = recvUnixMessage(buf, MAXBUF);

	if (debug) {
		printf("unixMessage: buf %d, rc %d\n", buf[0], rc);
	}
	switch (buf[0]) {
        case U_BEACONS:
                sendUnixMessage(MOBYSOCK_CLIENT,
                        (char *)&icmp_stats, sizeof (struct icmp_stats));
                break;
	case U_HAVLIST:
		ha_dumpVlist();
		break;
	case U_FAVLIST:
		fa_dumpVlist();
		break;
	case U_STAT:
		sendUnixMessage(MOBYSOCK_CLIENT,
			(char *)&ha_stats, sizeof (struct ha_stats));
		sendUnixMessage(MOBYSOCK_CLIENT,
			(char *)&fa_stats, sizeof (struct fa_stats));
		sendUnixMessage(MOBYSOCK_CLIENT,
			(char *)&udp_errors, sizeof (struct udp_errors));
#ifdef MADRP
                sendUnixMessage(MOBYSOCK_CLIENT,
                        (char *)&sr_metric, sizeof (struct sr_metric));
#endif
		break;
#ifdef MADRP
        case U_HLIST:
                dumpHlist();
                break;
#endif

	}
}

static
mipd_shutdown()
{
	faRouteWipe();
	haRouteWipe();
	unzapClones();
	unzapArp();
#ifdef MADRP
        hlist_delete_all(); /* flush hlist */
#endif
	closelog();
	mip_closelog();
}

/*
 * sigusr1 needs to be unblocked in terms of its signal
 * mask as we may have restarted
 */
static
void
unblockSignals()
{
	sigset_t s;
	sigset_t o;
	int rc;

	sigemptyset(&s);
	sigaddset(&s, SIGUSR1);
	rc = sigprocmask(SIG_UNBLOCK,&s,&o);
	if (rc < 0) {
		syslog(LOG_INFO, "sigprocmask failed\n");
		exit(1);
	}
}

/*
 * ask state module to delete routes
 */
void
onhup()
{
	struct vlist *hhead;


	hhead   =  (struct vlist *) vlist_gethahead();
	
#ifdef DIVERT
	while(hhead){

		/* if the ip-udp tunnelling is set remove the firewall rules */
		if (hhead->routeset == IPUDPTUNNEL)	  
			delFireWallRule( ((struct vlist *)hhead)->fr_number );
		  
		hhead = ((struct vlist *)hhead)->next;

	}
#endif
	
	syslog(LOG_INFO, "Daemon exiting - received HUP signal." );
	mip_log(MLOG_INFO, "Daemon exiting - received HUP signal." );
	mipd_shutdown();
	exit(0);
}


/*
 * restart daemon
 */
void
onusr1()
{
	signal(SIGALRM, SIG_IGN);

	syslog(LOG_INFO, "Daemon restarting - received USR1 signal." );
	mip_log(MLOG_INFO, "Daemon restarting - received USR1 signal." );
	mipd_shutdown();
#ifdef MADRP
        close(madSocket);
#endif
#ifdef HARP
	close(harpSocket);
	close(harpRegSocket);
#endif
	close(udpSocket);
	close(icmpSocket);
	close(unixSocket);
        if (execvp(progname, progargs) < 0) {
            syslog(LOG_ERR, "onusr1(): execvp failed (%m)");
            exit(1);
        }
}


void 
fatal(char *s)
{
	syslog( LOG_ERR, "Fatal Error: %s\n", s);
	exit(1);
}

logperror(char *s) 
{
	syslog( LOG_ERR, "Fatal Error: %s -- %m\n", s);
	exit(1);
}

#ifdef SAVEENERGY
static u_long mytick=0;
u_long changestate=0;
#endif


/* mip timer */
void
mipTick()
{       
        pending_tick++;
#ifdef SAVEENERGY
        mytick++;
        changestate=mytick/300;
#endif

	startAlarm();
}               

startAlarm() 
{
	alarm(1);
}

stopAlarm()
{
	alarm(0);
}


#ifdef KAMEIPSEC
static char * outbypassPolicy = "out bypass";
static char * inbypassPolicy = "in bypass";
/* enable kame ipsec 
 *
 * reality check: this really disentangles Mobile-IP and the KAME
 * IPSEC policy code; i.e., lets Mobile-IP send packets to Mobile-IP
 * parties without IPSEC policy involved.  
 *
 * idea here:
 *	udpSocket and icmpSocket are set free for input/output; i.e.,
 * no IPSEC applied.
*/
void
enableKameIpsec()
{
  char *outbypassbuf, *inbypassbuf;
  int outbypasslen, inbypasslen;
  int     ipsec_policy = IPSEC_POLICY_BYPASS;

  outbypassbuf = ipsec_set_policy( outbypassPolicy, strlen(outbypassPolicy));
  if( outbypassbuf == NULL )
    {
      syslog(LOG_ERR, "enableKameIpsec() : ipsec_set_policy failed, May be kernel is not compiled with IPSEC enabled\n");
      mip_log(MLOG_SEC, "enableKameIpsec() : ipsec_set_policy failed, May be kernel is not compiled with IPSEC enabled\n");
      if (debug)
	printf("enableKameIpsec() : ipsec_set_policy failed, May be kernel is not compiled with IPSEC enabled\n");
      exit(1);
    }
  
  outbypasslen = ipsec_get_policylen(outbypassbuf);
  

  inbypassbuf = ipsec_set_policy( inbypassPolicy, strlen(inbypassPolicy));
  if( inbypassbuf == NULL )
    {
      syslog(LOG_ERR, "enableKameIpsec() : ipsec_set_policy failed, May be kernel is not compiled with IPSEC enabled\n");
      mip_log(MLOG_SEC, "enableKameIpsec() : ipsec_set_policy failed, May be kernel is not compiled with IPSEC enabled\n");
      if (debug)
	printf("enableKameIpsec() : ipsec_set_policy failed, May be kernel is not compiled with IPSEC enabled\n");
      exit(1);
    }
  
  inbypasslen = ipsec_get_policylen(inbypassbuf);
  

  /* set in and out bypass policy for udpSocket */

  if (setsockopt(udpSocket, IPPROTO_IP, IP_IPSEC_POLICY,
		 (char *)outbypassbuf, outbypasslen) < 0) {
    syslog(LOG_ERR, "enableKameIpsec(): udpSocket sockopt IP_IPSEC_POLICY failed");
    mip_log(MLOG_SEC, "enableKameIpsec(): udpSocket sockopt IP_IPSEC_POLICY failed");
    if (debug) {
      printf("enableKameIpsec(): udpSocket sockopt IP_IPSEC_POLICY failed");
    }
    exit(1);
  }

  if (setsockopt(udpSocket, IPPROTO_IP, IP_IPSEC_POLICY,
		 (char *)inbypassbuf, inbypasslen) < 0) {
    syslog(LOG_ERR, "enableKameIpsec(): udpSocket sockopt IP_IPSEC_POLICY failed");
    mip_log(MLOG_SEC, "enableKameIpsec(): udpSocket sockopt IP_IPSEC_POLICY failed");
    if (debug) {
      printf("enableKameIpsec(): udpSocket sockopt IP_IPSEC_POLICY failed");
    }
    exit(1);
  }

  /* set in and out bypass policy for icmpSocket */
  
  if (setsockopt(icmpSocket, IPPROTO_IP, IP_IPSEC_POLICY,
		 (char *)outbypassbuf, outbypasslen) < 0) {
    syslog(LOG_ERR, "enableKameIpsec(): icmpSocket udpSocket sockopt IP_IPSEC_POLICY failed");
    mip_log(MLOG_SEC, "enableKameIpsec(): icmpSocket sockopt IP_IPSEC_POLICY failed");
    if (debug) {
      printf("enableKameIpsec(): icmpSocket sockopt IP_IPSEC_POLICY failed");
    }
    exit(1);
  }

  if (setsockopt(icmpSocket, IPPROTO_IP, IP_IPSEC_POLICY,
		 (char *)inbypassbuf, inbypasslen) < 0) {
    syslog(LOG_ERR, "enableKameIpsec(): icmpSocket udpSocket sockopt IP_IPSEC_POLICY failed");
    mip_log(MLOG_SEC, "enableKameIpsec(): icmpSocket sockopt IP_IPSEC_POLICY failed");
    if (debug) {
      printf("enableKameIpsec(): icmpSocket sockopt IP_IPSEC_POLICY failed");
    }
    exit(1);
  }
}
#endif /* KAMEIPSEC */











