
/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * mipstat.c
 *
 * query the mipd for stats and for visitor list info.
 *
 * syntax:
 *	mipstat [-shflb]
 *		-s: print stats
 *		-h: ha visitor list
 *		-f: fa visitor list
 *		-b: print icmp info
 *
 * default is to print usage
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <netinet/in.h>

#include "mipconfig.h"
#include "vlist.h"
#ifdef MADRP 
#include "sr_mad.h"
#endif

#include "unixsock.h"
#include "mipstats.h" 
#include "icmpstats.h" 

void parseArgs(int argc, char **argv);

int stats = 0;
int haflag = 0;
int hflag = 0;
int beacon = 0;
int longflag = 0;
int faflag = 0;
char *progname;

main(argc, argv)
int argc;
char **argv;
{
	if (getuid()) {
		printf("mipstat: root only\n");
		exit(0);
	}
	progname = argv[0];
	parseArgs(argc, argv);
	doit();
	return(0);
}

void
parseArgs(argc, argv)
	int             argc;
	char          **argv;
{
	extern int      errno;
	extern char    *optarg;
	extern int      optind;
	int             ch, fd;
	int 		rc;
	int 		sawFlag = 0;

#ifdef MADRP
	while ((ch = getopt(argc, argv, "bhfslH")) != EOF) {
#else
	while ((ch = getopt(argc, argv, "bhfsl")) != EOF) {
#endif
		switch (ch) {
 		case 'l':
                        longflag = 1;
                        break;
		case 'h':
			sawFlag = haflag = 1;
			break;
		case 'f':
			sawFlag = faflag = 1;
			break;
		case 's':
			sawFlag = stats = 1;
			break;
		case 'b':
			sawFlag = beacon = 1;
			break;
#ifdef MADRP
		case 'H':
			sawFlag = hflag = 1;
			break;
#endif
		case '?':
		default:
			usage();
		}
	}
	if (sawFlag == 0) {
		usage();
	}
}

usage()
{
#ifdef MADRP
	fprintf(stderr,"mobystat: [-hfsblH]\n");
#else
	fprintf(stderr,"mobystat: [-hfsbl]\n");
#endif
	fprintf(stderr,"-s: stats\n");
	fprintf(stderr,"-[l]f: fa visitor list \n");
	fprintf(stderr,"-[l]h: ha visitor list \n");
	fprintf(stderr,"-[l]b: print (long) beacon stats\n");
#ifdef MADRP
        fprintf(stderr,"-H: print host route list\n");
#endif

	exit(1);
}

doit()
{
	int rc;
	int vcount = 0;
	char buf[1024];
	struct vlist v;
	int i;
	/* sent in order as follows */
	struct ha_stats ha_stats;
	struct fa_stats fa_stats;
	struct icmp_stats icmp_stats;
	struct udp_errors udp_errors;
#ifdef MADRP
	struct sr_metric sr_metric;
#endif

	/* open the unix socket to talk to the server
	*/
	rc = initUnixSocket(MOBYSOCK_CLIENT, TRUE);
	if (rc < 0) {
		perror("mipstat: init unix socket failed");
		goto err;
	}

	/* send debug message */
	if (stats) {
		buf[0] = U_STAT;
		rc = sendUnixMessage(MOBYSOCK_SERVER, buf, 1);
		if ( rc < 0 ) {
			perror("mipstat: send unix message failed");
			goto err;
		}
		rc = recvUnixMessage((char *)&ha_stats, sizeof(struct ha_stats));
		if ( rc < 0 ) {
			perror("mipstat: recv unix message failed");
			goto err;
		}
		rc = recvUnixMessage((char *)&fa_stats, sizeof(struct fa_stats));
		if ( rc < 0 ) {
			perror("mipstat: recv unix message failed");
			goto err;
		}
		rc = recvUnixMessage((char *)&udp_errors, sizeof(struct udp_errors));
		if ( rc < 0 ) {
			perror("mipstat: recv unix message failed");
			goto err;
		}
#ifdef MADRP
		rc = recvUnixMessage((char *)&sr_metric, sizeof(struct sr_metric));
		if ( rc < 0 ) {
			perror("mipstat: recv unix message failed");
			goto err;
		}
#endif
		dump_hastats(&ha_stats);
		dump_fastats(&fa_stats);
		dump_udperrs(&udp_errors);
#ifdef MADRP
      		dumpsr_metric(sr_metric);
#endif

	}
	/* send vlist message */
	else if (haflag) {

		buf[0] = U_HAVLIST;
		sendUnixMessage(MOBYSOCK_SERVER, buf, 1);
		/* the 1st message back tells us the number of elements
		*/
		rc = recvUnixMessage((char *)&vcount, sizeof(int));
		if (longflag) 
			printf("vcount %d\n", vcount);
		else if (vcount > 0)
			vlist_shortHeader();
		for ( i = 0; i < vcount; i++) {
			rc = recvUnixMessage((char *)&v, sizeof(struct vlist));
			if (rc != sizeof(struct vlist)) {
				fprintf(stderr,"recv size wrong!\n");
				exit(1);
			}
			vlist_dumpone(&v, longflag);
		}
	}
	else if (faflag) {

		buf[0] = U_FAVLIST;
		sendUnixMessage(MOBYSOCK_SERVER, buf, 1);
		/* the 1st message back tells us the number of elements
		*/
		rc = recvUnixMessage((char *)&vcount, sizeof(int));
		if (longflag) 
			printf("vcount %d\n", vcount);
		else if (vcount > 0)
			vlist_shortHeader();
		for ( i = 0; i < vcount; i++) {
			rc = recvUnixMessage((char *)&v, sizeof(struct vlist));
			if (rc != sizeof(struct vlist)) {
				fprintf(stderr,"recv size wrong!\n");
				exit(1);
			}
			vlist_dumpone(&v, longflag);
		}
	}
#ifdef MADRP
        else if (hflag){
                buf[0] = U_HLIST;
                if (sendUnixMessage(MOBYSOCK_SERVER, buf, 1)<0){
                        perror("mnstat: socket send message failed");
                        goto err;
                }
                vcount = 0;
                if (recvUnixMessage((char *)&vcount, sizeof(int))<0){
                        perror("mipstat: socket recv message for count failed");
                        goto err;
                }
                for (i=0; i<vcount; i++){
                        struct SR_hlist h;
                        if(recvUnixMessage((char*)&h,sizeof(struct SR_hlist))<0)
			{
                                fprintf(stderr, "recv of hlist member failed");
                                exit(1);
                        }
                        if (i==0) srhlist_header();
                        srhlist_dumpone(&h);
                }
        }
#endif
 	else if (beacon) {
                buf[0] = U_BEACONS;
                rc = sendUnixMessage(MOBYSOCK_SERVER, buf, 1);
                if ( rc < 0 ) {
                        perror("mipstat: socket send message failed");
                        goto err;
                }       
                rc = recvUnixMessage((char *)&icmp_stats, 
			sizeof(struct icmp_stats));    
                if ( rc < 0 ) {
                        perror("mipstat: socket recv message failed");
                        goto err;
                }
                dumpBeacons(&icmp_stats, longflag);
        }

err:
	unlink(MOBYSOCK_CLIENT);
}

dump_hastats(struct ha_stats *s) 
{
	printf("ha statistics\n");
	printf("\ttimeouts: %d\n", s->timeouts);
	printf("\tinvalid mn auth: %d\n", s->invalidmnauth);
	printf("\tmissing mn auth: %d\n", s->missingmnauth);
	printf("\tinvalid fa auth: %d\n", s->invalidfaauth);
	printf("\tmissing fa auth: %d\n", s->missingfaauth);
	printf("\ttotal requests: %d\n", s->requests);
	printf("\ttotal deregistrations: %d\n", s->dereg);
	printf("\ttotal acks: %d\n", s->acks);
	printf("\ttotal naks: %d\n", s->naks);
	printf("\ttotal idmismatches: %d\n", s->idmismatches);
	printf("\ttotal idspoofs: %d\n", s->idspoofs);

	/* harp info */

	printf("\ttotal HARP forward packets sent: %d\n", s->harpforward);
	printf("\ttotal HARP forward packets recv: %d\n", s->harpforwrecv);
	printf("\ttotal HARP PING packets recv: %d\n", s->harpping);
	printf("\ttotal HARP ACK packets recv: %d\n", s->harpack);
	printf("\ttotal HARP UDP packets recv: %d\n", s->harpudpinput);
	printf("\ttotal HARP DUMP registrations recv: %d\n", s->harpdump);
	printf("\ttotal HARP DUMP MIP nat extensions recv: %d\n", s->harpdumpnat);
	printf("\ttotal HARP DUMP harp/nat extensions recv: %d\n", s->harpdumpharpnat);
	printf("\ttotal HARP DUMP connects: %d\n", s->harpconnects);
	printf("\ttotal HARP DUMP accepts: %d\n", s->harpaccepts);
	printf("\ttotal HARP forward has nat: %d\n", s->harpforwardhasnat);
	printf("\ttotal HARP forward has harp/nat: %d\n", s->harpforwardhasharpnat);
}

dump_fastats(struct fa_stats *s) 
{
	printf("fa statistics\n");
	printf("\ttotal requests: %d\n", s->requests);
	printf("\ttotal replies: %d\n", s->replies);
	printf("\tpending mns: %d\n", s->pending);
	printf("\tvisitor list count: %d\n", s->visitors);
	printf("\tmn timeouts: %d\n", s->timeouts);
	printf("\treplys with no mn present: %d\n", s->replydropped);
	printf("\tbad lifetimes: %d\n", s->badlifetime);
	printf("\tha reply ident mismatch: %d\n", s->identmismatch);
	printf("\tha denied mn: %d\n", s->hadenied);
	printf("\tmn requests, invalid authentication: %d\n", s->invalidmnauth);
	printf("\tmn requests, missing authentication: %d\n", s->missingmnauth);
	printf("\tha replys, invalid authentication: %d\n", s->invalidhaauth);
	printf("\tha replys, missing authentication: %d\n", s->missinghaauth);
}

dump_udperrs(struct udp_errors *u) 
{
	printf("overall registration/reply errors\n");
	printf("\tinvalid type code: %d\n", u->badtype);
	printf("\tregistration too short: %d\n", u->regtooshort);
	printf("\treply too short: %d\n", u->repltooshort);
	printf("\tregistration with no authentication extension: %d\n", 
		u->noauthent);
	printf("\ttoo many authentication extensions in packet: %d\n", 
		u->toomanyauthexts);
	printf("\tha got bogus request: %d\n", u->badrequest);
	printf("\tha got bogus reply: %d\n", u->badreply);
	printf("\tha sent fa registration: %d\n", u->hafaregistration);
	printf("\tfa got registration across wrong interface: %d\n", 
		u->fainvalidreg);
}
