/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * mipstate.c 
 *
 * input state machine for mipd to process fa/ha events.
 *
 * state machine + ha state machine code.  fa code
 * is in mip_fastate.c
 * 
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/times.h>
#include <errno.h>
#include <signal.h>
#include <string.h>
#include <net/if.h>
#include <netinet/ip_fw.h>

#include "mipconfig.h"
#include "mipstate.h"
#include "udp_reg.h"
#include "faha_pinfo.h"
#include "route.h"
#include "vlist.h"
#include "mipstats.h"
#include "unixsock.h"
#include "mip_log.h"
#include "alist.h"

#ifdef HARP
#include "harp.h"
#endif

#ifdef MADRP
#include "sr_mad.h"
#endif

struct ha_stats ha_stats;
struct fa_stats fa_stats;


/*
 * prototypes
 */

#ifdef HARP
void harpForward( int, struct packet_info *, int );	
extern harpSocket;
#endif

void
stateMachine(int, int, int, u_long *, struct sockaddr_in *, 
		struct packet_info *, struct mip_ifs *, int);

void 
stateIcmp( struct in_addr, unsigned short, unsigned short, int,
	char *, int, char *, int, int, char *, struct in_addr, u_long, u_long);

#ifdef DIVERT
void 
delFireWallRule(int number);
#endif

void 
haUdp(struct mip_ifs *, struct sockaddr_in *, struct packet_info *, int);

void
faUdp(struct mip_ifs *, u_long *, struct sockaddr_in *, struct packet_info *);

void
storeMd5(char *, int , unsigned char *, unsigned char *);

void
computeMd5(char *, int, char *, char *);

void haTick();

static void
sendUdpReply (struct sockaddr_in *, struct packet_info *, int, u_short, int);

void
sendOneUdp(int, char *, int , struct in_addr, u_short);

void
sendOneHaUdp(char *, int , struct in_addr, u_short);

static int
checkTimestamp(struct sockaddr_in *, struct vlist *, struct packet_info *);

char *
reverseAddress(struct in_addr);

char * 
reverseAddress1(struct in_addr);

int
ha_setroute(struct vlist *, struct mip_ifs *, struct sockaddr_in *, struct packet_info *, int);

static void route_delete_tunnel(struct vlist *v);

static void route_delete(struct vlist *v);

static void ha_tunnel_to_coa(struct vlist *v, struct in_addr, struct in_addr 
);

static void
route_do_arp_proxy( struct vlist *v, struct in_addr mn);

/* 
 * lists, one for ha and one for fa
 */
struct vlist *hahead = NULL;
#define IN             1
#define OUT            2
int udpsock;
extern harp2;
int doIpUdp = 0;

void
stateMachine(int sock, int event, int haFlag, u_long *if_addr, 
	struct sockaddr_in *from, struct packet_info *pinfo, 
	struct mip_ifs *mifs, int faifbutha)
{
	int rc = 0;
	extern int debug;

	udpsock = sock;

	if (debug) {
		if (event == UDPEVENT)
		printf("stateMachine TOP: event %d, haFlag %d, if %lx,  from %s faifbutha %d\n", 
			event, haFlag, *if_addr, inet_ntoa(from ->sin_addr),
			faifbutha);
	}

	/* print out packet info if debug is on
	*/
	if (debug && pinfo) {
		if (pinfo->reg) {
			dumpUdpRequest(pinfo->reg);
		}
		if (pinfo->rep) {
			dumpUdpReply(pinfo->rep);
		}
		if (pinfo->mn_ha_auth) {
			dumpMse(pinfo->mn_ha_auth);
		}
#ifdef DIVERT
		if (pinfo->nat_ext_req) {
		  	dumpNatReq(pinfo->nat_ext_req);
		}
#endif
		if (pinfo->fa_x_auth) {
			dumpMse(pinfo->fa_x_auth);
		}
	}

#ifdef DIVERT
	/* Set the hasNat flag if a Nat Registration Request Extension is found */
	if (pinfo) {
		if (pinfo->nat_ext_req) {
			pinfo->hasNat = 1;
		}
	}
#endif

	  switch(event) {
	case UDPEVENT:
		/* make security checks on packet; i.e., actually check
		 * the mse in terms of authentication.
		 * if a packet fails authentication, we don't tell
	 	 * anybody, but we log it as a security exception.
		 */
		rc = securityCheck(haFlag, from, pinfo);
		if (rc < 0) {
			if (debug) {
				printf("stateMachine(): securityCheck failed\n");
			}
			return;	/* NOTE: all errors logged within function */
		}
		if (haFlag) {
			if (mip_config.sawHa) {
				if (debug) {
					printf("stateMachine, HAUDP\n");
				}
				haUdp(mifs, from, pinfo, faifbutha);
			}
		}
		else {
			if (mip_config.sawFa) {
				if (debug) {
					printf("stateMachine, FAUDP\n");
				}
				faUdp(mifs, if_addr, from, pinfo);
			}
		}
		break;
	case TICKEVENT:
		if (mip_config.sawHa)
			haTick();
		if (mip_config.sawFa)
			faTick();
		break;
	}
}

void
stateTick()
{
	stateMachine(0, TICKEVENT, 0, NULL, NULL, NULL, NULL, 0);
}

/* icmp beacon from Mobile Node
 *
 * invariants:
 *	we know this beacon was authenticated.
 *	we know that we are not using arp
 *	we do not change real (non-initial) ha/fa states or do routing
 *		UNLESS the mn has switched interfaces on us and we are
 *		multihomed
 *	if node is new in terms of vlist, we create BEACON_NOSTATE state 
 *	which simply contains mac address for real state when UDP
 *	registration shows up elsewhere - MAC address is stored.
 *	
 * overview:
 *	find config structure
 *	determine if ha/fa list appropriate
 *	store info in existing vlist entry.  create and set state to
 *		BEACON_NOSTATE if it doesn't exist.
 *	if entry exists and mac address has changed
 *		delete link layer entry
 *		create new one
 */
struct mip_ifs *getConfigIf(struct in_addr *);

void
stateIcmp(
	struct in_addr bip, 	/* ip address of sender */
	unsigned short sequence,/* sequence number */
	unsigned short lifetime,/* lifetime requested */
	int auth_type,		/* auth bitmask (type of) */
	char *location,		/* location/name info for mn */
	int lsize,		/* size of location string */
	char *url,		/* url string */
	int usize,		/* sizeof url string */
	int sawMac,		/* mac is valid flag */
	char *macaddr,		/* mac address */
	struct in_addr ifaddr,  /* input if addr */
	u_long tssecs,		/* auth timestamp in secs */
	u_long tsfract          /* auth timestamp in fract */
        ) 
{

	struct mip_ifs *mifs;
	struct mip_ifs *mips;
	int i;
	struct vlist *v;
	extern struct vlist *fahead;
	extern int debug;
	int isha = 0;	/* beacon matches HA subnet on one interface */
	int new = 0;

	/* map ifaddr to interface pointer
	 */
	mifs = getConfigIf(&ifaddr);
	if (!mifs)
		return;

	/* find out if beacon matches any HA interface present
	 * here.  If so, we put MAC address in ha vlist.
	 * note that the interface pointer mifs points to
	 * the *input* i/f, which might be that of a FA.
	 */
	mips = &mip_config.mip_ifs[0];
 	for ( i = 0; i < mip_config.mip_ifcount; i++, mips++ ) {
        	if (mips->amfa) /* (1 if fa) */
                	continue;
		if ((bip.s_addr & mips->netmask.sin_addr.s_addr) ==
			(mips->if_sin.sin_addr.s_addr & mips->netmask.sin_addr.s_addr))  {
			isha = 1;
			break;
		}
	} 

        /* see if we are actually configured to be a ha/fa and
         * if not, toss the beacon acc. to whether we are its ha
         * or not.
         */

        /* we know we the HA for that system
        */
        if (isha) {
                /* but we aren't a HA period, so toss it 
                 * This check is probably worthless.
                */
                if (!mip_config.sawHa) {
                        return;
                }
        }
        /* are we a fa? */
        else {
                /* if we know we have no FA functionality, then we don't care
                */
                if (!mip_config.sawFa) {
                        return;
                }
        }

	if (debug) {
		printf("STATEICMP: isha %d\n", isha);
	}

	/* lookup in vlist acc. to ha/fa
	 */
	if (isha) {
		v = vlist_search(hahead, bip);
		if (!v) {
			v = vlist_create(&hahead, bip);
			ha_stats.visitors++;
			if (v == NULL) {
				fatal("stateIcmp(): out of memory in vlist");
			}
			new = 1;
			/* no state, but beacon info exists */
			v->vstate = BEACON_NOSTATE;	
			if (debug) {
				printf("STATEICMP: HA state is BEACON_NOSTATE\n");
			}
		}
	}
	else {
		v = vlist_search(fahead, bip);

		if (!v) {
			v = vlist_create(&fahead, bip);
			fa_stats.visitors++;
			if (v == NULL) {
				fatal("stateIcmp(): out of memory in vlist");
			}
			new = 1;
			/* no state, but beacon info exists */
			v->vstate = BEACON_NOSTATE;	
			if (debug) {
				printf("STATEICMP: FA state is BEACON_NOSTATE\n");
			}
		}
	}

#ifdef MADRP
        if (v && v->routeset == ROUTEHOST){
                struct SR_hlist * h;
                v->vstate = BEACON_NOSTATE;
                new = 1;
                h=(struct SR_hlist *)srhlist_search(v->mn);
                if(h){
                        srhlist_delete(h->destHost); 
                        route_delete_host(h->destHost,h->nextHop);
                } else {
#ifdef DEBUG
                        if(debug)
                        printf("stateicmp: can't find %s in hlist\n",
                                inet_ntoa(v->mn));
#endif
                }
        }
#endif 


	/* at this point we know we have a list entry
	*/

	/* fill cache entry in, with a few deviations 
	*/

	/* we only care about link_lifetime if the vlist entry
	 * is new else use the udp lifetime for timeout
	 * if new, log it.
	 */
	if (new) {
		char *whom = "fa";
		if (isha)
			whom="ha";
		mip_log( MLOG_INFO, 
		  "MN %s(%s) ICMP beacon has been received by %s",
		  reverseAddress(bip), inet_ntoa(bip), whom);

		v->lifetime = lifetime;
		if (debug) {
			printf("ICMP beacon lifetime %d\n", lifetime);
		}
	}

	/* if mifs has changed, we need to change link layer route
	 * TBD. think about this some more.
	*/
	if (!new && (v->vmifs != mifs) && (v->routeset == ROUTELINK)) {
		route_delete_link(bip, v->mac);
		route_add_link(bip, macaddr,
			mifs->ifname );
	}

	/* TBD. link_sequence check.  if overflow, refresh route entry
	 * can't figure out how this would be useful?
	if ((sequence < v->link_sequence) && (sequence <= 255)) {
	}
	v->link_sequence = sequence;
	*/

	v->vmifs = mifs;
	v->auth_type = auth_type;
	v->sawBeacon = 1;

	/* NOTE: sawMac is always == 1 */
	if (sawMac) {
		/* if route exists and mac has changed, refresh it
		*/
		if (bcmp(macaddr, v->mac, 6) != 0) {
			if (v->routeset == ROUTELINK) {
				route_delete_link(bip, v->mac);
				route_add_link(bip, macaddr,
					mifs->ifname);
				v->routeset = ROUTELINK;
				/* TBD. put mac addr in here old/new */
				mip_log( MLOG_SEC, 
				  "MN %s(%s) via ICMP changed MAC addr ",
				  reverseAddress(bip), inet_ntoa(bip));
			}
		}

		/* don't forget to store it.  After all, that's
		 * the major reason why this routine is here.
		 */
		bcopy(macaddr, v->mac, 6);
		v->sawMac = TRUE;
	}


	/* deal with location
	*/
	v->lsize = lsize;
	if (lsize) {
		if (lsize > MAXLOCATION)
			lsize = MAXLOCATION;
		bcopy(location, v->location, lsize);
		v->location[lsize] = 0;
	}
}

/* 
 *	This Function examines and determines if 2 IP 
 *	Address are equal. Used to determine if the MN is 
 *	registering from behind a NAT Machine.
 */
int 
isThereDiscrepancy(struct in_addr ad1,struct in_addr ad2)
{
	if ( (ad1.s_addr != 0) && (ad2.s_addr != 0) ) {          

		if (ad1.s_addr == ad2.s_addr)	  
			return 0;		  
		else
			return 1;		  

	}
	else{

			return -1;	  

	}
}

/* 
 * haUdp - manage MN state changes at HA
 * 
STATES:
------

HA_HOME
        the mn is at home
AT_COA
        the mn is at another HA that is a co-ha
HA_AWAY
        the mn is away (at a coa)
BEACON_NOSTATE
        ad hoc mode is in use, and the mn has sent
                a beacon with a MAC address; i.e., a vlist
                entry exists.  Not used with arp mode, but
                this info is per interface.
HA_NOSTATE
        the mn has no previous state

variables:
---------

v: valid if a vlist entry (index by ip) exists
dereg: boolean: true if mn is trying to deregister.  lifetime==0
arp: boolean: true if reg came in on arp/if as opposed to ad hoc i/f
COHA forward:  true if packet came from coha (not mn)
vlist timer: when 0, vlist entry is deleted.

events:
------

time out on vlist timeout
udp reg. (or dereg) packet came in

state transitions are driven by events + variables associated
        with events
=============================================================
transitions:
        X means can't happen.
        only variables that drive state change are noted.

HA_HOME to
        HA_HOME:
                dereg = 1
                forward = 0
        HA_NOSTATE:
                timeout (or arp mode)
        AT_COHA:
                forward = 1
                dereg = 1
                arp doesn't matter (at other ha)
        BEACON_NOSTATE:
                X
        HA_AWAY:
                dereg = 0
                forward = don't care
=============================================================
HA_NOSTATE to
        v = 0 by definition
        HA_HOME:
                dereg = 1
                forward = 0
        HA_NOSTATE:
                timeout (or HA_HOME in arp mode)
        AT_COHA:
                dereg = 1
                forward = 1
        BEACON_NOSTATE:
                ICMP message causes this.  no vlist entry.
                        mac stored. NO routes set.
        HA_AWAY:
                dereg = 0
                forward = don't care
=============================================================
AT_COHA to
        HA_HOME:
                dereg = 1
                forward = 0
        HA_NOSTATE:
                timeout
        AT_COHA:
                forward = 1
                dereg = 1
        BEACON_NOSTATE:
                X ( may enter through HA_HOME in arp mode )
        HA_AWAY:
                dereg = 0
                forward = don't care
=============================================================
BEACON_NOSTATE to
        HA_HOME:
                dereg = 1
                forward = 0
        HA_NOSTATE:
                timeout ( or HA_HOME in arp mode )
        AT_COHA:
                dereg = 1
                forward = 1
        BEACON_NOSTATE:
                icmp message only
        HA_AWAY:
                dereg = 0
                forward = don't care
=============================================================
HA_AWAY to
        HA_HOME:
                dereg = 1
                forward = 0
        HA_NOSTATE:
                timeout ( or HA_HOME in arp mode )
        AT_COHA:
                dereg = 1
                forward = 1
        BEACON_NOSTATE:
                X
        HA_AWAY:
                dereg = 0
                forward = don't care
---------------------------------------------------
pseudo code:

-----------------------------------------------------
HA_HOME
        if dereg
                if forward
                        state = AT_COHA
                        delete route
                        create tunnel route to co-HA
                        update timestamp
                        return
                else if !forward
                        state = HA_HOME
                        if doingArp
                                delete route
                                delete vlist entry
                                send reply
                                return  
                        if !doingArp
				if tunnel route || rosl changed
					delete route
					restore link route
                                update timestamp
				send reply
				return
	else if !dereg
		state = HA_AWAY
		delete route
		create tunnel route
		update lifetime
		if !forward
			send reply
		return
-----------------------------------------------------
AT_COHA
	if dereg
		if forward
			state = AT_COHA
			update lifetime
			if tunnel route || rosl changed
				delete route
				add tunnel route to co-HA
			return
		else if !forward
			state = HA_HOME
			if doingArp
				delete route
				delete vlist entry
			else if !doingArp
				delete route
				add link route
				update lifetime
			if !forward
				send reply
			return
	else if !dereg
		state = HA_AWAY
		delete route
		create tunnel route
		update lifetime
		if !forward
			send reply
		return
-------------------------------------------------------
HA_AWAY
	if dereg
		if forward
			state = AT_COHA
			delete route
			create tunnel route
			update lifetime
			return
		else if !forward
			state = HA_HOME
			if doingArp 
				delete route
				delete vlist entry
				send reply
				return
			if !doing Arp
				delete route
				update lifetime
				create link route
				send reply
				return
	else if !dereg
		state = HA_AWAY
		if coa is the same
			update lifetime
			if rosl changed
				if tunnel route
					delete tunnel
					create tunnel
				else if faifbutha
					delete link
					create link
		else if coa has changed
			delete route
			create new route
			update lifetime
		if !forward
			send reply
		return
----------------------------------------------------------
BEACON_NOSTATE
	if dereg
		if forward
			state = AT_COHA
			update lifetime
			create route
			return
		else if !forward
			state = HA_HOME
			update lifetime
			if !doingArp
				create link route
			send reply
			return
	else if !dereg
		state = HA_AWAY
		update lifetime
		create route
		if !forward
			send reply
			return

----------------------------------------------------------
HA_NOSTATE
	create vlist entry
	if dereg
		if forward
			state = AT_COHA
			update lifetime
			create route
			return
		else if !forward
			state = HA_HOME
			update lifetime
			if !doingArp
				create link route
			send reply
			return
	else if !dereg
		state = HA_AWAY
		create route
		update lifetime
		if !forward
			send reply
		return
------------------------------------------------------

 */
void
haUdp(
	struct mip_ifs *mifs,        /* the i/f pointer for packet */
	struct sockaddr_in *from,    /* ip/port actually sent us the pkt */
	struct packet_info *pinfo,   /* parsed reg. packet */
	int faifbutha	/* multihomed HA/FA */
	)
{
	struct vlist *v = NULL;
	int rc;
	int dereg = 0;		/* mn wants to deregister */
	char buf[1024];
	char rbuf[1024];
	extern int debug;
        extern struct vlist *fahead;
        int adhoc_create = 0;
	int tmproute = 0;
#ifdef HARP
	/* harp additions */  
	int forward = FALSE;
#endif
	int doingArp = FALSE;
	int state = FALSE;

	/* dns reverse lookup
	*/
	strncpy(rbuf, reverseAddress(from->sin_addr), 1024);

	ha_stats.requests++;
	if (debug) {
		printf("HAUDP: TOP: pkt from %s\n", rbuf);
	}

	/*
		mip_log( MLOG_INFO, "haUdp(): received registration packet from %s(%s)", 
		rbuf,
		inet_ntoa(from->sin_addr));
	*/

	if (pinfo->reg == NULL) {
		/* LOG an error message.  this shouldn't happen */
		mip_log( MLOG_ERR, "haUdp(): packet_info pointer is NULL.");
		return;
	}
	/* put name in local buffer for logging purposes
	*/
	strcpy(buf, inet_ntoa(from->sin_addr));

        /* access control: if the COA needs to be denied, send ADMIN PROHIBIT
           to MN, modified by Zhong to implement access control
        */ 
        if (pinfo->reg->coa.s_addr != 0) {
                if (access_control(pinfo->reg->coa) == DENIED) {
                        ha_stats.naks++;
                        sendUdpReply(from, pinfo, REG_FAIL_HA_PROHIBIT,
                                                from->sin_port, 0);
                        return;
                }
        }


	/* mn wants to deregister
	*/
	if (ntohs(pinfo->reg->lifetime) == 0) {
		dereg = 1;
		if (debug) {
			printf("mn wants to deregister\n");
		}
		ha_stats.dereg++;
	}


	/* look the mn up in our visitor's list
	 * if we don't find it, our state is NOSTATE
	 */
	v = vlist_search(hahead, pinfo->reg->home_address);

	if (debug && v) {
	    printf("HAUDP: Mn found in visitor list, state %d, hasmac %d\n", 
		v->vstate, v->sawMac);
	}

	/* if no vlist entry and arp is off on incoming interface
	 * and this is a dereg packet (on HA i/f) OR the packet
	 * came in on a FA interface but was at home (the MN
	 * supplies us with our own COA, therefore it's on the link).
	 * we can't do anything until there is a vlist entry
	 * which has to come in via an authenticated ICMP beacon 
	 * as we don't have a MAC address yet.
	 * Once we get a beacon, we can process the udp request.
	 */
	if (mifs && (v == NULL) && (mifs->doArp == FALSE) &&
	    (dereg || (faifbutha && 
		(pinfo->reg->coa.s_addr == mip_config.fa_coa.sin_addr.s_addr)))) {
#ifdef MADRP
                if( !srhlist_search(pinfo->reg->home_address))
#endif
                {
                        if (debug) {
                                printf( 
                                "HAUDP: ha tosses deregistration as ad hoc and no mac\n");
                        }
                        ha_stats.nonetauth++; 
                        return;
                }

	}

	/* mn wants flags.  we get grumpy as we don't
	 * support ANY flags but D_BIT, which is mn's job anyway
	 */
	if (pinfo->reg->flags) {
		/* let D_BIT (mn doing coa) through though	
		*/
		if (pinfo->reg->flags == D_BIT) {
			;
		}
		/* else we can't do it and we aren't going to explain why
		*/
		else {
 			/* temporary route for ad hoc mode 
			*/
                        if (v && v->vstate == BEACON_NOSTATE && !v->routeset){
                                tmproute = 1;
                                route_add_link(pinfo->reg->home_address,
                                    v->mac, v->vmifs->ifname);
                                v->routeset = ROUTELINK;
                        }
                        sendUdpReply(from, pinfo, REG_FAIL_HA_UNSPEC,
                                from->sin_port, 0);
			/* ad hoc - tear route down
			*/
                        if (tmproute) {
                                route_delete_link(v->mn, v->mac);
                                v->routeset = 0;
                        }
			return;
		}
	}
	/* determine what we think
	 * about timestamp freshness
	 *
	 * 3 possible results:
	 * 1. TS from MN is judged fresh and valid.
 	 * 2. TS from MN is judged out of whack timewise.
	 *	We send NAK.
	 * 3. TS from MN is judged non-fresh.  We drop the packet
	 *	and log the result.
	 *
	 */

	if (pinfo->timeflag && mip_config.doTimestamp 
#ifdef HARP
		&& !pinfo->forward 
#endif
		) {
		rc = checkTimestamp(from, v, pinfo);
		if (rc == -1) {
		       /* if BEACON state (ad hoc), set tmp route else
			* we can't talk to the mn
		        */
			if (v && (v->vstate == BEACON_NOSTATE) && v->sawMac
				&& !v->routeset) {
				if (debug) {
					printf("HAUDP: timestamp nak: BEACON_NOSTATE\n");
				}
				route_add_link(pinfo->reg->home_address,
				    v->mac, v->vmifs->ifname);
				v->routeset = ROUTELINK;
				tmproute = 1;
			}
			if (debug) {
				printf("HAUDP: ha timestamp failure\n");
			}
			sendUdpReply(from, pinfo, 
				REG_FAIL_HA_IDMISMATCH, from->sin_port, 0);
			if (tmproute) {
                                route_delete_link(v->mn, v->mac);
                                v->routeset = 0;
			}
			return;
		}
		/* else drop packet, already logged */
		else if (rc == -2) {
				return;
		}
	}
	/* 
	 * else NONCE code goes here 
	 * TBD:  
	 * this code should make a weak check (in the sense that we do
	 * not assume stable storage.  E.g., make sure nonce is always
	 * increasing assuming you have it.
	 */

	/* if we are in the list, that means we have a state
	 * note: if arp is off and registration, we need to install
	 * link entry, and set state HA_HOME.  This is only in case
	 * of deregistration packet.
	 *
	 * in all cases where the route to the MN is not deleted for
	 * other reasons, check whether the ipsec policy for the node has
	 * changed in case the route must be refreshed with the new setting
	 * before sending more data.
	 */

	if ( v ){
		state = v->vstate;
#ifdef HARP
		if ( debug && forward ){
			printf( "HARP: haUdp(): forwarded packet, updating information for %s (in vlist)\n", inet_ntoa( v->mn ));
		}
#endif
	}
	else {
		state = HA_NOSTATE;
	}

	if ( mifs && mifs->doArp ){
		doingArp = TRUE;
	}
	else {
		doingArp = FALSE;
	}
#ifdef HARP
	if ( pinfo->forward ){
		forward = TRUE;
		if ( debug ){
			printf( "HARP: haUdp(): forwarded udp packet\n" );
		}
	}
	else {
		forward = FALSE;
	}

#endif

	switch( state ){
		case HA_HOME:
			if ( dereg ){ 
#ifdef HARP
				if ( forward && mip_config.harp_tunnel ){ /* going to AT_COHA state */
					if ( debug ){
						printf( "HARP:haUdp():STATE: transition from HA_HOME to AT_COHA\n" );
					}
					v->vstate = AT_COHA;
					route_delete( v );
					ha_setroute( v, mifs, from, pinfo,
						faifbutha );
					v->orig_lifetime = ntohs( pinfo->reg
						->lifetime );
					v->lifetime = HALIFETIME;
					v->i.l[0] = pinfo->reg->i.l[0];
					v->i.l[1] = pinfo->reg->i.l[1];

					/* change to udp input addr? */
					/*v->coa = pinfo->reg->coa;*/
					v->coa = mip_config.coha.sin_addr;
					v->vmifs = mifs;
					return;
					/* send no reply */

				}
				else { /* still in HA_HOME */
#endif
					if ( debug ){
						printf( "haUdp():STATE: steady state in HA_HOME\n" );
					}
					              /* refresh route */	
					if ( doingArp ){ /* TBD fix in arp/ipsec */
						route_delete( v );
						vlist_delete( &hahead, 
							pinfo->reg->home_address );
						ha_stats.visitors--;
						sendUdpReply( from, pinfo,
							REG_REPLY_OK, from->sin_port, v->lifetime );
						return;
					}
					else if ( !doingArp ){
						if ( v->routeset == ROUTETUNNEL)
						{
							route_delete( v );
							route_add_link(
								pinfo->reg->home_address,
								v->mac,
								v->vmifs->ifname
								);
							v->routeset = ROUTELINK;
						}
						v->orig_lifetime = ntohs(
							pinfo->reg->lifetime );
						v->lifetime = HALIFETIME;
						v->i.l[0] = pinfo->reg->i.l[0];
						v->i.l[1] = pinfo->reg->i.l[1];
						sendUdpReply( from, pinfo, 
							REG_REPLY_OK,
							from->sin_port,
							v->lifetime );
						return;
					}
#ifdef HARP
				}
#endif
			}
			/* going to HA_AWAY
			*/
			else if ( !dereg ){
				if ( debug ){
					printf( "haUdp():STATE: transition from HA_HOME to HA_AWAY\n" );
				}
				v->vstate = HA_AWAY;
				route_delete( v );
				v->orig_lifetime = ntohs(pinfo->reg->lifetime);
				v->lifetime = mip_config.mobileLifetime;
				if (v->lifetime > v->orig_lifetime) {
					v->lifetime = v->orig_lifetime;
				}
				v->i.l[0] = pinfo->reg->i.l[0];
				v->i.l[1] = pinfo->reg->i.l[1];
				v->coa = pinfo->reg->coa;
				v->vmifs = mifs;
				ha_setroute(v, mifs, from, pinfo, faifbutha);
#ifdef HARP
				if ( !forward ){
#endif
					sendUdpReply( from, pinfo, REG_REPLY_OK,
						from->sin_port, v->lifetime );
#ifdef HARP
				}
#endif
				return;
			}
			break;

#ifdef HARP
		case AT_COHA:
			if ( dereg ){
				if ( forward ){ /* still AT_COHA */
					if ( debug ){
						printf( "HARP: haUdp():STATE: steady state in AT_COHA\n" );
					}
					v->orig_lifetime = ntohs(
						pinfo->reg->lifetime );
					v->lifetime = HALIFETIME;
					v->i.l[0] = pinfo->reg->i.l[0];
					v->i.l[1] = pinfo->reg->i.l[1];
					v->vmifs = mifs;
					return;
					/* send no reply */
				}
				else if ( !forward ){ /* going to HA_HOME */
					if ( debug ){
						printf( "haUdp():STATE: state transition from AT_COHA to HA_HOME\n" );
					}
					if ( doingArp ){
						route_delete( v );
						vlist_delete( &hahead, 
							pinfo->reg->home_address );
						ha_stats.visitors--;
					}
					else if ( !doingArp ){
						v->vstate = HA_HOME;
						route_delete( v );
						v->vmifs = mifs;
						route_add_link( pinfo->reg->home_address,
							v->mac, v->vmifs->ifname
						);
						v->routeset = ROUTELINK;
						v->orig_lifetime = ntohs(
							pinfo->reg->lifetime );
						v->lifetime = HALIFETIME;
						v->ha = pinfo->reg->home_agent;
						v->mn = pinfo->reg->home_address;
						v->coa.s_addr = 0;
						v->i.l[0] = pinfo->reg->i.l[0];
						v->i.l[1] = pinfo->reg->i.l[1];
					}
					sendUdpReply( from, pinfo, REG_REPLY_OK,
						from->sin_port, v->lifetime );
					return;
				}	
			}
			/* going to HA_AWAY */
			else if ( !dereg ){
				if ( debug ){
					printf( "haUdp():STATE: transition from AT_COHA to HA_AWAY\n" );
				}
				v->vstate = HA_AWAY;
				route_delete( v );
				ha_setroute( v, mifs, from, pinfo, faifbutha );
				v->orig_lifetime = ntohs(pinfo->reg->lifetime);
				v->lifetime = mip_config.mobileLifetime;
				if (v->lifetime > v->orig_lifetime) {
                                        v->lifetime = v->orig_lifetime;
                                }
				v->i.l[0] = pinfo->reg->i.l[0];
				v->i.l[1] = pinfo->reg->i.l[1];
				v->coa = pinfo->reg->coa;
				v->vmifs = mifs;
				return;
			}
			break;
#endif
		case HA_AWAY:
			if ( dereg ){
#ifdef HARP
				if ( forward && mip_config.harp_tunnel ){  /* going to AT_COHA */
					if ( debug ){
						printf( "HARP: haUdp():STATE: transition from HA_AWAY to AT_COHA\n" );
					}
					route_delete( v );
					v->vstate = AT_COHA;
					v->orig_lifetime = ntohs( pinfo->reg
						->lifetime );
					v->lifetime = HALIFETIME;
					v->i.l[0] = pinfo->reg->i.l[0];
					v->i.l[1] = pinfo->reg->i.l[1];
					/*v->coa = pinfo->reg->coa;*/
					v->coa = mip_config.coha.sin_addr;
					v->vmifs = mifs;
					ha_setroute( v, mifs, from, pinfo,
						faifbutha );
					/* send no reply */
					return;
				}
				else 
#ifdef HARP
				if ((!forward) || (!mip_config.harp_tunnel))
#endif
				{ /* going to HA_HOME */
#endif
					if ( debug ){
						printf( "haUdp():STATE: transition from HA_AWAY to HA_HOME\n" );
#ifdef HARP
						printf( "forward=%d mip_config.harp_tunnel=%d\n", forward, mip_config.harp_tunnel );
#endif
					}
					if ( doingArp ){
						route_delete( v );
						vlist_delete( &hahead, 
							pinfo->reg->home_address );
						ha_stats.visitors--;
					}
					else if ( !doingArp ){
						v->vstate = HA_HOME;
						route_delete( v );
						v->routeset = ROUTELINK;
						v->orig_lifetime = ntohs( 
							pinfo->reg->lifetime );
						v->lifetime = HALIFETIME;
						v->i.l[0] = pinfo->reg->i.l[0];
						v->i.l[1] = pinfo->reg->i.l[1];
						v->vmifs = mifs;
						v->ha = pinfo->reg->home_agent;
						v->mn = pinfo->reg->home_address;
						v->coa.s_addr = 0;
						route_add_link( pinfo->reg->home_address,
							v->mac, v->vmifs->ifname
							);
						if ( v->vmifs == NULL ){
							mip_log( MLOG_ERR, "HARP:haUdp(): going to AT_HOME state with null interface. Is harp_tunnel set?\n" );
						}  
					}
					sendUdpReply( from, pinfo, REG_REPLY_OK,
						from->sin_port, v->lifetime );
					return;
#ifdef HARP
				}
#endif
			}			
			else if ( !dereg ){
				/* stable state, coa still the same */
				if ( v->coa.s_addr == pinfo->reg->coa.s_addr ){
					if ( debug ){
						printf( "haUdp():STATE: steady state in HA_AWAY\n" );
#ifdef HARP
						if ( forward ){
							printf( "HARP: haUdp(): HA_AWAY: forwarded packet reveals stable state\n" );
						}
#endif
					}
					v->orig_lifetime = ntohs(pinfo->reg->lifetime);
					v->lifetime = mip_config.mobileLifetime;
#ifdef DIVERT  
					v->hasNat  = pinfo->hasNat;
#endif
					if (v->lifetime > v->orig_lifetime) {  
                                                v->lifetime = v->orig_lifetime;
                                        }

#ifdef HARP
#ifdef DIVERT
					if (forward) {
						if (pinfo->hasNat) {
							v->udp_src.sin_addr.s_addr  =
								ntohl(pinfo->harp_nat_ext->natSrc.s_addr);
							v->udp_port = ntohs(pinfo->harp_nat_ext->srcPort);

							v->length= pinfo->nat_ext_req->length;
							v->reserved_1= pinfo->nat_ext_req->reserved_1;
							v->reserved_2= pinfo->nat_ext_req->reserved_2;
							v->Encapsulation= pinfo->nat_ext_req->Encapsulation;
							v->reserved_3= pinfo->nat_ext_req->reserved_3;

                                                  }
					}
#endif
#endif
	
					if (!forward) {
						pinfo->udp_src = *from;
                                                pinfo->udp_port = from->sin_port;
                                	}

#ifdef DIVERT
					/* Steady State handling of the Routing setup 
					 *
					 *  IPUDPTUNNEL <---> IPUDPTUNNEL / IPIPTUNNEL
					 *
					 * If NatReqExt exits and Current Routing Setup is IPUDPTUNNEL
					 *	F_BIT Set
					 *		setup parameters for IPUDPTUNNEL.
					 *
					 *	F_BIT Not Set
					 *		There is discrepancy between CAO and Ip Src
					 *			setup parameters for IPUDPTUNNEL.
					 *		Else (do IPIP TUNNEL)
					 *			Reset parameters that setup IPUDPTUNNEL
					 *			Delete Firewall Rules/MN.
					 *			setup IPIP Tunnel.
					 */
					if ( (pinfo->hasNat) && (v->routeset == IPUDPTUNNEL) ){

						if (pinfo->nat_ext_req->reserved_2 == F_BIT) {
							v->doIpUdp        = 1;	
							pinfo->sendNatRep = 1; 
						} else{
#ifdef HARP
							if ( forward ) {
								if ( isThereDiscrepancy(v->udp_src.sin_addr
												,v->coa)) {
									
									v->doIpUdp        = 1;
									pinfo->sendNatRep = 1;

								} else {

									v->doIpUdp        = 0;
									pinfo->sendNatRep = 0;
									delFireWallRule(v->fr_number);
									ha_setroute( v, mifs, from, pinfo, faifbutha );

								}
							} 
#endif
							if (!forward) {
								if ( isThereDiscrepancy(from->sin_addr,v->coa) ) {

									v->doIpUdp        = 1;	
									pinfo->sendNatRep = 1;
 
								} else {

									v->doIpUdp        = 0;	
									pinfo->sendNatRep = 0;
									delFireWallRule(v->fr_number);
									ha_setroute( v, mifs, from, pinfo, faifbutha );

								} 
							}
						}

					}
					
					/* Steady State handling of the Routing setup 
					 *
					 * IPIPTUNNEL <--> IPIPTUNNEL / IPUDPTUNNEL
					 *
					 * if Nat Extension is Present Current Routing Setup is IPIPTUNNEL
					 *	F_BIT is Set:
					 *		set parameters that enable NAT Tunnel
					 *		setup IPUDPTUNNEL.
					 *
					 *	F_BIT is not Set:
					 *		Turn-off NAT Tunnel setup parameters	
					 */
					if ( (pinfo->hasNat) && (v->routeset == ROUTETUNNEL)) {

						if (pinfo->nat_ext_req->reserved_2 == F_BIT) {
							v->doIpUdp        = 1;	
							pinfo->sendNatRep = 1; 
							v->fr_number      = 0;
							ha_setroute( v, mifs, from, pinfo, faifbutha );
						} else {
							v->doIpUdp        = 0;
							pinfo->sendNatRep = 0; 
						}

					}
#endif

					/* store identity
					*/
					v->i.l[0] = pinfo->reg->i.l[0];
					v->i.l[1] = pinfo->reg->i.l[1];
					v->vmifs = mifs;
				}
				else {	/* coa has changed */
					char buf2[1024]; 
					char buf3[1024];
					strcpy(buf2, inet_ntoa(v->mn));

					v->vstate = HA_AWAY;
					if (debug) {
						printf("HAUDP: not dereg: coa changed \n");
#ifdef HARP
						if ( forward ){
							printf( "HARP: haUdp(): HA_AWAY: forwarded packet indicates COA changed\n" );
						}
#endif
					}
					/* update list entry including timer
					*/
					v->orig_lifetime = ntohs(pinfo->reg->lifetime);
					v->lifetime = mip_config.mobileLifetime;
#ifdef DIVERT
					v->hasNat   = pinfo->hasNat;
#endif
					if (v->lifetime > v->orig_lifetime) {
                                                v->lifetime = v->orig_lifetime;
                                        }
#ifdef HARP
#ifdef DIVERT
					if (forward){
						if (pinfo->hasNat){
							v->udp_src.sin_addr.s_addr = ntohl(pinfo->harp_nat_ext
											->natSrc.s_addr);
							v->udp_port = ntohs(pinfo->harp_nat_ext->srcPort);

							v->length= pinfo->nat_ext_req->length;
							v->reserved_1= pinfo->nat_ext_req->reserved_1;
							v->reserved_2= pinfo->nat_ext_req->reserved_2;
							v->Encapsulation= pinfo->nat_ext_req->Encapsulation;
							v->reserved_3= pinfo->nat_ext_req->reserved_3;
						}
					}
#endif 
#endif 
					
					if (!forward) {
#ifdef HARP
						pinfo->udp_src = *from;
						pinfo->udp_port = from->sin_port;
#endif

						v->udp_src = *from;
						v->udp_port = from->sin_port;

					}

					/* store identity
					*/
					v->i.l[0] = pinfo->reg->i.l[0];
					v->i.l[1] = pinfo->reg->i.l[1];
					v->coa = pinfo->reg->coa;
					strcpy(buf3, inet_ntoa(v->coa));
					v->vmifs = mifs;

#ifdef DIVERT
					/* delete previous route/firewall rules and add new one
					*/
					if (v->routeset == IPUDPTUNNEL){
						if (v->fr_number)
							delFireWallRule(v->fr_number);
					}
#endif

					if (v->routeset == ROUTETUNNEL)
						route_delete(v);

#ifdef DIVERT
					/* if Nat Extension is Present
					 *	F_BIT is Set:
					 *		set parameters that enable NAT Tunnel
					 *
					 *	F_BIT is not Set:
					 *		There is Discrepancy between IP Src and COA
					 *			set parameters that enable NAT Tunnel
					 *		else
					 *			Turn-off NAT Tunnel setup parameters
					 */
					if (pinfo->hasNat){
						if (pinfo->nat_ext_req->reserved_2 == F_BIT) {
							v->doIpUdp        = 1;	
							pinfo->sendNatRep = 1; 
							v->fr_number      = 0;
						} else{
							if ( isThereDiscrepancy(from->sin_addr,v->coa) ) {
								v->doIpUdp        = 1;	
								pinfo->sendNatRep = 1; 
								v->fr_number      = 0;
							} else {
								v->doIpUdp        = 0;	
								pinfo->sendNatRep = 0; 
							} 
						}
					}
#endif

					if (debug) {
					    printf("HAUDP: call on ha_setroute, faifbutha %d\n", faifbutha);
					}

					
					ha_setroute(v, mifs, from, pinfo, faifbutha);
					mip_log( MLOG_INFO, 
						  "haUdp(): MN %s(%s) is away, new coa is %s",
						  reverseAddress(v->mn),
				  		  buf2, buf3);
	
				}
#ifdef HARP
				if ( !forward ){
#endif
					sendUdpReply( from, pinfo, REG_REPLY_OK,
						from->sin_port, v->lifetime );
#ifdef HARP
				}
#endif
				return;
			}
		case BEACON_NOSTATE:
			if ( dereg ){
#ifdef HARP
				if ( forward && mip_config.harp_tunnel ){ /* going to AT_COHA */
					if ( debug ){
						printf( "HARP: haUdp():STATE: forwared packet, state transition from BEACON_NOSTATE to AT_COHA\n" );
					}
					route_delete( v );
					v->vstate = AT_COHA;
 					/*v->coa = pinfo->reg->coa;*/
					v->coa = mip_config.coha.sin_addr;
					v->vmifs = mifs;
					v->orig_lifetime = ntohs( pinfo->reg->
						lifetime );
					ha_setroute( v, mifs, from, pinfo, 
						faifbutha );
					v->lifetime = HALIFETIME;
					v->i.l[0] = pinfo->reg->i.l[0];
					v->i.l[1] = pinfo->reg->i.l[1];
					return;
				}
				else if ((!forward ) || (!mip_config.harp_tunnel)){ /* going to HA_HOME */
#endif
					if ( debug ){
						printf( "haUdp():STATE: state transition from BEACON_NOSTATE to HA_HOME\n" );
					}
					route_delete( v );
					v->vstate = HA_HOME;
					v->orig_lifetime = ntohs( 
						pinfo->reg->lifetime );
					v->lifetime = HALIFETIME;
					if ( !doingArp ){
						route_add_link( pinfo->reg->home_address,
							v->mac, v->vmifs->ifname
							);
						v->routeset = ROUTELINK;
						v->i.l[0] = pinfo->reg->i.l[0];
						v->i.l[1] = pinfo->reg->i.l[1];
						v->vmifs = mifs;
						v->ha = pinfo->reg->home_agent;
						v->mn = pinfo->reg->home_address;
						v->coa.s_addr = 0;
					}
					sendUdpReply( from, pinfo, REG_REPLY_OK,
						from->sin_port, v->lifetime );
#ifdef HARP
				}
#endif
				return;
			}
			else if ( !dereg ){
				if ( debug ){
					printf( "haUdp():STATE: state transition from BEACON_NOSTATE to HA_AWAY\n" );
				}
				v->vstate = HA_AWAY;
				ha_stats.visitors++;
				route_delete( v );
				v->udp_src = *from;
				v->udp_port = from->sin_port;
				v->ha = pinfo->reg->home_agent;
				v->mn = pinfo->reg->home_address;
				v->coa = pinfo->reg->coa;
				v->vmifs = mifs;
				v->orig_lifetime = ntohs( pinfo->reg->lifetime );
				v->lifetime = mip_config.mobileLifetime;
				if (v->lifetime > v->orig_lifetime) {
                                        v->lifetime = v->orig_lifetime;
                                }                       
				v->i.l[0] = v->i.l[0];
				ha_setroute( v, mifs, from, pinfo, faifbutha );
#ifdef HARP
				if ( !forward ){
#endif
					sendUdpReply( from, pinfo, REG_REPLY_OK, from->sin_port, v->lifetime );
#ifdef HARP
				}
#endif
				return;
			}
		case HA_NOSTATE:
			v = vlist_create( &hahead, pinfo->reg->home_address );
#ifdef HARP
			if ( debug && forward ){
				printf( "HARP: haUdp(): creating vlist entry for forwarded packet mn %s\n", inet_ntoa( pinfo->reg->home_address ));
			}
#endif
			if ( v == NULL ){
				fatal( "haUdp(): out of memory in vlist" );
			}
			if ( dereg ){
#ifdef HARP
				if ( forward && mip_config.harp_tunnel ){  /* going to AT_COHA */
					if ( debug ){
						printf( "HARP: haUdp():STATE: forwarded packet, state transition from HA_NOSTATE to AT_COHA\n" );
					}
					v->vstate = AT_COHA;
					/*v->coa = pinfo->reg->coa;*/
					v->coa = mip_config.coha.sin_addr;
					v->vmifs = mifs;
					v->orig_lifetime = ntohs( pinfo->reg->
						lifetime );
					ha_setroute( v, mifs, from, pinfo, 
						faifbutha );
					v->lifetime = HALIFETIME;
					v->i.l[0] = pinfo->reg->i.l[0];
					v->i.l[1] = pinfo->reg->i.l[1];
					/* send no reply */
					return;
				}
				else if ((!forward) || (!mip_config.harp_tunnel)){ /* going to HA_HOME */
#endif
					if ( debug ){
						printf( "haUdp():STATE: state transition from HA_NOSTATE to HA_HOME\n" );
					}
					v->vstate = HA_HOME;
					v->orig_lifetime = ntohs( 
						pinfo->reg->lifetime );
					v->lifetime = HALIFETIME;
					if ( !doingArp ){
						route_add_link( pinfo->reg->home_address,
							v->mac, v->vmifs->ifname
							);
						v->routeset = ROUTELINK;
#ifdef MADRP                    
                                                if(srhlist_search(v->mn)){
                                                        v->routeset=ROUTEHOST;
                                                } else
#endif
                                                /* added by Jen */ 
                                                ha_stats.visitors++;

						v->i.l[0] = pinfo->reg->i.l[0];
						v->i.l[1] = pinfo->reg->i.l[1];
						v->vmifs = mifs;
						v->ha = pinfo->reg->home_agent;
						v->mn = pinfo->reg->home_address;
						v->coa.s_addr = 0;
					}
#ifdef HARP
					if ( !forward ){
#endif
						sendUdpReply( from, pinfo, REG_REPLY_OK,
							from->sin_port, v->lifetime );
#ifdef HARP
					}
#endif
					return;
#ifdef HARP
				}
#endif
			}
			else if ( !dereg ){  /* going to HA_AWAY */
				if ( debug ){
					printf( "haUdp():STATE: state transition from HA_NOSTATE to HA_AWAY\n" );
#ifdef HARP
					if ( forward ){
						printf( "HARP: haUdp(): forward packet, state transition from HA_NOSTATE to HA_AWAY\n" );
					}
#endif
				}
				v->vstate = HA_AWAY;  
				ha_stats.visitors++;
				/* if we are doing harp and there is a NAT Extension following the 
				 * MIP Registration...
				 *  
				 * If (forwarded by COHA) {
				 *          if (hasNat) {
				 *                 Grab the Nat IP Src from the HARP_NAT Extension.
				 *                 Grab the Src Port from the HARP_NAT Extension.
				 *                 Feed these to the vlist.
				 *          }
				 * }
				 */
#ifdef HARP
#ifdef DIVERT
				if (forward) {
					if (pinfo->hasNat) {
						v->udp_src.sin_addr.s_addr  = ntohl(pinfo->harp_nat_ext
											->natSrc.s_addr);
						v->udp_port = ntohs(pinfo->harp_nat_ext->srcPort);

						v->length= pinfo->nat_ext_req->length;
                                        	v->reserved_1= pinfo->nat_ext_req->reserved_1;
                                        	v->reserved_2= pinfo->nat_ext_req->reserved_2;
                                        	v->Encapsulation= pinfo->nat_ext_req->Encapsulation;
                                        	v->reserved_3= pinfo->nat_ext_req->reserved_3;
							
					} 
				  
				}
#endif
#endif
				if ( !forward ){
#ifdef HARP
					pinfo->udp_src = *from;
					pinfo->udp_port = from->sin_port;
#endif

					v->udp_src  = *from;
					v->udp_port = from->sin_port; 
				}

#ifdef DIVERT
				v->hasNat  = pinfo->hasNat;
#endif
				v->ha = pinfo->reg->home_agent;
				v->mn = pinfo->reg->home_address;
				v->coa = pinfo->reg->coa;
				v->vmifs = mifs;
				v->orig_lifetime = ntohs( pinfo->reg->lifetime );
				v->lifetime = mip_config.mobileLifetime;
				if (v->lifetime > v->orig_lifetime) {
                                        v->lifetime = v->orig_lifetime;
                                }                       

				if ( debug ){
					printf("haUdp(): orig_lifetime set to %d, lifetime set to %d\n", v->orig_lifetime, v->lifetime);
					printf("The Value of pinfo->hasNAt --> %d \n",pinfo->hasNat);
				}
				v->i.l[0] = pinfo->reg->i.l[0];
				v->i.l[1] = pinfo->reg->i.l[1];

#ifdef DIVERT
#ifdef HARP
				if (!forward && pinfo->hasNat) {
					v->length            = pinfo->nat_ext_req->length;
					v->reserved_1        = pinfo->nat_ext_req->reserved_1;
					v->reserved_2        = pinfo->nat_ext_req->reserved_2;
					v->Encapsulation     = pinfo->nat_ext_req->Encapsulation;
					v->reserved_3        = pinfo->nat_ext_req->reserved_3;
				}
#endif

				/* if Nat Extension is Present
				 *	F_BIT is Set:
				 *		set parameters that enable NAT Tunnel
				 *
				 *	F_BIT is not Set:
				 *		There is Discrepancy between IP Src and COA
				 *			set parameters that enable NAT Tunnel
				 *		else
				 *			Turn-off NAT Tunnel setup parameters
				 */
				if (pinfo->hasNat){
					if (pinfo->nat_ext_req->reserved_2 == F_BIT) {
						v->doIpUdp        = 1;	
						pinfo->sendNatRep = 1; 
						if (v->fr_number > 0)
							delFireWallRule(v->fr_number);
						v->fr_number      = 0;
					} else{
#ifdef HARP
						if (forward) {
							if ( isThereDiscrepancy(v->udp_src.sin_addr,v->coa)) {
						
								v->doIpUdp        = 1; 
								pinfo->sendNatRep = 1;
								v->fr_number      = 0;
							} else {

								v->doIpUdp        = 0;
								pinfo->sendNatRep = 0;
							}
						}
#endif
						if (!forward ) {
							if ( isThereDiscrepancy(from->sin_addr,v->coa) ) {

								v->doIpUdp        = 1;	
								pinfo->sendNatRep = 1; 
								v->fr_number      = 0;

							} else{

								v->doIpUdp        = 0;	
								pinfo->sendNatRep = 0;
 
							}
						}
					} 
				}
#endif  /* DIVERT */
				
				ha_setroute( v, mifs, from, pinfo, faifbutha );

#ifdef HARP
				if ( !forward ){
#endif
					sendUdpReply( from, pinfo, REG_REPLY_OK, from->sin_port, v->lifetime );
#ifdef HARP
				}
#endif
				return;
			}
		}
}  /* end haUdp() */

/* access control check                         
 *  return DENIED  if the COA is denied
 *         ALLOWED if the COA is allowed
 */                     
 
int access_control(struct in_addr coa)
{
        char ip[16];
        char name[80], rname[80];
        struct element * list; 
        int allowed, len, iplen, namelen, permission, found;
 
        strcpy(ip, inet_ntoa(coa));
        strcpy(name, reverseAddress(coa));     
        strrcpy(rname, name);       
 
        iplen = strlen(ip);
        namelen = strlen(name);
 
        allowed = mip_config.access_control_default_permission;
                                        
        if (mip_config.access_control_order == DENY_ALLOW) {
                list = mip_config.denylist;
                permission = DENIED; 
        }
        else { 
                list = mip_config.allowlist;
                permission = ALLOWED; 
        }
 
        found = FALSE;
        while (list && !found) {
                len = strlen(list->address);
                if (list->type == ACCESS_CONTROL_IP_TYPE) {
                        if (len <= iplen &&
                                strncmp(list->address, ip, len) == 0) {
                                allowed = permission;
                                found = TRUE;
                        }
                }
                else {
                        if (len <= namelen &&
                                strncasecmp(list->address, rname, len) == 0) {
                                allowed = permission;
                                found = TRUE;
                        }
                }
                list = list->next;
        }

        if (mip_config.access_control_order == DENY_ALLOW) {
                list = mip_config.allowlist;
                permission = ALLOWED;
        }
        else {
                list = mip_config.denylist;
                permission = DENIED;
        }

        found = FALSE;
        while (list && !found) {
                len = strlen(list->address);
                if (list->type == ACCESS_CONTROL_IP_TYPE) {
                        if (len <= iplen &&
                                strncmp(list->address, ip, len) == 0) {
                                allowed = permission;
                                found = TRUE;
                        }
                }
                else {
                        if (len <= namelen &&
                                strncasecmp(list->address, rname, len) == 0) {
                                allowed = permission;
                                found = TRUE;
                        }
                }
                list = list->next;
        }

        return allowed;
}

/*
 * from ha side, setup route to mn that is AWAY
 *     
 * typically a tunnel but there is a    
 * special case: at multi-homed fa/ha, MN actually
 * at FA, so we treat as AWAY, and steal info
 * from fa interface.  In this case the route
 * is a link-layer route, not a tunnel as we can't
 * tunnel to ourselves.  In addition, if a proxy arp 
 * route is called for at the HA side, it is put in.
 *
 * returns:
 *	0: failed
 *	1: ok
 *
 * NOTE: this function never fails.
 */

int
ha_setroute(
	struct vlist *v,	     /* ha vlist entry */
	struct mip_ifs *mifs,        /* the i/f pointer for packet */
	struct sockaddr_in *from,    /* who actually sent us the pkt */
	struct packet_info *pinfo,   /* parsed reg. packet */
	int faifbutha	/* multihomed HA/FA */
	)
{
	char *mac = NULL;
	extern int debug;
	struct in_addr srcAddr ;
	struct in_addr dstAddr ;

	/* "tunnel" to the FA on the same machine
	 *  but only if the mipconfig COA matches the mn coa,
	 *  else the MN may be acting as its own coa.
	 */
	if (faifbutha && (mip_config.fa_coa.sin_addr.s_addr == v->coa.s_addr)) {
		route_add_link(pinfo->reg->home_address,
                     v->mac, v->vmifs->ifname);
		v->routeset = ROUTELINK;
	}
#ifdef DIVERT
	else if (v->doIpUdp && pinfo->sendNatRep){

		if (debug){	  
			printf("Setting Up An IP-UDP-IP Tunnel\n");
			printf("The home address %s \n",inet_ntoa(v->mn));
			printf("The home address %d \n",htons(v->udp_port));
		}
		
		if (v->fr_number == 0){

			srcAddr.s_addr = 0;		  
			setupFirewallRule(srcAddr,v->mn,IN,DIVERTPORTIN,0,mip_config.fr_number);
			
			setupFirewallRule(v->udp_src.sin_addr,v->ha,IN,DIVERTPORTOUT
					  ,IPPROTO_UDP,mip_config.fr_number);
			
			route_add_host(v->mn,v->ha,v->ha);

			
			v->routeset  = IPUDPTUNNEL;
			v->fr_number = mip_config.fr_number;
			mip_config.fr_number++;

			mip_log(MLOG_INFO, "ipfw (nat) rules in for %s",
				inet_ntoa(v->mn));
		}
	}
#endif
	/* else FA or COA are far away
	*/
	else {
	  ha_tunnel_to_coa(v, v->mn,
			   v->coa );
	  v->routeset = ROUTETUNNEL;
	}
	
	/* do proxy arp if necessary
         */
        route_do_arp_proxy(v, v->mn);
	
	return(1);
}  /* end ha_setroute() */

/*
 * encapsulate tunnel route
 */
void
ha_tunnel_to_coa(
	struct vlist *v,
	struct in_addr dst, 
	struct in_addr coa
)
{

	ha_route_add_tunnel(dst, coa
		);
}  /* end ha_tunnel_to_coa() */

/* add a proxy arp entry for the MN if the interface IP address
 * matches one in the i/f list for the HA with proxy configured
 */
static void
route_do_arp_proxy( struct vlist *v, struct in_addr mn)
{
        struct mip_ifs  *mips = &mip_config.mip_ifs[0];
        int             i;
        extern int debug;
 
#define nmask (mips->netmask.sin_addr.s_addr)
 
        for (i = 0; i < mip_config.mip_ifcount; i++, mips++) {
                if (!mips->amfa && mips->doProxy &&
                        (mn.s_addr & nmask) ==
                                (mips->if_sin.sin_addr.s_addr & nmask) )
                {



                        v->proxyArp = 1;
                        if (debug) {
                                printf("HAUDP: proxy arp for %s\n",
                                        inet_ntoa(mn));
                        }
                        route_add_proxy(mn, mips->mac, mips->ifname);
                        break;
                }
        }
#undef nmask
}


/*
 * stub routine.
 *
 * package call to route.c ha_route_delete_tunnel
 * deal with proxy arp here only.
 *
 * TBD. this routine and next should be merged
 */
static void
route_delete_tunnel(struct vlist *v)
{
        extern int debug;

        ha_route_delete_tunnel(v->mn, v->coa);
        if (v->proxyArp) {
                if (debug) {
                        printf("HAUDP: proxy arp for %s DELETED\n",
                                inet_ntoa(v->mn));
                }
                route_delete_proxy(v->mn);
                v->proxyArp = 0;
        }
}


static void 
route_delete(struct vlist *v)
{
	if (v->routeset == ROUTETUNNEL) {
		route_delete_tunnel(v);
	}
	else if (v->routeset  == ROUTELINK) {
		route_delete_link(v->mn, v->mac);
	}
	v->routeset = 0;
}

/*
 * ha uses this function to send reply 
 */
static void
sendUdpReply (struct sockaddr_in *from, struct packet_info *pinfo, 
	int code, u_short uport, int lifetime)
{
	char packet[1024];
	struct mse_authext *mse;
	struct mse_nat_ext_rep *nat;
	extern int debug;
	unsigned long dummy;
	int dummy2;
	char *mn_key;
	int  mn_keysize;
	u_long  mn_spi,
		spi;
	int  mn_transform;
	struct udp_reply *rep = (struct udp_reply *) packet;
	int psize = sizeof(struct udp_reply);

#ifdef HARP
	/*  make sure this is not a forwarded packet */
	if ( pinfo->forward ){
		mip_log( MLOG_ERR, "HARP: sendUdpReply(): should not send reply to forwarded HARP packet\n" );
		return;
	}
#endif
	/* tally up codes
	*/
	switch(code) {
	case REG_REPLY_OK:
		ha_stats.acks++;
#ifdef HARP
		if ( mip_config.harp ){
			harpForward( harpSocket, pinfo, HARP_FORWARD );   /* forward packet to coha */
		}
#endif
		break;
	case REG_FAIL_HA_IDMISMATCH:
		ha_stats.idmismatches++;
		ha_stats.naks++;
		break;
	}

	rep->type = UDP_REPLY_TYPE;
	rep->code = (u_char) code;
	rep->lifetime = htons(lifetime);
	rep->home_address = pinfo->reg->home_address;
	rep->home_agent = pinfo->reg->home_agent;
	rep->i.l[0] = pinfo->reg->i.l[0];
	rep->i.l[1] = pinfo->reg->i.l[1];

	/* we must tack on the mn-ha registration
	*/
	mse = (struct mse_authext *) &packet[psize];
	mse->type = MN_HA_AUTHENTICATOR_TYPE;
	mse->length = 4 + 16;

	mn_key = get_key(
		mip_config.sa[MNHASECURITY].kl,
		(unsigned long) rep->home_address.s_addr,  
		&mn_keysize,   
		&mn_spi,       
		&mn_transform,
		&dummy, &dummy2); 

	if (mn_key == NULL) {
		mip_log( MLOG_SEC, "MN %s failed to have key in MNHA keylist.", inet_ntoa(rep->home_address));
		return;
	}
	spi = htonl(mn_spi);
	bcopy(&spi, &mse->spi, sizeof(int));

	/* TBD. acc. to transform, for now MD5 only 
	*/
	storeMd5((char *)rep, psize + AUTHEXT_HASHEDBYTES, mse->hash, mn_key);

	/* now update size accordingly
	*/
	psize += (mse->length + 2);

	/* append faha security if desired but not if DBIT set
	*/
	if (((pinfo->reg->flags & D_BIT) == 0) &&
		mip_config.faha_security) {
		mse = (struct mse_authext *)
                        (((char *)mse) +  2 + mse->length);   /* Zhong Chen */
		mse->type = FA_HA_AUTHENTICATOR_TYPE;
		mse->length = 4 + 16;
		mn_key = get_key(mip_config.sa[FAHASECURITY].kl,
			from->sin_addr.s_addr, /* ip addr, index */
			&mn_keysize,   /* size of key returned */
			&mn_spi,       /* spi returned */
			&mn_transform, /* transform returned */
			&dummy, &dummy2);  

		if (mn_key == NULL) {
			mip_log( MLOG_SEC, "FA %s failed to have key in FAHA keylist.", inet_ntoa(from->sin_addr));
			return;
		}

		spi = htonl(mn_spi);
		bcopy(&spi, &mse->spi, sizeof(int));
		storeMd5((char *)rep, psize + AUTHEXT_HASHEDBYTES,
			mse->hash, mn_key);
		psize += mse->length + 2;
	}

	/* Setup the IP-UDP Tunnel if appropriate flags are set */
	if ( (pinfo->hasNat) && (pinfo->sendNatRep)) { 

		nat                    = (struct mse_nat_ext_rep *) &packet[psize];	  
		nat->type              = MN_HA_NAT_TYPE_REP;
		nat->length            = 6;
		nat->sub_type          = MN_HA_NAT_SUBTYPE;
		nat->replyCode         = MN_HA_NAT_REPLY_CODE;
		nat->reserved          = F_BIT;   
		nat->keepAliveInterval = htons(lifetime);
		psize                  += sizeof(struct mse_nat_ext_rep);		
 
		if ( debug )
			dumpNatRep(nat);		  
	}
	
	/* ship it
	*/
	if (debug) {
		dumpUdpReply((struct udp_reply *) packet);
		dumpMse(mse);

	}
	/* package up "udp" packet to send back
	 * to MN
	*/
	sendOneHaUdp(packet, psize , from->sin_addr, uport);
}



/*
 * walk of visitor list for mipstat
 */
void
ha_dumpVlist()
{
	struct vlist *v;
	int count = 0;

	vlist_initwalk(hahead);
	while (v = vlist_nextwalk()) {
		count++;
	}

	/* write the number of vlist entries 
	*/
	sendUnixMessage(MOBYSOCK_CLIENT, (char *)&count, sizeof(int));	

	/* write the vlist nodes
	*/
	vlist_initwalk(hahead);
	while (v = vlist_nextwalk()) {
		sendUnixMessage(MOBYSOCK_CLIENT, (char *)v,
						sizeof(struct vlist));
	}
}
#ifdef MADRP            
dumpHlist(){    
       struct SR_hlist * h; 
       int rc;  
       int srhCount=srhlist_count();
       if (sendUnixMessage(MOBYSOCK_CLIENT, (char*)&srhCount, sizeof(int))<0)
                return; 
       if (srhCount==0) return;
       /* walk the srhlist and write the structures out */
       srhlist_initwalk();
       while (h = (struct SR_hlist *) srhlist_nextwalk()) {
         rc=sendUnixMessage(MOBYSOCK_CLIENT,(char *)h,sizeof(struct SR_hlist)); 
         if (rc < 0) return;
       }                
       return;  
}
#endif          


/* ha state machine tick
 */
void
haTick()
{
	struct vlist *v;
	extern int debug;
#ifdef HARP
	static int harp_pingClock = 0;	/* tick timer for use with harp pings */

	extern harp_ackTimer;  /* used in other modules */ 

	if ( mip_config.harp ){
		if ( harp_pingClock > mip_config.harp_ping ){
			if ( mip_config.harp ){
				harpPing();
			}
			harp_pingClock = 0;
		}
		else{
			harp_pingClock++;
		}

#define ACK_TIMEOUT 100
		if ( !harp_ackTimer ){  /* times up */
			harp_ackTimer = ACK_TIMEOUT;
			mip_log( MLOG_ERR, "HARP: haTick(): No ACK response from co-HA\n" );
			if ( harp2 == TRUE ){
				harp2 = FALSE;
				harp2ArpOn();
			}
			if ( debug ){
				printf( "HARP: haTick(): The co-HA seems to be dead.\n" );
			}
		}
		else {
			harp_ackTimer--;
		}
	}
#endif

	vlist_initwalk(hahead);
	while (v = vlist_nextwalk()) {
		/* no timeout if lifetime set to infinity
		*/
		if (v->lifetime == 0xffff) {
			continue;
		}
		v->lifetime--;	
		/* if list timeout
		*/
		if (v->lifetime <= 0) {
			ha_stats.timeouts++;
			if (debug) {
				printf("haTick: timeout on %s\n",
					inet_ntoa(v->mn));
			}
			/* HOME->NOSTATE
			 * this should have been deleted by an mn
			 * deregistration if arp in use, else
			 * all we have to do is delete the entry
			 */
			if (v->vstate == HA_HOME) {
				if (v->routeset == ROUTETUNNEL) {
					if(debug) {
						printf("haTick: CROCK! vstate HA_HOME but tunnel set\n");
					}
					mip_log(MLOG_INFO, "haTick: total CROCK!, vstate HA_HOME but tunnel set\n");
				}
				/* FIX -- this is at least one place wehre
				 * a LL route would have to be replaced
				 * if running arp/ipsec and clone is gone */
				if (v->routeset == ROUTETUNNEL) {
					route_delete_tunnel(v);
				}
				else {
					route_delete_link(v->mn, v->mac);
				}
				vlist_delete(&hahead,v->mn);
				ha_stats.visitors--;
				mip_log( MLOG_INFO, "timeout: dropped MN %s which was in HOME state from HA visitor list.", inet_ntoa(v->mn));
			}
			/* AWAY->NOSTATE
			*/
			else if (v->vstate == HA_AWAY
#ifdef HARP	
				|| v->vstate == AT_COHA
#endif
			) {

#ifdef DIVERT
				/* If routeset is IP-UDP Tunnel Delete the Firewall Rules 
				 * of the corresponding MN 
				*/
			        if (v->routeset == IPUDPTUNNEL){

					if ( v->fr_number)
						delFireWallRule(v->fr_number);
					  
				  	route_delete_host(v->mn,mip_config.ha_sin.sin_addr);
				}
#endif
				if (v->routeset == ROUTETUNNEL) {
					route_delete_tunnel(v);
				}
				/*
				else {
					route_delete_link(v->mn, v->mac);
				}
				*/
				vlist_delete(&hahead,v->mn);
				ha_stats.visitors--;
				mip_log( MLOG_INFO, "timeout: dropped MN %s that was AWAY from HA visitor list.", inet_ntoa(v->mn));
			}
                        else if (v->vstate == BEACON_NOSTATE) {
				/* FIX -- make check for v->routeset!=0? */
                                if (v->routeset == ROUTETUNNEL) {
                                        route_delete_tunnel(v);
                                }
                                else {
                                        route_delete_link(v->mn, v->mac);
                                }
                                vlist_delete(&hahead,v->mn);
                                ha_stats.visitors--;
                                mip_log( MLOG_INFO, "timeout: dropped MN %s that was in BEACON_NOSTATE state from HA visitor list.", inet_ntoa(v->mn));
                        }
		}
	}
}

/* checkTimestamp - sanity check timestamp value
 * returns:
 *	-2: timestamp possible spoof
 *	-1: timestamp invalid, send NAK
 *	 0: ok
 */
static int
checkTimestamp(struct sockaddr_in *from, struct vlist *v, struct packet_info *pinfo)
{
	int rc;
	char *ident = NULL;
	unsigned long hasecs;

	if (v) {
		ident = (char *) &v->i.l[0];
	}
	/* see if pkt timestamp is less than current HA time
	 *
	 * 		(NOTE: errors logged within function)
	 */
	rc = check_ntpsys((char *)&pinfo->reg->i.l[0], ident, 
			mip_config.ntp_slop, &hasecs);
	if (rc == -1) {
		char *t = (char *) dumptime(ntohl(pinfo->reg->i.l[0]));
		/* we need to manipulate the pinfo buffer
		 * to return the time from the HA
		 */
		pinfo->reg->i.l[0] = htonl(hasecs);  /* copy in ha time */
		mip_log(MLOG_SEC, "time from MN %s(%s) corrected as out of sync by > %d secs.  MN time=<%s>", 
			reverseAddress(pinfo->reg->home_address), 
			inet_ntoa(pinfo->reg->home_address),
			mip_config.ntp_slop, 
			t );
		free( (void *) t);
	}
	else if (rc == -2) {
		ha_stats.idspoofs++;
		mip_log(MLOG_SEC, "possible timestamp spoof OR simply delayed packet from MN %s(%s)", 
			reverseAddress(pinfo->reg->home_address), 
			inet_ntoa(pinfo->reg->home_address));
	}
	return(rc);
}

/* delete routes for ha on shutdown
*/
void
haRouteWipe()
{
	struct vlist *v;

	if (mip_config.sawHa == 0)
		return;
	vlist_initwalk(hahead);
	while (v = vlist_nextwalk()) {
		if (v->vstate == HA_AWAY) {
			/* delete host route entry
			 * including any arp entries grat/otherwise
			 */
			if (v->routeset == ROUTETUNNEL) {
				route_delete_tunnel(v);
			}
			else {
				route_delete_link(v->mn, v->mac);
			}
		}
		else if (v->vstate == HA_HOME) {
			if (v->routeset == ROUTETUNNEL) {
				route_delete_tunnel(v);
			}
			else {
				route_delete_link(v->mn, v->mac);
			}
		}
	}
}

/*
 * if this is HA packet, we look at our configuration
 *	1. must check mn_ha authentication according to spi type
 *		if nonce spi or if timestamp spi
 *	2. if mse for fa exists AND we  have configured in 
 *		fa_ha info, check that too.
 * TBD:
 *	configuration option to silently discard errors or return
 *	error.  For both fa/ha
 * if this is FA packet, registration or reply
 *	1. if packet is reply, check for configured in fa_ha auth.
 *		if it exists, check packet
 *		if packet has auth, and we are not configured in,
 *			log error
 *		TBD: what to do in latter case?
 *	2. if packet is request, assume from MN and check for
 *		configured in mn-fa association and mse in packet.
 *
 * logging:
 *	log any errors
 *
 * returns:
 *	0: check was ok
 *	-1: error
 */

/* securityCheck - check md5 hash on incoming packet according
 *	to keys stored in config file
 *
 * if ha and registration packet
 *	check mn ha security
 *	if desired
 *		check fa ha security
 * if fa
 *	if registration packet and famn security flag
 *		check mn fa security
 *	if reply packet and faha security flag
 *		check fa ha security
 *
 * TBDs: 1. what role does spi play here?
 *
 * returns:
 *	0: ok
 *	-1: failed authentication
 */
int
securityCheck(int haFlag, struct sockaddr_in *from, 
	struct packet_info *pinfo)
{
	int size;
	char hash[16];
	char rbuf[1024];
	char *mn_key;
	int  mn_keysize;
	int  mn_transform;
	u_long  mn_spi;
	unsigned long dummy;
	int timeflag;
	char *who = inet_ntoa(from->sin_addr);
	extern int debug;

	strncpy(rbuf, reverseAddress(from->sin_addr), 1024);

	/* assumed true unless per MN config off at HA 
	*/
	pinfo->timeflag = 1;		

	/* we are ha and this is registration packet
	*/
	if (haFlag && pinfo->reg) {
		if (mip_config.mnha_security && pinfo->mn_ha_auth) {
			if (pinfo->mn_ha_auth_count != 1) {
				mip_log( MLOG_SEC, "HA MN authentication not present in UDP registration packet from %s(%s)", 
				rbuf,
				who);
				return(-1);
			}
			/* compute the checksum over the packet with
			 * our key.  get size and then call appropriate
			 * function. Take Care of size if Nat Request is present.
			 */
			if (pinfo->nat_ext_req)
				size = sizeof(struct udp_register) + sizeof(struct mse_nat_ext_req) + 			  
				  pinfo->premnha_size + AUTHEXT_HASHEDBYTES;
			else
				size = sizeof(struct udp_register) + 			  
				  pinfo->premnha_size + AUTHEXT_HASHEDBYTES;

			/* lookup key according to ip index */
			mn_key = get_key(
				mip_config.sa[MNHASECURITY].kl,
				(unsigned long) pinfo->reg->home_address.s_addr,  
				&mn_keysize,   
				&mn_spi,       
				&mn_transform,
				&dummy, &timeflag); 
			pinfo->timeflag = timeflag; 

			if (mn_key == NULL) {
				ha_stats.missingmnauth++;
				mip_log( MLOG_SEC, "MN %s(%s), via COA %s(%s) authentication failed at HA - key is missing", 
					reverseAddress(pinfo->reg->home_address),
					inet_ntoa(pinfo->reg->home_address),
					rbuf,
					who);
				return(-1);
			}
			/*
			spi = htonl(mip_config.sa[MNHASECURITY].spi);
			*/
			/* compute key and store computation in hash array
			*/
			computeMd5(pinfo->packet, size, 
				mn_key,
				hash);
			/* compare packet hash and computed hash
			 * if == 0, then ok
			 */
			if (bcmp(pinfo->mn_ha_auth->hash, hash, 16) != 0) {
				/* authentication failed
				*/
				if (debug) {
					printf("mn ha authentication failed\n");
				}
				ha_stats.invalidmnauth++;
				mip_log( MLOG_SEC, "MN %s(%s), via COA %s(%s) authentication failed at HA - key is wrong", 
					reverseAddress(pinfo->reg->home_address),
					inet_ntoa(pinfo->reg->home_address),
					rbuf,
					who);
				return(-1);
			}
		}
		if (mip_config.faha_security && pinfo->mn_ha_auth) {
			if (pinfo->fa_x_auth_count != 1) {
				mip_log( MLOG_SEC, "HA FA authentication required in configuration, but not present in packet from %s(%s)", 
				reverseAddress(from->sin_addr),	
				inet_ntoa(from->sin_addr));
				return(-1);
			}
			size = sizeof(struct udp_register) +
				pinfo->premnha_size  +
				pinfo->mn_ha_auth_size +
				pinfo->postmnha_size +
				AUTHEXT_HASHEDBYTES;
			/*
			spi = htonl(mip_config.sa[FAHASECURITY].spi);
			*/
			mn_key = get_key(
				mip_config.sa[FAHASECURITY].kl,
				(unsigned long) from->sin_addr.s_addr,
				&mn_keysize,   
				&mn_spi,       
				&mn_transform,
				&dummy, &timeflag); 

			if (mn_key == NULL) {
				ha_stats.missingfaauth++;
				mip_log( MLOG_SEC, "FA %s(%s) authentication failed at HA, key is missing.", reverseAddress(from->sin_addr), who);
				return(-1);
			}
			computeMd5(pinfo->packet, size, 
				mn_key,
				hash);
			/* compare packet hash and computed hash
			 * if == 0, then ok
			 */
			if (bcmp(pinfo->fa_x_auth->hash, hash, 16) != 0) {
				/* authentication failed
				*/
				if (debug) {
					printf("fa ha authentication failed\n");
				}
				ha_stats.invalidfaauth++;
				mip_log( MLOG_SEC, "FA %s(%s) authentication failed at HA - key is wrong.", reverseAddress(from->sin_addr), who);
				return(-1);
			}
		}
	}
	/* else fa */
	/* if registration packet, mn to fa */
	else if (pinfo->reg) {
		/* if security relationship exists between FA and
		 * MNs
		 */
		if (mip_config.famn_security && pinfo->mn_ha_auth) {
			/* before we commit suicide, make sure that
			 * there actually is a famn mse in the packet
			 */
			if (pinfo->fa_x_auth_count != 1) {
				mip_log( MLOG_SEC, "fa_mn authentication required by configuration, but not present in packet from %s(%s)", 
				reverseAddress(from->sin_addr),
				inet_ntoa(from->sin_addr));
				return(-1);
			}
			size = sizeof(struct udp_register) + 
				pinfo->premnha_size  +
				pinfo->mn_ha_auth_size +
				pinfo->postmnha_size +
				+ AUTHEXT_HASHEDBYTES;

			/* TBD. sanity check, compare to actual packet size
			*/
			/*
			spi = htonl(mip_config.sa[FAMNSECURITY].spi);
			*/
			mn_key = get_key(
				mip_config.sa[FAMNSECURITY].kl,
				(unsigned long) pinfo->rep->home_agent.s_addr,  
				&mn_keysize,   
				&mn_spi,       
				&mn_transform,
				&dummy, &timeflag); 

			if (mn_key == NULL) {
				fa_stats.missingmnauth++;
				mip_log( MLOG_SEC, "MN %s(%s) FA authentication failed at FA - key is missing.", reverseAddress(from->sin_addr),who);
				return(-1);
			}
			computeMd5(pinfo->packet, size, 
				mn_key,
				hash);
			/* compare packet hash and computed hash
			 * if == 0, then ok
			 */
			if (bcmp(pinfo->fa_x_auth->hash, hash, 16) != 0) {
				/* authentication failed
				*/
				if (debug) {
					printf("mn fa authentication failed\n");

				}
				fa_stats.invalidmnauth++;
				mip_log(MLOG_SEC, "MN %s(%s) FA authentication failed at FA - key is wrong.", reverseAddress(from->sin_addr), who);
				return(-1);
			}
		}
	}
	/* else reply, sec. association exists with ha
	*/
	else if (pinfo->rep) {
		if (mip_config.faha_security && pinfo->mn_ha_auth) {
			if (pinfo->fa_x_auth_count != 1) {
				mip_log( MLOG_SEC, "fa_ha authentication required by configuration, but not present in packet from %s(%s)", 
				reverseAddress(from->sin_addr),
				inet_ntoa(from->sin_addr));
				return(-1);
			}
			size = sizeof(struct udp_reply) +
				pinfo->premnha_size  +
				pinfo->mn_ha_auth_size +
				pinfo->postmnha_size +
				+ AUTHEXT_HASHEDBYTES;
			/*
			spi = htonl(mip_config.sa[FAHASECURITY].spi);
			*/
			mn_key = get_key(
				mip_config.sa[FAHASECURITY].kl,
				(unsigned long) pinfo->rep->home_agent.s_addr,
				&mn_keysize,   
				&mn_spi,       
				&mn_transform,
				&dummy, &timeflag); 

			if (mn_key == NULL) {
				if (debug) {
                                        printf("securityCheck(): FAHA authentication from HA (%s) failed -- key is missing.\n",
                                                who);
                                }

				fa_stats.missinghaauth++;
				mip_log( MLOG_SEC, "securityCheck(): FAHA authentication from HA (%s) failed -- key is missing.",
                                        who);
				return(-1);
			}

			computeMd5(pinfo->packet, size, 
				mn_key,
				hash);

			/* compare packet hash and computed hash
			 * if == 0, then ok
			 */
			if (bcmp(pinfo->fa_x_auth->hash, hash, 16) != 0) {
				/* authentication failed
				*/
				if (debug) {
					printf("fa ha authentication failed\n");
				}
				fa_stats.invalidhaauth++;
				mip_log( MLOG_SEC, "HA %s(%s) FA authentication failed at FA - key is wrong.", reverseAddress(from->sin_addr),who);
				return(-1);
			}
		}
	}
	return(0);
}

int noDefaultroute()
{
        return(0);
}
struct vlist * vlist_gethahead(){ return hahead;}
struct vlist ** vlist_gethahead_pointer(){ return &hahead;}


