/*
 * Copyright (c) 1996, 2003 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * mipudpinput.c
 *
 * overview:
 *	functions related to input of udp message (request or reply)
 *	at mip agent (mipd).  read it, "parse" it (break it up) 
 *	and weed out errors, and prepare messages for input
 *	to mip fa/ha state machines.
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/times.h>
#include <errno.h>
#include <signal.h>

#include "mipconfig.h"
#include "mipstate.h"
#include "udp_reg.h"
#include "faha_pinfo.h"
#include "mipstats.h"
#include "mip_log.h"

#ifdef HARP
#define MIPSEC
#include "harp.h"
#endif

/* prototypes
*/
void    
stateMachine(int, int, int, u_long *, struct sockaddr_in *,
                struct packet_info *, struct mip_ifs *mifs, int faifbutha);

static int 
determineHaFa(struct sockaddr_in *, u_long *, struct udp_register *,
	struct udp_reply *, int *, int *, struct mip_ifs **);

int recvMsg(int sd, char *buf, int bufsize, struct sockaddr *from, 
	int *fromsize, int *ifindex);

/* udp error stats
*/
struct udp_errors udp_errors;

extern struct ha_stats ha_stats;
struct packet_info pinfo;

/* handleUdpInput
 * given a socket with a pending udp message,
 * we parse and sanitize the message.  by parse,
 * we mean recoginize & break up into sub-components including
 * any security associations.  We sanitize by
 * checking sizes (typically too short).
 * type indicates whether a given packet is a regular MIP UDP message
 * or a forwarded HARP messsage
 *
 * Note:
 *	kernel code will add 4 byte in_addr at end
 *	of this message to tell us ip address of incoming
 *	interface.  We need this so that FA can know
 *	what interface to bind a MN to.
 */
static int expect_minreg_size = sizeof(struct udp_register) + 
		sizeof(struct mse_authext);
static int expect_minrep_size = sizeof(struct udp_reply) + 
		sizeof(struct mse_authext);

handleUdpInput(int s
#ifdef HARP
	,int type   
#endif
)
{
	int size;
	int rc;
#ifdef HARP
	extern harp2;	/* secondary flag */
	extern harp2ArpOn();	/* turn ARP on and off at secondary */
	extern harp2ArpOff();
#ifdef MIPSEC
	extern haha_key;	/* HARP auth key */
#endif
	char hash[16];
#endif
	struct sockaddr_in from;
	unsigned char packet[MAXUDPPACKET];
	struct udp_register *udp_reg = 0;
	struct udp_reply *udp_rep = 0;
	unsigned long if_addr;
	unsigned long *input_addr;
	extern int debug;
	int err = 0;
	int offset = 0;
	int haFlag = 0;
	int faifbutha = 0;
	int before_mn_ha;
	struct mse_authext *msep;
	struct mse_nat_ext_req *nat;
	struct harpNat_ext *harpNat;
	char *p;
	struct mip_ifs *mifs;
	int ifindex = 0;
	u_short t;
	input_addr = &if_addr;  /* XXX. idiotic */
	size = sizeof(struct sockaddr_in);
#ifdef HARP
	if ( type == HARP_UDP_SOC ){	/* a forwarded reg. packet */
		char buf[MAXUDPPACKET];
		extern harp_ackTimer;

		size = sizeof from;		/* paranoia? */
		rc = recvfrom(s, buf, sizeof buf, 0, 
			( struct sockaddr* )&from, &size );
		if (rc < 0) {
			syslog( LOG_ERR, "handleUdpInput(): HARP recvfrom failed (%m).");
			exit(1);
		}
		ha_stats.harpudpinput++;

#ifdef MIPSEC
		/* check authentication */
		rc -= 16;  /* adjust end of packet to exclude auth hash */
		computeMd5( buf, rc, haha_key, hash );
		if ( debug ){
			printf( "dumping hash:\n" );
			dumpme( hash, 16);
		}
		
		if ( bcmp( hash, buf+rc, 16 ) != 0 ){
			mip_log( MLOG_ERR, "HARP: handleUdpInput(): authentication failed on harp message. type=%d\n", ntohs(*(u_short*)buf) );
			if ( debug ){
				printf( "HARP: handleUdpInput(): auth failed (%x) (%x)\n", hash, buf+rc );
			}
			return; 
		}
#endif
		
		/* coHA must be alive -- a packet just arrived from it.  Don't
		 * really care what the packet type is.
		 */
		harp_ackTimer = HARP_ACK_TIMELIMIT;



		switch ( ntohs(*(u_short*)buf )){   /* switch on HARP type */
			case HARP_FORWARD:
				if ( debug ){
					printf( "HARP: handleUdpInput():  processing forwarded packet from %s\n", inet_ntoa( from.sin_addr ));
				}
				memcpy( packet, buf+4, rc-4 );
				rc -= 4;
				if ( harp2 == FALSE ){
					harp2 = TRUE; /* I am secondary */
					harp2ArpOff();
				}
				ha_stats.harpforwrecv++;
				break;
			case HARP_PING:  /* send ack & reset timer */
				if ( debug ){
					printf( "HARP: handleUdpInput(): processing ping packet from %s\n", inet_ntoa( from.sin_addr ));
				}
				ha_stats.harpping++;
				harpAck();
				return;
			case HARP_ACK:  /* reset timer */
				if ( debug ){
					printf( "HARP: handleUdpInput(): processing ack packet from %s\n", inet_ntoa( from.sin_addr ));
				}
				ha_stats.harpack++;
				return;
			default:
				if (debug) {
					printf("handleUdpInput(): fubar %x from coHA %s\n",
						ntohs(*(u_short*) buf),
						inet_ntoa(from.sin_addr));
				}
				return;
		}

	}
	else {
		if ( harp2 == TRUE ){ /* receiving udp packets directly */
			harp2 = FALSE;
			harp2ArpOn();
		}
#endif
		rc = recvMsg(s, packet, MAXUDPPACKET, 
			(struct sockaddr *)&from, &size, &ifindex);
		if_addr = indextoif(ifindex);  /* get if ip address */
      		if (rc < 0) {
                	syslog( LOG_ERR, "handleUdpInput(): recvfrom failed (%m).");
                	exit(1);
        	}

		if (rc < 0) {
			syslog( LOG_ERR, "handleUdpInput(): recvfrom failed (%m).");
			exit(1);
		}
		if_addr = indextoif(ifindex);  /* get if ip address */
#ifdef HARP
	}
#endif
	if (debug) {
		printf( "handleUdpInput(): UDP recvfrom: addr=%s, port=%d, rc=%d, if(%d)=%x \n", 
			inet_ntoa(from.sin_addr),
			ntohs(from.sin_port), rc, ifindex, *input_addr
			);
		dumpme(packet, rc);
	}

	/* zero out the packet info structure
	*/
	memset(&pinfo, 0, sizeof(struct packet_info));

	pinfo.packet = packet;
	pinfo.psize = rc;

	/* figure out what kind of packet it is and check min. size
	 * sans mses
	 */
	udp_reg = (struct udp_register *) packet;
	if (udp_reg->type == UDP_REGISTER_TYPE) {
		if ( debug ){
			printf( "a udp registration packet\n" );
		}
		udp_rep = NULL;
		pinfo.reg = udp_reg;
		pinfo.rep = NULL;
		if (pinfo.psize < expect_minreg_size) {
			udp_errors.regtooshort++;
			mip_log( MLOG_ERR, "handleUdpInput(): UDP registration packet from %s(%s) too short.", reverseAddress(from.sin_addr),
				inet_ntoa(from.sin_addr));
			if (debug) {
				printf("udp registration packet -- too short\n");
			}
			return;
		}
		offset += (sizeof(struct udp_register));
	}
	else if (udp_reg->type == UDP_REPLY_TYPE) {
		pinfo.reg = udp_reg = NULL;
		pinfo.rep = udp_rep = (struct udp_reply *) packet;
		if (pinfo.psize < expect_minrep_size) {
			udp_errors.repltooshort++;
			mip_log( MLOG_ERR, "handleUdpInput(): UDP reply packet from %s(%s) too short.",
				reverseAddress(from.sin_addr),
				inet_ntoa(from.sin_addr));
			if (debug)
				printf("udp reply packet -- too short\n");
			return;
		}
		offset += (sizeof(struct udp_reply));
	}
	/* bad type */
	else {
		if ( debug ){
			printf( "handleUdpInput(): type = %d\n", udp_reg->type ); 
		}
		udp_errors.badtype++;
		mip_log( LOG_ERR, "handleUdpInput(): UDP packet from %s(%s) received with bad type %d\n",
		    reverseAddress(from.sin_addr),
		    inet_ntoa(from.sin_addr),
		    udp_reg->type);
		if (debug)
			printf("udp pkt recv, bad type %d\n", udp_reg->type);
		return;
	}

	/* at this point, we have enough info to determine if
	 * this packet is to a HA or a FA.
	 *
	 * inputs:
	 *	from - who sent it
	 *	udp_reg - if non-NULL, udp registration packet
	 *		(either from FA or MN)
	 *	udp_rep - if non-NULL, either from HA and we had
	 *		better be FA
	 */
	err = determineHaFa(&from, input_addr, udp_reg, udp_rep, &haFlag,
		&faifbutha,
		&mifs);
	if (err < 0) {
		return;		/* NOTE: all errors logged prior to return */
	};

	/* now parse the packet for extensions.  
	 * We leave authentication for the state machine code
	 */

	/* extension ordering rules for MN->HA registration message
	 *	.may have 0 or more non-auth messages to HA.  FA should 
	 *		forward.
	 *	.MUST have MN-HA auth. extension. FA should forward.
	 *  if FA, we may have following. FA should remove before
	 *  forwarding to HA.
	 *	.may have 0 or more MN-FA non-auth. messages.  If so,
	 *		the FA should strip them.
	 *	.may have MN-FA auth. extension.  FA should strip
	 *		before forwarding.
	 * FA may append after MN-HA the following fields for HA:
	 *	.may have 0 or more FA-HA non-auth. messages.  HA
	 *		should strip.
	 *	.may have FA-HA auth. extension.  HA should strip
	 *		but pre
	 */

	/* rules for HA->MN reply message
	 *	.may have 0 or more non-auth messages.  FA may or
	 *		may not care but should forward.
	 *	.MUST have MN-HA auth. extension.  FA should forward.
	 *	.may have 0 or more HA-FA non-auth. extensions.
	 *		FA should strip before forwarding to MN.
	 *	.may have HA-FA auth. extension.  FA should strip.
	 */
	before_mn_ha = 1;


	nat = (struct mse_nat_ext_req *) (packet + offset);
	if (nat->type == MN_HA_NAT_TYPE_REQ){
		pinfo.nat_ext_req = nat;	  
		offset += sizeof(struct mse_nat_ext_req);
#ifdef HARP
		pinfo.hasNat = 1;
#endif

	}else {
		nat = NULL;
#ifdef HARP
		pinfo.hasNat = 0; 
#endif	  
	}
	msep = (struct mse_authext *)(packet + offset);	
	while(offset < pinfo.psize) {
		switch (msep->type) {

		case MN_HA_AUTHENTICATOR_TYPE:
			before_mn_ha = 0;
			pinfo.mn_ha_auth = msep;
			/* +2 because we add in type/length */
			pinfo.mn_ha_auth_size = (msep->length + 2);
			pinfo.mn_ha_auth_count++;
			if (pinfo.mn_ha_auth_count > 1) {
				udp_errors.toomanyauthexts++;
				mip_log( MLOG_SEC, "handleUdpInput(): MN %s(%s) HA security association, too many authentication extensions", 
				reverseAddress(from.sin_addr),
				inet_ntoa(from.sin_addr));
				return;
			}
			break;

		case MN_FA_AUTHENTICATOR_TYPE:
			if (before_mn_ha) {
				mip_log(MLOG_SEC, "handleUdpInput(): MN-FA auth ext found before MN-HA auth ext!  Discarding packet.  (from=%s)",
					inet_ntoa(from.sin_addr));
				return;
			}
			pinfo.fa_x_auth = msep;
			pinfo.fa_x_auth_size = (msep->length + 2);
			pinfo.fa_x_auth_count++;
			if (pinfo.fa_x_auth_count > 1) {
				udp_errors.toomanyauthexts++;
				mip_log( MLOG_SEC, "handleUdpInput(): MN %s(%s) FA security association, too many authentication extensions", 
				reverseAddress(from.sin_addr),
				inet_ntoa(from.sin_addr));
				return;
			}
			break;

		case FA_HA_AUTHENTICATOR_TYPE: 
			if (before_mn_ha) {
				mip_log(MLOG_SEC, "handleUdpInput(): FA-HA auth ext found before MN-HA auth ext!  Discarding packet.  (from=%s)",
					inet_ntoa(from.sin_addr));
				return;
			}
			pinfo.fa_x_auth = msep;
			pinfo.fa_x_auth_size = (msep->length + 2);
			pinfo.fa_x_auth_count++;
			if (pinfo.fa_x_auth_count > 1) {
				udp_errors.toomanyauthexts++;
				mip_log( MLOG_SEC, "handleUdpInput(): packet from %s(%s), FA HA security association, too many authentication extensions", 
				reverseAddress(from.sin_addr),
				inet_ntoa(from.sin_addr));
				return;
			}
			break;
		default:
			mip_log(MLOG_SEC, "handleUdpInput(): ignoring unrecognized extention %s MN-HA auth ext (from=%s)",
				before_mn_ha?"before":"after",
				inet_ntoa(from.sin_addr));
			if (before_mn_ha) {
				pinfo.premnha_size += (msep->length + 2);
			}
			else {
				pinfo.postmnha_size += (msep->length + 2);
			}
		}



		offset += (msep->length + 2);
		
		/* If there is a NAT Registration Request Extension
		 * and we are doing HARP then there should be a 
		 * HARP NAT Extension .
		 */
#ifdef HARP
	/*  pass forwarded registration packet to haUdp() */ 
	if ( type == HARP_UDP_SOC && mip_config.harp ){
	  

		if ( debug ){
			printf( "HARP: handleUdpInput(): passing forwarded packet to haUdp()\n" );
		}
		pinfo.forward = TRUE;
		if (pinfo.hasNat){

			harpNat = (struct harpNat_ext *)(packet + offset);
			if (harpNat->type ==  HARP_NAT_EXT) {

				if (debug) {
					printf("HARP_NAT Extension Found %x \n",harpNat->type);
				}
				
				pinfo.harp_nat_ext = harpNat;

			} else {

				mip_log( MLOG_ERR, "handleUdpInput(): NAT Registration Ext Found, But No HARP NAT Extension");
			}
		}

		haUdp( 0, &from, &pinfo, 0 );
		return;
	}
#endif

		p = (char *) msep;
		p += (msep->length + 2);
		msep = (struct mse_authext *) p;
	}

	/* if message recv. sans mn/ha authentication ext., we toss it
	*/
	if (pinfo.mn_ha_auth_count == 0) {
		if (debug) {
			printf("tossing msg: no mn/ha mse\n");
		}
		udp_errors.noauthent++;
		mip_log( MLOG_ERR, "handleUdpInput(): tossing message from %s(%s), no MH/HA security extension present",
			reverseAddress(from.sin_addr),
			inet_ntoa(from.sin_addr));

		return;
	}

	stateMachine(s, UDPEVENT, haFlag, input_addr, &from, &pinfo, mifs,
		faifbutha);

}  /* end handleUdpInput() */


/*
 * determineHaFa
 *	given inputs and configuration info, TRY to determine
 * 	if this packet was sent to us as HA or FA.
 *
 * Results:
 *	1. we can tell if this is for a HA.
 *	2. we weed out any obvious errors.
 *	3. else we assume for a FA and if the packet survives
 *		the error weeding phase, post any authentication,
 *		we will try to feed it later to the FA state machine.
 *
 * general error semantics:
 * if we are HA, we could get a registration packet,
 *	but not reply packet
 * if we are FA, we could get both.  
 *
 * returns:
 *	-1: error
 *	0: error
 */
static
int
determineHaFa(struct sockaddr_in *from, 
	u_long *input_addr,
	struct udp_register *reg,
	struct udp_reply *rep, 
	int *haflag,
	int *faifbutha,	/* HA match but pkt came through FA interface */
	struct mip_ifs **mifs)
{
	struct mip_ifs *mips;
	struct mip_ifs *ifptr;
        extern struct mip_ifs mip_ifs[MAX_INTERFACES];
        extern int mip_ifcount;
	int itsHa = 0;
	int itsFa = 0;
	int bogus = 0;
	int i;

	/* find a pointer to an interface structure and return that
	 * indirectly.  Note implicit assumption that each interface
	 * has unique ip address.  Note also that it is possible
	 * that the packet may not be directly from an MN.  We
	 * only care about which interface in that case.
	 * Examples:
	 *	MN -> FA: should get interface match.
	 *	MN -> HA: should get interface match.
	 *	FA -> HA: don't care
	 * 	HA -> FA: don't care
	 * result:
 	 * if we matched an interface, mifs is NOT NULL.
	 * else it is NULL.
	 */
	*mifs = ifptr = NULL;
	mips = &mip_config.mip_ifs[0];
       	for ( i = 0; i < mip_config.mip_ifcount; i++, mips++ ) {
		if (*input_addr == mips->if_sin.sin_addr.s_addr) {
			*mifs = ifptr = mips;  /* match */
			break;
		}
	}

	*haflag = 0;
	*faifbutha = 0;

	/* if registration packet,
	 * see if we have any HA interfaces and this packet
	 * is a registration packet, see if the packet's
	 * HA address matches one of our beacon interfaces
	 */
	if (reg) {
		/* loop through configured interfaces looking
		 * for HA address that matches incoming packet
		 */
	        mips = &mip_config.mip_ifs[0];
        	for ( i = 0; i < mip_config.mip_ifcount; i++, mips++ ) {
			if (mips->amfa) /* (1 if fa) */
				continue;
			/* match HA if address to registration HA addr
			*/
			if (mips->if_sin.sin_addr.s_addr ==
				(u_long) reg->home_agent.s_addr) {
				/* if we found an interface binding
				 * and the interface was NOT the HA
				 * and the i/f is a fa i/f (only)
				 * note that.
				 */
				if (ifptr && (mips != ifptr) && ifptr->amfa) {
					*faifbutha = 1;
				}
				*haflag = 1;
				return(0);
			}
		}
	}
	/* done with HA check, FA from here on out */

	/* this is an explicit error, we can toss the packet.
	 * we aren't a FA and we got any registration/reply packet
	 * at this point.
	 */
	if (!mip_config.sawFa) {
		if (reg) {
			udp_errors.badrequest++;
			mip_log( MLOG_ERR, "determineHaFa(): Bad MN UDP registration packet from %s(%s)",
				reverseAddress(from->sin_addr),
				inet_ntoa(from->sin_addr));
		}
		if (rep) {
			udp_errors.badreply++;
			mip_log( MLOG_ERR, "determineHaFa(): Bad reply packet from %s(%s)",
				reverseAddress(from->sin_addr),
				inet_ntoa(from->sin_addr));
		}
		return(-1);
	}

	/* another error we can weed out here.  If we got
	 * a registration request and the 
	 * udp src and home agent match, shouldn't happen.
	 * That would mean the HA has decided to send the FA
	 * a registration.  Cute.
	 */
	if (reg && from->sin_addr.s_addr == (u_long) reg->home_agent.s_addr) {
		udp_errors.hafaregistration++;
		mip_log( MLOG_ERR, "determineHaFa(): UDP src and home agent match. HA %s(%s) confused",
			reverseAddress(from->sin_addr),
			inet_ntoa(from->sin_addr));
		return(-1);
	}

	/* if it is a registration message,
	 * we can check the input address against our list
	 * of FA addresses.  It must have come from one of those ifs
	 * or we might somehow be getting registrations across
	 * some other interface, which might be dangerous
	 * security-wise.
 	 *
	 * This is one more way to weed things out, and not
	 * necessarily know what we are doing.
	 */
	if (reg) {
		char ifaddr[1024];
		mips = &mip_config.mip_ifs[0];
        	for ( i = 0; i < mip_config.mip_ifcount; i++, mips++ ) {
			if (!mips->amfa) {
				continue;
			}
			if (*input_addr == mips->if_sin.sin_addr.s_addr) {
				return(0);
			}
		}
		udp_errors.fainvalidreg++;
		mip_log( MLOG_SEC, 
		    "determineHaFa(): Invalid registration from %s(%s) received by FA across wrong i/f",
			reverseAddress(from->sin_addr),
			inet_ntoa(from->sin_addr));
		return(-1);
	}
	return(0);
}

