/*
 * Copyright (c) 1996, 2001 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * mnconfig.c - config info for mn mobile node daemon (mnd)
 * 
 * purpose: read/parse/store tag values from /etc/mnd.conf mnd configuration
 * file.
 * 
 * file syntax: ascii lines, one tag per line, may have N values depending on
 * tag skip lines/and comments, # in 1st column is comment
 * 
 * tag syntax:
 * 
 * 1. mobile node <ifname> <ip address>
 *      e.g.,
 *      mn wlp0 <ip address>
 * 
 *      This may be a dns name but it is probably best configured 
 *      as an ip address just in case dns is not available when the
 *      mobile node boots.
 *
 * 2. home agent ip address
 *      ha <ip address>
 * 
 *      ip address for home agent.  can be dns name but probably
 *      should be ip address in case dns is not available when the
 *      mobile node boots.
 * 
 * 3. spi/keys for UDP messages
 * 
 * mnha_key <ha_ipaddr>	<transform> <spi> <hexstring> 
 *
 * famn_key <fa_ipaddr>	<transform> <spi> <hexstring>
 * (more than one line per fa)
 * 
 * The above should be exactly the same as in the mipd config file/s. The
 * existance of one of the lines above means a security association is
 * required between the two parties involved.  Keyed MD5 is used as the
 * authentication and integrity mechanism.
 * 
 * At a minimum there should be a security association between the mn and the
 * ha.  fa to ha and fa to mn security associations may optionally exist (but
 * the mn wouldn't know or care about the fa ha security association).  The
 * hex string used for the key (5th column) can be generated with the md5gen
 * program that is provided or via some other mechanism.  For example,
 * one might invoke md5(1) with no arguments and invite a cat to walk
 * on the keyboard, then type in CNTRL-D and take the resulting md5 hash
 * as the key (make sure the cat does a good job of producing a key
 * though).
 * 
 * Example: mnha_key 1.2.3.4 md5 256 0x9700ca3a82f7cf9b2d0cc739abd5f33f
 *
 * 4. type of replay protection for UDP messages.
 *
 * replay timestamp | nonce
 *
 * Timestamp is default.  nonce is not implemented.
 * This keyword is optional.
 * 
 * 5. time stamp mode for UDP messages.
 * 
 * tsmode 0 | 1
 * 
 * The mode governs how the MN behaves when it receives a UDP NAK from the HA
 * telling it that is TS is out of range; i.e. the HA thinks the time is
 * wrong.  The default value is tsmode 1.  The MN will try and correct the
 * system time acc. to information recv. from the HA.  If tsmode is set to 0,
 * the MN will not try to correct the system time.  The latter should be used
 * if a second mode of time is available.  If the mode is 0, the MN will
 * NEVER try to fix the time.  This keyword is optional.  
 *
 * 6. arp on | off
 *
 * arp on is the default.  If the system is setup to use authenticated
 * ICMP beacons to pass link-layer address info in an authenticated manner,
 * arp should be sent to off and at least one key should be specified
 * in the config file (normally the netauth key).  
 * 
 * 7. ICMP beacon timer in seconds 
 *
 * beacon <seconds> <multicast addr>
 *
 * If arp is off, we will send a adhoc/router advertisement every N seconds.  
 * We expect such a message to have at least one of two possible
 * keys (netauth or adhoc) used to authenticate it as a shared
 * secret between the sender and the recv.  The default is 2 seconds.
 * This is NOT a mobile-ip mechanism, and is experimental.
 * Note to turn this on: arp must be off and at least one key must
 * be present in the configuration.
 *
 * If an optional third parameter is supplied, multicast beacons
 * are sent to that address.  The address should be 224.0.0.1 by default.
 * 
 * 8. netauth_key - authentication key for logical network beacons 
 *
 * netauth_key <transform> <spi> <hexstring> 
 *
 * Example: netauth_key md5 256 0x9700ca3a82f7cf9b2d0cc739abd5f33f
 * 
 * If arp is off, we send ICMP beacons with mac and IP addresses in them.
 * This is a simple form of ad hoc routing.  Subnet differences are ignored.
 * We assume we can talk to a system if we can hear it.  However all
 * beacons must be authenticated with either the netauth_key (the default)
 * and an optional adhoc key which may be generated, used, and discarded
 * for short face to face meetings.  At least one key must be shared
 * for a beacon to be accepted.  It is assumed that agents (HAs/FAs)
 * and mns for a certain network will share the netauth_key and MNs
 * from external enterprises will NOT share the netauth_key.  
 * This sort of network authentication makes IP address spoofing difficult,
 * unless the MAC address is shared.   It also allows systems lacking
 * a shared IP subnet to directly correspond.  An important idea here
 * is that this shared key basically defines a logical link.  An IP
 * subnet does NOT define the logical link.
 * 
 * 9. adhoc_key - authentication used between mns and fas(in a temporary 
 * logical network)
 *
 * adhoc_key	<transform> <spi> <hexstring>
 *
 * Example: adhoc_key md5 256 0x9700ca3a82f7cf9b2d0cc739abd5f33f
 * 
 * The adhoc_key is intended to be used as a secondary supplemental
 * link key in addition to the netauth_key.  This key might be generated
 * and used for a short meeting between systems coming from different
 * management domains.  The key would be discarded at the end of the
 * meeting.
 * 
 * 10. type of replay protection for ICMP beacons.  This stuff is
 * not implemented.
 * 
 * beacon_replay <timestamp>|<nonce>  on | off
 * 
 * Timestamp is default.  nonce is not implemented. This keyword is optional.
 * 
 * 11. timeslop allowed in replay protection for address authentication
 *	with ICMP beacons.  This is not currently implemented.
 * 
 * beacon_timeslop <secs>
 * 
 * Amount of time to tolerate +- in timestamp of authentication received in
 * beacon from either agent or MN.  Default is 2 secs. 
 *
 * 12: netauth accept | reject
 *
 * This switch sets a simple policy for how authenticated ICMP beacons
 * are to be treated.  If beacons are recv. with the ICMP beacon and
 * "accept" is set (the default) and there is a shared netauth_key,
 * then addressing will be setup from this system to the link layer
 * system that sent the packet; i.e., we will accept its ip/mac pair
 * and stick them in the arp cache.  If reject is set, all beacons
 * with netauth keys are ignored.  reject might be used by the
 * paranoid during a meeting where the adhoc_key was being used.
 * Note that beacons with no keys at all are always ignored.
 * accept is the default.  This switch is optional.
 *
 * 13: adhoc accept | reject
 *
 * This switch sets a simple policy for how the adhoc_key is treated.
 * By default it is set to reject.  However for an ad hoc meeting
 * between systems from different management domains, one might choose
 * to generate a shared key, distribute it by out of band mechanisms
 * (easy to do actually:  echo <passphrase> | md5
 * and use the output of md5 as the key).
 * If reject is set, beacons authenticated with the adhoc key are ignored.
 * However the netauth_key has priority over the adhoc_key.
 * If a beacon has a valid netauth_key hash, but the adhoc_key is
 * set to reject, it will be accepted.  
 * This switch is optional.
 *
 * 14: udp_port <number>
 * 
 * The MN may change its port number from 434 (default) to N (435) 
 * or any other port number for that matter.  mipd should note this
 * and send the reply to the proper port.  
 *
 * 15. logIPOnly <off | on>
 *
 * default is "on".  setting this to "on" prevents IP addresses printed to
 * the logfile from being resolved to DNS names, avoiding a lookup with the
 * local DNS server which may be inaccessible in situations of low-bandwidth
 * or lacking connectivity.  this tag is optional.
 * 
 * 21.  dhcp mvifname [renewing_period]
 *      
 * Defaults: on, mvif0, 600 (10 minutes in seconds).
 *
 * Make the mobile node DHCP capable, "renewing period" is the time period to 
 * renew lease with DHCP server, default to 600 seconds, miniumum 5 seconds,
 * if 0, means use lease provide by DHCP server acc to DHCP specs.
 * We then reup in 1/2 the lease time, unless the DHCP server tells us
 * T1 and T2 times.
 *
 * Summary:
 *	defaults to 600.  If the server tells us > 600, we use 600.
 *		if the server tells us < 600, we use what the server tells us.
 *	cannot be less than 5 seconds.
 *	0 - means always use what server tells us.  MUST BE SUPPLIED.  The
 *	default is ten minutes.  
 *
 * mvifname is the name of the mvif interface, home address will be bound
 * to this interface in DHCP mode and outgoing traffic will be reverse 
 * tunneled through this interface to the Home Agent.  We use a IPIP
 * tunnel to the HA this way.
 *
 * experimental DHCP options:
 ********************************************************** 
#ifdef KAMEIPSEC
 * 22. ipsec bypass
 * mobile node is FreeBSD (kame) ipsec capable. 
 * Basically applies ipsec bypass policy for both UDP socket (  
 * mobile ip registration packets) and ICMP socket.( icmp router
 * advertisements from mnd).  This gets KAME ipsec and Mobile-IP
 * into separate spaces so they cannot interfere with each other.
#endif
 *
 * 23. doNat - turn nat capability on.  on by default
 * This is sub-mode for DHCP.  We dynamically detect agent case,
 * no agent case (ordinary DHCP), or NAT situation (behind DHCP/NAT
 * firewall.
 *
 * 24. dhcp_tunnel_policy [ notunnel | ipip | nat ]
 *
 * How does mnd reverse-tunnel (or not) to Home Agent?
 * This sets the default dhcp tunnel policy used by the MN
 * when it is in auto mode.  DHCP must be used for this policy
 * to be in effect.  Note that if you desire to manually
 * force the mode change, use mnstat -m to force a runtime
 * manual tunnel setup.  The default is ipip.
 *
 * In all cases, the MN MIP IP is pushed on the mvif device,
 * and the DHCP COA IP is bound to the mobile interface.
 * Encapsulated packets from the HA to the COA are sent to
 * the mvif0 device to be decapsulated and delivered normally
 * to the MN.  
 *
 * notunnel - means that the DHCP-advertised router is used as
 * the default router.  There is no reverse tunnel.
 *  
 * ipip - means the MN sets up a reverse IPIP tunnel to the HA.
 * 	IP {IP datagram} is the form of packets sent out the default route.
 *
 * nat - means setup a IP/UDP/MIP {IP datagram) reverse tunnel to the HA.
 * 
 * 25. dhcp_dns
 *
 * Assuming dhcp is enabled, this option which is NOT on by default
 * will have DHCP get the DNS info provided and overwrite /etc/resolv.conf
 * with it.  Note that if this option is on, mnd will always
 * copy /etc/resolv.conf to /etc/resolv.conf.mnd.bak before it
 * overwrites /etc/resolv.conf the first time.  If mnd is HUPPED
 * or otherwise exits cleanly it will attempt to restore /etc/resolv.conf
 * at exit.  Do NOT use the filename /etc/resolv.conf.mnd.bak.
 *
 * 26. fr_number <number>
 * 
 * In order to do NAT, mnd uses the BSD divert mechanism which
 * in turn requires ipfw.  In order for this mechanism to work
 * an ipfw rule must be inserted into the firewall table.
 * The default rule is 301.  This command allows a different rule number
 * if needed.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "mnconfig.h"
#include "mip_log.h"
#include "mnstats.h"
#include "route.h"
#include "mnstate.h"

/* global config variables *************************** */

struct mn_config mn_config;
struct mn_stats mn_stats;

/*************************************************/

extern int debug;

/*
 * function prototype declarations
 */
#ifdef BADIDEA
static		checkHints();
#endif
static          switchOnOp();
int             resolveAddress(char *addr, struct sockaddr_in * to);
int             pack_string(char *, unsigned char *, int);
static          dokey(char *, char *, int, int, int);
char 		*getmac();
static void 	checkMulticast(char *s);

void
configInit()
{
	/* udp messages use timestamp replay with 2 sec timeslop */
	mn_config.doTimestamp = TRUE;  /* use timestamp replay protection */
	mn_config.tsmode = 1;	       /* actually set time acc to HA NAK */
	/* icmp does not do this yet */
	mn_config.beacon_replay = FALSE;  /* it is off for ICMP packets */
	mn_config.beacon_timestamp = TRUE; /* TBD */
	mn_config.beacon_timeslop = 2;    /* TBD */
	mn_config.doNetauth = NETAUTHACCEPT;  /* use net auth key */
	mn_config.doAdhoc = NETAUTHREJECT;    /* don't require ad hoc key */
	mn_config.doArp = TRUE;               /* arp is default */
#ifdef DHCP
        mn_config.dhcpEnable = TRUE;         /* dhcp on by default */
#ifdef DIVERT
	mn_config.doNat = 1;
	mn_config.fw_rule = 301;		 
#endif
	mn_config.doDns = 0;		
        mn_config.dhcp_renewing_period = DHCPRENEWINGPERIOD;
        strcpy(mn_config.mvifname, "mvif0"); 
#endif

#ifdef KAMEIPSEC
	mn_config.kameipsecEnable = FALSE;    /* not kame ipsec capable by default */
#endif
	mn_config.adhoc_security = FALSE;     /* no key as of yet */
	mn_config.netauth_security = FALSE;   /* no key as of yet */
	mn_config.mnha_security = FALSE;      /* no key as of yet */
	mn_config.famn_security = FALSE;      /* no key as of yet */
	mn_config.beaconRate = 5;             /* default ICMP beacon rate */
	mn_config.sawHa = FALSE;	      /* no ha addr yet */
	mn_config.sawMn = FALSE;              /* no mn addr yet */
	mn_config.beacon_seq = 0;             /* initial ICMP seq no */
	mn_config.sawLocation = 0;	      /* no location string */
	mn_config.udp_port = UDP_PORT;	      /* host byte order */
#ifdef DIVERT
	mn_config.divertPortIn = DIVERTPORTIN; /* In-coming Packets are diverted to this port */
	mn_config.divertPortOut = DIVERTPORTOUT;/* Out going Packets are diverted to this port */
#endif
	mn_config.dhcp_policy = IPIP;	       /* default policy for DHCP */
        mn_config.logIPOnly = TRUE;           /* do NOT DO DNS lookups 
						for logged IPaddrs */
	mn_config.useMulticast = FALSE;       /* broadcast is default */
	mn_config.maddr.s_addr = inet_addr("255.255.255.255"); 

	/* initialize keylists
	*/
	mn_config.sa[MNHASECURITY].kl = init_keylist();
	mn_config.sa[FAMNSECURITY].kl = init_keylist();
}

/*
 * read the config file and load its values
 * 
 * return: 0: ok -1: fatal error (although we return)
 */
int
configRead(configName)
	char           *configName;	/* name of config file */
{
	FILE           *fd;
	char            linebuf[1024];
	char            opbuf[1024];
	char           *opword;
	char           *result;
	int             lineNumber = 1;
	int 		i, j;
	char 		*ptr;

	fd = fopen(configName, "r");
	if (fd == NULL) {
		syslog(LOG_ERR,
		       "ConfigRead(): fopen of config file %s failed\n",
		       configName);
		return (-1);
	}

	configInit();

	/*
	 * while not at EOF read a line, parse it
	 */
	while (feof(fd) == 0) {
		result = fgets(linebuf, MAXLINE, fd);
		if (result == NULL)
			break;
		/*
		 * skip comment
		 */
		if (linebuf[0] == '#') {
			lineNumber++;
			continue;
		}
		/*
		 * skip empty lines
		 */
		if (linebuf[0] == '\n' || linebuf[0] == '\r') {
			lineNumber++;
			continue;
		}
		/*
		 * skip whitespace in line
		 */
		while (*result == ' ' || *result == '\t') {
			result++;
		}
		/*
	         * find first token 
		 */
		strncpy(opbuf, result, 1024);
		opword = strtok(opbuf, " \t\n");
		if (opword == NULL || (strlen(opword) == 0))
			continue;
		/*
		 * sscanf(result, "%s %s", op, value); printf("op value %s
		 * %s\n", op, value);
		 */
		switchOnOp(opword, result, lineNumber);
		lineNumber++;
	};
	fclose(fd);

	/* walk interface/s
	*/

	getIfs();

	/*
	 * semantics
	 */
	if (mn_config.sawHa == 0) {
		syslog(LOG_ERR, "readConfig(): must have ha address\n");
		exit(1);
	}
	if (mn_config.sawMn == 0) {
		syslog(LOG_ERR, "readConfig(): must have mn address\n");
		exit(1);
	}

	/* if arp is off, there should at least one of the netauth_key
	 * and adhoc_key found 
	 */
	if ((mn_config.doArp == FALSE)
	    &&(mn_config.netauth_security==FALSE)
	    &&(mn_config.adhoc_security == FALSE)) {
			syslog(LOG_ERR, "readConfig(): arp is off, but neither netauth_key nor adhoc_key is found.");
			exit(1);
	}

#ifdef DHCP
	/* a few dhcp sanity checks
	*/
	/* likely: dhcp must use arp mode and be "normal", 
 	 * unauthenticated access makes no sense in key-o-mania land.
	 */
        if (mn_config.dhcpEnable && !mn_config.doArp) {
                syslog(LOG_ERR, "dhcp requires arp mode, not AAH");
		exit(1);
        }
#endif

	/* get the mac address */
	ptr=getmac(mn_config.ifname);
	if (ptr == NULL) {
		syslog(LOG_ERR, "no mac address found for if %s", mn_config.ifname);
		exit(1);
	}

	/* copy the mac address into config
	 * structure
	 */ 
	for (i=0;i<6;i++) 
		mn_config.mac[i]=ptr[i];

	return (0);

}  /* end configRead() */

#ifdef BADIDEA

/*
 * 802.11 hints, which can modify AAH mode in the config file
 * and/or are passed to 802.11 module to tell it how to start up
 * in its state machine
 */

static
checkHints() 
{
	int l3mode, l2mode, channel;
	
	readHints(&l3mode, &l2mode, &channel);

	/* bail if no hints
	*/
	if (l3mode < 0)
		return;

	/* force arp mode
	*/
	if (l3mode == L802ARP) {
		mn_config.doArp = 1;
	}
	/* force aah mode (hope we have keys ...)
	*/
	/* L802AAH */
	else {
		mn_config.doArp = 0;
	}
	pass802hints(l2mode, channel);
}
#endif


static
switchOnOp(op, linebuf, line)
	char           *op;
	char           *linebuf;
	int             line;
{
	char            dc[1024];
	char            v1[1024];
	char            v2[1024];
	char            v3[1024];
	int             rc;
	struct sockaddr_in sa;

	/* mn ifname ip */
	if (strcmp(op, "mn") == 0) {
		if (mn_config.sawMn) {
			syslog(LOG_ERR, "only one mn line per file\n");
			exit(1);
		}
		sscanf(linebuf, "%s %s %s", dc, v1, v2);
		if (strlen(v1)==0 || strlen(v2) == 0) {
			syslog(LOG_ERR, "switchOnOp(): config error line%d, mn line needs 3 parameters: mn <ifname> <if address>\n", line);
			exit(1);
		}
		strncpy(mn_config.ifname, v1, MAXIFNAME);
		rc = resolveAddress(v2, &sa);
		if (rc <= 0) {
			syslog(LOG_ERR,
			       "switchOnOp(): invalid mn address %s\n", v2);
			exit(1);
		}
		mn_config.mn.s_addr = sa.sin_addr.s_addr;
		mn_config.sawMn = 1;
	/* ha ip */
	} else if (strcmp(op, "ha") == 0) {
		if (mn_config.sawHa) {
			syslog(LOG_ERR, "only one ha line per file\n");
			exit(1);
		}
		sscanf(linebuf, "%s %s", dc, v1);
		rc = resolveAddress(v1, &sa);
		if (rc <= 0) {
			syslog(LOG_ERR, 
				"switchOnOp(): invalid ha address %s\n", v1);
			exit(1);
		}
		mn_config.ha.s_addr = sa.sin_addr.s_addr;
		mn_config.sawHa = 1;
	/* mnha_key transform spi key */
	} else if (strcmp(op, "mnha_key") == 0) {
		mn_config.mnha_security = 1;
                dokey(op, linebuf, line, MNHASECURITY, 1);
#ifdef DHCP
#ifdef DIVERT 
	} else if (strcmp(op, "doNat") == 0) {
	        sscanf(linebuf,"%s %s",dc,v1);
		mn_config.doNat = atoi(v1);
		if ( (mn_config.dhcpEnable == 0) && (mn_config.doNat > 0)) {
			syslog( LOG_ERR, 
				"Exiting, NAT enabled but DHCP not enabled ");
			exit(1);
		}
        } else if (strcmp(op, "fr_number") == 0) {
                rc = sscanf(linebuf, "%s %s ", dc, v1);
		mn_config.fw_rule = atoi(v1);
#endif
#endif
	/* famn_key transform spi key */
	} else if (strcmp(op, "famn_key") == 0) {
		mn_config.famn_security = 1;
                dokey(op, linebuf, line, FAMNSECURITY, 1);
	/* NEW: netauth_key transform spi key */
	} else if (strcmp(op, "netauth_key") == 0) {
		mn_config.netauth_security = 1;
                dokey(op, linebuf, line, NETAUTHSECURITY, 0);
	/* NEW: adhoc_key transform spi key */
	} else if (strcmp(op, "adhoc_key") == 0) {
		mn_config.adhoc_security = 1;
                dokey(op, linebuf, line, ADHOCSECURITY, 0);
	/* tsmode 0 | 1  UDP registration */
	} else if (strcmp(op, "tsmode") == 0) {
		sscanf(linebuf, "%s %s", dc, v1);
		if (atoi(v1) == 0) {
			mn_config.tsmode = 0;
		}
	/* replay timestamp | nonce - UDP registration */
        } else if (strcmp(op, "replay") == 0) {
                sscanf(linebuf, "%s %s", dc, v1);
                if (strcmp(v1, "timestamp") == 0) {
                        mn_config.doTimestamp = 1;
                }
                else {
                        syslog( LOG_ERR,
                            "switchOnOp(): invalid replay style (%s)\n", v1);
                        exit(1);
                }
	/* NEW: beacon seconds */
	} else if (strcmp(op, "beacon") == 0) {
		rc = sscanf(linebuf, "%s %s %s", dc, v1, v2);
		mn_config.beaconRate = (unsigned short) atoi(v1);
		if (mn_config.beaconRate <= 0) {
                        syslog( LOG_ERR,
                            "switchOnOp(): invalid beacon rate (%d)\n", 
				mn_config.beaconRate);
                        exit(1);
		}
		if (rc == 3) {
			checkMulticast(v2);
		}
	/* NEW: beacon_timeslop seconds */
	} else if (strcmp(op, "beacon_timeslop") == 0) {
		sscanf(linebuf, "%s %s", dc, v1);
		mn_config.beacon_timeslop = atoi(v1);
	/* NEW: beacon_replay timestamp | nonce  on | off */
        } else if (strcmp(op, "beacon_replay") == 0) { 
		sscanf(linebuf, "%s %s %s", dc, v1, v2);
		if (strcmp(v1, "timestamp") == 0) 
			mn_config.beacon_timestamp = 1;
		else if (strcmp(v1, "nonce")==0) {
			syslog(LOG_ERR, "switchOnOp(): config error line [%d]: nonce replay not supported\n", line);
                        exit(1);
		}
		else mn_config.doTimestamp =1;
		if ((strcmp(v1, "off") ==0)||(strcmp(v2, "off")==0))
			mn_config.beacon_replay=0;
		else mn_config.beacon_replay=1;
	/* NEW: arp on | off */
	/* NOTE: arp is actually disabled here on the mobile i/f
	 * if so requested 
	 */
	} else if (strcmp(op, "arp") == 0) {
		sscanf(linebuf, "%s %s", dc, v1);
		if (strcmp(v1, "off") == 0) {
			mn_config.doArp = FALSE;
			mn_config.dhcpEnable = FALSE;         
#ifdef DIVERT
			mn_config.doNat = 0;
#endif
			arpOff();
		} else if (strcmp(v1, "on") == 0) {
			mn_config.doArp = TRUE;
			arpOn();
		} else {
			syslog(LOG_ERR, "Usage: arp <on | off>\n");
                        exit(1);
		}
	/* NEW: netauth accept | reject */
	} else if (strcmp(op, "netauth") == 0) {
		sscanf(linebuf, "%s %s", dc, v1);
		if (strcmp(v1, "accept") == 0) 
			mn_config.doNetauth = NETAUTHACCEPT;
		else if (strcmp(v1, "reject") == 0) 
			mn_config.doNetauth = NETAUTHREJECT;
		else {
			syslog(LOG_ERR, "Usage: netauth <accept | reject>\n");
                        exit(1);
		}
	/* NEW: adhoc accept | reject */
	} else if (strcmp(op, "adhoc") == 0) {
		sscanf(linebuf, "%s %s", dc, v1);
		if (strcmp(v1, "accept") == 0) 
			mn_config.doAdhoc = NETAUTHACCEPT;
		else if (strcmp(v1, "reject") == 0) 
			mn_config.doAdhoc = NETAUTHREJECT;
		else {
			syslog(LOG_ERR, "Usage: adhoc <accept | reject>\n");
			exit(1);
		}
	} 
 	/* location "string" */
        else if (strcmp(op, "location") == 0) {
                rc = extractString(linebuf, mn_config.location,
                                   MAXLOCATION);
                if (rc < 0) { 
			syslog(LOG_ERR, "Usage: location string \n");
                        exit(1);
                } 
                mn_config.sawLocation = 1;
	}
        else if (strcmp(op, "udp_port") == 0) {
		sscanf(linebuf, "%s %s", dc, v1);
		if (atoi(v1) == 0) {
			syslog(LOG_ERR, "Usage: udp_port may not be zero\n");
                        exit(1);
		}
		mn_config.udp_port = atoi(v1);
        /* logIPOnly <on|off> */
        } else if (strcmp(op, "logIPOnly") == 0) {
                sscanf(linebuf, "%s %s", dc, v1);
                if (strcmp(v1, "off") == 0) {
                        mn_config.logIPOnly = FALSE;
                } else if (strcmp(v1, "on") == 0) {
                        mn_config.logIPOnly = TRUE;
                } else {
                        syslog(LOG_ERR, "Usage: logIPOnly <on | off>\n");
                        exit(1);
                }
#ifdef DHCP
	/* dhcp config options:
	*/
        } else if (strcmp(op, "dhcp") == 0) {
                mn_config.dhcpEnable = TRUE;
                rc = sscanf(linebuf, "%s %s %s", dc, v1, v2);
                if (rc <= 1) {
                        syslog(LOG_ERR, "line %d of configuration file, mvif "
                                        "name must be specified", line);
                        exit(1);
                }
		/* hack to turn dhcp off
		*/
		if (strcmp(v1, "off") == 0) {
			mn_config.dhcpEnable = FALSE;
		}
		/* get interface name, probably mvif0, but maybe
		 * something else (unlikely to work)
		 */
                strcpy(mn_config.mvifname, v1); 
		/* get optional period to reup for DHCP
		*/
                if (rc == 3) {
                        rc = atoi(v2);
                        if (rc == 0 || rc >= 5)
                                mn_config.dhcp_renewing_period = rc;
                        else {
                                syslog(LOG_ERR, "line %d of configuration file,"
                                        " DHCP renewing period out of range (0 "
                                        "or >=5)", line);
                                exit(1);
                        }
                }
	/* dhcp config options:
	*/
        } else if (strcmp(op, "dhcp_dns") == 0) {
		mn_config.doDns = 1;
		if (!mn_config.dhcpEnable){
			syslog(LOG_ERR, "line %d, dhcp not enabled ", line);
			exit(1);
		}
	/* dhcp config options:
	 */
        } else if (strcmp(op, "dhcp_tunnel_policy") == 0) {
                rc = sscanf(linebuf, "%s %s ", dc, v1);
                if (rc <= 1) {
                        syslog(LOG_ERR, "line %d, invalid dhcp option", line);
                        exit(1);
                }
		if (mn_config.dhcpEnable){
			if ( strcmp(v1,"ipip") == 0){
				mn_config.dhcp_policy = IPIP;			  
			}
			else if ( strcmp(v1,"notunnel") == 0){
				mn_config.dhcp_policy = NOTUNNEL;			  
			}
#ifdef DIVERT
			else if ( strcmp(v1,"nat") == 0){
				if (mn_config.doNat) {			  
					mn_config.dhcp_policy = NAT;
				}
				else{
					syslog(LOG_ERR, "line %d,doNat not enabled ", line);
					exit(1);
				}
			}
#endif
			else {
				syslog(LOG_ERR, "line %d,unknown dhcp tunnel policy ", line);
				exit(1);
			}

		}else{
			syslog(LOG_ERR, "line %d, dhcp not enabled ", line);		  
			exit(1);
		}
#endif 
#ifdef KAMEIPSEC
	} else if(strcmp(op, "ipsec") == 0) {
		mn_config.kameipsecEnable = TRUE;
#endif  	
	/* fatal 
	*/
	} else {
		syslog(LOG_ERR, "switchOnOp(): line [%d], unknown config op in config file\n", line);
		exit(1);
	}
}

arpOff()
{
	set_arp(mn_config.ifname, "-arp", 128);
}

arpOn()
{
	set_arp(mn_config.ifname, "arp", -128);
}

/*
 * debug - dump configuration variables
 */
configDump()
{
	int             i;

	printf("---------config dump-------------------\n");
        printf("logIPOnly: %s\n", mn_config.logIPOnly ? "on" : "off");
	printf("sawMn: %s\n", mn_config.sawMn ? "yes": "no");
	printf("sawHa: %s\n", mn_config.sawHa ? "yes":"no");
	printf("mn name: %s\n", mn_config.ifname);
	printf("mn netmask: %s\n", inet_ntoa(mn_config.netmask));
	printf("mn clone route: %s\n", inet_ntoa(mn_config.cloneip));
	printf("mn udp port: %d\n", mn_config.udp_port); 
	printf("mn ip addr: %s\n", inet_ntoa(mn_config.mn));
	printf("mn use multicast %d\n", mn_config.useMulticast);
	if (mn_config.useMulticast) {
		printf("multicast addr is %s\n", inet_ntoa(mn_config.maddr));
	}
	printf("mn mac addr: ");
	for (i=0;i<6;i++)
		printf("%02x%c", mn_config.mac[i], (i<5)?':':' ');
	printf("\nha addr: %s\n", inet_ntoa(mn_config.ha));
	printf("tsmode: %s\n", mn_config.tsmode ? "fix time": "other");
	printf("doTimestamp : %s\n", mn_config.doTimestamp ? "timestamp": "nonce");
	printf("beacon rate: %d\n", mn_config.beaconRate);
	printf("beacon replay: %s\n", mn_config.beacon_replay?"on":"off");
	if (mn_config.beacon_replay) 
		printf("replay protection mode:%s\n", mn_config.beacon_timestamp?"timestamp":"nonce");
	printf("beacon_timeslop: %d\n", mn_config.beacon_timeslop);
	printf("arp: %s\n", (mn_config.doArp==TRUE)?"on":"off");
	printf("netauth: %s\n", (mn_config.doNetauth==NETAUTHACCEPT)?"accept":"reject");
	printf("adhoc: %s\n", (mn_config.doAdhoc==NETAUTHACCEPT)?"accept":"reject");
	if (mn_config.sawLocation) {
		printf("location string <%s>\n", mn_config.location);
	}
	/*
	 * print out security values
	 */
	if (mn_config.mnha_security) {
		printf("mnha keylist:\n");
		dump_keylist(mn_config.sa[MNHASECURITY].kl);
	}
	if (mn_config.famn_security) {
		printf("famn keylist:\n");
		dump_keylist(mn_config.sa[FAMNSECURITY].kl);
	}
	if (mn_config.adhoc_security) {
		i = 0;
		printf("adhoc spi %d, key=", mn_config.sa[ADHOCSECURITY].spi);
		while (i < 16) {
			printf("%2.2x", mn_config.sa[ADHOCSECURITY].key[i]);
			i++;
		}
		printf("\n");
	}
	if (mn_config.netauth_security) {
		i = 0;
		printf("netauth spi %d, key=", mn_config.sa[NETAUTHSECURITY].spi);
		while (i < 16) {
			printf("%2.2x", mn_config.sa[NETAUTHSECURITY].key[i]);
			i++;
		}
		printf("\n");
	}
#ifdef KAMEIPSEC
	printf("kameipsecEnable: enable %d\n", mn_config.kameipsecEnable);
#endif
#ifdef DHCP
	printf("dhcp: enable %d\n", mn_config.dhcpEnable);
	printf("dhcp: mvif name %s\n", mn_config.mvifname);
	printf("dhcp: period %d\n", mn_config.dhcp_renewing_period);
#ifdef DIVERT
	printf("doNat: on/off %d\n", mn_config.doNat);
	printf("fr_tunnel policy: %d\n", mn_config.fw_rule);
#endif
	printf("dhcp tunnel policy: %d\n", mn_config.dhcp_policy);
	printf("doDns switch : %d\n", mn_config.doDns);
#endif
	printf("---------config dump-------------------\n");
}

/*
 * take in security association key
 * 
 * example: mnha_key md5 256 9700ca3a82f7cf9b2d0cc739abd5f33f
 *
 * note: if ipflag is set we assume that an ipaddr follows the
 * op type.
 */
static
dokey(char *op, char *linebuf, int line, int type, int ipflag)
{
	char            dc[1024];	/* coa line, don't care */
	char		ipline[1024];   /* ipaddr */
	char            v1[1024];	/* transform */
	char            v2[1024];	/* spi */
	char            v3[1024];	/* key */
	int             l;
	int             rc;
	int             spi_wannabe;
	char           *p;
	struct 		sockaddr_in  sa;

	v1[0] = v2[0] = v3[0] = 0;
	/* resolve ip address if provided
	*/
	if (ipflag) {
		rc = sscanf(linebuf, "%s %s %s %s %s", dc, ipline, v1, v2, v3);
                if (rc != 5 || strlen(ipline) == 0) {
                        syslog(LOG_ERR, "config error line %d, invalid key line.\n", line);
                        syslog(LOG_ERR, "example: mnha_key ipaddr md5 256 hexkey...\n");
                        exit(1);
                }
                rc = resolveAddress(ipline, &sa);
                if (rc <= 0) {
                        syslog(LOG_ERR, "config error line %d, invalid key line.\n", line);
                        syslog(LOG_ERR, "ipaddr %s failed to resolve",
 ipline);
                        exit(1);
                }
	}
	else {
		rc = sscanf(linebuf, "%s %s %s %s", dc, v1, v2, v3);
		if (rc != 4) {
			syslog(LOG_ERR, "config error line %d, invalid key line\n", line);
			syslog(LOG_ERR, "should only have 4 parameters");
			exit(1);
		}
	}
	if (strlen(v1) == 0 || strlen(v2) == 0 || strlen(v3) == 0) {
		syslog(LOG_ERR, "config error line %d, invalid key line\n", line);
		exit(1);
	}

	/*
	 * spi cannot be 0.
	 */
	spi_wannabe = atoi(v2);
	if (spi_wannabe <= 255) {
		syslog(LOG_ERR, "config error line %d, bad spi\n", line);
		exit(1);
	}

	l = strlen(v3);
	p = v3;
	/*
	 * skip over initial hex 0x if present
	 */
	if (v3[0] == '0' && v3[1] == 'x' || v3[1] == 'X') {
		p += 2;
		l -= 2;
	}
	/*
	 * length of key string should be 32 for md5
	 */
	if (l != 32) {
		syslog(LOG_ERR, "config error line %d, key string should be 32 chars long\n", line);
		exit(1);
	}
	rc = pack_string(p, mn_config.sa[type].key, 32);
	if (rc < 0) {
		syslog(LOG_ERR, "config error line %d, invalid key string\n", line);
		exit(1);
	}
	mn_config.sa[type].spi = spi_wannabe;

	/* store ip addr and key material in keylist
	*/
	if (ipflag) {
		u_long	spi[4];
		bzero(spi, sizeof(u_long)*4);
		spi[0] = spi_wannabe;

 		rc = enter_keylist(
                        mn_config.sa[type].kl, /* keylist pointer */
                        sa.sin_addr.s_addr,     /* ip addr in net order */
                        mn_config.sa[type].key,  /* the key */
                        16,     /* HARDWIRED to MD5 for now */
                        spi,
                        1,      /* HARDWIRED to MD5 for now */
			0L,	/* ipaddr2: field not used for mobileIP keys */
			1,	/* inuse: field not used for mobileIP keys */
			0	/* timestamp - don't care here */
                        );
		if (rc <= 0) {
			syslog(LOG_ERR, "config error line %d, invalid key string\n", line);
			if (rc < 0)
				syslog(LOG_ERR, "malloc - out of space");
			else
				syslog(LOG_ERR, "duplicate ip address");
			exit(-1);
		}
	}
}

/*
 * following are stubs necessary to allow icmpmsg.c to be linked in TBD. fix
 * this somehow
 */
char           *
configLocation()
{
	if (mn_config.sawLocation)
		return(mn_config.location);
	return (NULL);
}

int 
getBeaconRate()
{
	return(mn_config.beaconRate);	
}

int 
getNetauthSecurity()
{
	return(mn_config.netauth_security);	
}

int 
getAdhocSecurity()
{
	return(mn_config.adhoc_security);	
}

char	
*getNetauthKey()
{
	if ((mn_config.doArp==FALSE)&&(mn_config.netauth_security))
		return(mn_config.sa[NETAUTHSECURITY].key);	
	else return(NULL);
}

char	
*getAdhocKey()
{
	if ((mn_config.doArp==FALSE)&&(mn_config.adhoc_security))
		return(mn_config.sa[ADHOCSECURITY].key);	
	else return(NULL);
}

int	
getNetauthSpi()
{
	if (mn_config.netauth_security)
		return(mn_config.sa[NETAUTHSECURITY].spi);	
	else return(0);
}

int	
getAdhocSpi()
{
	if (mn_config.adhoc_security)
		return(mn_config.sa[ADHOCSECURITY].spi);	
	else return(0);
}

/*
 * query: do we use arp on this interface
 *
 * returns:
 *	-1: no such interface
 *	 0: arp is off
 *	 1: arp is on
 */
int	
getConfigArp(struct in_addr *ifaddr)
{
	/* check the interface, if this didn't match, bail
	*/
#ifndef DHCP
	/* TBD. jrb.  this code doesn't make sense.
	*/
        /* in DHCP_MODE, beacons come in from the interface with temp addr 
	if (ifaddr->s_addr != mn_config.mn.s_addr)
		return(-1);
	*/
#endif
	/* now check the arp value
	*/
	if (mn_config.doArp)
		return(1);
	else 
		return(0);
}

getConfiglogIPOnly()
{
	return(mn_config.logIPOnly);
}

int 	
getConfigMac(char *router)
{
	if (mn_config.doArp)
		return(0);
	else return(1);
}

/*
 * returns:
 *	1 - if netauth is in reject mode
 *	0 - if accept
 */
int
checkNetauthReject()
{
	if (mn_config.doNetauth == NETAUTHREJECT)
		return(1);
	return(0);
			
}

void
setNetauthReject()
{
	mn_config.doNetauth = NETAUTHREJECT;
}

void
setNetauthAccept()
{
	mn_config.doNetauth = NETAUTHACCEPT;
}

int
checkAdhocReject()
{
	if (mn_config.doAdhoc == NETAUTHREJECT)
		return(1);
	return(0);
			
}

void
setAdhocReject()
{
	mn_config.doAdhoc = NETAUTHREJECT;
}

void
setAdhocAccept()
{
	mn_config.doAdhoc = NETAUTHACCEPT;
}

/*
 * callback from getIfs() - store netmask acc. to interface
 *
 * in this case we are only interested in the mobile node interface
 */
void
configInterface(char *ifname, struct sockaddr_in *sin, 
	struct sockaddr_in *netmask, int mcast, int link_index)   
{
        int i;

	if (strcmp(mn_config.ifname, ifname) == 0) {
		mn_config.netmask = netmask->sin_addr;
		mn_config.cloneip.s_addr = mn_config.mn.s_addr 
			& mn_config.netmask.s_addr;
		mn_config.link_index = link_index;
	}
}

/*
 * map string name of interface to internal if index number and ip address
 */
void
mapIfname(char *ifname, int *index, unsigned long *ifaddr)
{
	*index = mn_config.link_index;
	*ifaddr = mn_config.mn.s_addr;
}
/*
 * routine makes sure that input string is valid limited broadcast
 * or multicast address
 */
static void
checkMulticast(s)
char *s;
{
	struct in_addr maddr;
	int rc;

	/* checking for multicast address here
	*/
	rc = inet_aton(s, &maddr);
	if (rc == 0) {
		syslog(LOG_ERR, "address part (%s) of beacon <seconds> addr invalid", s);
		exit(1);
	}

	/* 255.255.255.255 is valid */
	if (maddr.s_addr == 0xffffffff) {
		mn_config.useMulticast = 0;
		return;
	}

	/* else address should be multicast */
	if(IN_MULTICAST(ntohl(maddr.s_addr))) {
		mn_config.useMulticast = 1;
		mn_config.maddr = maddr;
		return;
	}
	syslog(LOG_ERR, "address part (%s) of beacon <seconds> addr invalid", s);
	exit(1);
}

int
indextoif(int index)
{
	if (mn_config.link_index == index)
		return(mn_config.mn.s_addr);
	return(0);
}

#ifdef MADRP

u_long getmyaddr() { return mn_config.mn.s_addr; }
int IsAgent(){ return FALSE;}

#endif  /* MADRP */


#ifdef TESTMAIN
int debug = 1;
int ipsecKeysInuse = FALSE;

/*
 * syntax: mnconfig <config file>
 */

logperror(char *s)
{
	printf("%s\n", s);
}

stopAlarm()
{
}

startAlarm()
{
}

main(argc, argv)
int argc; 
char **argv;
{
	/* open a local log file */
        if ( mip_openlog( "./mnd.log.local" ) < 0 ) {
            syslog( LOG_ERR,
                "Failed to open local logfile \"%s\".", "./mnd.log.local" );
            exit(1);
        }

	configRead(argv[1]);
	configDump();

	mip_closelog();
}
#endif

