
/*
 * Copyright (c) 1996, 2001 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * mnmain.c
 *
 * mobile ip "mobile node" daemon.
 *
 * DEBUG:
 *	-d for printfs
 *	define TIMEDEBUG for 1 time clock tick per 
 *		icmp message recv. otherwise you
 *		have to deal with an alarm
 */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/signal.h>
#include <termios.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/ip_var.h>
#ifdef KAMEIPSEC
#include <netinet6/ipsec.h>
#endif
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>

#include "mnconfig.h"
#include "mip_icmp.h"	/* icmp router advert + appropriate mse extension/s */
#ifdef MADRP
#include "sr_mad.h"
struct sr_config  sr_config;
#endif
#include "mnstate.h"
#include "mnstats.h"
#include "unixsock.h"
#include "mip_log.h"
#include "mndversion.h"
#include "route.h"
#include "alist.h"
#include "icmpstats.h"

struct mn_stats mn_stats;
struct mn_state mns;
struct icmp_stats icmp_stats;

static struct 	sockaddr_in ifaddr;	/* mobile interface */
static struct 	sockaddr_in maddr;	/* multicast dest, if used */
static char 	configFile[1024];
static int	noticeFunction = 2;
char *		progname;
char **		progargs;
int 		debug = 0;
int             pubOrpriv = 0; 
int		errorexit = 0;	/* non-zero if onhup() is called to cleanup
				 * after a runtime error is discovered */


/* prototypes
*/

void write_mndconfig();

void mnTick();

void parseArgs(int, char **);

void icmpBeacon();
void pushBeacon();

void
mnStateTick(int udpSocket);

int handleIcmpAdvert(int icmpSocket );

void
sendBeacon(struct sockaddr_in *router, struct sockaddr_in *dst,
      struct sockaddr_in *coa, int mse_lifetime, int identity,
      int bothAgents, u_short seq, u_char *mac_addr, int mcastflag,
	struct sockaddr_in *ipsrc, char *ifname);

static void unblockSignals();

#ifdef KAMEIPSEC
void enableKameIpsec();
#endif

main(argc, argv)
int argc;
char **argv;
{
	void initState();
	extern void onusr1();
	extern void onusr2();
	extern void onhup();

	/* check status - root only
	*/
	if (getuid()) {
		fprintf(stderr, "mnd: not super user\n");
		exit(1);
	}
	progname = argv[0];
	progargs = argv;

        /* open log files: logging via syslogd for fatal errors, startup,
         * logging to MND_LOGFILE for security specific
         * errors and events, and for packet parsing errors.
         */
        openlog("mnd", LOG_CONS | LOG_PID , LOG_DAEMON);
        if ( mip_openlog( MND_LOGFILE ) < 0 ) {
            syslog( LOG_ERR,
                "Failed to open local logfile \"%s\".", MND_LOGFILE );
            exit(1);
        }

	/*
	 * configuration file read in here, but linkage to
	 * ifconfig not checked yet.
	 */
	parseArgs(argc, argv);

	initSockets();
	initState();
	
	/* become a daemon
	*/
	if (!debug) {
		if (daemon(0,0) < 0) {
			syslog(LOG_ERR, "main(): daemon (%m)\n");
			exit(1);
		}
	}

	signal(SIGALRM, mnTick);
	signal(SIGUSR1, onusr1);
	unblockSignals();
	signal(SIGHUP, onhup);
	signal(SIGINT, onhup);	/* allows cleanup even when debugging */
        signal(SIGTERM, onhup);

	storepid();

	syslog(LOG_INFO, "Mobile node daemon started [%d].  (%s)",
		getpid(), MNDVERSION);
	mip_log(MLOG_INFO, "Mobile node daemon started [%d].  (%s)",
		getpid(), MNDVERSION);


	/* loop forever
	*/ 
	while(1) 
	{ 
		doit(); 
	} 
}


/*
 * 1. open a udp socket at port 434, the mobile-ip port.
 * We recv. authentication replies on this port
 *
 * 2. open an ICMP raw socket, we get router adverts on this
 * port.
 */
static int icmpSocket;
int udpSocket;

static int unixSocket;
#ifdef DIVERT
static int divertSocketIn;
static int divertSocketOut;
#endif
#ifdef MADRP
int madSocket;
#endif

extern int dhcpSocket;

static int routeSocket;

static int bpfSocket;	/* used for icmp input only */
static int useBpf;	/* 1 if bpf in use */

initSockets()
{
	if (mn_config.useMulticast) {
		/* send multicast, maddr is group addr, mn is if address */
		icmpSocket = initIcmp(0, 1, mn_config.maddr, mn_config.mn);
		maddr.sin_addr.s_addr = mn_config.maddr.s_addr;
	}
	else {
		/* broadcast sent only */
		icmpSocket = initIcmp(1, 0, 0, 0);
	}
	udpSocket = initUdp(mn_config.udp_port);
	unixSocket = initUnixSocket(MNDSOCK_SERVER, FALSE);
	
#ifdef KAMEIPSEC
	if ( mn_config.kameipsecEnable ) {
		enableKameIpsec();
	}
#endif

#ifdef MADRP
	madSocket= init_sr_Udp();
#endif
#ifdef DHCP
        initDhcpSocket();
#endif
	routeSocket = routeInit();

	/* if we are using arp, we must use the bpf input
	 * mechanism for icmp packets
	 */
	if (mn_config.doArp) {
		char errbuf[2048];
		int rc;

		/* bpf only used in arp mode currently.
		 * icmp input filter defined in bpf module.
		 * note: MADRP needs icmp unreachables as well as
		 * router discovery packets needed by MIP in general
		 * "icmp[0] = 9" works for the latter.  But then
		 * MADRP cannot use arp ... 
		 */
		bpfSocket = init_bpf(mn_config.ifname, errbuf, 1); 
		if (bpfSocket < 0) {
			syslog(LOG_INFO, "%s", errbuf);
			mip_log(MLOG_INFO, "%s", errbuf);
			exit(1);
		}
		/* possible but not likely that we might have
		 * a non ARP device.  probably wouldn't work anyway
		 * but might be uesful for testing.
		 */
		rc = bpf_needed();
		if (rc == FALSE) {
			bpf_close();
			useBpf = 0;
		}
		else {
			useBpf = 1;
			shutdown(icmpSocket, 0);	/* no reads */
		}
	}
	else {
		char errbuf[2048];
		int rc;
		/* we only use the bpf for writing icmp packets to limited   
		 * broadcast 
		*/
		bpfSocket = init_bpf(mn_config.ifname, errbuf, 0); 
		useBpf = 0;
		if (bpfSocket < 0) {
			syslog(LOG_INFO, "%s", errbuf);
			mip_log(MLOG_INFO, "%s", errbuf);
			exit(1);
		}
	}
}


static
mnd_shutdown()
{
        /* config the network to default setting before exiting */
#ifdef DHCP  
        if (mns.state == CO_LOCATED) {
		//disable_dhcp();
		dhcp_tunnel_policy(NOSTATE);
        }
#endif
	/* delete current default route */
	route_nofail();
	if (mns.haroute) {
#ifdef MADRP
		if(mns.agent && mns.agent->sawMac)
#endif
			mn_route_delete_default(mn_config.ha,1);
	    	mns.haroute = 0;
	}
	else if ( mns.state == AWAY ) {
#ifdef MADRP
		if(mns.agent && mns.agent->sawMac)
#endif
			mn_route_delete_default(mns.fa, 1);
	}
	route_fail();

	/* delete mns in mnlist */
	mn_delete_all();
#ifdef MADRP
	hlist_delete_all();
#endif
	/* if arp was off, and we are exiting, restore clone
	 * route, and make sure arp is turned back on for interface
	 */
	if (mn_config.doArp == FALSE) {
		clone_route_add(mn_config.cloneip, mn_config.mn,
			mn_config.netmask );
		mns.cloneroute = 1;
		arpOn();
	}

	/* replace default route.  assume it was to the ha over the
	 * interface stated in mnd.conf
	 */
	route_nofail();
	mn_route_add_default(mn_config.ha, mn_config.mn, 0, NULL,
		mn_config.ifname );
	route_fail();

	/* close the logs files and sockets */
	closelog();
	mip_closelog();

	close(routeSocket);
#ifdef MADRP
	close(madSocket);
#endif
	close(unixSocket);
	close(udpSocket);
	close(icmpSocket);
#ifdef DHCP
#ifdef DIVERT
	if (mn_config.doNat){
		close(divertSocketIn);
	  	close(divertSocketOut);
	}
#endif
        close(dhcpSocket);
	restoreResolver();
#endif  
	close(bpfSocket);
}

/*
 * sigusr1 needs to be unblocked in terms of its signal
 * mask as we may have restarted
 */
static
void
unblockSignals()
{
	sigset_t s;
	sigset_t o;
	int rc;

	sigemptyset(&s);
	sigaddset(&s, SIGUSR1);
	rc = sigprocmask(SIG_UNBLOCK,&s,&o);
	if (rc < 0) {
		syslog(LOG_INFO, "sigprocmask failed\n");
		exit(1);
	}
}

void
onhup( void )
/*
 *  --	delete default and host route before exiting under controlled
 *	circumstances.
 */
{
	char	*message = "Daemon exiting (mnd) - received HUP signal.";

	if (errorexit) {
		message = "Daemon exiting (mnd) -- runtime error.";
	}
	syslog(LOG_INFO, "%s", message);
	mip_log(MLOG_INFO, "%s", message);

	mnd_shutdown();
	
	if (errorexit)
		exit(1);
	else
		exit(0);

}  /* end onhup() */

reboot_mnd(char *msg, char *args, char *switches)
{
	static char *vec[5];
	syslog(LOG_INFO, msg);
	mip_log(MLOG_INFO, msg);

	mnd_shutdown();

	if (args == NULL) {
		if (execvp(progname, progargs) < 0) {
		    syslog(LOG_ERR, "mnd reboot(): execvp failed (%m)");
		    exit(1);
		}
		exit(0);
	}
	vec[0] = progname;
	vec[1] = switches;
	vec[2] = "-c";
	vec[3] = args;		/* expect config file name */
	vec[4] = 0;		/* null-terminated */

	if (execvp(progname, vec) < 0) {
	    syslog(LOG_ERR, "mnd reboot(): 802.11 execvp failed (%m)");
	    exit(1);
	}
}

void
onusr1( void )
/*
 *  --	reset daemon
 */
{
	signal(SIGALRM, SIG_IGN);
	reboot_mnd("Daemon restarting (mnd) - received USR1 signal.",
			NULL, NULL);
}


void
parseArgs(argc, argv)
	int             argc;
	char          **argv;
{
	extern int      errno;
	extern char    *optarg;
	extern int      optind;
	int             ch, fd;
	int 		rc;
	int parseOnly = 0;

	strncpy(configFile, CONFIGDEFAULT, 1024);
	while ((ch = getopt(argc, argv, "pdc:")) != EOF) {
		switch (ch) {
		case 'p':
			parseOnly = 1;
			break;
		case 'd':
			debug = 1;
			break;
		case 'c':
			strncpy(configFile, optarg, 1024);
			break;
		case '?':
		default:
			usage();
		}
	}

	/* read in config file
	*/
	rc = configRead(configFile);
	if (rc < 0) {
		exit(1);	/* NOTE: all errors logged within function */
	}
#ifdef MADRP
	sr_initconfig(&mn_stats.sr_metric);
#endif

	if (parseOnly) {
		debug = 1;
	}

	if (debug) {
		printf("configFile %s\n", configFile);
		configDump();
		if (parseOnly) {
			printf("parse mode successfully completed - exit\n"); 
			exit(0);
		}
	}
}

usage()
{
	fprintf(stderr,"mnd: -p -d -c <config file>\n");
	exit(1);
}

void 
initState()
{
	struct in_addr	unknownRoute;
	int i,j;

	bzero(&mns, sizeof(struct mn_state));

	mns.cloneroute = 1;		/* assumed to exist at boot */
	mns.halifetime = HALIFETIME;
	mns.orig_lifetime = 30;		/* something non-zero just in case */

	/*
	 * be sure no default route (and associated host route) exists
	 * before starting.
	 *
	 * TBD: find link route before deleting default
	 *      also, don't delete routes that don't exist.
	 *
 	 * note: XXX. this is a HACK!.  we can't delete the link
	 * layer route, but we can delete the default.
	 */
	route_nofail();
	unknownRoute.s_addr = 0;
	mn_route_delete_default(unknownRoute, 0);
	/* delete the HA's link layer route, else
	 * we may have problems with Kame IPSEC as a consequence
	 */
	route_delete_link(mn_config.ha, NULL);

	route_fail();

	/* setup socket needed for sending beacons
	*/
        ifaddr.sin_family = AF_INET;
        ifaddr.sin_addr.s_addr = mn_config.mn.s_addr;

	/* multicast dst if used */
        maddr.sin_family = AF_INET;

	/* if arp is off, delete clone route
	 * note: we must put it back on a hup/exit.
	 */
	if (mn_config.doArp == FALSE) {
		route_nofail();
		clone_route_delete(mn_config.cloneip);
		mns.cloneroute = 0;
		route_fail();
	}
        
	/* dhcp mode setup.  mns.mode tracks whether or
	 * not we are actively doing dhcp.
	 * auto_mode tracks whether or not we manually force dhcp.
	*/
        mns.mode = AGENT_MODE;		/* not doing dhcp yet */
        if (mn_config.dhcpEnable)
                mns.auto_mode = TRUE;	/* do dhcp automagically */
        else {
                mns.auto_mode = -1;	/* no dhcp at all */
	}

}  /* end initState() */


static int pending_tick;

doit()
{
	fd_set readmask;
	int rc;
	extern int errno;
	struct in_addr zero;

	pending_tick = 0;
	zero.s_addr = 0;
#ifndef TIMEDEBUG
	/* normal alarm - 1 tick per second
	*/
	startAlarm();
#endif
	while(1) {
#ifndef TIMEDEBUG
		/* alarm only sets this pending_tick flag
		*/
		if (pending_tick) {
			if (debug)
				printf("clock tick\n");
			pending_tick--;
			mnStateTick(udpSocket);

                      	/* if arp is off, we send beacons to provide our mac address
			*/
                        if (mn_config.doArp==0) {
                                icmpBeacon();
			}
			mn_stats.ticks++;
#ifdef MADRP
			mn_tickBackoffTimer();
#endif
		}
#endif
		/* select on icmp and udp sockets
		 * select on 1 second interval.  This
		 * is heartbeat clock for daemon and timer/s
		 * are driven from it. This allows mnd to 
		 * react to lack of agent, respond to registration timer, etc.
		 */
		FD_ZERO(&readmask);
		FD_SET(udpSocket, &readmask);
		FD_SET(unixSocket, &readmask);
#ifdef DIVERT
		FD_SET(divertSocketIn,&readmask);
		FD_SET(divertSocketOut,&readmask);
#endif
#ifdef MADRP
		FD_SET(madSocket, &readmask);
#endif
                FD_SET(dhcpSocket, &readmask);
		FD_SET(routeSocket, &readmask);
		if (useBpf) {
			FD_SET(bpfSocket, &readmask);
		}
		else {
			FD_SET(icmpSocket, &readmask);
		}
		
		/*
		 * select must get the most recently created socket, which
		 * we mysteriously "know" (see initSockets()).
		 */
		rc = select(100,(fd_set *) &readmask,NULL,NULL,NULL);
		if (rc < 0) {
			if (errno == EINTR) {
				continue;
			}
			syslog( LOG_ERR, "doit(): select (%m)");
			exit(1);
		}
		if (FD_ISSET(icmpSocket, &readmask)) {
			/* got icmp advert message */
			handleIcmpAdvert(icmpSocket );
		}
		if (FD_ISSET(bpfSocket, &readmask)) {
			/* got icmp advert message */
			handleBpfInput(bpfSocket);
		}
		if (FD_ISSET(udpSocket, &readmask)) {
			handleUdpInput(udpSocket);
			mn_stats.replies++;
		}
		if (FD_ISSET(unixSocket, &readmask)) {
			handleUnixInput(unixSocket);
		}
		if (FD_ISSET(routeSocket, &readmask)) {
			routeRead();
		}
#ifdef MADRP
		if (FD_ISSET(madSocket, &readmask)) {
			handleMadInput(madSocket, &mn_stats.sr_metric);
		}
#endif
                if (FD_ISSET(dhcpSocket, &readmask)) {
                        dhcpStateMachine(1);
                }
#ifdef TIMEDEBUG
		/* just do a tick per icmp message
		*/
		mnStateTick(udpSocket);
#endif
	}
}

#ifdef SAVEENERGY
static u_long mytick=0;
u_long changestate=0;
#endif

/* 1. reregistration timer:
 *	set by handleIcmpAdvert
 *	determines how long we wait until we try and resend registration.
 *	if timeout
 *		resend registration
 * 2. request timer:
 *	set by handleIcmpAdvert
 *	determines how long to wait until we resend request, if we get
 *		a UDP reply, then we registered
 */
void
mnTick()
{
	pending_tick++;
#ifdef SAVEENERGY
        mytick++;
	changestate=mytick/300;
#endif  

	startAlarm();
}

startAlarm()
{
	alarm(1);
}

stopAlarm()
{
	alarm(0);
}

int
handleUnixInput(int unixSocket)
{
 	int rc;
        char buf[MAXBUF];
	struct hintmsg *h;
	struct changekey *k;
	struct alist	curAgent;
	int mndpid;
	char testkey[16];

	bzero( &curAgent, sizeof(struct alist) );
	bzero( testkey, 16);

        rc = recvUnixMessage(buf, MAXBUF);
        
        if (debug) { 
                printf("unixMessage: buf %d, rc %d\n", buf[0], rc);
        }
        switch (buf[0]) {
	case WRITECONF:
		write_mndconfig();
		break;

	case U_MNBEACONS:
                sendUnixMessage(MNDSOCK_CLIENT,
                        (char *)&icmp_stats, sizeof (struct icmp_stats));
                break;
	case X_MNBEACONS:
                sendUnixMessage(XMNDSOCK_CLIENT,
                        (char *)&icmp_stats, sizeof (struct icmp_stats));
                break;
        case U_MNSTAT:
                sendUnixMessage(MNDSOCK_CLIENT,
                        (char *)&mn_stats, sizeof (struct mn_stats));
                break;
	case X_MNSTAT:
		sendUnixMessage(XMNDSOCK_CLIENT,
			(char *)&mn_stats, sizeof (struct mn_stats));
		break;
        case U_MNSTATE:
                sendUnixMessage(MNDSOCK_CLIENT,
                        (char *)&mns, sizeof (struct mn_state));
                break;
	case X_MNSTATE:
		sendUnixMessage(XMNDSOCK_CLIENT,
			(char *)&mns, sizeof (struct mn_state));
		break;
	case U_MNHINT:
		h = (struct hintmsg *) buf;

		/* load the hint timer
		*/
		memcpy((char *)&mns.hintaddr, h->ip, 4);

		/*
		 *  if hintaddr is not zero, turn on hint_flag.  Else
		 *  turn it off.  If hint_set is already set, that means
		 *  an old HINT is being overridden by a new hint, so turn
		 *  off the hint_set flag to notify the daemon in
		 *  mnStateTick().
		 */
		if ( mns.hintaddr.s_addr != 0 ) {
		    mns.hint_flag = 1;		/* HINT turned on */
		    if (mns.hint_set) {
			mns.hint_set = 0;	/* HINT reset */
		    }
		}
		else {
		    mns.hint_flag = 0;		/* HINT turned off */
		}

		if (debug) {
			printf("hint flag %d, hint set %d, hint address %s\n",
				mns.hint_flag, mns.hint_set,
				inet_ntoa(mns.hintaddr));
		}
		break;
	case U_ALIST:
		mnstate_dumplist(MNDSOCK_CLIENT, TYPE_ALIST);
		break;
	case X_ALIST:
		mnstate_dumplist(XMNDSOCK_CLIENT, TYPE_ALIST);
		break;
	case U_MNLIST:
		mnstate_dumplist(MNDSOCK_CLIENT, TYPE_MNLIST);
		break;
	case X_MNLIST:
		mnstate_dumplist(XMNDSOCK_CLIENT, TYPE_MNLIST);
		break;
#ifdef MADRP
	case U_HLIST:
		mnstate_dumplist(MNDSOCK_CLIENT, TYPE_HLIST);
		break;
#endif
	case AUTHCHANGEKEY:
		k = (struct changekey*)(buf+4);
		enterNowhereState(mns.state, TRUE);

		switch(k->change_type){
			case CHANGE_MNHA:
				if (memcmp(k->key, testkey, 16)==0){
					mn_config.mnha_security = OFF;
					if (debug)
						printf("turn mnha off\n");
					break;
				}
				mn_config.mnha_security = ON;
				mn_config.sa[0].spi = 256;
				memcpy(mn_config.sa[0].key, k->key, 16);
				break;
			case CHANGE_FAMN:
				if (memcmp(k->key, testkey, 16)==0){
					mn_config.famn_security = OFF;
					break;
				}
				mn_config.famn_security = ON;
				mn_config.sa[1].spi = 256;
				memcpy(mn_config.sa[1].key, k->key, 16);
				break;
			case CHANGE_NETAUTH:
				if (!mn_config.doArp && mn_config.doNetauth){ /* doing auth & accepting netauth */
					mn_delete_all(); /* flush mnlist */
					alist_delete_all(); /* flush agentlist */
#ifdef MADRP
					hlist_delete_all(); /* flush hlist */
#endif
				}
				if (memcmp(k->key, testkey, 16)==0){
					mn_config.netauth_security = OFF;
					break;
				}
				mn_config.netauth_security = ON;
				mn_config.sa[2].spi = 256;
				memcpy(mn_config.sa[2].key, k->key, 16);
				break;
			case CHANGE_ADHOC:
				if (!mn_config.doArp && mn_config.doAdhoc){ /* doing auth & accepting adhoc */
					mn_delete_all(); /* flush mnlist */
					alist_delete_all(); /* flush agentlist */
#ifdef MADRP
					hlist_delete_all(); /* flush hlist */
#endif
				}
				if (memcmp(k->key, testkey, 16)==0){
					mn_config.adhoc_security = OFF;
					break;
				}
				mn_config.adhoc_security = ON;
				mn_config.sa[3].spi = 256;
				memcpy(mn_config.sa[3].key, k->key, 16);
				break;
			}
		break;
	case AUTHPOLICY:
		enterNowhereState(mns.state, TRUE);
		switch(buf[1]){
			case ACCEPT_NETAUTH:
				mn_config.doNetauth = NETAUTHACCEPT;
				break;
			case REJECT_NETAUTH:
				mn_config.doNetauth = NETAUTHREJECT;
				mn_delete_all(); /* flush mnlist */
				alist_delete_all(); /* flush agentlist */
#ifdef MADRP
				hlist_delete_all(); /* flush hlist */
#endif
				break;
			case ACCEPT_ADHOC:
				mn_config.doAdhoc = NETAUTHACCEPT;
				break;
			case REJECT_ADHOC:
				mn_config.doAdhoc = NETAUTHREJECT;
				mn_delete_all(); /* flush mnlist */
				alist_delete_all(); /* flush agentlist */
#ifdef MADRP
				hlist_delete_all(); /* flush hlist */
#endif
				break;
		}
		break;
	case LINKMODE:
		if (buf[1] == ARP_TO_AUTH){
			clone_route_delete(mn_config.cloneip);
			mns.cloneroute = 1;
			mn_config.doArp = FALSE;
			mn_config.doNetauth = TRUE;
		}
		else if (buf[1] == AUTH_TO_ARP){
			/* FIX -- what happens here if going to arp/ipsec?
			 * how about no clone period... */
			clone_route_add(mn_config.cloneip,
				mn_config.mn,
				mn_config.netmask
				); /* add clone route */
			mns.cloneroute = 1;
			mn_config.doArp = TRUE;
			mn_config.doNetauth = FALSE;
		}
		enterNowhereState(mns.state, TRUE);  
		arpOn();
		flushArp();
		mn_delete_all(); /* flush mnlist */
		alist_delete_all(); /* flush agentlist */
#ifdef MADRP
		hlist_delete_all(); /* flush hlist */
#endif
		break;
	case GETMNDPID:
		mndpid = getpid();
		sendUnixMessage(XMNDSOCK_CLIENT, (char*)&mndpid, sizeof(int));
		break;
        case U_MODESWITCH:
		/* auto mode on
		*/
                if (mn_config.dhcpEnable && buf[1] == 0) {
                        mns.auto_mode = TRUE;
                        mip_log(MLOG_INFO, "User request: enable AUTOmatic "
                                "mode switching\n");
                }
		/* auto mode off, force agent mode (not dhcp)
		*/
                else if (mn_config.dhcpEnable && buf[1] == 1) {
                        mns.auto_mode = FALSE;
                        if (mns.mode != AGENT_MODE) {
                                enterNowhereState(mns.state, TRUE);
                                mns.mode = AGENT_MODE;
                        }
                        mip_log(MLOG_INFO, "User request: disable automatic "
                                "mode switching, set to AGENT_MODE\n");
                }
		/* auto mode off, force dhcp mode
		*/
                else if (mn_config.dhcpEnable && buf[1] == 2) {
                        mns.auto_mode = FALSE;
                        if (mns.mode != IPIP) {
                                enterNowhereState(mns.state, TRUE);
                                mns.mode = IPIP;
                        }
                        mip_log(MLOG_INFO, "User request: disable automatic "
                                "mode switching, set to IPIP\n");
                }
		/* auto mode off, force dhcp nat  mode
		*/
                else if (mn_config.dhcpEnable && buf[1] == 3) {
                        mns.auto_mode = FALSE;
                        if (mns.mode != NAT ) {
                                enterNowhereState(mns.state, TRUE);
                                mns.mode = NAT;
                        }
                        mip_log(MLOG_INFO, "User request: disable automatic "
                                "mode switching, set to NAT \n");
		}
		/* auto mode off, force no tunnel mode
		*/
                else if (mn_config.dhcpEnable && buf[1] == 4) {
                        mns.auto_mode = FALSE;
                        if (mns.mode != NOTUNNEL) {
                                enterNowhereState(mns.state, TRUE);
                                mns.mode = NOTUNNEL;
                        }
                        mip_log(MLOG_INFO, "User request: disable automatic "
                                "mode switching, set to NOTUNNEL \n");
                }		
                else {
                        mip_log(MLOG_INFO, "User request on mode switching "
                                "rejected: MND not DHCP capable\n");
                }
                sendUnixMessage(MNDSOCK_CLIENT, (char *)&mn_config.dhcpEnable,
                                                                sizeof(int));
                break;
	}

}  /* end handleUnixInput() */



void 
fatal(char *s)
{
	syslog( LOG_ERR, "Fatal Error -- %s", s);
	exit(1);
}

logperror(char *s) 
{
	syslog( LOG_ERR, "%s (%m)", s);
	exit(1);
}


/*
 * send a beacon every beaconClock seconds
 * This is only done if arp is off.
 */

void
icmpBeacon()    
{
        static int beaconClock=0;

        /* if virtual beacon clock times out, send beacons
         * and reload counter   
        */
        if (beaconClock <= 0) { 
                beaconClock = mn_config.beaconRate;
		/* XXX. lifetime is 10.  TBD what should it be?
		*/
		if (debug)
			printf("sendBeacon\n");
		pushBeacon();
        }
        beaconClock--;  
}

/*
 * send mn beacon
 */
void
pushBeacon()
{
	sendBeacon(
		&ifaddr, 	/* router, in this case SELF */
		(mn_config.useMulticast) ? &maddr : &ifaddr,  /* dst */
		NULL,  /* FA, not applicable here */
		3 * mn_config.beaconRate,   /* rate to beacon */
		SELF_MN, 		/* who am I */
		0,                      /* neither HA nor FA */
		getSequence(&mn_config.beacon_seq),  /* sequence number */
		mn_config.mac, /* mac address */
		mn_config.useMulticast,  /* multicast or unicast */
		&ifaddr,	/* ip src */ 
		mn_config.ifname   /* if name as string */
	);
}

#ifdef KAMEIPSEC
static char * outbypassPolicy = "out bypass";
static char * inbypassPolicy = "in bypass";
/* enable kame ipsec 
 *
 * gist of this:
 *	udpSocket and icmpSocket are allowed to bypass any
 * IPSEC policy for packets in/out.
*/
void
enableKameIpsec()
{
  char *outbypassbuf, *inbypassbuf;
  int outbypasslen, inbypasslen;
  int     ipsec_policy = IPSEC_POLICY_BYPASS;

  outbypassbuf = ipsec_set_policy( outbypassPolicy, strlen(outbypassPolicy));
  if( outbypassbuf == NULL )
    {
      syslog(LOG_ERR, "enableKameIpsec() : ipsec_set_policy failed, May be kernel is not compiled with IPSEC enabled\n");
      mip_log(MLOG_SEC, "enableKameIpsec() : ipsec_set_policy failed, May be kernel is not compiled with IPSEC enabled\n");
      if (debug)
	printf("enableKameIpsec() : ipsec_set_policy failed, May be kernel is not compiled with IPSEC enabled\n");
      exit(1);
    }
  
  outbypasslen = ipsec_get_policylen(outbypassbuf);
  

  inbypassbuf = ipsec_set_policy( inbypassPolicy, strlen(inbypassPolicy));
  if( inbypassbuf == NULL )
    {
      syslog(LOG_ERR, "enableKameIpsec() : ipsec_set_policy failed, May be kernel is not compiled with IPSEC enabled\n");
      mip_log(MLOG_SEC, "enableKameIpsec() : ipsec_set_policy failed, May be kernel is not compiled with IPSEC enabled\n");
      if (debug)
	printf("enableKameIpsec() : ipsec_set_policy failed, May be kernel is not compiled with IPSEC enabled\n");
      exit(1);
    }
  
  inbypasslen = ipsec_get_policylen(inbypassbuf);
  

  /* set in and out bypass policy for udpSocket */

  if (setsockopt(udpSocket, IPPROTO_IP, IP_IPSEC_POLICY,
		 (char *)outbypassbuf, outbypasslen) < 0) {
    syslog(LOG_ERR, "enableKameIpsec(): udpSocket sockopt IP_IPSEC_POLICY failed");
    mip_log(MLOG_SEC, "enableKameIpsec(): udpSocket sockopt IP_IPSEC_POLICY failed");
    if (debug) {
      printf("enableKameIpsec(): udpSocket sockopt IP_IPSEC_POLICY failed");
    }
    exit(1);
  }

  if (setsockopt(udpSocket, IPPROTO_IP, IP_IPSEC_POLICY,
		 (char *)inbypassbuf, inbypasslen) < 0) {
    syslog(LOG_ERR, "enableKameIpsec(): udpSocket sockopt IP_IPSEC_POLICY failed");
    mip_log(MLOG_SEC, "enableKameIpsec(): udpSocket sockopt IP_IPSEC_POLICY failed");
    if (debug) {
      printf("enableKameIpsec(): udpSocket sockopt IP_IPSEC_POLICY failed");
    }
    exit(1);
  }

  /* set in and out bypass policy for icmpSocket */
  
  if (setsockopt(icmpSocket, IPPROTO_IP, IP_IPSEC_POLICY,
		 (char *)outbypassbuf, outbypasslen) < 0) {
    syslog(LOG_ERR, "enableKameIpsec(): icmpSocket udpSocket sockopt IP_IPSEC_POLICY failed");
    mip_log(MLOG_SEC, "enableKameIpsec(): icmpSocket sockopt IP_IPSEC_POLICY failed");
    if (debug) {
      printf("enableKameIpsec(): icmpSocket sockopt IP_IPSEC_POLICY failed");
    }
    exit(1);
  }

  if (setsockopt(icmpSocket, IPPROTO_IP, IP_IPSEC_POLICY,
		 (char *)inbypassbuf, inbypasslen) < 0) {
    syslog(LOG_ERR, "enableKameIpsec(): icmpSocket udpSocket sockopt IP_IPSEC_POLICY failed");
    mip_log(MLOG_SEC, "enableKameIpsec(): icmpSocket sockopt IP_IPSEC_POLICY failed");
    if (debug) {
      printf("enableKameIpsec(): icmpSocket sockopt IP_IPSEC_POLICY failed");
    }
    exit(1);
  }
}
#endif /* KAMEIPSEC */
