
/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * mnstat.c
 *
 * query the mnd for stats and for the current mn state
 *
 * syntax:
 *	mnstat [-blasASM] [-r (on|off)]
 *		-a <ipaddr> | (auto|0) :  hint to the mnd that you wish
 *			to use ipaddr as your agent.  an ipaddr of zero
 *			or the string "auto" will turn off the current
 *			hint in the daemon.
 *		-s: print stats (non beacon)
 *		-b: beacon stats
 *		-l: long form of stats/State/beacon/mn or agent lists
 *		-S: print state 
 *		-A[l]: print agent list
 *		-M[l]: print mn list
#ifdef DHCP
 *               -m (agent|ipip|nat|notunnel): set mode to AGENT_MODE,IPIP,
 *                               NAT, NOTUNNEL
 *                      automatic switching between AGENT_MODE and 3 DHCP modes:
 *			IPIP, NAT, NOTUNNEL  
#endif
 *
 * These switches [aAsS] are mutually exclusive; i.e., you
 * can only give one per command iteration.
 *
 * examples:
 *	mnstat -ls
 *	mnstat -lS
 *	mnstat -S
 *
 * note: all switches are mutually exclusive.
 * default (no switches): print usage
 *
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/time.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <netinet/in.h>
#ifdef MADRP
#include "sr_mad.h"
#endif
#include "unixsock.h"
#include "mnstats.h" 
#include "mnstate.h" 
#include "icmpstats.h"
#include "alist.h"

void parseArgs(int argc, char **argv);
void dumpMnsAgent( struct alist *a );

struct mn_state mns;
struct icmp_stats icmp_stats;
int debug = 0;

int stats = 0;
int state = 0;
int agent = 0;
int beacon = 0;
int alist = 0;
int mnlist =0;
#ifdef MADRP
int hlist = 0;
#endif
int longflag = 0;
struct sockaddr_in agentsock;
#ifdef DHCP
int modeSwitch = 0;
int mode = 0;           
#endif

char *progname;

#define HINTOFF	"auto"

main(argc, argv)
int argc;
char **argv;
{
	if (getuid()) {
		printf("mnstat: root only\n");
		exit(1);
	}
	progname = argv[0];
	
	parseArgs(argc, argv);
	doit();
	return(0);
}

void
parseArgs(argc, argv)
	int             argc;
	char          **argv;
{
	extern int      errno;
	extern char    *optarg;
	extern int      optind;
	int             ch, fd;
	int 		rc;
	int 		sawFlag = 0;
	char agentip[1024];
#ifdef MADRP
#ifdef DHCP
	char	*optstring = "bAMHlsSa:m:";
#else
	char	*optstring = "bAMHlsSa:";
#endif
#else
#ifdef DHCP
        char    *optstring = "bAMlsSa:m:";
#else
	char	*optstring = "bAMlsSa:";
#endif
#endif

#ifdef DHCP
        optstring = "bAMlsSa:r:m:";
#else
	optstring = "bAMlsSa:r:";
#endif

	while ((ch = getopt(argc, argv, optstring)) != EOF) {
		switch (ch) {
		case 'l':
			longflag = 1;
			break;
		case 's':
			sawFlag = stats = 1;
			break;
		case 'S':
			sawFlag = state = 1;
			break;
		case 'b':
			sawFlag = beacon = 1;
			break;
		case 'A':
			sawFlag = alist = 1;
			break;
		case 'M':
			sawFlag = mnlist = 1;
			break;
#ifdef MADRP
		case 'H':
			sawFlag = hlist = 1;
			break;
#endif
		case 'a':
			strncpy(agentip, optarg, 1024);
			sawFlag = agent = 1;
			break;
#ifdef DHCP
                case 'm':
                        sawFlag = modeSwitch = 1;
                        if (strcmp("auto", optarg)==0)
                                mode = 0;
                        else if (strcmp("agent", optarg)==0)
                                mode = 1;
                        else if (strcmp("ipip", optarg)==0)  
                                mode = 2;
                        else if (strcmp("nat", optarg)==0)  
                                mode = 3;
                        else if (strcmp("notunnel", optarg)==0){  
                                mode = 4;
			}
                        else
                                usage();
                        break;
#endif
		case '?':
		default:
			usage();
		}
	}
	/* longflag isn't enough 
	*/
	if (sawFlag == 0) {
		usage();
	}
	if (agent) {
		if ( strcmp( agentip, HINTOFF ) == 0 )
		    sprintf(agentip, "0");
		if (resolveAddress(agentip, &agentsock) == 0) {
			printf("agent address could not be resolved\n");
			exit(1);
		}
	}
}

usage()
{
	fprintf(stderr,"mnstat: [-ls][-lS][-lb] [-A[l]] [-M[l]] [ [-a <ipaddr>|auto] ");
#ifdef MADRP 
	fprintf(stderr," [-H]");
#endif
#ifdef DHCP                     
        fprintf(stderr,"\n\t[-m (auto|agent|dhcp)]");
#endif
	fprintf(stderr,"\n");
	fprintf(stderr,"-s: print out mn stats\n");
	fprintf(stderr,"-S: print out current mn state\n");
	fprintf(stderr,"-b: print out beacon stats\n");
	fprintf(stderr,"-l: longer -s -S -A -M (more info)\n");
	fprintf(stderr,"-A[l]: print agent list\n");
	fprintf(stderr,"-M[l]: print mn list\n");
#ifdef MADRP
	fprintf(stderr,"-H: print host route list\n");
#endif
	fprintf(stderr,"-a <agent ipaddr>: suggest agent to mn\n");
	fprintf(stderr,"\t ipaddr can be a DNS name, but ip addr is better\n");
	fprintf(stderr,"\t using \"%s\" or <ipaddr>==0 turns off hint mechanism\n", HINTOFF);
#ifdef DHCP
	fprintf(stderr, "-m (auto | agent | ipip | nat | notunnel): set mode to AUTOMATIC, or 1 of 4 MANUAL MODES: MIP:AGENT_MODE, DHCP:IPIP, DHCP:NAT, DHCP:NOTUNNEL \n");
#endif
	exit(1);
}

/*
 * doit toit
 */
doit()
{
	int rc;
	char buf[1024];
	int i;
	struct hintmsg hmsg;
	int acount;
	struct alist	mnsAgent;

	/* sent in order as follows */
	struct mn_stats mn_stats;

	/* open the unix socket to talk to the server
	*/
	rc = initUnixSocket(MNDSOCK_CLIENT, TRUE);
	if ( rc < 0) {
		perror("mnstat: socket setup failed");
		goto err;
	}

	/* send stats message to get stats structure
	*/
	if (stats) {
		buf[0] = U_MNSTAT;
		rc = sendUnixMessage(MNDSOCK_SERVER, buf, 1);
		if ( rc < 0 ) {
			perror("mnstat: socket send message failed");
			goto err;
		}
		rc = recvUnixMessage((char *)&mn_stats, sizeof(struct mn_stats));
		if ( rc < 0 ) {
			perror("mnstat: socket recv message failed");
			goto err;
		}
		dump_mnstats(&mn_stats);
	}
	/* send state message, get structure back */
	else if (state) {
		buf[0] = U_MNSTATE;
		rc = sendUnixMessage(MNDSOCK_SERVER, buf, 1);
		if ( rc < 0 ) {
			perror("mnstat: socket send message failed");
			goto err;
		}
		rc = recvUnixMessage((char *)&mns, sizeof(struct mn_state));
		if ( rc < 0 ) {
			perror("mnstat: socket recv message failed");
			goto err;
		}
		dumpMnState(longflag);
		/* TBD: redundant now, but perhaps not later...?
		rc = recvUnixMessage(&mnsAgent);
		if ( rc < 0 ) {
			perror("mnstat: socket recv message failed");
			goto err;
		}
		if ( longflag )
		    dumpMnsAgent(&mnsAgent);
		*/
	}
	else if (beacon) {
		buf[0] = U_MNBEACONS;
		rc = sendUnixMessage(MNDSOCK_SERVER, buf, 1);
		if ( rc < 0 ) {
			perror("mnstat: socket send message failed");
			goto err;
		}
		rc = recvUnixMessage((char *)&icmp_stats, sizeof(struct icmp_stats));
		if ( rc < 0 ) {
			perror("mnstat: socket recv message failed");
			goto err;
		}
		dumpBeacons(&icmp_stats, longflag);
	}
	/* send hint/agent message, nothing back
	*/
	else if (agent) {
		hmsg.code = U_MNHINT;
		memcpy((char *)hmsg.ip, (char *) &agentsock.sin_addr.s_addr, 4);
		rc = sendUnixMessage(MNDSOCK_SERVER, (char *)&hmsg, sizeof (struct hintmsg));
		if ( rc < 0 ) {
			perror("mnstat: socket send message failed");
			goto err;
		}
	}
	else if (alist) {
		buf[0] = U_ALIST;
		rc = sendUnixMessage(MNDSOCK_SERVER, buf, 1);
		if ( rc < 0 ) {
			perror("mnstat: socket send message failed");
			goto err;
		}
		acount = 0;
 		rc = recvUnixMessage((char *)&acount, sizeof(int));
		if ( rc < 0 ) {
			perror("mnstat: socket recv message for count failed");
			goto err;
		}
		for ( i = 0; i < acount; i++) {
			struct alist a;
 			rc = recvUnixMessage((char *)&a, sizeof(struct alist));
                        if (rc < 0) {
                                fprintf(stderr,"recv of alist member failed\n");
                                exit(1);
                        }
			if (i==0) alist_header(longflag);
                        alist_dumpone(&a, longflag);
		}
#ifdef MADRP
		/* sr_hagentlist */
		acount = 0;
		if (recvUnixMessage((char *)&acount, sizeof(int))<0){
			perror("mnstat: socket recv message for count failed");
			goto err;
		}
		for (i=0; i<acount; i++){
			struct alist h;
			if(recvUnixMessage((char*)&h,sizeof(struct alist))<0)
			{
				fprintf(stderr, "recv of alist member failed");
				exit(1);
			}
			if (i==0) alist_header(longflag);
			alist_dumpone(&h,longflag);
		}
#endif
	}
	else if (mnlist){
		buf[0] = U_MNLIST;
		if (sendUnixMessage(MNDSOCK_SERVER, buf, 1)<0){
			perror("mnstat: socket send message failed");
			goto err;
		}
		acount = 0;
		if (recvUnixMessage((char *)&acount, sizeof(int))<0){
			perror("mnstat: socket recv message for count failed");
			goto err;
		}
		for (i=0; i<acount; i++){
			struct alist a;
			if (recvUnixMessage((char*)&a, sizeof(struct alist))<0){
				fprintf(stderr, "recv of mnlist member failed");
				exit(1);
			}
			if (i==0) alist_header(longflag);
			alist_dumpone(&a, longflag);
		}
	}
#ifdef MADRP
	else if (hlist){
		buf[0] = U_HLIST;
		if (sendUnixMessage(MNDSOCK_SERVER, buf, 1)<0){
			perror("mnstat: socket send message failed");
			goto err;
		}
		acount = 0;
		if (recvUnixMessage((char *)&acount, sizeof(int))<0){
			perror("mnstat: socket recv message for count failed");
			goto err;
		}
		for (i=0; i<acount; i++){
			struct SR_hlist h;
			if(recvUnixMessage((char*)&h,sizeof(struct SR_hlist))<0){
				fprintf(stderr, "recv of hlist member failed");
				exit(1);
			}
			if (i==0) srhlist_header();
			srhlist_dumpone(&h);
		}
	}
#endif
#ifdef DHCP
        else if (modeSwitch) {
                buf[0] = U_MODESWITCH;  
                buf[1] = (char) mode;
                rc = sendUnixMessage(MNDSOCK_SERVER, buf, 2);
                if ( rc < 0 ) {
                        perror("mnstat: socket send message failed");
                        goto err;
                }
                rc = recvUnixMessage((char *)&modeSwitch, sizeof(int));
                if ( rc < 0 ) {
                        perror("mnstat: socket recv message failed");
                        goto err;
                }
                if (!modeSwitch) {
                        fprintf(stderr, "mode setting request rejected: "
                                        "mnd not DHCP capable\n");
                }
        }
#endif /* DHCP */
	/* zap our socket file
	*/
err:
	unlink(MNDSOCK_CLIENT);
}

dump_mnstats(struct mn_stats *s)
{
	printf("mn statistics\n");

	printf("UDP registrations sent %d\n", s->reg);
	printf("UDP replies received %d\n", s->replies);
	printf("UDP error: ident didn't match %d\n", s->identnomatch);
	printf("UDP error: invalid ha authentication %d\n", s->invalidhaauth);
	printf("UDP error: invalid fa authentication %d\n", s->invalidfaauth);
	printf("UDP error: missing ha authentication %d\n", s->missinghaauth);
	printf("UDP error: missing fa authentication %d\n", s->missingfaauth);
	printf("UDP error: ha made time update %d\n", s->timeupdate);
	
#ifdef MADRP
	dumpsr_metric(s->sr_metric);
#endif

	if (!longflag)
		return;

	printf("heartbeat ticks %d\n", s->ticks);
	printf("UDP error: mn got registration? %d\n", s->invalidregistration);
	printf("UDP error: reply too short %d\n", s->repltooshort);
	printf("UDP error: unexpected type %d\n", s->udpbadtype);

} 

#ifdef NOTUSED
void
dumpMnsAgent( struct alist *a )
/*
 *  --	print a little data about the agent pointed to by mns.agent
 *	it should only be defined when mns.state != NOWHERE and it should
 *	jive with mns.
 *
 *  TBD: NOT USED.  perhaps worth keeping?
 */
{
    char	*s;

    printf("current agent is:");
    if ( ! a->sigtime ) {
	printf("\t<none>\n");
	return;
    }
    switch(a->identity) {
    case OURHA:
	s = "OURHA";
	break;
    case OURFA:
	s = "OURFA";
	break;
    case THEIRFA:
	s = "THEIRFA";
	break;
    case MOBILENODE:
	s = "MOBILENODE";
	break;
    default:
	s = "<unknown>";
	break;
    }
    /*
    printf("\tidentity:\t%s\n", s);
    printf("\tbip:\t\t%s\n", inet_ntoa(a->bip));
    */
    printf("\t%s (%s)\n", s, inet_ntoa(a->bip) );


}  /* end dumpMnsAgent() */
#endif

/* stub functions
*/

stopAlarm()
{
}

startAlarm()
{
}

getConfiglogIPOnly()
{
	return(1);
}

