
/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * mnstate.c
 *
 * state machine for mobile node
 *
 * 3 kinds of events make up inputs to the state machine.
 *
 *	1. icmp router advertisements
 *	2. 1 second clock ticks that drive 3 timers
 *	3. udp reply messages from the HA or FA (on error)
 *
 */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/signal.h>
#include <termios.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/ip_var.h>
#include <netdb.h>
#include <unistd.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <md5.h>

#include "mnconfig.h"
#include "mip_icmp.h"	/* icmp router advert + appropriate mse extension/s */
#include "mnstate.h"
#include "mnstats.h"
#include "icmpstats.h"
#include "udp_reg.h"
#include "faha_pinfo.h"
#include "mip_log.h"
#include "route.h"
#include "alist.h"
#include "unixsock.h"

static int	noticeFunction = 2;
extern int	debug;
extern int	errorexit;

/* local mode boolean: are we running 80211 adaptivity or not
 * 0 means off (it is only on during NOWHERE state)
*/
static  int     link_running = 0;	

#define IN              1
#define OUT             2
#ifdef MADRP
extern struct alist * sr_hagentlist;
extern int srhagentcount;
#endif

/* prototypes
*/
void
mnStateMachine(int udpSocket, struct alist *a 
#ifdef DHCP
, struct in_addr co_located_coa);
#else 
);
#endif

char *
getStateString(int state);

void
mnStateTick(int udpSocket);

void 
reArp();

void 
deArp();

void
sendRegistration(int udpSocket, u_short lifetime, struct in_addr coa,
	struct in_addr mn, struct in_addr ha, struct in_addr dst);

void
mnDoRoute(int prevstate, int newstate, struct in_addr coa, struct in_addr fa);
#ifdef MADRP
void
srDoRoute(int prevstate, int newstate, struct in_addr coa, struct in_addr fa);
#endif

static struct alist * registerTilAck(int udpSocket);

static void initRetryTimer();

static void resendRegistration(int);

static void doIdentity();

void enterNowhereState(int prevstate, int nolog );

struct alist *findHint( void );

static int securityCheck(struct sockaddr_in *from, struct packet_info *pinfo);

static
int enter_signal_strength(struct alist *a, char *ifname, struct in_addr bip, int mnflag);

static void mnDeleteList();

void delFireWallRule();

/* agent list defines/and list head
*/
static struct alist *agenthead = NULL;
static struct alist *mnhead = NULL;
static int agentCount = 0;

int mnCount = 0;

/* 
 * mnStateMachine - set connectivity state according to input and the
 *		    current form of the alist.
 *
 * Basic idea: make state transititions, which really
 * means routing table changes.
 *
 * if current agent is ha, make sure it is our ha, if not, ignore.
 *
 * check user hint flag and move to state NOWHERE if beacon
 *	is from desired agent
 *
 * if NOWHERE state  
 *	if current agent is from HOME, move to HOME state
 *	if current agent is from AWAY, move to AWAY state
 *      note: if DHCP in force, as opposed to AWAY, we may go to COLOCATED
 *	if adaptive link change in force, turn that off as we are not NOWHERE.
 * if AWAY state
 *	store sequence number and check for MIP rollover
 * if HOME state
 *	keep lifetime set, we don't really care about the
 *	sequence number because we are home
 *
 *  EXIT ASSUMPTION:	state is never NOWHERE.
 *
 */

void
mnStateMachine(int udpSocket, struct alist *a 
#ifdef DHCP
, struct in_addr co_located_coa) 
#else 
)
#endif
{
	int prevstate = mns.state;	/* really current state */
	struct in_addr zero;

	int sawha; 			/* if icmp, true if ha, else fa */
	struct in_addr router;		/* router to use */
	unsigned short seqno;		/* sequence number if icmp */
	unsigned short lifetime; 	/* lifetime if icmp */
	struct in_addr coa;		/* coa if icmp */
	int sawAuth;			/* auth was included */
	int auth_ok;			/* auth was valid or no */
	char *l; 			/* pointer to beacon location */
	int lsize;			/* size of location string */

#ifdef MADRP
	int sr_agent;
#endif
	zero.s_addr = 0; 

top:
	if (debug) {
		printf("mnStateMachine: checking new agent entry\n");
		if ( a == NULL ) {
		    printf("mnStateMachine(): agent pointer is NULL! \n");
		}
	}

#ifdef MADRP
	/* if multi ad-hoc (experimental)
	 * determine if source routing agent is known
	*/
    	if (a) {
		if (alist_search(agenthead, a->bip)) 	{
			sr_agent=0;
		}
		else 	{
			sr_agent=1;
			prevstate=mns.sr_state;
		}
	}
#endif

	/*
	 *  initialize
	 */
    	if (a) {
#ifdef DHCP
		/* make sure dhcp is off as we have agent
		 */
		if (prevstate == NOWHERE && mns.mode >= IPIP) {
			mns.mode = AGENT_MODE;
		}
		/* if previous state was dhcp state, disable dhcp.
		 * this else branch should never be executed beacuse we always
		 *  switch to NOWHERE state before entering ATHOME or AWAY state 
		 */
		else if (prevstate == CO_LOCATED && mns.mode >= IPIP) {
                        mns.mode = AGENT_MODE;
                        dhcp_tunnel_policy(NOSTATE);
	 	}
#endif  
		/* normal init for non dhcp MIP case
		*/
		sawha = (a->identity == OURHA);
		router = a->bip;
		seqno = a->seqno;
		lifetime = a->lifetime;
		coa = a->coa;
		sawAuth = a->sawAuth;
		auth_ok = a->auth_ok;
		/*
		auth_timestamp = a->auth_timestamp;
		*/
		l = a->location;
		lsize = a->lsize;
    	}
    	else {  
#ifdef DHCP
		/* the only parameter with DHCP mode is the COA
		 */
		sawha = FALSE;
        	coa = co_located_coa;
#endif
    	}

	switch(prevstate) {
	case NOWHERE:
		/* NOWHERE -> ATHOME */
		if (sawha) {
#ifdef MADRP
			if(sr_agent) {
				mns.state = NOWHERE;
				mns.sr_state = ATHOME;
			}
			else
			{
				mns.sr_state = NOWHERE;
				mns.state = ATHOME;
			}
#else
			mns.state=ATHOME;
#endif
			mns.agent = a;
			mns.seqno = seqno;
			mns.orig_lifetime = lifetime;
 			mip_log( MLOG_INFO, "mobile node state: MN is at HOME (HA)");
#ifdef MADRP
			if(sr_agent)
				srDoRoute(prevstate, mns.sr_state, zero, zero);
			else
#endif
			/* set routing up
			 */
			mnDoRoute(prevstate, mns.state, zero, zero);

			/* set identity for freshness */
			doIdentity();

			/* send msg to ha to deregister */
			sendRegistration(udpSocket, 0, zero, 
				mn_config.mn, mn_config.ha, mn_config.ha);

			/* load retry timer so that we 
			 * send reg until we get ACK
			 */
			initRetryTimer();
			if (debug) {
			    printf("NOWHERE->ATHOME\n");
			    dumpMnState(1);
			}
			break;
		}

		/* NOWHERE -> AWAY */
		else if (a) {
#ifdef MADRP
			if(sr_agent) {
				mns.state = NOWHERE;
				mns.sr_state = AWAY;
			}
			else
			{
				mns.sr_state = NOWHERE;
				mns.state = AWAY;
			}
#else
			mns.state = AWAY;
#endif
			mns.agent = a;
			mns.seqno = seqno;
			mns.orig_lifetime = lifetime;
 			mip_log( MLOG_INFO, "mobile node state: MN is AWAY (at FA, %s)", inet_ntoa(router));

#ifdef MADRP
			if(sr_agent)
				srDoRoute(prevstate, mns.sr_state, coa, router);
			else
#endif
			/* set routing up
			 */
			mnDoRoute(prevstate, mns.state, coa, router);

			/* set identity for freshness */
			doIdentity();

			/* send registration */
			sendRegistration(udpSocket, mns.orig_lifetime, coa,
				mn_config.mn, mn_config.ha, router);

			/* set registration retry timer going */
			initRetryTimer();
			if (debug) {
			    printf("NOWHERE->AWAY\n");
			    dumpMnState(1);
			}
			break;
		}
#ifdef DHCP
                /* NOWHERE -> CO_LOCATED */
                else {
			mns.state = CO_LOCATED;
                        mns.coa = coa;
                        mns.orig_lifetime = HALIFETIME;

                        mip_log(MLOG_INFO, "mobile node state: MN is "
                                "CO_LOCATED, coa %s", inet_ntoa(coa));

                        /* set identity for freshness */
                        doIdentity();

                        /* send registration */
                        sendRegistration(udpSocket, mns.orig_lifetime, coa,
				mn_config.mn, mn_config.ha, mn_config.ha);                                

                        /* set registration retry timer going */
                        initRetryTimer();
                        break;
                }
#endif
		break;

	case AWAY:
		/* if beacon is from our current fa, store 
		 * sequence number and check it for rollover 
		 */
		if (mns.fa.s_addr == router.s_addr) {
			/* we want to check for sequence number
			 * rollover.  if this happened we need to
			 * reregister.  The sequence range 0..255
			 * is only used by an agent at boot.
			 */
			if ((seqno < mns.seqno) && (seqno <= 255)) {
				mip_log(MLOG_INFO, "mnd detected agent reboot");

				enterNowhereState(prevstate, FALSE);
				prevstate = NOWHERE;
				goto top;
			}
			mns.seqno = seqno;
		}
				
		break;

	case ATHOME:	
		mns.seqno = seqno;
		break;

#ifdef DHCP     
        case CO_LOCATED:
                mip_log (MLOG_INFO, "we should not reach here because we "
                        "always change to NOWHERE state before contacting "
                        "a new agent\n");
                break;
#endif
	}
}



/*
 * mnStateTick
 *
 * 1 tick per second clock tick.
 *
 * determine which agent to use when a hint is active:
 *	if the hint it not set, set bestAgent to the hint given
 *	(though of course it may not be the "best" one)
 *	
 *	if the hint flag is turned off, but the hint is set, this means
 *	that the hinted at agent should be discared, so go to NOWHERE
 *	state
 *
 * decrement the agent timer for the current agent.  this heuristic
 * tells mnd to look for a possibly better agent every agent_timer seconds.
 * the timer is current set to 1 for the best possible feedback.  (TBD--
 * it may not even be necessary...)
 *
 *	if there is no hint set, find the "best" agent (currently the one
 *	at the top of the ordered agent list).
 *	
 *	if mns.agent is already set, and the found agent is not waiting
 *	for deletion (TBD-- which may be impossible of mns.agent is not
 *	NULL), and if it is different from the mns.agent (the current agent)
 *	see whether it's signal strength is better by the fudge factor
 *	SIGFUDGE.  use the newly found agent if the difference is great
 *	enough.
 *
 * if the hint is off and an agent still has not been chosen, try finding the
 * "best" agent one more time.
 *
 * if the "best" agent is different from the current agent, transition
 * to NOWHERE state in preparation for calling mnStateMachine.  preserve
 * the hint setting flags in case the switch is due to a hint.
 *
 * call mnStateMachine with the current agent, or the "best" one (in case
 * we are NOWHERE).  if there is no agent chosen by this time then all
 * agents are bad for some reason (or none exist...).  reset the wait timers
 * on any existing agents so they can be tried again in the next pass
 * even if it is likely that they will fail.
 *
 * decrement agent_timer and go to NOWHERE state when it reaches zero
 *  	to trigger check for an agent with a stronger signal.
 *
 * if state is NOWHERE, we return.  no state change is possible until
 *	an appropriate ICMP beacons arrive.
 * 
 * if retry registration timer in effect
 *	count down and resend registration acc. to backoff - we haven't
 *	got an ACK
 * if registration timer goes off
 *	resend registration.  This is to keep us in the FA's
 *	visitor list.  at home, we do this to make sure our clock is up
 *	to date.
 * if lifetime timer
 *	if it goes off, we move to NOWHERE state.  we do not have 
 *	an agent.
 */

void
mnStateTick(int udpSocket)
{
	int		prevstate = mns.state;
#ifdef MADRP
	int		srprevstate= mns.sr_state;
#endif
	struct alist	*bestAgent = NULL;
#ifdef DHCP
        struct in_addr  zero;
                        
        zero.s_addr = 0;
#endif

#ifdef DHCP

	/* auto_mode means DHCP is allowed to occur automatically
         * AGENT_MODE means DHCP is not occuring now
	 */
	if (mns.auto_mode || mns.mode == AGENT_MODE) {
#endif
	/*
	 *  find the best agent according to mns.hintaddr if the hint
	 *  flag is on, but no attempt has been made to set mns.agent
	 *  appropriately.
	 *
	 *  if hint_set is on, but the hint flag is off, then turn off
	 *  the hint mechanism.
	 *
	 *  after this clause, hint_flag and hint_set will both be on
	 *  or both be off.
	 */
	if (mns.hint_flag && !mns.hint_set) {
		bestAgent = findHint();
	}
	else if (mns.hint_set && !mns.hint_flag) {
		enterNowhereState(mns.state, TRUE);
		if ( debug ) {
			printf("HINT turned off.\n");
		}
		mip_log( MLOG_INFO, "HINT turned off.\n");
	}
	else if (mns.hint_set && mns.hint_flag) {
		bestAgent = mns.agent;
	}
	/*
	 *  find the agent with the strongest signal if the agent_timer
	 *  decrements to zero.
	 */
	if ( --mns.agent_timer <= 0 ) {
		mns.agent_timer = AGENTCYCLE;

		if ( !mns.hint_set ) {
			bestAgent =
			    alist_findFirst(agenthead, MOBILENODE, FALSE);
			/*
			 *  reset bestAgent back to (non-NULL) mns.agent 
			 *  if bestAgent is not better by SIGFUDGE percent.
			 *  be sure current agent has not been deleted from
			 *  the alist; ignore if it has...
			 */
			if (mns.agent && !mns.agent->notinlist 
				&& bestAgent && bestAgent != mns.agent) {
				if (
#ifdef MADRP
				    (mns.agent->sawMac==bestAgent->sawMac) &&
#endif
				    (bestAgent->sigaverage - mns.agent->sigaverage 
					<= SIGFUDGE)) {
					bestAgent = mns.agent;
				}
			}
		}
	}


	/*
	 *  if bestAgent has not been determined from the HINT or
	 *  from a timeout of agent_timer and we are NOWHERE, search
	 *  the list for an agent.  (alist_findFirst may be invoked as many
	 *  as three times including the following, though this will only
	 *  happen if the alist is empty or the alist contains no match
	 *  to hintaddr or no agents.)
	 */
	if ( !mns.hint_set && !mns.agent && !bestAgent ) {
		bestAgent = alist_findFirst(agenthead, MOBILENODE, FALSE);
	}

#ifdef MADRP
	/* if still can't find a bestAgent, we will see if there is an agent
	 * several hops away, which is sorted in a list called sr_hagentlist
	 * according to their metrics
	 */
	if ( (!bestAgent) && sr_hagentlist) {
		/* mightnot need to findFirst */
		bestAgent=sr_hagentlist;
	}
#endif

	/*
	 *  if bestAgent is set and is not the current agent, transition to
	 *  NOWHERE in order to allow mnStateMachine() to setup
	 *  a relationship with the new agent.
	 *
	 *  if bestAgent==NULL then this goes immediately to NOWHERE state.
	 */
	if (bestAgent != mns.agent && (mns.state != NOWHERE
#ifdef MADRP
				       || mns.sr_state != NOWHERE
#endif
				       )) {
		int hinton = mns.hint_set;

		enterNowhereState(
			(mns.state == NOWHERE)? 
#ifdef MADRP
			mns.sr_state
#else
			mns.state
#endif
			:mns.state, 
				  bestAgent?TRUE:FALSE);
		if (bestAgent) {
			mns.hint_flag = mns.hint_set = hinton;
		}
	}

	/*
	 *  if we are not NOWHERE, check the current agent in case
	 *  its details have changed (eg, FA reboot).
	 *
	 *  if we are NOWHERE, and there is an agent has been chosen
	 *  1 by a hint, or 2 by an agent_timer timeout, or 3 because
	 *  mns.agent was unset, allow mnStateMachine() to transition
	 *  to the given agent.
	 *
	 *  else there are 2 possibilities: a) there is nothing to do
	 *  because the alist is empty, so just return; b) the alist has
	 *  entries but they are none valid for the moment.  Try clearing the
	 *  retry wait timers so some agent is caught on the next round and
	 *  tried (even if trying is futile).
	 */
#ifdef DHCP
        if ( mns.state != NOWHERE && mns.state != CO_LOCATED) {

                mnStateMachine(udpSocket, mns.agent, zero);
#else
	if ( mns.state != NOWHERE 
#ifdef MADRP
	     || mns.sr_state!= NOWHERE
#endif
	   ) {

		mnStateMachine(udpSocket, mns.agent );
#endif
	}

	else if( mns.state == NOWHERE && bestAgent ) {
#ifdef DHCP
                mnStateMachine(udpSocket, bestAgent, zero);
#else
		mnStateMachine(udpSocket, bestAgent );
#endif
	} else {
#ifdef MADRP		
		struct in_addr ifa;
		ifa.s_addr=getmyaddr();
		Send_SR_request(mn_config.ha,TRUE, ifa);
#endif
		alist_clrRetry(agenthead);
	}
#ifdef DHCP
	}
#endif

	/* registration retry timer, we have sent a registration
	 * message but have not yet gotten a response.
	 * we keep trying for some number of retries.
	 */  
	if (mns.retry_timer_flag) {
		struct alist *nextBest;

		/* note: this is a resend.  if we were using
		 * nonces, we do not advance
		 * the identity value.  We keep the one we have.
		 *
		 * if the return value is non-NULL, the returned agent
		 * is suggested as the next best and indicates that the
		 * current agent did not respond promptly enough to
		 * registration requests.  change state to this new agent.
		 */
		nextBest = registerTilAck(udpSocket);
		if (nextBest) {
#ifdef DHCP
                        mnStateMachine(udpSocket, nextBest, zero);
#else
			mnStateMachine(udpSocket, nextBest );
#endif
		}
	}

	/* timer to reregister when AWAY (at fa) 
	 * or at home
	 */
	if (mns.reg_timer_flag) {
		mns.reg_timer--;
		/* timeout on registration timer, resend
		 * current registration as long as we are NOT
		 * in the nowhere state
		 */
		if (mns.reg_timer <= 0) {
			/* cancel retry timer
			*/
			mns.retry_timer_flag = 0;
			resendRegistration(udpSocket);
			if (debug) {
				printf("reg timer expired\n");
				dumpMnState(1);
			}
		}
	}

	/* if we time out on HA/FA existance, we go to
	 * NOWHERE state and delete any current routes
	 * note: we could cache previous advert here
	 * and then use it
	 */
	if (mns.lifetime_flag && mns.lifetime != 0xffff) {
		mns.lifetime--;
		if (mns.lifetime <= 0) {
#ifdef MADRP
			if ( mns.state==NOWHERE && mns.sr_state!=NOWHERE) {
				enterNowhereState(srprevstate, FALSE);
			} else
#endif
				enterNowhereState(prevstate, FALSE);
			if (debug) {
				printf("lifetime timer expired\n");
				dumpMnState(1);
			}
		}
	}

	/* alistTick()
	 * decrement the agent/beacon list entries and timeout
	 * if = 0
	 */
	{
		struct alist	*a,
				*next = agenthead;
		while (a = next) {
			next = a->next;
			if (--a->ticks <= 0) {
				if (mns.agent == a) {
					enterNowhereState(mns.state, FALSE);
				}
				alist_delete(&agenthead, a->bip);
				agentCount--;
			}
			else if (a->retryWait > 0) {
			    a->retryWait -= 1;
			}
		}

		if (mnCount) {
			mnDeleteList();
		}
	}

        /* if dhcp is enabled in config, and in auto-mode or dhcp mode
	*/
	if (mn_config.dhcpEnable && (mns.auto_mode || mns.mode >= IPIP)){
		dhcpStateMachine(0);
	}
}  /* end mnStateTick() */


/*
 * delete mns in their per mn alist
 *
 * this only occurs in authenticated ad hoc mode.
 */
static
void
mnDeleteList()
{
	struct alist	*a,
			*next = mnhead;

	while (a = next) {
		next = a->next;
		if (--a->ticks <= 0) {
			if (a->mnroute) {
				route_delete_link(a->bip, a->mac);
			}
			alist_delete(&mnhead, a->bip);
			mnCount--;
		}
	}
}

/*
 * wipe the routes and MN list entries
 *
 * TBD: perhaps worth collapsing {mn,alist}_delete_all().
 */
void
mn_delete_all()
{
	while (mnhead) {
		if (mnhead->mnroute) {
			route_delete_link(mnhead->bip, mnhead->mac);
		}
		alist_delete(&mnhead, mnhead->bip);
		mnCount--;
	}
}

/*
 * wipe the routes and agent list entries
 */
void
alist_delete_all()
{
	while (agenthead) {
		if (agenthead->mnroute) {
			route_delete_link(agenthead->bip, agenthead->mac);
		}
		alist_delete(&agenthead, agenthead->bip);
		agentCount--;
	}
}

#ifdef MEASURE_ROUND_TRIP_TIME
extern struct timeval start_time, end_time;
#endif

/* mnStateUdp
 *
 * We got a udp reply message as input to state machine.  This is
 * basically one of the top event points for driving state changes,
 * in addition to timer events.
 *
 * set lifetime_timer to ICMP recv. time value (or lifetime here
 *	for that matter)
 * set reregister_timer to 1/3 of above
 * since we got reply we don't need timer to send registration anymore.
 *
 */
mnStateUdp(int udpSocket, struct sockaddr_in *from, struct packet_info *pinfo)
{
	/* sanity check
	*/
	if (!pinfo->rep) {
		return;
	}

	if (debug) {
		printf("mnStateUdp: pkt from %s\n",
			inet_ntoa(from->sin_addr));
	}


	/*
	 *  check that the UDP is arriving only from the current agent
	 *
	 *  (how might this happen anyways...?)
	 */
	if (mns.state == NOWHERE 
#ifdef MADRP
	    && mns.sr_state == NOWHERE
#endif
	    ) {
		if (debug) {
			printf("mnStateUdp(): Received UDP in NOWHERE state!\n");
		}
		mip_log(MLOG_ERR, "mnStateUdp(): Received UDP in NOWHERE state!\n");
		return;
	}

#ifdef NOTINUSE
	/* not clear if this is useful?
	*/
	if (mns.agent && from->sin_addr.s_addr != mns.agent->bip.s_addr) {
		char cursrc[MAXBUF];
		strcpy(cursrc, inet_ntoa(from->sin_addr));
		if (debug) {
			printf("mnStateUdp(): Received UDP from another agent, %s (cur=%s)!\n",
			cursrc, inet_ntoa(mns.agent->bip));
		}
		mip_log(MLOG_ERR, "mnStateUdp(): Received UDP from another agent, %s (cur=%s)!\n",
			cursrc, inet_ntoa(mns.agent->bip));
		return;
	}
#endif

	/* Like the others above, this has never happened yet, but there
	 * may be danger of a replay attack if you don't do this.  But then
	 * this provides a different sort of denial-of-service attack.
	 */
#ifdef WE_THINK_THIS_IS_ACTUALLY_A_WIN
	if (!mns.retry_timer_flag) {
		if (debug) {
			printf("mnStateUdp(): Received unexpected UDP from our FA.\n");
		}
		mip_log(MLOG_ERR, "mnStateUdp(): Received unexpected UDP from our FA.\n");
		return;
	}
#endif

	/*
	 * _first_, see if this is an error from the FA.  If so,
	 * there will be no authentication from the HA for us
	 * to check, which is what we want to do next.
	 */
	if (!REG_IS_FAIL_FA(pinfo->rep->code)
			&& securityCheck(from, pinfo) < 0) {
		return;
	}

	/* ignore if ident doesn't match our value
	*/
	if (mns.ident[1] != pinfo->rep->i.l[1]) {
		mn_stats.identnomatch++;
 		mip_log( MLOG_SEC, "mnStateUdp(): identity mismatch on UDP reply, got %x, expected %x", pinfo->rep->i.l[1], mns.ident[1]);
		return;
	}
	  
	/*
	 * At this point we know that the packet was authenticated ok (or 
	 * was a FA NAK) and is a response to our most recent request. 
	 * Now check for NAKs from HA or FA.  
	 */

	if (pinfo->rep->code > 1) {
		/* FA didn't like our lifetime
		*/
		if (pinfo->rep->code == REG_FAIL_FA_LIFETIME) {
			mns.orig_lifetime = mns.lifetime 
				= ntohs(pinfo->rep->lifetime);
 			mip_log( MLOG_INFO, 
			  "mnStateUdp(): updating lifetime according to NAK from FA, ip = %s, new lifetime=%d", inet_ntoa(from->sin_addr), mns.lifetime); 
			sendRegistration(udpSocket, mns.orig_lifetime, 
				mns.coa, mn_config.mn, 
				mn_config.ha, mns.fa);
			initRetryTimer();
		}
		/* HA zapped us with NAK on timestamp/nonce
		*/
		else if (pinfo->rep->code == REG_FAIL_HA_IDMISMATCH) {
			/* TBD. put time into this string
			*/
			/* note: this is of extreme interest security
			 * wise as some sort of spoof might possibly
			 * be based on an attack on network time.
			 * On the other hand, the HA will regulate the
			 * MN's time in our implementation so this will  
			 * happen.  Too many of them would be of interest
			 * however.
			 */

			/* update our system time */
			if (mn_config.tsmode) {
                                unsigned long ourtime;
                                gettimeofday(&mns.recvtime, NULL);
                                set_ntptime(ntohl(pinfo->rep->i.l[0]),
                                        &mns.sendtime, &mns.recvtime);
                                ourtime = get_ntptime_secs();
                                mip_log( MLOG_SEC, 
                                  "mnStateUdp(): received timestamp NAK from HA - our time was updated, HA time=<%s>, our time=<%s>", 
                                        dumptime(ntohl(pinfo->rep->i.l[0])),
                                        dumptime(ourtime));
                                mn_stats.timeupdate++;
			}

			/* else currently nonexistant NONCE method
			*/

			/* take system time and set current ident
			 * value in mns state
			 */
			if (mn_config.doTimestamp) {
				doIdentity();
			}
			/* resend registration */
#ifdef DHCP
                        if (mns.state == CO_LOCATED) {
                                sendRegistration(udpSocket, mns.orig_lifetime,
                                        mns.coa, mn_config.mn, mn_config.ha,
                                        mn_config.ha);
                        }
                        else {
#endif
			sendRegistration(udpSocket, mns.orig_lifetime, 
				mns.coa, mn_config.mn, 
				mn_config.ha, mns.fa);
#ifdef DHCP
                        }
#endif

			initRetryTimer();
			/* TBD. need loop detection here
			 * should fail if HA always fails us
			 */
		}
		else {
			/* for now just note NAKS if we should get
			 * any.
			 */
 			mip_log( MLOG_ERR, 
				"mnStateUdp(): recv nak code = %d from %s",
					pinfo->rep->code, 
					inet_ntoa(from->sin_addr));
		}
		return;
	}

	if (mns.retryTimeouts) { 
		struct alist *a = alist_search(agenthead, from->sin_addr); 
		if (a) { 
			a->retryWait = 0; 
		}
#ifdef MADRP 
		else { 
			a= alist_search(sr_hagentlist, from->sin_addr); 
			if (a) { 
				a->retryWait = 0; 
			} 
		}
#endif 
	} 
	mns.retry_timer_flag = 0; 
	mns.retryTimeouts = 0;
	  
	/* else ACK */
	/*
	 * set registration timer.  if ATHOME, we register on a fixed basis
	 * in order to keep the clock in sync.
	 */
	if (mns.state == AWAY
#ifdef MADRP
	    || mns.sr_state == AWAY
#endif
	   ) {
		mns.orig_lifetime = mns.lifetime = ntohs(pinfo->rep->lifetime);
	}
	else if (mns.state == ATHOME
#ifdef MADRP
	        || mns.sr_state== ATHOME
#endif
		) {
		mns.orig_lifetime = mns.lifetime = mns.halifetime;
	}
#ifdef DHCP
        else if (mns.state == CO_LOCATED) {

                mns.orig_lifetime = mns.lifetime = ntohs(pinfo->rep->lifetime);

		if (debug)
			printf("mns.mode %d \n",mns.mode);

		/* we have nat extension and HA said we are to do NAT
		*/
		if (pinfo->nat_ext_rep && 
			pinfo->nat_ext_rep->replyCode == MN_HA_NAT_REPLY_CODE) {

			/* setup a divert tunnel
			 */
			if (mns.auto_mode == TRUE) {
				dhcp_tunnel_policy(NAT);
			}	
			
			if (debug) {
				printf("##### alist_dump ###################\n");
				alist_dump(agenthead);
				printf("##### alist_dump ###################\n");
			}
				
			/* TBD FLOGME why?  */
			alist_delete_all();

			if (debug) {
				printf("Home agent address %s \n",
					inet_ntoa(pinfo->rep->home_agent));
				printf("Home address %s \n",
					inet_ntoa(pinfo->rep->home_address));
				printf("Home address %s \n",
					inet_ntoa(mns.coa));
			}
		} 
	}
#endif                   
	mns.ident[0] = pinfo->rep->i.l[0];
	mns.lifetime_flag = 1;
	if (mns.lifetime != 0) {
		mns.reg_timer = mns.lifetime / 3;
		mns.reg_timer_flag = 1;
	}
#ifdef        MEASURE_ROUND_TRIP_TIME 
	gettimeofday(&end_time, NULL); 
	mip_log(MLOG_INFO, "Round trip time is %d ms", 
		(end_time.tv_sec - start_time.tv_sec)*1000 + 
		(end_time.tv_usec - start_time.tv_usec)/1000);
#endif
}  /* end mnStateUdp() */

/* according to newstate, setup routing
 *
 * this routine manages the following mn state variables:
 *	fa
 *	coa
 * and manages the default route in the routing table.
 * it does *not* manage mns.state
 *
 * the caller should set mns.state before or after calling this
 * routine.
 *
 * re mns state.  if we are not at a fa, mns.fa and mns.coa will be
 * zeroed out.  mn_config.ha is never touched.  The state variable
 * SHOULD describe the state of the default route; e.g.,
 *
 * if current state is
 *	NOWHERE: there will be no default route
 *	AWAY: default route will be fa
 *	HOME: default route will be ha
 *
 * state variables maintained here:
 *	mns.fa
 *	mns.coa
 *	mns.haroute - 1 if ha route is set, 0 if not
 *
 * ASSUMED that mns.agent is defined if we are going *somewhere.*
 */
void
mnDoRoute(int prevstate, int newstate, struct in_addr coa, struct in_addr fa)
{
	char *mac = NULL;
	int sawMac = 0;
	if (mns.agent && mns.agent->sawMac) {
		sawMac = 1;
		mac = mns.agent->mac;
	} 

	if (debug) {
		char coas[1024];
		char fas[1024];
		char pst[1024];
		char nst[1024];
		sprintf(coas, "%s", inet_ntoa(coa));
		sprintf(fas, "%s", inet_ntoa(fa));
		sprintf(pst, "%s", getStateString(prevstate));
		sprintf(nst, "%s", getStateString(newstate));
		printf("mnDoRoute prevstate %s, new state %s, coa %s, fa %s\n",
			pst, nst, coas, fas);
	}

	switch(prevstate) {
	/* NOWHERE to
	 *	NOWHERE
	 *	ATHOME
	 *	AWAY
	 */
	case NOWHERE:
		switch(newstate) {
		case NOWHERE:
			return;
			break;
		case ATHOME:
			/* if NOWHERE to ATHOME
			 * basically make sure we put HA in as default
			 * route.  mac value is set for AAH, not for ARP.
			*/
			if (mns.haroute)
				return;
			/* 
			 * if ipsec/arp mode && if no clone route, add it
			 */
			if (mn_config.kameipsecEnable && mn_config.doArp) {
				if (mns.cloneroute == 0) {
					reArp();
				}	
			}
			mn_route_add_default(mn_config.ha, mn_config.mn,
				1, mac /* mac */, mn_config.ifname
				);
			mns.haroute = 1;	/* we have a HA route set */
			break;
		case AWAY:
			/* if we were at ha, delete that route 
		         * note, we may need to do magic for arp table
			 * entries here
			*/
			if (mns.haroute) {
				mn_route_delete_default(mn_config.ha,
					sawMac);
				mns.haroute = 0;
			}
			/* 
			 * if ipsec/arp mode && if clone route, 
			 * zap arp route entries so that tunnel
			 * to HA can work (HA is "local" ip).
			 */
			if (mn_config.kameipsecEnable && mn_config.doArp) {
				/* note: deArp sets cloneroute flag to 0
				*/
				if (mns.cloneroute == 1) {
					deArp();
				}	
			}
			mn_route_add_default(fa, mn_config.mn,
				1, mac /* mac */, mn_config.ifname
				);
			mns.fa = fa;
			mns.coa = coa;
			break;
		}
		break;
	/* AWAY to
	 *	NOWHERE
	 *	ATHOME
	 *	AWAY
	 */
	case AWAY:
		switch(newstate) {
		case NOWHERE:
			/* note: we do not restore ha route
			*/
			mn_route_delete_default(mns.fa, sawMac);
			mns.fa.s_addr = 0;
			mns.coa.s_addr = 0;
			break;
		case ATHOME:
			/* if ipsec/arp, put clone route back
			*/
			if (mn_config.kameipsecEnable && mn_config.doArp) {
				if (mns.cloneroute == 0) {
					reArp();
				}	
			}
			mn_route_delete_default(mns.fa, sawMac);
			mns.fa.s_addr = 0;
			mns.coa.s_addr = 0;
			mn_route_add_default(mn_config.ha, mn_config.mn,
				1, mac /* mac */, mn_config.ifname
				);
			mns.haroute = 1;
			break;
		case AWAY:
			/* if same fa, nothing to do
			 */
			if (coa.s_addr == mns.coa.s_addr)
				return;
			mn_route_delete_default(mns.fa, 
				sawMac);
			mn_route_add_default(fa, mn_config.mn,
				1, mac /* mac */, mn_config.ifname
				);
			mns.fa = fa;
			mns.coa = coa;
			break;
		}
		break;
	/* ATHOME to
	 *	NOWHERE
	 *	ATHOME
	 *	AWAY
	 */
	case ATHOME:
		switch(newstate) {
		case NOWHERE:
			if (mn_config.kameipsecEnable && mn_config.doArp) {
				if (mns.cloneroute == 1) {
					deArp();
				}	
			}
			/* remove default route
		        */
			mn_route_delete_default(mn_config.ha, 
				sawMac);
			/* ITBD.  remove clone route/arp table
			*/
			mns.haroute = 0;
			break;
		case ATHOME:
			return;
			break;
		case AWAY:
			/* if haroute present, delete it, 
			 * if ipsec/arp mode, zap clone route/arp table
			*/
			if (mns.haroute) {
				mn_route_delete_default(mn_config.ha,
					sawMac);
				mns.haroute = 0;
			}
			/* if kame enabled, and arp mode and cloneroute
			 * present, then zap clone route and arp table
			 */
			if (mn_config.kameipsecEnable && mn_config.doArp) {
				if (mns.cloneroute == 1) {
					deArp();
				}	
			}
			mn_route_add_default(fa, mn_config.mn,
				1, mac /* mac */, mn_config.ifname
				);
			mns.fa = fa;
			mns.coa = coa;
			break;
		}
		break;
	}
}
/*
 * registerTilAck - if registration timer indicates need,
 * we resend registration message through backoff scheme.
 *
 * if the current agent is unresponsive after a partial traversal of
 * the retryBackoff[] array, a new agent is chosen and returned to the
 * calling function.  in all other cases, NULL is returned.
 */
static struct alist *
registerTilAck(int udpSocket)
{
	struct in_addr zero;
	int prevstate = mns.state;
	extern int debug;
#ifdef MADRP
	int sr_prevstate = mns.sr_state;
#endif
	struct alist	*nextBest;

	if (debug) {
		printf("registerTilAck(), mns.retry_timer %d, mns state\n", mns.retry_timer,
			mns.state);
	}
	zero.s_addr = 0;
	mns.retry_timer--;
	/* if retry timer expires on reg. try again
	*/
	if (mns.retry_timer <= 0) {
		/* short circuit dhcp state.  we have no agents.
		 * reset retry index if needed, and go to
		 * resend part. AVOID code following this as it is
		 * for agent case.
		 */
                if (mns.state == CO_LOCATED) {
			if (mns.retryIndex >= SHORT_RETRY) {
				mns.retryIndex = 0;
				mns.retry_timer = retryBackoff[mns.retryIndex];
			}
			goto dhcp_case;
		}
		/*
		 *  if the current agent is not responding, mark it
		 *  as "retried," by setting retryWait to a period of
		 *  seconds to wait before again attempting an initial
		 *  Registration Request.  try the next agent in the list.
		 *  wait until all agents in the list (which could be
		 *  dynamically changing, btw) have been tried before
		 *  officially noting the situation as hopeless.
		 *
		 *  because the list is dynamic (dependent upon the
		 *  regular receipt of beacons), retried elements may
		 *  change position in the list according to their current
		 *  signal strength, and/or (untried) elements may appear in
		 *  the list at any time, and/or (tried and untried) elements
		 *  may dissappear from the list at any time.
		 *
		 *  if a hint has been set, skip this clause and perform
		 *  a full retry cycle before discarding the hint and
		 *  returning to dynamic agent searches.
		 *
		 *  SHORT_RETRY must be <= NORETRIES.
		 */
		if (!mns.hint_set && mns.retryIndex >= SHORT_RETRY) {
#ifdef NOTSURE
			/* added mns.agent test by Z.Chen so that in DHCP mode, this
			 *  block of code is guranteed not to be executed 
			 */
			if (mns.agent) {
				mns.agent->retryWait = RETRYWAIT;
			}
#endif
			mns.retryIndex = 0;
			mns.retry_timer = retryBackoff[mns.retryIndex];

			/* find the next best agent */
			nextBest =
				alist_findFirst(agenthead, MOBILENODE, FALSE);
#ifdef MADRP
			if(!nextBest)
				nextBest= sr_hagentlist;
#endif

			/* clear the list if all agents have been tried
			 * and return a NULL, else return the nextBest agent
			 */
			if (!nextBest) {
				if (debug) {
					printf("SHORT_RETRY failure: no response from any agent in agentlist.\n");
				}
				mip_log( MLOG_INFO, "SHORT_RETRY failure: no response from any agent in agentlist.\n");
				mns.retryTimeouts++;
#ifdef MADRP
				if(mns.state==NOWHERE && mns.sr_state!=NOWHERE)
					enterNowhereState(mns.sr_state, FALSE);
				else
#endif
					enterNowhereState(mns.state, FALSE);

				alist_clrRetry(agenthead);

				alist_delete_all();
				return(NULL);
			}

			else {
				if (debug) {
					printf("SHORT_RETRY timeout: trying next best agent (%s).\n",
						inet_ntoa(nextBest->bip));
				}
				mip_log( MLOG_INFO, "SHORT_RETRY timeout: trying next best agent (%s).\n",
					inet_ntoa(nextBest->bip));
#ifdef MADRP
				if(mns.state==NOWHERE && mns.sr_state!=NOWHERE)
					enterNowhereState(mns.sr_state, TRUE);
				else
#endif
					enterNowhereState(mns.state, TRUE);
				return(nextBest);
			}
		}

		/*  if we have sent too many tries, mn moves to 
		 *  NOWHERE state and we deinstall ha route
		 */
		else if (mns.retryIndex >= (NORETRIES-1)) {
			mns.retryTimeouts++;
			if ( mns.hint_set && mns.hint_flag ) {
				mip_log( MLOG_INFO,
				    "Refused HINT for %s: retry timeout.\n",
				    inet_ntoa(mns.hintaddr));
			}
#ifdef MADRP
			if(mns.state==NOWHERE && mns.sr_state!=NOWHERE)
				enterNowhereState(sr_prevstate, FALSE);
			else
#endif
				enterNowhereState(prevstate, FALSE);
			if (debug) {
				printf("ack timer expired, NOWHERE state\n");
				dumpMnState(1);
			}
			return(NULL);
		}

dhcp_case:
		/* if at home, send dereg again
		*/
		if (mns.state == ATHOME
#ifdef MADRP
		    || mns.sr_state == ATHOME
#endif
		   ) {
			if (mn_config.doTimestamp) {
				doIdentity();
			}
			/* ha is direct dest */
			sendRegistration(udpSocket, 0, zero, mn_config.mn, 
				mn_config.ha, mn_config.ha);
		}
		else if (mns.state == AWAY
#ifdef MADRP
		    || mns.sr_state == AWAY
#endif
		        ) {
			if (mn_config.doTimestamp) {
				doIdentity();
			}
			/* fa is direct dest, ha is indirect */
			sendRegistration(udpSocket, mns.orig_lifetime, mns.coa, mn_config.mn, 
				mn_config.ha, mns.fa);
		}
#ifdef DHCP     
                else if (mns.state == CO_LOCATED) {

                        if (mn_config.doTimestamp) {
                                doIdentity();
                        }
                        /* send registration directly to HA in IPIP
			*/
			if (debug) {
				printf("mns.state CO_LOCATED, mns.mode %d, send registration\n",
					mns.mode);
			}
                        sendRegistration(udpSocket, mns.orig_lifetime, mns.coa,
                                mn_config.mn, mn_config.ha, mn_config.ha);
                }
#endif                  

		/* update retry timer */
		mns.retryIndex++;
		mns.retry_timer = retryBackoff[mns.retryIndex];
		if (debug) {
			printf("registerTilAck - NO ACK: mns.retryIndex retry_timer %d %d\n",
				mns.retryIndex, mns.retry_timer);
			dumpMnState(1);
		}
	}
	return(NULL);
}

/*
 * registerTilAck - if registration timer indicates need,
 * we resend registration message through backoff scheme.
 *
 * if the current agent is unresponsive after a partial traversal of
 * the retryBackoff[] array, a new agent is chosen and returned to the
 * calling function.  in all other cases, NULL is returned.
 */
static struct alist *
registerTilAck_a(int udpSocket)
{
	struct in_addr zero;
	int prevstate = mns.state;
#ifdef MADRP
	int sr_prevstate = mns.sr_state;
#endif
	struct alist	*nextBest;

	zero.s_addr = 0;
	mns.retry_timer--;
	/* if retry timer expires on reg. try again
	*/
	if (mns.retry_timer <= 0) {
		/*
		 *  if the current agent is not responding, mark it
		 *  as "retried," by setting retryWait to a period of
		 *  seconds to wait before again attempting an initial
		 *  Registration Request.  try the next agent in the list.
		 *  wait until all agents in the list (which could be
		 *  dynamically changing, btw) have been tried before
		 *  officially noting the situation as hopeless.
		 *
		 *  because the list is dynamic (dependent upon the
		 *  regular receipt of beacons), retried elements may
		 *  change position in the list according to their current
		 *  signal strength, and/or (untried) elements may appear in
		 *  the list at any time, and/or (tried and untried) elements
		 *  may dissappear from the list at any time.
		 *
		 *  if a hint has been set, skip this clause and perform
		 *  a full retry cycle before discarding the hint and
		 *  returning to dynamic agent searches.
		 *
		 *  SHORT_RETRY must be <= NORETRIES.
		 */
                /* added mns.agent test by Z.Chen so that in DHCP mode, this
                   block of code is guranteed not to be executed */
		if (!mns.hint_set && mns.retryIndex >= SHORT_RETRY) {
			if (mns.agent) {
				mns.agent->retryWait = RETRYWAIT;
			}
			mns.retryIndex = 0;
			mns.retry_timer = retryBackoff[mns.retryIndex];


			/* find the next best agent */
			nextBest =
				alist_findFirst(agenthead, MOBILENODE, FALSE);
#ifdef MADRP
			if(!nextBest)
				nextBest= sr_hagentlist;
#endif

			/* clear the list if all agents have been tried
			 * and return a NULL, else return the nextBest agent
			 */
			if (!nextBest) {
				if (debug) {
					printf("SHORT_RETRY failure: no response from any agent in agentlist.\n");
				}
				mip_log( MLOG_INFO, "SHORT_RETRY failure: no response from any agent in agentlist.\n");
				mns.retryTimeouts++;
#ifdef MADRP
				if(mns.state==NOWHERE && mns.sr_state!=NOWHERE)
					enterNowhereState(mns.sr_state, FALSE);
				else
#endif
					enterNowhereState(mns.state, FALSE);

				alist_clrRetry(agenthead);

				alist_delete_all();
				return(NULL);
			}

			else {
				if (debug) {
					printf("SHORT_RETRY timeout: trying next best agent (%s).\n",
						inet_ntoa(nextBest->bip));
				}
				mip_log( MLOG_INFO, "SHORT_RETRY timeout: trying next best agent (%s).\n",
					inet_ntoa(nextBest->bip));
#ifdef MADRP
				if(mns.state==NOWHERE && mns.sr_state!=NOWHERE)
					enterNowhereState(mns.sr_state, TRUE);
				else
#endif
					enterNowhereState(mns.state, TRUE);
				return(nextBest);
			}
		}

		/*  if we have sent too many tries, mn moves to 
		 *  NOWHERE state and we deinstall ha route
		 */
		else if (mns.retryIndex >= (NORETRIES-1)) {
			mns.retryTimeouts++;
			if ( mns.hint_set && mns.hint_flag ) {
				mip_log( MLOG_INFO,
				    "Refused HINT for %s: retry timeout.\n",
				    inet_ntoa(mns.hintaddr));
			}
#ifdef MADRP
			if(mns.state==NOWHERE && mns.sr_state!=NOWHERE)
				enterNowhereState(sr_prevstate, FALSE);
			else
#endif
				enterNowhereState(prevstate, FALSE);
			if (debug) {
				printf("ack timer expired, NOWHERE state\n");
				dumpMnState(1);
			}
			return(NULL);
		}


		/* if at home, send dereg again
		*/
		if (mns.state == ATHOME
#ifdef MADRP
		    || mns.sr_state == ATHOME
#endif
		   ) {
			if (mn_config.doTimestamp) {
				doIdentity();
			}
			/* ha is direct dest */
			sendRegistration(udpSocket, 0, zero, mn_config.mn, 
				mn_config.ha, mn_config.ha);
		}
		else if (mns.state == AWAY
#ifdef MADRP
		    || mns.sr_state == AWAY
#endif
		        ) {
			if (mn_config.doTimestamp) {
				doIdentity();
			}
			/* fa is direct dest, ha is indirect */
			sendRegistration(udpSocket, mns.orig_lifetime, mns.coa, mn_config.mn, 
				mn_config.ha, mns.fa);
		}
#ifdef DHCP     
                else if (mns.state == CO_LOCATED) {

                        if (mn_config.doTimestamp) {
                                doIdentity();
                        }
                        /* send registration directly to HA in  IPIP */
                        sendRegistration(udpSocket, mns.orig_lifetime, mns.coa,
                                mn_config.mn, mn_config.ha, mn_config.ha);
                }
#endif                  

		/* update retry timer */
		mns.retryIndex++;
		mns.retry_timer = retryBackoff[mns.retryIndex];
		if (debug) {
			printf("registerTilAck - resend reg no ack yet\n");
			dumpMnState(1);
		}
	}

	return(NULL);
}

/* init retry timer.  we sent a registration and
 * will keep doing it for a bit until HA pings us
 * back.
 */
static void
initRetryTimer()
{
	mns.retry_timer_flag = 1;  /* retry until we get ack */
	mns.retryIndex = 0;
	mns.retry_timer = retryBackoff[mns.retryIndex];
}

/* 
 * resend registration when registration timer times out
 */
static void
resendRegistration(int udpSocket)
{
	struct in_addr zero;
	mns.reg_timer_flag = 0;
	zero.s_addr = 0;
	if (mns.state == NOWHERE
#ifdef MADRP
	    && mns.sr_state == NOWHERE
#endif
	    )
		return;

	/* set identity for freshness */
	doIdentity();
	if (mns.state == ATHOME
#ifdef MADRP
	    || mns.sr_state == ATHOME
#endif
	   ) {
		sendRegistration(udpSocket, 0, zero, 
			mn_config.mn, mn_config.ha, mn_config.ha);
	}
	else if (mns.state == AWAY
#ifdef MADRP
	         || mns.sr_state == AWAY
#endif
		) {
		/* always use the lifetime the FA most recently advertised */
		sendRegistration(udpSocket, mns.agent->lifetime, mns.coa,
			mn_config.mn, mn_config.ha, mns.fa);
	} 
#ifdef DHCP
        else {  /* mns.state == CO_LOCATED */

                sendRegistration(udpSocket, mns.orig_lifetime, mns.coa,
                        mn_config.mn, mn_config.ha, mn_config.ha);
        }
#endif
	initRetryTimer();
}

/*
 * force state machine to nowhere state.  nolog == TRUE when the log message
 * should be supressed.
 *
 * NOTE that secure route is not reset.  this only happens in unstable
 * situations (eg, knowing all agents are flaky) or in mnStateUdp() when an
 * agent does not sport a route security tuple.
 */
void
enterNowhereState(int prevstate, int nolog)
{
	struct in_addr zero;
	int sr=0;

	zero.s_addr = 0;
	mns.hint_flag = 0;
	mns.hint_set = 0;
	mns.reg_timer_flag = 0;  
	mns.retry_timer_flag = 0;
	mns.lifetime_flag = 0;

#ifdef MADRP
	if(mns.state==NOWHERE && mns.sr_state!=NOWHERE)		
		sr=1;
	mns.sr_state=NOWHERE;
#endif
	mns.state = NOWHERE;
	mns.retryIndex = 0;

#ifdef MADHUCHANGE
	/* free the pointer to the alist entry if it's not in the list
	*/
	if (mns.agent && mns.agent->notinlist ) {
		free((void *) mns.agent);
	}
	mns.agent = NULL;
	mns.agent_timer = 0;
#endif

	/* wipe the default route
	*/
#ifdef DHCP
        /* if mode >= IPIP, do not call mnDoRoute() beacause
           a call to dhcpEnterInitState() will take care of
           removing the default route */
    	if (mns.mode >= IPIP)
		dhcpEnterInitState();
    	else  {
#endif                  
#ifdef MADRP
	if(sr)
		srDoRoute(prevstate, NOWHERE, zero, zero);
	else
#endif
		mnDoRoute(prevstate, NOWHERE, zero, zero);
	}

#ifndef MADHUCHANGE
	/* free the pointer to the alist entry if it's not in the list
	*/
	if (mns.agent && mns.agent->notinlist ) {
		free((void *) mns.agent);
	}
	mns.agent = NULL;
	mns.agent_timer = 0;
#endif
	if (prevstate != NOWHERE && ! nolog) {
		mip_log( MLOG_INFO, "mobile node state: MN is NOWHERE.");
	}
}

/*
 *  search alist for an agent which has a beaconIP or a COA which matches
 *  mns.hintaddr.  return a pointer to that agent or NULL upon failure
 *  and set mns.hint_set or reset mns.hint_flag, respectively.  NOTE:
 *  will consider agents even if retryWait field not zero.
 *
 *  TBD: cache hints until appropriate agent arrives?  sticky agents?
 */
struct alist *
findHint( void )
{
	struct alist	*a;

entryMark( 1, "findHint" );


	/*
	 *  Find the first agent in the alist which such that
	 *  its beacon IP or its Care of Address matches 
	 *  mns.hintaddr
	 */
	alist_initwalk(agenthead);
	while ( a = alist_nextwalk() ) {
		if ( a->identity != MOBILENODE
		    && (a->bip.s_addr == mns.hintaddr.s_addr
			|| a->coa.s_addr == mns.hintaddr.s_addr) )
		break;
	}

	if ( !a ) {
		mns.hint_flag = 0;
		if (debug) {
			printf("Refused HINT for %s: no such agent.\n",
			    inet_ntoa(mns.hintaddr));
		}
		mip_log( MLOG_INFO, "Refused HINT for %s: no such agent.\n",
		    inet_ntoa(mns.hintaddr));
	}
	else {
	    mns.hint_set = 1;
	    if (debug) {
		    printf("Set HINT for %s\n", inet_ntoa(mns.hintaddr));
	    }
	    mip_log( MLOG_INFO, "Set HINT for %s.\n", inet_ntoa(mns.hintaddr));
	}


	return( a );

}  /* end findHint() */



/* according to spi/registration type, set identity
* to current freshness
*/
static void
doIdentity()
{
	if (mn_config.doTimestamp == 0) {
		mns.ident[1]++;
		return;
	}

	/* put NTP time into mn_state structure.
	 * This is done in NETWORK byte order.
	 * It will be copied to the packet by sendRegistration()
	 */
	store_ntptime((char *)&mns.ident[0]);
}


/* securityCheck - check md5 hash on incoming packet according
 *	to keys stored in config file
 *
 * for mn, 
 *	check mn ha security
 *	if desired
 *		check fa mn security
 *
 * returns:
 *	0: ok
 *	-1: failed authentication
 */
static int
securityCheck(struct sockaddr_in *from, struct packet_info *pinfo)
{
	int size;
	unsigned char hash[16];
	int spi;
	char *mn_key;
	int mn_transform;
	u_long mn_spi;
	int mn_keysize;
	unsigned long dummy;
	int dummy2;

	if (mn_config.mnha_security) {
		/* compute the checksum over the packet with
		 * our key.  get size and then call appropriate
		 * function.
		 */
		size = sizeof(struct udp_reply) + pinfo->premnha_size 
			+ AUTHEXT_HASHEDBYTES;
		
 		mn_key = get_key(
			mn_config.sa[MNHASECURITY].kl,
			(unsigned long) pinfo->rep->home_agent.s_addr,
			&mn_keysize,
			&mn_spi,
			&mn_transform,
			&dummy, &dummy2);

		/* probably a sw foobar, but given that we decided
		 * to get the ha address out of the packet ...
		 * it isn't a fatal error.
		 */
		if (mn_key == NULL) {
			if (debug) {
				printf("MISSING HA KEY!!!\n");
			}
			mip_log( MLOG_SEC, "missing MIP UDP MN/HA key for HA %s", inet_ntoa(pinfo->rep->home_agent));
			mn_stats.missinghaauth++;
			return(-1);
		}
		
		/* TBD. for now spi is ignored as there is only one 
		 * MAC algorithm.  We implicitly csum it however.
		 */
		spi = htonl(mn_config.sa[MNHASECURITY].spi);
		
		/* recompute MD5 hash, result in local hash[]
		*/
		computeMd5(pinfo->packet, size, 
			mn_key, hash);

		/* compare packet hash and computed hash
		 * if == 0, then ok
		 */
		if (bcmp(pinfo->mn_ha_auth->hash, hash, 16) != 0) {
			mip_log( MLOG_SEC, 
			  "security error: invalid MIP UDP MN/HA authentication from %s", inet_ntoa(from->sin_addr));

			if (debug) {
				printf("mn ha authentication failed\n");
			}
			mn_stats.invalidhaauth++;
			/* authentication failed
			*/
			return(-1);
		}
	}

	/* we are done if we are at home
	*/
	if (mns.state == ATHOME
#ifdef MADRP
	    || mns.sr_state == ATHOME
#endif
	    )
		return(0);

	if (mn_config.famn_security) {
		if (pinfo->fa_x_auth_count == 0 || !pinfo->fa_x_auth) {
			mip_log(MLOG_SEC, "securityCheck(): missing expected FA-MN hash from %s.",
				inet_ntoa(from->sin_addr));
			mn_stats.missingfaauth++;
			return(-1);
		}

		size = sizeof(struct udp_reply) + 
			pinfo->premnha_size  +
			pinfo->mn_ha_auth_size +
			pinfo->postmnha_size +
			+ AUTHEXT_HASHEDBYTES;
		if (debug) {
			printf("famn security size %d\n", size);
			printf("%d %d %d +%d\n", pinfo->premnha_size,
				pinfo->mn_ha_auth_size,
				pinfo->postmnha_size,
				AUTHEXT_HASHEDBYTES);
		}

		/* 
		 * Try the COA first, as it is more likely to be a "real" address
		 * than what is being used as the local interface address.
		 */
		if (mns.agent) {
			mn_key = get_key(
				mn_config.sa[FAMNSECURITY].kl,
				(unsigned long) mns.agent->coa.s_addr,
				&mn_keysize,
				&mn_spi,
				&mn_transform,
				&dummy, &dummy2);
		}

		if (mn_key == NULL) {
 			mn_key = get_key(
				mn_config.sa[FAMNSECURITY].kl,
				(unsigned long) from->sin_addr.s_addr,
				&mn_keysize,
				&mn_spi,
				&mn_transform,
				&dummy, &dummy2);
		}

		if (mn_key == NULL) {
			if (debug) {
				printf("MISSING FA KEY!!!\n");
			}
			mip_log(MLOG_SEC, "missing MIP UDP MN/FA key for FA %s", inet_ntoa(from->sin_addr));
			mn_stats.missingfaauth++;
			return(-1);
		}

		/* TBD. do something with spi
		spi = htonl(mn_config.sa[FAMNSECURITY].spi);
		*/

		computeMd5(pinfo->packet, size, 
			mn_key, hash);

		/* compare packet hash and computed hash
		 * if == 0, then ok
		 */
		if (bcmp(pinfo->fa_x_auth->hash, hash, 16) != 0) {
			mip_log( MLOG_SEC, 
			  "security error: invalid MN/FA authentication from %s", inet_ntoa(from->sin_addr));
			if (debug) {
				printf("mn fa authentication failed, size %d\n", size);
				printf("fa hash\n");
				dumpme(pinfo->fa_x_auth->hash, 16);	
				printf("computed hash\n");
				dumpme(hash, 16);
			}
			mn_stats.invalidfaauth++;
			/* authentication failed
			*/
			return(-1);
		}
	}
	return(0);
}

/*
 * stick beacons from agents in agent list, mns in mnlist
 * possibilities
 *	agents
 *		arp is on, therefore no authentication.  X
 *		authentication, if it fails auth, we still put it in alist.
 *	mns
 *		arp is on, we don't care about beacons.  X
 *		authentication, we install link layer route acc. to POLICY X
 *			if auth fails, we toss the beacon.  X
 */
void
checkAlist(	int sawha, 		/* ha or fa */
		struct in_addr bip, 	/* ip address associated with beacon */
		int seqno, 		/* seq no for beacon */
		int lifetime, 		/* lifetime for beacon */
		struct in_addr coa, 	/* coa for fa */
		int sawAuth, 		/* auth bit field, NET or ADHOC */
		int auth_ok,		/* auth was valid, (if none, then 0) */ 
		u_long auth_tssecs,	/* auth timestamp */
		u_long auth_tsfract,	/* auth timestamp */
		char *location,		/* location string */
		int lsize,		/* size of location string */
		int sawMac,		/* mac address was included */
		char *macaddr,		/* pointer to mac address */
		int sawmn,		/* true if beacon was from mn */
		int doArp,		/* true if doing arp on this interface */
		struct in_addr ifaddr    /* if ip address */
		)
/* checkAlist */
{
	register struct alist *a;
	int	rc;
	int new = 0;

entryMark( 1, "checkAlist" );

	/* no point if mn beacon and we speak arp or auth failed
	*/
	if (sawmn) {
		if (doArp)
			return;
		/* POLICY -- right now only says if auth fails,
		 * we ignore, needs to discriminate acc. to ui 
		 */
		if (auth_ok == FALSE) {
			return;
		}
	}

	/*
	 *  search list.  allocate a new entry if the beacon (new) is
	 *  not found.
	 * 
	 * NOTE: agents include mns in list. agentCount is incremented for them.
	 */
	if (sawmn)
		a = alist_search(mnhead, bip);
	else
		a = alist_search(agenthead, bip);

	if ( !a ) {
		if (sawmn) {
			a = alist_create(&mnhead, bip);
			mnCount++;
		}
		else {
		    	a = alist_create(&agenthead, bip);
	    		agentCount++;	/* # of entries in list */
		}
	        if (a == NULL) {
		    syslog(LOG_ERR, "malloc - out of space in agent or mn list");
		    exit(1);
	        }
		new = 1;
		/* store ts info if beaconer is new and not arp mode
		*/
		if (!doArp) {
		    a->auth_tssecs = auth_tssecs;
		    a->auth_tsfract = auth_tsfract;
		}

	}

	/* make weak denial of service check vis-a-vis timestamp
	 * info if beacon is not arp beacon (authenticated) and
	 * is already in cache
	 *
	 * idea: we cache 1st beacon.  if ts value is not increasing
	 * either by sec or fract part, we discard the beacon.
	 * We also update a counter and log that this is happening
	 */
	else if (!doArp) {
                if (debug) {
                        printf("ADHOCTS: bip %s: alist %x,%x, pkt %x,%x\n",
                            inet_ntoa(bip),
                            a->auth_tssecs, 
                            a->auth_tsfract, 
                            auth_tssecs, 
                            auth_tsfract 
                        );
                }

		/* if secs increasing, fine
                 * if not, check fractional part
                 * logic below:
                 *   if new value less than or equal to stored value,
                 *      get suspicious
                 */
		if (auth_tssecs <= a->auth_tssecs){
			/* if fract increasing, fine
			*/
			if (auth_tsfract <= a->auth_tsfract) {
				/* else ts is stuck ... */
				icmp_stats.ts_auth_failure++;
		    		mip_log(MLOG_SEC, 
                                  "timestamp from node ip=%s not increasing, possible denial of service attack (time in alist (sec/frac):%x,%x pkt %x,%x)\n",
                                    inet_ntoa(bip),
                                    a->auth_tssecs, a->auth_tsfract,
                                    auth_tssecs, auth_tsfract 
                                );
                                if (debug) {
                                        printf("ADHOCTS: ts failed\n");
                                }
				return;
			}
		}
                /* make alist values fresh */
                if (a) {
                        a->auth_tssecs = auth_tssecs;
                        a->auth_tsfract = auth_tsfract;
                }
	}

	/* (re)store identity as appropriate
	 */
	if (sawha) {
		a->identity = OURHA;
	}
	else if (sawmn) {
		a->identity = MOBILENODE;
	}
	else {
		if (auth_ok)
			a->identity = OURFA;
		else
			a->identity = THEIRFA;
		a->coa = coa;
	}
	a->seqno = seqno;
	a->lifetime = lifetime;
	/* timeout, for mn load lifetime */
	a->ticks = lifetime;
	a->sawAuth = sawAuth;
	a->auth_ok = auth_ok;
	a->lsize = lsize;
	if (lsize) {
		bzero(a->location, MAXLOCATION);
		strncpy(a->location, location, lsize);
	}
	else {
		a->location[0] = 0;
	}
	a->sawMac = sawMac;
	if (sawMac) {
		bcopy(macaddr, a->mac, 6);
	}
	a->notinlist = FALSE;


	/* we saw an MN beacon, if we are not doing arp, and auth worked,
	 * we need to get a link layer entry
	 * in here UNLESS the node is an agent, in which case we leave that
	 * work up to the state machine, which is managing default routes.
	 */
	if (sawmn) {
		/* we know auth was ok */
		if (new) {
			/* install mn route */
			route_add_link(bip, macaddr,
				mn_config.ifname
				);
			a->mnroute = TRUE;
		}
		rc = enter_signal_strength(a, mn_config.ifname, bip, 1);
		return;
	}

	/* remainder is for agents
	*/
	/* compute signal strength and sort if the return result
	 * of enter_signal_strength() is 1.  unsorted items will remain
	 * at the end of the list implementing FCFS.
	 */
	rc = enter_signal_strength(a, mn_config.ifname, bip, 0);
	if (rc) {
		alist_sort(&agenthead, a);
	}


        if (debug) {
		printf("#################### alist_dump ###################\n");
		alist_dump(agenthead);
		printf("#################### alist_dump ###################\n");
        }
}

/*
 * write agent list out to
 * client who wants to print it out
 */
void
mnstate_dumplist(char *sock, int listype)
{
	struct alist *a;
	int rc;
	int s = sizeof(struct alist);
	struct alist *listhead;

	/* write the number of agents */
	if (listype == TYPE_ALIST){
	 	rc = sendUnixMessage(sock, (char *)&agentCount, sizeof(int));
		if (rc < 0)
			return;
		if (agentCount == 0)
			return;
		listhead = agenthead;
	}
	else if (listype == TYPE_MNLIST){
		if (sendUnixMessage(sock, (char*)&mnCount, sizeof(int))<0)
			return;
		if (mnCount==0)
			return;
		listhead = mnhead;
	}
#ifdef MADRP
        else if (listype == TYPE_HLIST){
		struct SR_hlist *h;
                int srhCount=srhlist_count();
                if (sendUnixMessage(sock, (char*)&srhCount, sizeof(int))<0)
                        return;
                if (srhCount==0)
                        return;
                /* walk the srhlist and write the structures out
                */
                srhlist_initwalk();
                while (h = (struct SR_hlist *) srhlist_nextwalk()) {
                        rc = sendUnixMessage(sock, (char *)h,
				sizeof(struct SR_hlist));
                        if (rc < 0)
                                return;
                }
		return;
        }
#endif

	/* walk the agent list and write the structures out
	*/
	alist_initwalk(listhead);
	while (a = alist_nextwalk()) {
		rc = sendUnixMessage(sock, (char *)a, s);
		if (rc < 0)
			return;
	}

#ifdef MADRP
        if (listype == TYPE_ALIST){
		/* sr_hagentlist */
                if (sendUnixMessage(sock, (char*)&srhagentcount, sizeof(int))<0)
                        return;
                if (srhagentcount==0)
                        return;
                /* walk the srhagentlist and write the structures out
                */
                alist_initwalk(sr_hagentlist);
                while (a = (struct alist *) alist_nextwalk()) {
                        rc = sendUnixMessage(sock, (char *)a,
				sizeof(struct alist));
                        if (rc < 0)
                                return;
                }
	}
#endif
	
}

/*
 * routine is terrible hack: XXX
 *
 * wireless devices wlp0/wi0 have variable "signal strength",
 * which allows handoff based on signal (in case of wi0 this is
 * actually quality).  
 *
 * for all other devices,
 * the signal strength gathering and associated alist_sort'ing devolves
 * to an alist that behaves as first come first serve.  if compiled with
 * NOSIGNAL, all devices are treated equal and the alist is never sorted
 * (reducing some possibly unnecessary overhead).
 *
 * return 1 if the list entry should be sorted, 0 otherwise.
 *
 *    TBD:
 *	. differentiate per device
 *	. refine/change formula and input data used to differentiate
 *	  and rate beacon signals...
 */
static int
enter_signal_strength(struct alist *a, char *ifname, struct in_addr bip, int mnflag)
{
	extern int	wl_ip_to_signal(struct in_addr bip);
	time_t		curtime;
	int		i;


entryMark( 1, "enter_signal_strength" );


#ifdef NOSIGNAL
	a->sigstrength = a->sigaverage = MAX_SIGSTR;
	return(0);
#else
	/*
	 *  set raw signal strength from the kernel cache if the device
	 *  supports it, otherwise set raw signal strength and the average
	 *  to the maximum possible percentage
	 */

	a->sigstrength = MAX_SIGSTR;
	if (strcmp(ifname, "wlp0") == 0) {
		a->sigstrength = wl_ip_to_signal(bip);
	}
	else if (strcmp(ifname, "wi0") == 0) {
		a->sigstrength = wi_ip_to_signal(bip);
	}

	if (mnflag)
		return;

	/*
	 *  too many seconds have passed since the last beacon was
	 *  registered.  delete the entry and allow it to reinstantiate
	 *  itself.
	 */
	curtime = time(NULL);
	if ( a->sigcount && ((curtime - a->sigtime) >= MAX_NOBEACON) ) {
		if (mns.agent == a) {
			enterNowhereState(mns.state, FALSE);
		}
		alist_delete( &agenthead, bip );
		agentCount -= 1;
		return(0);
	}
	a->sigtime = curtime;

	/*
	 *  keep current average by adding the new sig strength and
	 *  substracting the oldest.  routine begins with sigindex of
	 *  the most recent element (recorded at last execution).
	 *
	 *  return 1 if the entry should be sorted, 0 otherwise.
	 */
	a->sigindex = (a->sigindex + 1) % MIN_GOODAVE;
	a->sigsum += a->sigstrength - a->sigarray[a->sigindex];
	a->sigarray[a->sigindex] = a->sigstrength;
	a->sigaverage = a->sigsum / MIN_GOODAVE;
	if ( ++a->sigcount >= MIN_GOODAVE ) {
		a->sigcount = MIN_GOODAVE;
		return(1);
	}

	return(0);

#endif
}  /* end enter_signal_strength() */

int     
noDefaultroute()
{
    return (mns.state == NOWHERE) ;
}

#ifdef MADRP
/* out put mnhead */
struct alist * alist_getmnhead(){
      return mnhead;
}

/* almost the same as mnDoRoute(), except that we will not add default route
 * for host route agents
 */
void
srDoRoute(int prevstate, int newstate, struct in_addr coa, struct in_addr fa)
{
	char *mac = NULL;
	int sawMac = 0;
	if( mns.agent && mns.agent->sawMac) {
		sawMac = 1;
		mac = mns.agent->mac;
	}

	switch(prevstate) {
	case NOWHERE:
		switch(newstate) {
		case NOWHERE:
			return;
			break;
		case ATHOME:
			if (mns.haroute)
				return;
			mns.haroute = 1;
			break;
		case AWAY:
			if (mns.haroute) {
				if(sawMac)
					mn_route_delete_default(mn_config.ha,1);
				mns.haroute = 0;
			}
			mns.fa = fa;
			mns.coa = coa;
			break;
		}
		break;
	case AWAY:
   		switch(newstate) {
		case NOWHERE:
			/* note: we do not restore ha route
			*/
			if (sawMac)
				mn_route_delete_default(mns.fa, 1);
			mns.fa.s_addr = 0;
			mns.coa.s_addr = 0;
			break;
		case ATHOME:
			if (sawMac)
				mn_route_delete_default(mns.fa, 1);
			mns.fa.s_addr = 0;
			mns.coa.s_addr = 0;
			mns.haroute = 1;
			break;
		case AWAY:
			/* if same fa, nothing to do
			 */
			if (coa.s_addr == mns.coa.s_addr)
				return;
			if (sawMac)
				mn_route_delete_default(mns.fa, 1);
			mns.fa = fa;
			mns.coa = coa;
			break;
		}
		break;
	case ATHOME:
		switch(newstate) {
		case NOWHERE:
			/* remove default route
		        */
			if (sawMac)
				mn_route_delete_default(mn_config.ha, 1);
			mns.haroute = 0;
			break;
		case ATHOME:
			return;
			break;
		case AWAY:
			if (mns.haroute) {
				if (sawMac)
					mn_route_delete_default(mn_config.ha,1);
				mns.haroute = 0;
			}
			mns.fa = fa;
			mns.coa = coa;
			break;
		}
		break;
	}
}

#endif

char *
getStateString(int state)
{
	switch (state) {
	case NOWHERE:
		return("NOWHERE");
		break;
	case ATHOME:
		return("ATHOME");
		break;
	case CO_LOCATED: 
		return("CO_LOCATED");
		break;
	case AWAY: 
		return("AWAY");
		break;
	default:
		return("UNKNOWN");
		break;
	}
}


/*
 * put the clone route back in dynamically
 */
void
reArp()
{
	route_nofail();
	clone_route_add(mn_config.cloneip, mn_config.mn,
                        mn_config.netmask
                        );
        mns.cloneroute = 1;
	route_fail();

}

/*
 * wipe clone route and flush arp table
 */
void
deArp()
{
	/* first, get rid of clone route
	*/
	route_nofail();
	clone_route_delete(mn_config.cloneip);
	mns.cloneroute = 0;
	route_fail();
	/* 2nd, zap arp table entries
	*/
	flushArp();
}

