
/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * mnudpinput.c
 *
 * overview:
 *	functions related to input of udp message (reply)
 *	at mip agent (mipd).  read it, "parse" it (break it up) 
 *	and weed out errors, and prepare messages for input
 *	to mip fa/ha state machines.
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/times.h>
#include <sys/time.h>
#include <errno.h>
#include <signal.h>
#include <arpa/inet.h>
#include <net/if.h>
#ifdef NOPE
#include <netinet/ip_fw.h>
#endif

#include "mnconfig.h"
#include "mnstate.h"
#include "udp_reg.h"
#include "faha_pinfo.h"
#include "mnstats.h"
#include "mip_log.h"

#define IN   1
#define OUT  2

/* prototypes
*/
void    
stateMachine(int, int, int, u_long *, struct sockaddr_in *,
                struct packet_info *);


int recvMsg(int sd, char *buf, int bufsize, struct sockaddr *from, 
	int *fromsize, int *ifindex);


struct packet_info pinfo;

/* handleUdpInput
 * given a socket with a pending udp message,
 * we parse and sanitize the message.  by parse,
 * we mean recoginize & break up into sub-components including
 * any security associations.  We sanitize by
 * checking sizes (typically too short).
 * Note:
 *	kernel code will add 4 byte in_addr at end
 *	of this message to tell us ip address of incoming
 *	interface.  We need this so that FA can know
 *	what interface to bind a MN to.
 *
 * TBD?: during packet extention parsing: note whether a given ext is
 * appearing on the proper side of the MN-HA auth ext?  are security
 * implications of unordered extentions that might not be caught which
 * create bit flipping holes?  perhaps unrecognized extentions should cause
 * the packet to be dropped... 
 */
static int expect_minrep_size = sizeof(struct udp_reply) + 
		sizeof(struct mse_authext);

handleUdpInput(int s)   
{
	int size;
	int rc;
	struct sockaddr_in from;
	unsigned char packet[MAXUDPPACKET];
	struct udp_register *udp_reg = 0;
	struct udp_reply *udp_rep = 0;
	int ifindex;
	extern int debug;
	int err = 0;
	int offset = 0;
	int haFlag = 0;
	int before_mn_ha;
	struct mse_nat_ext_rep *nat;
	struct mse_authext *msep;
	char *p;
	extern int errno;
	char buf[1024];

	size = sizeof(struct sockaddr_in);

	rc = recvMsg(s, packet, MAXUDPPACKET, (struct sockaddr *)&from, &size,
		&ifindex);

	if (rc < 0) {
		if (errno == EINTR)
			return;
		syslog( LOG_ERR, "handleUdpInput(): recvmsg (%m)");
		exit(1);
	}
	if (debug) {
		printf("UDP REPLY packet: addr=%s, port=%d, rc=%d, if(%d)%x\n", 
			inet_ntoa(from.sin_addr),
			ntohs(from.sin_port), rc, ifindex,
			indextoif(ifindex));
		dumpme(packet, rc);
	}

	/* zero out the packet info structure
	*/
	memset(&pinfo, 0, sizeof(struct packet_info));

	pinfo.packet = packet;
	pinfo.psize = rc;

	/* figure out what kind of packet it is and check min. size
	 * sans mses
	 */
	udp_reg = (struct udp_register *) packet;

	/* as a mobile node, we shouldn't be getting registration
	 * packets
	 */
	if (udp_reg->type == UDP_REGISTER_TYPE) {
		strcpy(buf, inet_ntoa(udp_reg->home_address));
		mn_stats.invalidregistration++;
		mip_log( MLOG_ERR, "handleUdpInput(): MN received registration packet from %s(%s)!",
			reverseAddress(udp_reg->home_address), buf);
		if (debug) {
			printf("handleUdpInput(): MN received registration packet!\n");
			dumpUdpRequest((struct udp_register *) packet);
			printf("Whole packet...\n");
			dumpme(packet);
		}
		return;
	}
	else if (udp_reg->type == UDP_REPLY_TYPE) {
		pinfo.reg = udp_reg = NULL;
		pinfo.rep = udp_rep = (struct udp_reply *) packet;
		
		/*
		 * Errors from FA may be unauthenticated
		 * and therefore shorter.  If it is a FA failure, compare
		 * the packet size against the size of a MIP reply; otherwise,
		 * compare it against a precomputed expected size.
		 */
		if (pinfo.psize < ((REG_IS_FAIL_FA(udp_rep->code))
				? (sizeof(struct udp_reply)) : expect_minrep_size)) {
			mn_stats.repltooshort++;
			mip_log( MLOG_ERR, "handleUdpInput(): UDP reply packet from %s too short"
				 , inet_ntoa(from.sin_addr));
			if (debug)
				printf("udp reply packet -- too short\n");
			return;
		}
		offset += (sizeof(struct udp_reply));
	}
	/* bad type */
	else {
		mn_stats.udpbadtype++;
		mip_log( MLOG_ERR,
		    "handleUdpInput(): UDP (mobile-ip) input pkt from %s, bad type %d\n",
		    inet_ntoa(from.sin_addr),
		    udp_reg->type);
		if (debug)
			printf("udp pkt recv, bad type %d\n", udp_reg->type);
		return;
	}

	/* now parse the packet for extensions.  
	 * We leave authentication for the state machine code
	 */

	/* extension ordering rules for MN->HA registration message
	 *	.may have 0 or more non-auth messages to HA.  FA should 
	 *		forward.
	 *	.MUST have MN-HA auth. extension. FA should forward.
	 *  if FA, we may have following. FA should remove before
	 *  forwarding to HA.
	 *	.may have 0 or more MN-FA non-auth. messages.  If so,
	 *		the FA should strip them.
	 *	.may have MN-FA auth. extension.  FA should strip
	 *		before forwarding.
	 * FA may append after MN-HA the following fields for HA:
	 *	.may have 0 or more FA-HA non-auth. messages.  HA
	 *		should strip.
	 *	.may have FA-HA auth. extension.  HA should strip
	 *		but pre
	 */

	/* rules for HA->MN reply message
	 *	.may have 0 or more non-auth messages.  FA may or
	 *		may not care but should forward.
	 *	.MUST have MN-HA auth. extension.  FA should forward.
	 *	.may have 0 or more HA-FA non-auth. extensions.
	 *		FA should strip before forwarding to MN.
	 *	.may have HA-FA auth. extension.  FA should strip.
	 */

	/* NOTE: all currently unrecognized extentions are ignored,
	 * although their size must be recorded to aid proper authentication
	 * checks later.  premnha and postmnha size values are therefore
	 * inclusive of any known and unknown extentions prior to the
	 * authentication extentions.
	 */

	/* use mse_authext as a template to easily access type and length */
	msep = (struct mse_authext *)(packet + offset);
	before_mn_ha = 1;	/* indicates which 1/2 of packet we are in */

	while(offset < pinfo.psize) {
		switch (msep->type) {

		case MN_HA_AUTHENTICATOR_TYPE:
			before_mn_ha = 0;
			pinfo.mn_ha_auth = msep;
			/* +2 because we add in type/length */
			pinfo.mn_ha_auth_size = (msep->length + 2);
			pinfo.mn_ha_auth_count++;
			/* this err also catches the case of MN-HA auth
			 * in the second 1/2 of the packet, by definition
			 */
			if (pinfo.mn_ha_auth_count > 1) {
				mip_log( MLOG_SEC, "handleUdpInput(): packet from %s, too many authentication extensions between MN and HA",
					inet_ntoa(from.sin_addr));
				/* udp_errors.toomanyauthexts++; */
				return;
			}
			break;

		case MN_FA_AUTHENTICATOR_TYPE:
			if (before_mn_ha) {	/* found in wrong 1/2 */
				mip_log(MLOG_SEC, "handleUdpInput(): MN-FA auth appears before MN-HA auth (from=%s).", inet_ntoa(from.sin_addr));
				return;
			}
			pinfo.fa_x_auth = msep;
			pinfo.fa_x_auth_size = (msep->length + 2);
			pinfo.fa_x_auth_count++;
			if (pinfo.fa_x_auth_count > 1) {
				mip_log( MLOG_SEC, "handleUdpInput(): packet from %s, too many authentication extensions between MN and FA",
					inet_ntoa(from.sin_addr));
				/* udp_errors.toomanyauthexts++; */
				return;
			}
			break;

		case FA_HA_AUTHENTICATOR_TYPE: 
			mip_log( MLOG_SEC, "handleUdpInput(): packet from %s, UDP with FA/HA authenticator type!", inet_ntoa(from.sin_addr));
			return;
			break;

		default:
			if (before_mn_ha) {
				pinfo.premnha_size += (msep->length + 2);
			}
			else {
				pinfo.postmnha_size += (msep->length + 2);
			}
			mip_log(MLOG_ERR, "handleUdpInput(): skipping unrecognized packet extension %s MN-HA auth ext. (from=%s)",
				before_mn_ha?"before":"after",
				inet_ntoa(from.sin_addr));
		}
			
		offset += sizeof(struct mse_authext);
		nat    = (struct mse_nat_ext_rep *)(packet + offset);
		if ( nat->type == MN_HA_NAT_TYPE_REP ) 
                	pinfo.nat_ext_rep = nat;

		offset += (msep->length + 2);
		p = (char *) msep;
		p += (msep->length + 2);
		msep = (struct mse_authext *) p;
	}
	/*
	 * If the reply had no HA-MN authentication extension,
	 * and the message is not the FA griping, chuck it.
	 */
	if (pinfo.mn_ha_auth_count == 0
			&& !REG_IS_FAIL_FA(udp_rep->code)) {
		if (debug) {
			printf("tossing UDP reply msg: no mn/ha mse\n");
		}
		mip_log( MLOG_SEC, "handleUdpInput(): packet from %s, no MN/HA authentication extension - packet ignored", inet_ntoa(from.sin_addr));
		/* TBD: udp_errors.noauthent++; */
		return;
	}

	if (debug) {
		dumpUdpReply(pinfo.rep);
		if (pinfo.mn_ha_auth) {
			dumpMse(pinfo.mn_ha_auth);
		}
		if (pinfo.fa_x_auth) {
			dumpUdpReply(pinfo.fa_x_auth);
		}
		if (pinfo.nat_ext_rep){
			if (debug) {
				printf("Nat Ext Found \n");
				dumpNatRep(pinfo.nat_ext_rep);
			}
                }
		else{
			if (debug) {
				printf("Nat Ext Not Found\n");
			}                			
		}	
	}

	mnStateUdp(s, &from, &pinfo);
}

