
/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * mnudprequest.c
 * 
 * mn sends udp registration to HA (possibly via FA)
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/times.h>
#include <sys/time.h>
#include <errno.h>
#include <signal.h>

#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <netinet/ip_var.h>
#include <netinet/udp_var.h>
#include "mnconfig.h"

#include "mnconfig.h"
#include "mnstate.h"
#include "mnstats.h"
#include "udp_reg.h"
#include "mip_log.h"
#include "route.h"
#include "alist.h"

/* prototypes
*/
void storeMd5(char *, int, unsigned char *, unsigned char *);

extern struct mn_stats mn_stats;
extern int debug;

#ifdef MEASURE_ROUND_TRIP_TIME
struct timeval start_time;
struct timeval end_time;
#endif

/*
 * open raw socket for creation of HA/UDP output
 * messages.  
 *
 * returns:
 *	int s - udp socket descriptor
 */
static int
initUdpOut()
{
        int     sockd,
		one = 1,
		rc;
	struct sockaddr_in server;

	if((sockd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0){
		perror("raw socket");
	  	exit(1);
	}

	/* Set close-on-exec flag to prevent fd leaks. */
	if (fcntl(sockd, F_SETFD, 1) < 0) {
		syslog(LOG_INFO, "initUdp(): F_SETFD (%m)");
		/* not catastrophic, don't exit */
	}

	/* we want to be able to write the IP header
	*/
	if (setsockopt(sockd, IPPROTO_IP, IP_HDRINCL, (char *)&one
		       ,sizeof(one)) < 0){
	  perror("IP_HDRINCL");
	  exit(1);
	}
	return sockd;
}

/* write udp registration packet
 *
 * caller: must set ident field in mns mn_state appropriately 
 * before calling this function.
 *
 * parameters:
 *	int udpSocket - udp socket descriptor
 *	u_short lifetime - lifetime value from icmp message
 *	in_addr coa - care of address, mn_ip if deregistration
 *	in_addr mn - mn's ip address
 *	in_addr ha - home agent's ip address
 *	in_addr dst - who to send message to as ip destination (fa/ha)
 *
 * indirect parameters:
 *	ident fields are just loaded from mn_state.  Assume caller
 *	has set those appropriately (time or nonce)
 *
 *	key/spi security parameters come from mn_config structure.
 */

void
sendRegistration(int udpSocket, u_short lifetime, struct in_addr coa,
	struct in_addr mn, struct in_addr ha, struct in_addr dst)
{
	char packet[1024];
	int pubOrpriv;
	struct udp_register *reg;
	struct mse_authext *mse = NULL;
#ifdef DIVERT
	struct mse_nat_ext_req *nat = NULL;
#endif
	int psize = sizeof(struct udp_register);
	int spi;
	unsigned long dummy;
	int dummy2;
	struct sockaddr_in to;
	int rc;
	char *mn_key;
	int mn_transform;
	u_long mn_spi;
	int mn_keysize;
	int dereg = 0;

#ifdef	MEASURE_ROUND_TRIP_TIME
	gettimeofday(&start_time, NULL);
#endif
	/* if we are not doing arp, push out a beacon before
	 * the udp message is sent.  Pro: This will maximize odds
	 * that we have a MAC address, therefore we can cut back
	 * on the beacon rate for MNs.  Con: 2 packets per registration.
	 */
	if (mn_config.doArp == FALSE) {
		pushBeacon();
	}

	bzero(packet, 1024);
	reg = (struct udp_register *) packet;

	/* set up registration packet
	*/
	reg->type = UDP_REGISTER_TYPE;
	reg->flags = 0;
#ifdef DHCP
	/* if dhcp and co-located, indicate that in registration
	*/
	if ( mns.state == CO_LOCATED )
		reg->flags = D_BIT;
#endif

	/* if deregistration set lifetime to 0
	*/
	if (coa.s_addr == 0) {
		reg->lifetime = htons(0);
		dereg = 1;
	}
	else {
		reg->lifetime = htons(lifetime);
	}

	/* my mn/ip address 
	*/
	reg->home_address.s_addr = mn.s_addr;

	/* my ha's ip address 
	*/
	reg->home_agent.s_addr = ha.s_addr;

	/* set coa, if deregistration set it to mn address
 	 * see RFC 2002, 3.6.1.2
	*/
	if (!dereg) {
		reg->coa.s_addr = coa.s_addr;
	}
	else {
		reg->coa.s_addr = mn.s_addr;	/* dereg all coas */
	}

	/* set identity according to values
	 * already placed in mn_state field
	 */
	reg->i.l[0] = mns.ident[0];
	reg->i.l[1] = mns.ident[1];

#ifdef JUSTFORDEBUG
        mip_log(MLOG_INFO, "mn sends registration mesg: ident %x %x, lifetime %d, dst %s\n", 
                mns.ident[0], mns.ident[1], ntohs(reg->lifetime),
		inet_ntoa(dst));
#endif

	if (debug){
		printf("mns.auto_mode mns.mode ---> %d  %d \n",mns.auto_mode,mns.mode);
	}

#ifdef DIVERT
	/* Fill The Nat Message Extenstion if Private and doNat */
	if (mn_config.dhcpEnable && 
		(mn_config.doNat) && (mns.state == CO_LOCATED) ) {

		nat                = (struct mse_nat_ext_req *) &packet[psize];	  
		nat->type          = MN_HA_NAT_TYPE_REQ;
		nat->length        = 6; 
		nat->sub_type      = MN_HA_NAT_SUBTYPE;
		nat->reserved_1    = 0x00;
		/* if auto_mode is FALSE meaning manual and mns.mode
		 * manually set to NAT (force NAT)
		 */
		if ( (mns.auto_mode==FALSE) && (mns.mode == NAT)){
			/* Forcing a 2 way IP-UDP Tunnel */
			nat->reserved_2    = F_BIT;
		/* else HA tells us if we are behind NAT
		*/
		}else{
			nat->reserved_2    = 0x00;
		}
		nat->Encapsulation = IPVERSION;
		nat->reserved_3    = htons(0x0000);
		psize              += sizeof(struct mse_nat_ext_req);
	  
		if (debug){
			dumpNatReq(nat);
		}
	}
#endif
	
  	/* we must tack on the mn-ha registration
        */
	mse = (struct mse_authext *) &packet[psize];
        mse->type = MN_HA_AUTHENTICATOR_TYPE;
        mse->length = 4 + 16;

 	mn_key = get_key(
                mn_config.sa[MNHASECURITY].kl,
                (unsigned long) reg->home_agent.s_addr,
                &mn_keysize,
                &mn_spi,
                &mn_transform,
		&dummy, &dummy2);
 
	/* insane if we don't have HA key
	*/
        if (mn_key == NULL) {
                mip_log(MLOG_SEC, "MN failed to have key in HA keylist for HA %s",
			inet_ntoa(reg->home_agent));
		exit(1);
        }

        spi = htonl(mn_spi);
        bcopy(&spi, &mse->spi, sizeof(int));

        storeMd5((char *)packet, psize + AUTHEXT_HASHEDBYTES,
		mse->hash, mn_key);

        /* now update size accordingly
	*/
	psize += 2 + mse->length;

	if (debug) {
		printf("sending registration packet to %s:\n",
			inet_ntoa(dst));
		dumpUdpRequest(reg);
		dumpMse(mse);
	}

	/* if needed tack on optional mn/fa security association
	 * note: we only put this on if we are 
	 * 1. we were manually configured to do this and
 	 * 2. we are NOW talking through a FA
	 * (actually we have a coa) 
	 */
	if (mn_config.famn_security && mns.state == AWAY &&
		mns.agent && coa.s_addr != 0) {
                mse = (struct mse_authext *)
                        ((char *)mse + 2 + mse->length);    /* Zhong Chen */
                mse->type = MN_FA_AUTHENTICATOR_TYPE;
                mse->length = 4 + 16;

		/* we try dst first, then COA to get key
		 */
		mn_key = get_key(
			mn_config.sa[FAMNSECURITY].kl,
			(unsigned long)dst.s_addr,
			&mn_keysize,
			&mn_spi,
			&mn_transform,
			&dummy, &dummy2);

		if (mn_key == NULL) {
			mn_key = get_key(
				mn_config.sa[FAMNSECURITY].kl,
				(unsigned long)reg->coa.s_addr,
				&mn_keysize,
				&mn_spi,
				&mn_transform,
				&dummy, &dummy2);
		}
	 
		/* insane if we don't have FA key
		*/
		if (mn_key == NULL) {
			char dstbuf[1024];
			strncpy(dstbuf, inet_ntoa(dst), 1024);
			mip_log( MLOG_SEC, "MN failed to have FA dst=%s OR coa=%s key in FA list when specified\n",  dstbuf, inet_ntoa(coa));
			return;
		}

		spi = htonl(mn_spi);
		bcopy(&spi, &mse->spi, sizeof(int));
                storeMd5((char *)packet, psize + AUTHEXT_HASHEDBYTES,
			mse->hash, mn_key);

                psize += 2 + mse->length;
		if (debug) {
			dumpMse(mse);
		}
        }
	

	/* send the packet to the  udp port 434
	*/
        to.sin_family = AF_INET;
        to.sin_addr.s_addr = dst.s_addr;
	to.sin_port = htons(UDP_PORT);


	rc = sendto(udpSocket, packet, psize,
		0,  (struct sockaddr *) &to, 
		sizeof (struct sockaddr_in));


        /* record time we sent packet
        */
        gettimeofday(&mns.sendtime, NULL);

	/* do not want to exit here as mnd might be
	 * disconnected
	 */
	if (rc < 0) {
	        syslog( LOG_ERR, "sendRegistration(): sendto (%m)");
	}
	/* increment number of reg. packets sent
	*/
	mn_stats.reg++;

	if (debug) {
		printf("mn send of registration packet to %s\n",
			inet_ntoa(to.sin_addr));
		dumpme(packet, psize);
	}
}

