
/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * ntptime.c
 *
 * functions needed to manipulate time for timestamp replay protection
 * in UDP registration packets
 *
 * Used by both MN and HA.
 */
#include <sys/time.h>
#include "ntp_time.h"
#include "mip_log.h"

int timedebug = 0;

char *humandate(unsigned long);

char *
dumptime(unsigned long t)
{
	return(humandate(t));
}

/*
 * store the current time as a byte-swapped 64 bit NTP
 * time in a buffer (i.e., the sending packet)
 *
 * MN will use this when sending registration/reply pkts
 */
void
store_ntptime(char *buf)
{
	l_fp mytime;  	
	get_systime(&mytime);
	if (timedebug) {
		printf("store_ntptime:<%s>\n", dumptime(mytime.l_ui));
	}
	HTONL_FP(&mytime, (l_fp *) buf);
}

/* get ntp time in secs
 * don't byte swap it
 */
static l_fp mytime;  	

unsigned long
get_ntptime_secs()
{
	get_systime(&mytime);
	return(mytime.l_ui);
}

/*
 * should be called immediately after call to get_ntptime_secs
 */
unsigned long
get_ntptime_fract()
{
	return(mytime.l_uf);
}

/*
 * HA must take recv. npt time from MN and determine
 * if it is accurate time-wise or no
 *
 * params: 
 *	current - recv. time from MN
 *	previous - if not null, previous list time for MN
 *		NULL if MN not previously registered
 *	slop - allowed deviation in seconds
 *	hasecs - if MN timestamp is fresh, but exceeds slop,
 *		we return ha's time in seconds
 *
 * returns:
 *	0 - ok
 *     -1 - off by too much acc. to slop value, send NAK
 *     -2 - time is before previous time, upper level decides.
 *
 *  NOTE: return errors are logged prior to returning.
 */
int
check_ntpsys(char *current, char *previous, int slop, unsigned long *hasecs)
{
	l_fp hatime;  	/* current ha time as NTP value */
	l_fp mncur;     /* current MN value acc. to incoming packet */
	l_fp mnprev;    /* MN NTP acc. to previous registration */
	struct timeval hatv;  /* ha time as unix timeval */
	struct timeval mntv;  /* mn time as unix timeval */
	unsigned long secs, usecs;
	long diffsecs;
	extern int debug;

	/* convert current mn time to host byte order 
	*/
	NTOHL_FP((l_fp *)current, &mncur);
	if (timedebug) {
		printf("check_ntpsys: mn input time: <%s>\n", dumptime(mncur.l_ui));
	}

	/* if previous time is passed in, we check current to make
	 * sure it is greater than previous.  if this fails,
	 * we have an INVALID packet that might be an attack.
	 * Of course it might merely be late due to a retry.
	 */
	if (previous) {
		NTOHL_FP((l_fp *)previous, &mnprev);
		if ((mncur.l_ui < mnprev.l_ui) || 
		    (mncur.l_ui == mnprev.l_ui && mncur.l_uf <= mnprev.l_uf)) {
			if (timedebug) {
				printf("check_ntpsys, SPOOF return (-2)\n");
			}
			return(-2);  /* pkt time precedes last vlist time */
		}
	}
	/* now check for sanity time wise
	*/
	/* 1st get current time */
	get_systime(&hatime);
	if (timedebug) {
		printf("check_ntpsys: get_systime, %s\n", dumptime(hatime.l_ui));
		printf("hatime: %x %x\n", hatime.l_ui, hatime.l_uf);
		printf("mncur: %x %x\n", mncur.l_ui, mncur.l_uf);
	}

	/* convert ntp times to timestamps */
	TSTOTV(&mncur,&mntv);
	TSTOTV(&hatime,&hatv);

	if (timedebug) {
		printf("mntv 0x%x 0x%x\n", mntv.tv_sec, mntv.tv_usec);
		printf("hatv 0x%x 0x%x\n", hatv.tv_sec, hatv.tv_usec);
	}
	/* make sanity check
	*/
	secs = hatv.tv_sec - mntv.tv_sec;

	if (secs < 0)
		secs = -secs;
	if (timedebug) {
		printf("secs %d slop %d\n", secs, slop);
	}
	if (secs >= slop) {
		*hasecs = hatv.tv_sec;
		if (timedebug) {
			printf("check_ntpsys: mn time off by too much\n");
		}
		return(-1);
	}
	if (timedebug) {
		printf("check_ntpsys: ha judges mn time ok\n");
	}
	return(0);
}





/*
 * MN uses this function to set system time from HA suggestion
 */
void
set_ntptime(unsigned long secs, struct timeval *send, struct timeval *recv)
{
	l_fp mntime;
	l_fp mytime;  	
	l_fp delta;  	
        unsigned long roundtrip = 0;
	int rc;
        int ssecs, rsecs;

	get_systime(&mytime);

        ssecs = rsecs = 0;
        /* very rough smoothing function at second granularity only
        */
        if (send && recv && (recv->tv_sec > send->tv_sec)) {
                        roundtrip = (recv->tv_sec - send->tv_sec) / 2;
                        ssecs = send->tv_sec;
                        rsecs = recv->tv_sec;
        }

	mntime.l_ui = (secs);
	mntime.l_uf = 0;
	delta.l_ui = secs - mytime.l_ui + roundtrip;
	delta.l_uf = 0;

#ifdef JUSTFORDEBUG
        mip_log(MLOG_INFO, "time debug: delta value %x, secs %x mytime %x roundtrip %x, send secs %x, recv secs %x",
                delta.l_ui, secs, mytime.l_ui, roundtrip, ssecs, rsecs );
#endif
	
	step_systime(&delta);
	if (timedebug) {
		printf("set_ntptime, time set to %s\n", dumptime(mntime.l_ui));
		get_systime(&mytime);
		printf("set_ntptime, time NOW is %s\n", dumptime(mytime.l_ui));
	}
}

#ifdef TEST
int debug;
main()
{
	char buf[1024];
	bzero(buf, 1024);
	store_ntptime(buf);
}
#endif
