/*
 * 80211 dynamic adaptability module
 *
 * This is experimental; i.e., massively NOT debugged.
 *
 *	needs per device support.  currently only supports
 *	wi0 device on FreeBSF.  aironet should be possible.
 *
 *	synopsis:
 *	dynamically change between the following states until
 *	a higher-level entity (e.g., MN state machine) tells us
 *	to stop:
 *
 *	state1: managed 
 *	state2: adhoc channel start 
 *	state3: adhoc next channel
 *	stateN; adhoc channel end ... 
 *		back to managed
 *
 * 	a timer is supplied in the config file and is used to
 *	changed states above only when in NOWHERE mode.  At this
 *	point we have at least two things possibly going on in
 *	NOWHERE mode, layer 3 adaptibility and layer 2 adapatibility.
 *	This module implements layer 2 adaptibility.  Layer 3 
 *	adaptibility exists iff DHCP is configured on.  When in 
 *	NOWHERE mode, mnd may then attempt to start DHCP, and will
 *	change from NOWHERE Mode to CO-LOCATION mode if that works.
 *	It will fall back to MIP modes if any agent is detected.
 *	Currently there is no attempt to synchronize between these
 *	two modes.  It MIGHT be worthwhile to have layer 2 tell DHCP
 *	that it should start over, but possibly they don't need to
 *	cross-communicate.
 */

#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h> 
#include <netinet/in.h>
#include <arpa/inet.h>
#include "mnconfig.h"
#include "mip_log.h"
#include "mnstats.h"

/* possible link hw modes
*/
#define MANAGED	1
#define ADHOC	2

static int link_running;	/* 1 if in NOWHERE state and active */
static int link_channel;	/* current ad hoc channel when in adhoc mode */
static int link_timer;		/* current link_timer value */
/* note: this sets a default STATE machine start state
*/
static int link_mode=MANAGED;	/* current mode */
static int link_fail;		/* always fail if no known device */

extern struct mn_config mn_config;
static int link_changeState();
static int link_managed();
static int link_adhoc();
static void link_execState();
extern int debug;

/* if the device is ok, force managed mode
 * we do 2 things here (logically):
 *	1. put mobile device in managed mode
 *	2. put NOWHERE/link state machine in managed mode
 * 
 * Note: these two routines (forcemanaged/forcelink) should
 * only be called after link_init has been called at boot
 * and has succeeded
*/
link_forcemanaged()
{
	if (link_fail)
		return;

	link_mode = MANAGED;
	link_managed(mn_config.ifname);
	if (debug) {
		printf("link_forcemanaged called\n");
	}
}

link_forceadhoc(int channel)
{
	if (link_fail)
		return;

	link_mode = ADHOC;
	link_channel = channel;
	link_adhoc(mn_config.ifname);
	if (debug) {
		printf("link_forceadhoc called\n");
	}
}

/* config routine.  called at config read time (after)
 * to make sure everything in here defaults to something
 */
link_defaultsetup()
{
	link_channel = mn_config.link_chan_start;
	link_timer = mn_config.link_timer;
}

/*
 * initialize the link and state machine to a known start state
 *
 * called at any beginning of state machine cycle by link_start.
 *
 * algorithm:
 *	
 *	determine if mobile device can do this trick.
 *		fail if no acceptable device (wi0 is it for now)
 *	fail if config never turns us on.
 *
 * TBD: sanity check link_running variable.  can we get rid of it?
 */
link_init()
{
	char *dev = mn_config.ifname;

	link_running = 0;
	link_timer = mn_config.link_timer;

	/* sanity check for known useable devices, wi0 only choice so far
	*/
	if (strcmp(dev, "wi0") == 0) {
		link_fail = 0;		/* we will work */
		if (debug) {
			printf("link: link_init dev %s is OK\n", dev);
		}
	}
	else {
		link_fail = 1;	
		if (debug) {
			printf("link: link_init FAILS dev %s NO match\n", dev);
		}
	}

	/* we also fail if we were never turned on
	*/
	if (link_timer == 0) {
		link_fail = 1;	
		if (debug) {
			printf("link: link_init FAILS - timer %d not on\n", link_timer);
		}
	}
}

/* called if new transition to NOWHERE state
 * we start the state machine going.
*/
link_start()
{
/*
	if (debug) {
		printf("link_start called: link_running %d, link_fail %d\n",
			link_running, link_fail);
	}
*/
	if (link_running)
		return;

	/* ok, we were NOT running.  try and make us run
	*/

	/* 1st initialize and make sure we can actually do
	 * something
	 */
	link_init();
	if (link_fail)
		return;

	/* we are running
	*/
	link_running = 1;

	/* set initial state
	*/
	link_execState();

	if (debug) {
		printf("link_start: link_fail %d, run %d\n", 
			link_fail, link_running);
	}
}

/* called if NOT in NOWHERE state.  we stop the state machine.
 * actually we stop the timer.
*/
link_stop()
{
	/* never successfully initialized. return
	*/
	if (link_fail)
		return;

	link_timer = 0;
	link_running = 0;
	if (debug) {
		printf("link_stop: run %d\n", link_running);
	}
}


/*
 * note: our state machine model is like so:
 *
 *	we execute the initial state at link_start
 *	we tick tick tick
 *	when timer is timed out
 *		we change state and execute again
 */
link_tick()
{
	/* never successfully initialized. return
	*/
	if (link_fail)
		return;

	/* bail if we are not active
	*/
	if (link_running == 0)
		return;

	/* we are active, decrement timer and make state
	 *  changes
	*/
	if (debug) {
		printf("link: link_tick %d\n", link_timer);
	}
	link_timer--;
	if (link_timer <= 0) {
		link_changeState();
		link_execState();
		link_timer = mn_config.link_timer;
	}
}

/* remaining routines are not visible outside this module
*/

/*
 * state machine model
 *
 *	start out in MODE  MANAGED for N ticks
 *	if called again, change states as follows
 *		MANAGED
 *		ADHOC chan start
 *		      chan start+1 ... until
 *		      chan end
 *
 * basic idea: set the state, then carry it out.
 */	

static
link_changeState()
{
	
	/* based on current state, go to next state
	*/
	switch(link_mode) {
	case MANAGED:
		link_mode = ADHOC;
		link_channel = mn_config.link_chan_start;
		break;
	case ADHOC:
		/* if current channel is last channel,
		 * switch to MANAGED     
		 */
		if (link_channel == mn_config.link_chan_end) {
			link_mode = MANAGED;
			link_channel = mn_config.link_chan_start;
		}
		else {
			link_channel++;
		}
		break;
	}

	if (debug) {
		printf("link change state next mode is:\n");
		if (link_mode == MANAGED) {
			printf("link: managed\n");
		}
		else {
			printf("link: ad hoc channel %d\n", link_channel);
		}
	}

}

/*
 * actually implement the change
 */
static void
link_execState()
{
	char *dev = mn_config.ifname;

	/* having changed states, activate them
	*/
	switch(link_mode) {
	case MANAGED:

		/* if we were not in arp mode, and we are to force arp mode
		 * exec mnd with certain params that force this to happen.
		 * the config must use arp ... else you have big trouble ...
		 */
		if (mn_config.doArp == 0 && mn_config.link_forcearp) {
			/* exec mnd with a certain config file
			*/
			reboot_mnd("reboot in managed mode",
				mn_config.managed_reboot, "-m" );
		}
		link_managed(dev);
		break;

	case ADHOC:
		/* if forcearp is set and we are doing arp ... then
		 * reconfigure
		 */
		if (mn_config.doArp && mn_config.link_forcearp) {
			/* exec mnd with a certain config file to get us
			 * to AAH substitute for arp
			*/
			reboot_mnd("reboot in adhoc mode",
				mn_config.adhoc_reboot, "-a");
		}
		/* set to current channel
		*/
		link_adhoc(dev);
		break;
	}
}


/* set managed mode for device
 * /usr/sbin/wicontrol -i wi0 -p 1
 * /usr/sbin/wicontrol -i wi0 -n essid
 */
static
link_managed(char *dev)
{
	char buf[1024];

	/* to managed mode
	*/
	sprintf(buf, "%s -i %s -p 1\n", "/usr/sbin/wicontrol", dev);
	system(buf);

	/* then set essid
	*/
	sprintf(buf, "%s -i %s -n %s\n", "/usr/sbin/wicontrol",
		dev,
		mn_config.link_essid);
	system(buf);

	if (debug) {
		printf("link: link_managed called\n");
	}
}

/* set adhoc mode for device, and then set the channel
 * /usr/sbin/wicontrol -i wi0 -p 3
 * /usr/sbin/wicontrol -i wi0 -f <chan_number>
 */
static
link_adhoc(char *dev)
{
	char buf[1024];

	if (debug) {
		printf("link: link_adhoc %d channel\n", link_channel);
	}
	sprintf(buf, "%s -i %s -p 3\n", "/usr/sbin/wicontrol",
		dev);
		
	system(buf);

	sprintf(buf, "%s -i %s -f %d\n", "/usr/sbin/wicontrol",
		dev, link_channel);

	system(buf);
}

#ifdef TESTMAIN

struct mn_config mn_config;

int debug = 1;

/*
 * note: alarm tick is simulated for debug by carriage return!
 */
main()
{
	int count;

	/* hardwire mn_config structure for testing
	*/
	strcpy(mn_config.ifname, "wi0");
	mn_config.link_timer = 2;	/* set small for testing */
	mn_config.link_chan_start = 3;
	mn_config.link_chan_end = 4;
	/* note this test setup is 3 states max */
	strcpy(mn_config.link_essid, "any");

	count = mn_config.link_timer * 3  /* states */ + 1;

	link_init();

	link_start();

	while(count--) {
		link_tick();   /* tick ... drive state machine */
	}

	link_stop();

}
#endif

