/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef IPSEC
/*
 *  keysock.c
 *
 *  application level management of IPSEC route keys used by MIP
 */

#include "mip_log.h"

#include <sys/socket.h>
#include <sys/socketvar.h>
#include <netinet/in.h>

#include <netsec/ipsec.h>
#include <netkey/key.h>

#include "ipsec.h"

char * reverseAddress(struct in_addr);

/* debugging */
extern int debug;
static int noticeFunction = 1;


/* key message/socket globals */
#define MAXRCVBUFSIZE	(8 * 1024)	/* from ipkey/key.c */
static char key_message[sizeof(struct key_msghdr)
		+ (MAX_SOCKADDR_SZ * 3) + MAX_KEY_SZ + MAX_IV_SZ];
static struct sockaddr_in sin_dummy = { sizeof(sin_dummy), AF_INET };
static int keyseqno = 1;
static int keymsgver = 1;		/* PF_KEY version */
static int keysock;


/*
 *  open and close socket on PF_KEY domain
 */
void
openKeySocket(void)
{
	int rcvsize;

entryMark(0, "openKeySocket");

	if ((keysock = socket(PF_KEY, SOCK_RAW, 0)) < 0) {
		syslog(LOG_ERR, "openKeySocket(): socket (%m)\n");
		exit(1);
	}

	/* Set close-on-exec flag to prevent fd leaks. */
	if (fcntl(keysock, F_SETFD, 1) < 0) {
		syslog(LOG_INFO, "openKeySocket(): F_SETFD (%m)");
		/* not catastrophic, don't exit */
	}

	/* stolen from usr.sbin/ipkey.  shrinks socket receive buffer
	 * presumably for security reasons.
	 */
	for (rcvsize = MAXRCVBUFSIZE; rcvsize; rcvsize -= 1024) {
		if (setsockopt(keysock, SOL_SOCKET, SO_RCVBUF, &rcvsize, 
			sizeof(rcvsize)) <= 0)
		break;
	}

}  /* end openKeySocket() */

void
closeKeySocket(void)
{
	close(keysock);
}


/*
 *  keyMsg:	
 *  
 *  ASSUMES that (global) key_message has already been properly constructed.
 *
 *  supported key message types--
 *	KEY_ADD		attempts add of key_message
 *	KEY_GET		gets key with characteristics of key_message
 *			  and overwrites that with the result received
 *
 *  NOTE that write() on key socket may return useful errno values:
 *	ESRCH	could not find key
 *	EEXIST	key already exists
 *
 *  keyMsg return values:
 *	0	success
 *	else 	return errno value
 */
static int
keyMsg(int msgtype)
{
	struct key_msghdr	*kmsg = (struct key_msghdr *) key_message;
	int seqno, pid, type;
	int nbytes, cc;
	char *buf;

entryMark(0, "keyMsg");

	/* build a key message
	*/
	kmsg->key_msglen = sizeof(struct key_msghdr)
		+ (sizeof(struct sockaddr_in)*3) + kmsg->keylen + kmsg->ivlen;
	kmsg->key_msgvers = keymsgver;
	kmsg->key_pid = pid = getpid();
	kmsg->key_seq = seqno = keyseqno++;
	type = kmsg->type;	/* save type for setting read() size */

	switch(kmsg->key_msgtype = msgtype) {
	case KEY_GET:
	case KEY_ADD:
		break;
	default:
		mip_log(MLOG_ERR, "keyMsg(): unsupported key message type.");
		exit (1);
	}


	/* write key message and read out result (which may be
	 * a simple echo)
	 */
	buf = key_message;
	cc = kmsg->key_msglen;
	errno = 0;
	while (cc > 0) {
		nbytes = write(keysock, buf, cc);
		if (nbytes < 0) {
		    if (errno == ESRCH || errno == EEXIST) {
			    return(errno);
		    }
		    else {
			    mip_log(MLOG_ERR,
				"keyMsg(): write failed (errno=%d).\n", errno);
			    exit(1);
		    }
		}
		cc -= nbytes;
		buf += nbytes;
	}

	do {
		bzero(key_message, sizeof(key_message));
		buf = key_message;
		cc = sizeof(struct key_msghdr) +
			(sizeof(struct sockaddr_in) * 3) +
				((type==KEY_TYPE_ESP)? 12 : 16);
				
		while (cc > 0) {
			if ((nbytes = read(keysock, buf, cc)) < 0) {
				mip_log(MLOG_ERR, "keyMsg(): read failed.\n");
				exit(1);
			}
			cc -= nbytes;
			buf += nbytes;
		}
	} while (! (kmsg->key_pid == pid && kmsg->key_seq == seqno
			&& kmsg->key_msgtype == msgtype) );


	return 0;

}  /* end keyMsg() */



/*
 *  verifyKey:	build a message for the Key Socket and perform a key
 *		GET on the SA-tuple given by input arguments.  Return
 *		value indicates whether or not the key exists in the kernel.
 *
 *		all keys checked will be either for AH or ESP, and all
 *		keys apply to either outgoing or incoming traffic.
 *
 *  RETURNS:
 *	0	success -- key found in kernel
 *	-1	err
 */
int
verifyKey(	struct in_addr src,
		struct in_addr dst,
		u_long spi,
		int isah,		/* is AH or ESP */
		int outgoing)		/* TRUE, else incoming */
{
	char			addr1[MAXBUF], *addr2;
	int			ptype;	/* policy type */
	struct key_msghdr	*kmsg = (struct key_msghdr *)key_message;

entryMark(0, "verifyKey");


	if (isah)	ptype = AHSEC;
	else		ptype = ESPSEC;

	if (outgoing) {
		createKeyMsg(src, dst, spi, ptype);
		sprintf(addr1, "%s", reverseAddress(src));
		addr2 = reverseAddress(dst);	
	}
	else {
		createKeyMsg(dst, src, spi, ptype);
		sprintf(addr1, "%s", reverseAddress(dst));
		addr2 = reverseAddress(src);	
	}

	if (keyMsg(KEY_GET) || kmsg->key_errno) {
		syslog(LOG_ERR, "verifyKey(): Missing %s key, <%s,%s,0x%x,%s> (errno %d,key_errno %d)",
			outgoing?"outbound":"inbound",
			addr1, addr2, spi, isah?"ah":"esp",
			errno, kmsg->key_errno);
	    	return -1;
	}


	return(0);

}  /* end verifyKey() */


/*
 *  createKeyMsg:
 *
 *  transform is either AHSEC or ESPSEC.  these are converted
 *  to type/algorithm values from netsec/ipsec.h where ESP is hardwired
 *  to 1DES and AH hardwired to md5.
 */
static void
createKeyMsg(	struct in_addr srcAddr,
		struct in_addr dstAddr,
		u_long spi,
		int transform)
{
	struct sockaddr_in	src,
				dst;
	struct key_msghdr	*kmsg = (struct key_msghdr *) key_message;
	char			*kdata = (char *) (key_message +
						sizeof(struct key_msghdr));

entryMark(0, "createKeyMsg");

	bzero(key_message, sizeof(key_message));

	kmsg->spi = spi;	/* spi is host byte order */

	switch (transform) {
	case ESPSEC:
		kmsg->type = KEY_TYPE_ESP;
		kmsg->algorithm = IPSEC_ALGTYPE_ESP_DES_CBC;
		break;

	case AHSEC:
		kmsg->type = KEY_TYPE_AH;
		kmsg->algorithm = IPSEC_ALGTYPE_AH_MD5;
		break;

	default:
		mip_log(MLOG_ERR, "createKeyMsg(): invalid transform value (%d).  Oops.", transform);
		exit(1);
	}

	bcopy(&sin_dummy, &src, sizeof(struct sockaddr_in));
	bcopy(&srcAddr, &src.sin_addr, sizeof(struct in_addr));
	bcopy(&sin_dummy, &dst, sizeof(struct sockaddr_in));
	bcopy(&dstAddr, &dst.sin_addr, sizeof(struct in_addr));

	bcopy(&src, kdata, sizeof(struct sockaddr_in));
	ADVANCE(kdata, sizeof(struct sockaddr_in));
	bcopy(&dst, kdata, sizeof(struct sockaddr_in));
	ADVANCE(kdata, sizeof(struct sockaddr_in));
	bcopy(&sin_dummy, kdata, sizeof(sin_dummy.sin_len));

}  /* end createKeyMsg() */

#endif /* IPSEC */

