
/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * rand_number.c
 *
 * produce 32 bit random number.  The assumption is that
 * this may be used at the application level by
 * 1. a boot daemon that needs a random number for some
 * reason OR
 * 2. application level key generation programs.
 *
 * This is not very fast.  About 32 bits per second.
 *
 * 1. we use BSD random(3) which is purported to
 *	be a good pseudo-random number source.  It
 *	is seeded with srandom(3).
 * 2. we get the time (number of seconds).
 * 3. we consult a number of files that are mostly
 * system log files (but not always) and extract
 * various "random" values, #3 and #2 are used
 * as input to srandom(3).
 * 4. we also (#define) optionally do an additive checksum
 * on netstat -in (no DNS, count packet i/o).
 * 5. we plug all of that into srandom.
 * 6. we return the ulong value from random().
 *
 * This code runs on FREEBSD 2.1R.  It is not portable currently.
 *
 * Assumptions:
 *	1. srandom/random(3) available.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>

#define MAILLOG  "/var/log/maillog.0" 

/* slower if this is used as process is exec'ed, written
 * to file, and result added up.
 */
#define USECOMMAND	1

static unsigned int get_varlogmessages();
static unsigned int get_varlogmaillog();
static unsigned int get_netstat();

unsigned long
get_random()
{
	long l;
	time_t t;
	struct timeval tv;
	int uid = getuid();
	unsigned int sum1;

	unsigned int sum2 = get_varlogmessages();
	unsigned int sum3 = get_varlogmaillog();
	unsigned int sum4 = (getpid() + uid) << 20;
	gettimeofday(&tv, 0);
#ifdef USECOMMAND
	sum1 = get_netstat();
#else
	srandom(tv.tv_usec ^ tv.tv_sec);
	sum1 = random();
#endif
	srandom((tv.tv_usec << 16) + tv.tv_usec + tv.tv_sec + sum1 + sum2 + sum3 + sum4);
	/* random only returns a long
	*/
	{
		unsigned long l = random();
		return(l ^ (random() << 1));
	}
}

/* yesterday's maillog.  cron rotates the maillog
 * produced by sendmail once a day.
 * more or less produce an additive sum and throw
 * in an xor to make it interesting.
 */
static 
unsigned int
get_varlogmaillog()
{
	unsigned char c;
	unsigned int sum;
	FILE *fd = fopen(MAILLOG, "r");
	if (fd == NULL) {
		return(random() ^ random());
	}
	while( fread(&c, 1, 1, fd) == 1) {
		sum = (sum + c) ^ c;
	}
	fclose(fd);
	return(sum);
}

/*
 * add up values returned by netstat -in
 * (network interface packet counts with no DNS name translation)
 *
 * we will take the output of netstat
 * and add it up.  This may not be a good
 * source of randomness, but on the other hand
 * if the local networks are busy, it may not
 * be bad.
 *
 * XXX. There is some bizarre dyn. lib loading problem here
 * after a number of iterations.  Should probably
 * not invoke netstat but use the same kernel hooks
 * and read values of out kernel directly.
 *
 * ps -alx might not
 * be a bad model either for something like this
 * as you would get random timing values and that
 * might even be useful at boot.
 */
static 
unsigned int
get_netstat()
{
	FILE *p;
	time_t t;
	unsigned char c;
	unsigned int sum = 0;
	int pid = getpid();
	char *name;
	char command[1024];
	char template[1024];
	extern int errno;

	template[0] = 0;
	strcat(template, "/tmp/tmp.XXXXX");

	mkstemp(template);
	name = template;
	/* fallback strategy in case of failure
	*/
	if (name == NULL) {
		time(&t);
		return((unsigned int) t);
	}
	sprintf(command, "/usr/bin/netstat -in > %s", name);
	system(command);
	p = fopen(name, "r");
	if (p == NULL) {
		unlink(name);
		time(&t);
		return((unsigned int) t);
	}
	while( fread(&c, 1, 1, p) == 1) {
		/*
		printf("%c",c);
		*/
		sum += (int) c;
	}
	unlink(name);
	fclose(p);
	/*
	printf("sum %x\n", sum);
	*/
	return(sum);
}

/* read the current system log.
 * it is rotated on a weekly basis, however
 * it is added to for every reboot.
 * we count the number of characters and xor it with the time.
 */
static 
unsigned int
get_varlogmessages()
{
	struct stat s;
	struct timeval tv;
	int rc = open("/var/log/messages", 0);  /* read only */
	gettimeofday(&tv, 0);
	if (rc < 0) {
		srandom((unsigned int) tv.tv_usec);
		return(random() << 4);
	}
	fstat(rc, &s);
	return(s.st_size ^ tv.tv_usec);	
}

#ifdef TESTMAIN
main(int argc, char **argv)
{
	unsigned long r;
	int i, count;
	
	if (argc == 1)
		count = 1;
	else
		count = atoi(argv[1]);

	for ( i = 0; i < count; i++) {
		r = get_random();
		printf("0x%8.8x\n", r);
	}
}
#endif
