/*
 * Copyright (c) 2000 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * ripv2.c
 *
 * 2 times per minute send RIPv2 HA network out
 * a certain interface.
 *
 * This module is capable of sending ONE ripv2 packet out
 * one interface.  It is intended to be used with HARP
 * when two Home Agents are in use.  The co-HAs can thus
 * dynamically advertise their existance to an IGP mesh.
 * We assume that the advertise Home Agent network is one
 * hop away.  
 *
 * mipd only would use this facility and then only when HARP
 * is in use, and on a Home Agent.
 *
 * mipd config:
 *
 *	ripv2	if_addr  ha_net  ha_mask
 *
 * if_addr - interface thru which we spew RIP multicast
 * ha_net - network we advertise via RIP
 * ha_mask - associated network netmask
 */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/signal.h>
#include <termios.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/ip_var.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <unistd.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>

#define RIPVERSION 2

#include <protocols/routed.h>

#include "mip_log.h"

#define TRUE	1
#define FALSE	0

extern int debug;

/* the rip socket descriptor UDP/socket */
static
int ripSD;

/* the multicast socket target address (dst) */
static
struct in_addr mip;

/* the interface ip address (unicast) that we use to
 * aim the multicast write
 */
static
struct in_addr rip_mif;  

/* advertised Home Agent network addr and netmask
*/
static
struct in_addr ha_net;

static
struct in_addr ha_mask;

static void sendMulticast(int beaconSd);

/*
 * initIcmp
 *
 * open raw icmp socket so that we can send and recv router adverts
 * on it.  should be able to send solicitations too.
 *
 * Note: if and when we can send multicast (224.0.0.1) as well as
 * limited broadcast, we can't do both on the same socket descriptor.
 * The code in netinet/ip_output.c won't support that.
 * 
 * returns:
 *	int beaconSd - socket descriptor
 */ 
int
initRIP (maddr, mif, hanet, hamask)
struct in_addr maddr;  /* multicast address for output */
struct in_addr mif;  /* if non-zero, a default multicast ip interface */
struct in_addr hanet;
struct in_addr hamask;
{
	struct sockaddr_in ripport;
	int on;
	int rc;

	ha_net = hanet;
	ha_mask = hamask;

	mip = maddr;
	rip_mif = mif;

	if ((ripSD = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
 		syslog(LOG_ERR, "initRIP(): socket (%m)");
		exit(1);
	}
	
	/* Set close-on-exec flag to prevent fd leaks. */
	if (fcntl(ripSD, F_SETFD, 1) < 0) {
		syslog(LOG_INFO, "initRIP(): F_SETFD (%m)");
		/* not catastrophic, don't exit */
	}

	/*
	recvMulticast(ripSD, maddr, mif);
	*/

	sendMulticast(ripSD);

	/* setup port/520 bind 
	*/
	ripport.sin_family = AF_INET;
	ripport.sin_addr.s_addr = INADDR_ANY;   /* ok from any machine */
	ripport.sin_port = htons(520);     /* specific port */

	/* bind protocol to socket
	*/
	if (bind(ripSD, (struct sockaddr *)&ripport, sizeof(ripport))) {
 		syslog(LOG_ERR, "initRIP(): rip bind failed (%m)");
		exit(1);
	}

	return(ripSD);
}


#ifdef NOWAYCHARLIEP
/*
 * recvMulticast 
 *	1. merely enables the reception of
 * 	multicast addresses on the MIP default multicast port.
 * 	This allows us to recv icmp advert/solicitations from
 * 	hosts using multicast.
 *
 *	2. disables loopback of multicast packets
 *
 * Note: this code does NOT cause a fatal exit if the attempt
 * to add a multicast group on a particular interface fails.
 * It is possible that a given i/f might not support multicast.  
 * A message will be logged however.
 */
static void
recvMulticast(int beaconSd, struct in_addr maddr, struct in_addr mif)
{
	struct ip_mreq mreq;
	unsigned char loop = 0;

 	mreq.imr_multiaddr.s_addr = maddr.s_addr;
	if (mif.s_addr)
		mreq.imr_interface.s_addr = mif.s_addr;
	else
		mreq.imr_interface.s_addr = htonl(INADDR_ANY);

        if (setsockopt(beaconSd, IPPROTO_IP, 
		IP_ADD_MEMBERSHIP, &mreq, sizeof(mreq)) < 0){
 		syslog(LOG_ERR, "initIcmp(): IP_ADD_MEMBERSHIP failed (%m).");
		return;
	}
        if (setsockopt(beaconSd, IPPROTO_IP, 
		IP_MULTICAST_LOOP, &loop, sizeof(loop))) {
 		syslog(LOG_ERR, "initIcmp(): IP_MULTICAST_LOOP failed (%m).");
		exit(1);
	}
}
#endif

/*
 * sendMulticast
 *
 * setup send side.  recvMulticast already joined the mcast group.
 * We set the TTL to 1.  Setting the i/f either is only done
 * in the recv routine BECAUSE we only support one interface
 * for mns (mnd).  mipd has to explicitly set the output i/f
 * for each interface a multicast beaocn is sent out on.
 */
static void
sendMulticast(int beaconSd)
{
	unsigned char ttl = 1;

 	if (setsockopt(beaconSd, IPPROTO_IP, 
		IP_MULTICAST_TTL, &ttl, sizeof(ttl)) < 0) {
 		syslog(LOG_ERR, "initIcmp(): IP_MULTICAST_TTL failed (%m).");
		exit(1);
	}
}

/*
 * aim multicast socket at a given interface
 */
static
setMcastIf(int ripSD, struct in_addr addr)
{
	int rc;
     	rc = setsockopt(ripSD, IPPROTO_IP, IP_MULTICAST_IF, &addr, sizeof(addr));
	if (rc < 0) {
 		syslog(LOG_ERR, "setMcastIF(): IP_MULTICAST_IF failed (%m).");
		exit(1);
	}
}

sendRIP()
{
	int rc = 0;
	char pkt[2000];
	struct sockaddr_in to;
	int size = 0;
	struct rip *rip = (struct rip *) pkt;

	/* send the packet to the  udp port THEIRUDP
	*/
        to.sin_family = AF_INET;
        to.sin_addr.s_addr = mip.s_addr;
	to.sin_port = htons(520);

	rip->rip_cmd = 2;
	size += 1;
	rip->rip_vers = 2;
	size += 1;
	rip->rip_res1 = 0;
	size += 2;

	rip->ripun.ru_nets[0].n_family = RIP_AF_INET;
	rip->ripun.ru_nets[0].n_tag = 0;
	rip->ripun.ru_nets[0].n_dst = ha_net.s_addr;
	rip->ripun.ru_nets[0].n_mask = ha_mask.s_addr;
	rip->ripun.ru_nets[0].n_nhop = 0;
	rip->ripun.ru_nets[0]. n_metric = htonl(1);

	size += sizeof (struct netinfo);

	setMcastIf(ripSD, rip_mif);

	rc = sendto(ripSD, pkt, size, 0,  (struct sockaddr *) &to, 
		sizeof (struct sockaddr_in));

	if (rc < 0) {
 		syslog( LOG_ERR, "sendRIP(): sendto failed (%m).");
		exit(1);
	}
	return(1);
}

#ifdef TESTMAIN

int debug;

main()
{
	initRIP( inet_addr("224.0.0.9"), 
			inet_addr("131.252.222.83"),
			inet_addr("131.252.215.64"), 
			inet_addr("255.255.255.192")
			);
	sendRIP();
}
#endif
