
/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * udp_debug.c
 *
 * overview:
 *	functions to print out udp request/reply/mse submessages
 */
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/times.h>
#include <errno.h>
#include <signal.h>

#include "mipconfig.h"
#include "mipstate.h"
#include "udp_reg.h"

/* 
 * dumpUdpRequest()
 * debug or verbose logging
 *
 * print out udp request message contents.
 */
void
dumpUdpRequest(struct udp_register *reg)
{
	unsigned char flags;


	if (!reg) {
		printf("dumpUdpRequest(): reg == NULL!\n");
		return;
	}

	flags = reg->flags;

	printf("Mobile-IP: UDP registration message dump:\n");
	printf("type=%d:(", reg->type);
	if (reg->type == UDP_REGISTER_TYPE) {
		printf("REGISTER)\n");
	}
	else {
		printf("UNKNOWN)\n");
	}
	printf("flags=0x%x\n", flags);
	if (flags & S_BIT)
		printf("\tS BIT set, mn requests simultaneous bindings\n");
	if (flags & B_BIT)
		printf("\tB BIT set, mn requests forwarding of broadcasts\n");
	if (flags & D_BIT)
		printf("\tD BIT set, mn acting as own coa\n");
	if (flags & M_BIT)
		printf("\tM BIT set, mn requests minimal encapsulation\n");
	if (flags & G_BIT)
		printf("\tM BIT set, mn requests Cisco encapsulation\n");
	if (flags & V_BIT)
		printf("\tM BIT set, mn requests Van Jacobson compression\n");
	printf("lifetime=%d secs\n", ntohs(reg->lifetime));
	if (ntohs(reg->lifetime) == 0)
		printf("\t0 means mn wants to deregister\n");
	printf("mn home address=%s (0x%x)\n", inet_ntoa(reg->home_address),
			reg->home_address);
	printf("HA address=%s (0x%x)\n", inet_ntoa(reg->home_agent),
			reg->home_agent);
	printf("COA address=%s (0x%x)\n", inet_ntoa(reg->coa),
			reg->coa);
	printf("ident in hex=(%x,%x)\n", reg->i.l[0], reg->i.l[1]);
} 

void
dumpUdpReply(struct udp_reply *rep)
{
	if (!rep) {
		printf("dumpUdpRequest(): rep == NULL!\n");
		return;
	}

	printf("Mobile-IP: UDP reply message dump:\n");
	printf("type=%d:(", rep->type);
	if (rep->type == UDP_REPLY_TYPE) {
		printf("REPLY)\n");
	}
	else {
		printf("UNKNOWN)\n");
	}
	printf("code=%d:", rep->code);
	switch(rep->code) {
	case REG_REPLY_OK:
		printf("(REPLY OK)\n");
		break;
	case REG_OK_NOMB:
		printf("(REPLY OK, no multiple bindings)\n");
		break;
	/****************************************************/
	/* fa denies registration */
	case REG_FAIL_FA_UNSPEC:
		printf("(FA/FAIL: reason unspecified)\n");
		break;
	case REG_FAIL_FA_PROHIB: 	
		printf("(FA/FAIL: administratively prohibited)\n");
		break;
	case REG_FAIL_FA_RESOURCES:
		printf("(FA/FAIL: insufficient resources)\n");
		break;
	case REG_FAIL_FA_MN_AUTH:
		printf("(FA/FAIL: mobile node failed authentication)\n");
		break;
	case REG_FAIL_FA_HA_AUTH:
		printf("(FA/FAIL: home agent failed authentication)\n");
		break;
	case REG_FAIL_FA_LIFETIME:
		printf("(FA/FAIL: requested lifetime too long)\n");
		break;
	case REG_FAIL_FA_REQUEST:	
		printf("(FA/FAIL: poorly formed request)\n");
		break;
	case REG_FAIL_FA_REPLY:	
		printf("(FA/FAIL: poorly formed reply)\n");
		break;
	case REG_FAIL_FA_ENCAPS:
		printf("(FA/FAIL: requested encapsulation unavailable)\n");
		break;
	case REG_FAIL_FA_VJ:
		printf("(FA/FAIL: requested VJ compression unavailable)\n");
		break;
	case REG_FAIL_FA_NETUNREACH:
		printf("(FA/FAIL: home network unreachable)\n");
		break;
	case REG_FAIL_FA_HAUNREACH:
		printf("(FA/FAIL: home agent unreachable)\n");
		break;
	case REG_FAIL_FA_PORTUNREACH:
		printf("(FA/FAIL: home agent port unreachable)\n");
		break;
	case REG_FAIL_FA_OTHERUNREACH:
		printf("(FA/FAIL: home agent (other icmp) unreachable?)\n");
		break;
	/****************************************************/
	/* ha denies registration */
	case REG_FAIL_HA_UNSPEC:
		printf("(HA/FAIL: reason unspecified)\n");
		break;
	case REG_FAIL_HA_PROHIBIT:
		printf("(HA/FAIL: access prohibited)\n");
		break;
	case REG_FAIL_HA_RESOURCES:
		printf("(HA/FAIL: insufficient resources)\n");
		break;
	case REG_FAIL_HA_MN_AUTH:	
		printf("(HA/FAIL: mobile node failed authentication)\n");
		break;
	case REG_FAIL_HA_FA_AUTH:	
		printf("(HA/FAIL: FA failed authentication)\n");
		break;
	case REG_FAIL_HA_IDMISMATCH:	
		printf("(HA/FAIL: registration ID mismatch)\n");
		break;
	case REG_FAIL_HA_REQUEST:
		printf("(HA/FAIL: poorly formed request)\n");
		break;
	case REG_FAIL_HA_BINDINGS:
		printf("(HA/FAIL: too many simultaneous bindings)\n");
		break;
	case REG_FAIL_HA_UNKNOWN:	
		printf("(HA/FAIL: unknown home agent address)\n");
		break;
	default:
		printf("(unknown reply code)\n");
		break;
	}
	printf("lifetime=%d\n", ntohs(rep->lifetime)); 
	printf("mn home address=%s (0x%x)\n", inet_ntoa(rep->home_address),
			rep->home_address);
	printf("HA address=%s (0x%x)\n", inet_ntoa(rep->home_agent),
			rep->home_agent);
	printf("ident in hex=(%x,%x)\n", rep->i.l[0], rep->i.l[1]);
};

/*
 * dumpMse
 *
 * dump ANY extension, not just security ones.
 * These sub-messages are in the form of Tag/Length/Value.
 * We print out the "value" (data) in hex, 16 bytes a line.
 * the first two fields, type and length, are ASSUMED to be one byte
 * large.
 *
 * msep MUST point to something.
 */
void 
dumpMse(char *msep)
{
	int l;		/* length of "value" (data) */
	char *pt;	/* pointer to beginning of "value" (data) */
	int i;
	int type;
	int length;
	struct mse_authext	*authext;
	int spi;
	struct mse_ipsecext	*ipsecext;

	if (!msep) {
		printf("dumpMse(): msep == NULL!  (Good thing I checked...)\n");
		return;
	}

	type = ((int) *msep) & 0xff;
	length = ((int) *(msep+1)) & 0xff;

	printf("Mobile-IP: Mobile UDP Extension message dump:\n");
	printf("type=0x%x:", type);

	switch (type) {
	case MN_HA_AUTHENTICATOR_TYPE:
		printf("(MN-HA authenticator)\n");
		break;
	case MN_FA_AUTHENTICATOR_TYPE:
		printf("(MN-FA authenticator)\n");
		break;
	case FA_HA_AUTHENTICATOR_TYPE:
		printf("(FA-HA authenticator)\n");
		break;
	default:
		/* TBD. are there other types that are known?
		 * The subnet msg?
		 */
		printf("(UNKNOWN TYPE!)\n");
	}

	printf("length=%d (bytes)\n", length);

	switch (type) {
	case MN_HA_AUTHENTICATOR_TYPE:
	case MN_FA_AUTHENTICATOR_TYPE:
	case FA_HA_AUTHENTICATOR_TYPE:
		authext = (struct mse_authext *) msep;
		bcopy(&authext->spi, &spi, 4);
		printf("spi=0x%x\n", ntohl(spi));
		l = length - 4; /* minus 4 bytes for the spi */
		pt = authext->hash;
		break;
	default:
		l = length;
		pt = (char *)(msep + 2);
	}

	if (l != 0) {
		printf("value field (data) follows:\n\t");
		for ( i = 0; i < l; i++) {
			if (i > 0 && ((i % 16) == 0)) {
				printf("\n\t");
			}
			printf("%2.2x ", *pt++ & 0xff);
		}
		printf("\n");
	}

}  /* end dumpMse() */

void 
dumpNatReq(char *nat)
{
	int type;
	int sub_type;
	int length;
	int reserved_1;
	int reserved_2;
	int Encapsulation;
	int reserved_3; 

	if (!nat) {
		printf("dumpNatReq(): nat == NULL!  (Good thing I checked...)\n");
		return;
	}

	type            = ((int) *nat) & 0xff;
	length          = ((int) *(nat+1)) & 0xff;
	sub_type        = ((int) *(nat+2) ) & 0xff;
	reserved_1      = ((int) *(nat+3) ) & 0xff;
	reserved_2      = ((int) *(nat+4) ) & 0xff;
	Encapsulation   = ((int) *(nat+5) ) & 0xff;
	reserved_3      = ((int) *(nat+6) ) & 0xff;

	printf("Mobile-IP: Mobile UDP Nat Request Extension message dump:\n");
	printf("type           = 0x%x:\n", type);
	printf("length         = %d (bytes)\n", length);
	printf("sub_type       = 0x%x:\n", sub_type);
	printf("reserved_1     = 0x%x:\n", reserved_1);
	printf("reserved_2     = 0x%x:\n", reserved_2);
	printf("Encapsulation  = 0x%x:\n", Encapsulation);
	printf("reserved_3     = 0x%x:\n", reserved_3);
}


void 
dumpNatRep(char *nat)
{
	int type;
	int sub_type;
	int length;
	int replyCode;
	int reserved;
	int keepAliveInterval;

	if (!nat) {
		printf("dumpNatRep(): nat == NULL!  (Good thing I checked...)\n");
		return;
	}

	type              = ((int) *nat) & 0xff;
	length            = ((int) *(nat+1)) & 0xff;
	sub_type          = ((int) *(nat+2) ) & 0xff;
	replyCode         = ((int) *(nat+3) ) & 0xff;
	reserved          = ((int) *(nat+4) ) & 0xff;   
	//keepAliveInterval = ((int) *(nat+5) ) & 0xff;

	printf("Mobile-IP: Mobile UDP Nat Reply Extension message dump:\n");
	printf("type               = 0x%x:\n", type);
	printf("length             = %d (bytes)\n", length);
	printf("sub_type           = 0x%x:\n", sub_type);
	if (replyCode == 0)
		printf("replyCode          = 0x%x:(REPLY OK)\n", replyCode);
	else
		printf("replyCode          = 0x%x:\n", replyCode);
	printf("flags              = 0x%x:\n", reserved);
	printf("keepAliveInterval  = %d:\n"
	                             , ntohs(((struct mse_nat_ext_rep *)nat)
	                                ->keepAliveInterval));
}
