
/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * unixsock.c
 *
 * unix socket message passing system.
 *
 * 2 processes can use this to exchange messages via UNIX
 * family SOCK_DGRAM messages.  The assumption is that
 * the mobile ip daemon will be the server and the mobystat
 * utility will be the client.  The client will send a
 * request message to the server and the server will return
 * binary data so that the client can format it and print it out.
 *
 *
 * Basic assumption.  Each side has its own socket.  The server
 * sends messages to the client's socket.  The client sends
 * messages to the server's socket.  Both sides must a priori
 * know the socket filename.
 *
 * routines:
 *
 * initUnixSocket(char *whoami, int bepolite)
 *	initialize unix socket, bind whoami as socket filename.
 *	exit if socket exists and bepolite==TRUE, else unlink previously
 *	existing socket.
 * sendUnixMessage(char *sockname, char *buf, int size) 
 *	send message to socket associated with sockname.
 *	If client, send message to server socket (name) and vice versa.
 * recvUnixMessage(char *buf, int bufsize)
 *	if client, recv message from server, else vice versa
 *
 * See TESTMAIN main() below for runtime details.
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <string.h>
#include <stdio.h>

#include "unixsock.h"
#include "mip_log.h"

extern char *progname;

static int unixSocket;

/*
 * bind a unix datagram socket to sockfile
 *
 * returns:
 *	-1: failed
 *	0-n: socket descriptor value
 */
initUnixSocket(char *whoami, int bepolite)
{
	int sd;
	int rc;
	int mask;
	struct sockaddr_un self;
	extern int errno;
	struct stat sbuf;
 
	sd = socket(AF_UNIX, SOCK_DGRAM, 0);
	if (sd < 0) {
		syslog( LOG_ERR, "initUnixSocket(): socket (%m)");
		return(-1);
	}

	/* Set close-on-exec flag to prevent fd leaks. */
	if (fcntl(sd, F_SETFD, 1) < 0) {
		syslog(LOG_INFO, "initUnixSocket(): F_SETFD (%m)");
		/* not catastrophic, don't exit */
	}

	self.sun_len = strlen(whoami);
	self.sun_family = AF_UNIX;
	strcpy(self.sun_path, whoami);


	/* politeness is required by statistics gathering programs
	 * and is improper for daemons.
	 */
	if (bepolite) {
		/*
		 * politeness: exit if the socket exists.
		 */
		if (stat(whoami, &sbuf) == 0) {
		    printf("%s: socket busy.\n", progname);
		    exit(0);
		}
	} else {
		/* blast socket file, in case it is still there
		*/
		unlink(whoami);
	}

	errno = 0;
	rc = bind(sd, (struct sockaddr *)&self, sizeof(struct sockaddr_un));
	if (rc < 0) {
		syslog( LOG_ERR, "initUnixSocket(): bind (%m).");
		return(-1);
	}
	if ( chmod( whoami, US_MODE ) < 0 ) {
		syslog( LOG_ERR, "initUnixSocket(): chmod (%m).");
		return(-1);
	}

	unixSocket = sd;
	return(sd);
}

/*
 * send a message to the server bound to sockfile
 *
 * returns: -1, failed
 *	     0, ok
 */
int
sendUnixMessage(char *sockfile, char *buf, int size) 
{
	struct sockaddr_un to;
	int rc;
	extern int errno;

	to.sun_family = AF_UNIX;
	strcpy(to.sun_path, sockfile);
	to.sun_len = strlen(sockfile);

again:
	rc = sendto(unixSocket, buf, size, 0,  (struct sockaddr *)
		&to, sizeof (struct sockaddr_un));
	if ((rc < 0) && (errno == ENOBUFS))
		goto again;

	if (rc < 0) {
		syslog( LOG_ERR, "sendUnixMessage(): sendto failed (%m).");
		return(-1);
	}
	return(0);
}

/*
 * recvUnixMessage
 *
 * 	recv message on UNIX datagram socket that bind
 *	was performed on.
 *
 * returns: -1, failed
 *	     rc > 0, amount actually received
 */
int
recvUnixMessage(char *buf, int bufsize)
{
	int size = sizeof(struct sockaddr_un);
	int rc;
	struct sockaddr_un from;

	rc = recvfrom(unixSocket, buf, bufsize, 0, (struct sockaddr *)&from, &size);
	if (rc < 0) {
		syslog( LOG_ERR, "recvUnixMessage(): recvfrom failed (%m).");
		return(-1);
	}
	return(rc);
}

#ifdef TESTMAIN
main()  
{
	char sbuf[1024];
	char cbuf[1024];
	char rbuf[1024];
	int size;
	int rsize;
	char *smsg = "this is the server i have data for you!";
	char *cmsg = "this is the client! i have a request!\n";
	struct sockaddr_un client;
	struct sockaddr_un dummy;

	setbuf(stdout, NULL);

	/* parent is server and reads and writes a message
	*/
#ifdef PARTIALTEST
	initUnixSocket("1dirtystinky", FALSE);
#else
	initUnixSocket(MOBYSOCK_SERVER, FALSE);
#endif
#ifndef PARTIALTEST
	if (fork() == 0) {
		int rc;
		close(unixSocket);
		initUnixSocket(MOBYSOCK_CLIENT, FALSE);

		strcpy(sbuf, cmsg);
		size = strlen(cmsg);
		printf("client sends message\n");
		sendUnixMessage(MOBYSOCK_SERVER, sbuf, size);
		printf("client recvs message\n");
		rc = recvUnixMessage(cbuf, 1024);
		printf("client got: %d=%s\n", rc,cbuf);
		exit(0);
	}
	printf("parent about to recv message\n");
	rsize = recvUnixMessage(rbuf, 1024);
	printf("parent msg size %d, msg=%s\n", rsize, rbuf);
	strcpy(sbuf, smsg);
	size = strlen(smsg);
	printf("server sends message\n");
	sendUnixMessage(MOBYSOCK_CLIENT, sbuf, size);
#endif
}
#endif
