
/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * vlist.c
 *
 * visitor list module.  Implement a linked list of "visitor"
 * entries to be used by ha or fa agent.  The list head
 * is maintained by the caller.
 *
 * routines include:
 *	struct *vlist
 *	vlist_search(struct *vlist_head, struct in_addr mn);
 *
 * Module should hide how list is constructed from callers.
 * XXX. Current search mechanism is simple linear search.  Should
 * eventually be replaced with hash search on ip address.
 *
 * XXX. more important, sort entries according to delta lifetimes
 * ala callout list.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netinet/ip.h>

#include "mipstate.h"
#include "mipconfig.h"
#include "mip_icmp.h"
#include "route.h"
#include "vlist.h"

/*
 * search vlist specified to see if mn is in list.
 *
 * returns:
 *	struct vlist * - pointer to vlist entry if found
 *	NULL - no such entry.
 */
static struct vlist *prev;

struct vlist *
vlist_search(struct vlist *head, struct in_addr mn)
{
	prev = head;
	while(head) {
		if (((unsigned long)head->mn.s_addr) == ((unsigned long)mn.s_addr))
			return(head);
		prev = head;
		head = head->next;
	}
	return(NULL);
}

/* create an entry and insert the node with the key mn
 *
 * returns:
 *	struct *vlist - pointer to vlist entry if found
 *	NULL - out of buffer space
*/
struct vlist *
vlist_create(struct vlist **head, struct in_addr mn)
{
	struct vlist *n = NULL;
	struct vlist *v = malloc(sizeof(struct vlist));
	if (!v) {
		return(NULL);
	}
	bzero(v, sizeof(struct vlist));

	v->mn = mn;	/* store key */

	/* for now just put at head of list
	*/
	if (*head)
		n = *head;
	*head = v;
	v->next = n;
		
	return(v);
}

/*
 * vlist_delete
 *	delete a list entry
 * returns (indirectly only):
 *	**head - reset if it was pointing to deleted item
 */
void
vlist_delete(struct vlist **head, struct in_addr mn)
{
	struct vlist *v = vlist_search(*head, mn);
	if (v == NULL)
		return;
	if (*head == v) {
		*head = (*head)->next;
		free(v);
		return;
	}
	prev->next = v->next;
	free(v);
	return;
}

static struct vlist *whead;

void
vlist_initwalk(struct vlist *head)
{
	whead = head;
}

struct vlist *
vlist_nextwalk()
{
	struct vlist *cur = whead;
	if (whead == NULL)
		return(NULL);
	whead = whead->next;
	return(cur);
}

void
vlist_dump(struct vlist *head)
{
	int i = 0;
	vlist_initwalk(head);
	while(1) {
		head = vlist_nextwalk();
		if (!head)
			return;
		printf("---------------------\n");
		vlist_dumpone(head, 1);
		printf("---------------------\n");
	}
}

static
char *getState(int state)
{
	switch(state) {
	case BEACON_NOSTATE:
		return("B_NONE");
	case FA_PENDING:
		return("FA_PEND");
	case FA_PRESENT:
		return("FA_PRES");
	case HA_AWAY:
		return("HA_AWAY");
	case HA_HOME:
		return("HA_HOME");
#ifdef HARP	
	case AT_COHA:
		return("AT_COHA");
#endif
	}
	return(NULL);
}

/*
 * return a string describing the kind of security in use
 * for the link-layer.  This does not apply or make sense
 * for MNs that are AWAY at either the HA or FA.
 */
static
char 
*getAuth(int state, int auth)
{
	if (state == HA_AWAY)
		return("<notapp>");  /* not applicable */
	if (TST(auth,NETAUTH) && TST(auth, ADHOCAUTH))
		return("net+adhoc");
	else if (TST(auth, NETAUTH))
		return("net");
	else if (TST(auth, ADHOCAUTH))
		return("adhoc");
	else 
		return("arp");
}

vlist_shortHeader()
{
	char *s = "%-7s %-15s %-15s %-9s %4s %-16s\n";
	printf(s, "state", "mn", "coa/ha",
		"auth", "life", "location");
}


/*
 * print one vlist entry out, either short form (condensed)
 * or long form (stupid but everything)
 */
int
vlist_dumpone(struct vlist *v, int longflag)
{
	int i;
	char	*secstate = "";

	if (longflag) {
		printf("mn: %s\n", inet_ntoa(v->mn)); 
		printf("state: %d\n", v->vstate);
		printf("routeset: %d\n", v->routeset);
		printf("interface address: %s\n", inet_ntoa(v->i_f));
		printf("src: %s\n", inet_ntoa(v->udp_src.sin_addr));
		printf("ha: %s\n", inet_ntoa(v->ha));
		printf("coa: %s\n", inet_ntoa(v->coa));
		printf("orig lifetime: %d\n", v->orig_lifetime);
		printf("lifetime: %d\n", v->lifetime);
		printf("id[0]: %x\n", v->i.l[0]);
		printf("id[1]: %x\n", v->i.l[1]);
		printf("sequence: %d\n", v->link_sequence);
		printf("sawBeacon: %d\n", v->sawBeacon);
		printf("sawMac: %d\n", v->sawMac);
		printf("auth_type: %x\n", v->auth_type);
		if (v->lsize)
			printf("location %s\n", v->location);
		printf("mac:");
 		for (i=0;i<6;i++)
                	printf("%02x%c", v->mac[i], (i<5)?':':' ');
		printf("\n");
	}
/* HYAR */
	else {
		printf("%-7s ", getState(v->vstate));
		printf("%-15s ", inet_ntoa(v->mn)); 
		switch(v->vstate) {
		case HA_AWAY:
			printf("%-15s ", inet_ntoa(v->coa));
			break;
		case FA_PENDING:
		case FA_PRESENT:
			printf("%-15s ", inet_ntoa(v->ha));
			break;
#ifdef HARP
		case AT_COHA:
			printf("%-15s ", inet_ntoa(v->coa));
			break;
#endif
		default:
			printf("%-15s ", "<notapp>");
			break;
		}
		printf("%-9s ", getAuth(v->vstate, v->auth_type));
		printf("%4d ", v->lifetime);
		if (v->lsize)
			printf("%-16s ", (v->location));
		else
			printf("<none> ");
		printf("\n");
	}
}

#ifdef TESTMAIN
/*
 * test all primitives above
 */
main()
{
	struct vlist *myhead = NULL;
	struct vlist *myhead2 = NULL;
	struct in_addr mn;
	struct vlist *v;

	/* test case 1.  put 5 entries in list and dump
	*/
	mn.s_addr = 0x01010101;
	vlist_create(&myhead, mn);
	mn.s_addr = 0x02020202;
	vlist_create(&myhead, mn);
	mn.s_addr = 0x03030303;
	vlist_create(&myhead, mn);
	mn.s_addr = 0x04040404;
	vlist_create(&myhead, mn);
	mn.s_addr = 0x05050505;
	vlist_create(&myhead, mn);

	vlist_dump(myhead); 

	/* now find all the list entries
	 * but first search for one that is not there.
	 */
	mn.s_addr = 0x01020304;
	v = vlist_search(myhead, mn);
	if (v == NULL) 
		printf("not found, correct!\n");
	else
		printf("found %x\n", v->mn.s_addr);

	mn.s_addr = 0x01010101;
	v = vlist_search(myhead, mn);
	if (v == NULL) 
		printf("not found, incorrect!\n");
	else
		printf("found %x\n", v->mn.s_addr);

	mn.s_addr = 0x05050505;
	v = vlist_search(myhead, mn);
	if (v == NULL) 
		printf("not found, incorrect!\n");
	else
		printf("found %x\n", v->mn.s_addr);

	mn.s_addr = 0x03030303;
	v = vlist_search(myhead, mn);
	if (v == NULL) 
		printf("not found, incorrect!\n");
	else
		printf("found %x\n", v->mn.s_addr);

	/* test series 2.  put 1 entry in list and delete it
	*/

	mn.s_addr = 0x01010101;
	vlist_create(&myhead2, mn);
	vlist_delete(&myhead2, mn);
	vlist_dump(myhead2); 
	v = vlist_search(myhead2, mn);
	if (v == NULL) 
		printf("correct - didn't find it\n");
		
	/* test series 3.  put 2 entries in list and delete first
	*/
	mn.s_addr = 0x02020202;
	vlist_create(&myhead2, mn);
	mn.s_addr = 0x01010101;
	vlist_create(&myhead2, mn);
	vlist_delete(&myhead2, mn);
	printf("list should only have 02020202 in it\n");
	vlist_dump(myhead2); 
	mn.s_addr = 0x02020202;
	v = vlist_search(myhead2, mn);
	if (v == NULL) 
		printf("incorrect - didn't find it\n");
	else
		printf("found %x\n", v->mn.s_addr);

	/* test series 4. delete 3rd entry in original list
	*/
	mn.s_addr = 0x03030303;
	vlist_delete(&myhead, mn);
	printf("following lisit should have 5/4/2/1 in that order\n");
	vlist_dump(myhead); 

	/* test series 5. delete 1st and last entries in list
	*/
	mn.s_addr = 0x01010101;
	vlist_delete(&myhead, mn);
	mn.s_addr = 0x05050505;
	vlist_delete(&myhead, mn);
	printf("following list should only have 4/2 in it\n");
	vlist_dump(myhead); 

	/* final test. put entry back in and print out.
	 * list should have 3/4/2 in that order
	 */
	mn.s_addr = 0x03030303;
	vlist_create(&myhead, mn);
	printf("following list should have 3/4/2 in that order\n");
	vlist_dump(myhead); 
}
#endif

#ifdef TESTDUMP
main()
{
	struct vlist *head = NULL;
	struct vlist *v;
	struct in_addr mn;
	struct in_addr coa;

	mn.s_addr = 0xc1d1e1f1;
	v = vlist_create(&head, mn);
	v->vstate = 0;
	strncpy(v->location, "my pc", 5);
	v->lsize = 5;
	v->lifetime = 1;
	v->coa.s_addr = 0x01020304;
	v->auth_type = 0;

	mn.s_addr = 0x02020202;
	v = vlist_create(&head, mn);
	v->vstate = 1;
	strncpy(v->location, "your pc", 7);
	v->lsize = 7;
	v->coa.s_addr = 0xcccd02ce;
	v->auth_type = 0x1;
	v->lifetime = 2;

	mn.s_addr = 0x03030303;
	v = vlist_create(&head, mn);
	v->vstate = 2;
	strncpy(v->location, "bobs pc", 7);
	v->lsize = 7;
	v->coa.s_addr = 0xacbd02cd;
	v->auth_type = 0x2;
	v->lifetime = 3;

	mn.s_addr = 0x04040404;
	v = vlist_create(&head, mn);
	v->vstate = 3;
	strncpy(v->location, "tena pc", 7);
	v->lsize = 7;
	v->coa.s_addr = 0xcccccc12;
	v->auth_type = 0x3;
	v->lifetime = 4;

	mn.s_addr = 0x05050505;
	v = vlist_create(&head, mn);
	v->vstate = 4;
	strncpy(v->location, "xyxz pc", 7);
	v->lsize = 7;
	v->coa.s_addr = 0xcccccc11;
	v->auth_type = 0x0;
	v->lifetime = 5;

	vlist_shortHeader();
	vlist_initwalk(head);
	while(1) {
		head = vlist_nextwalk();
		if (!head)
			return;
		vlist_dumpone(head, 0);
	}
}
#endif
