/*
 * wisignal.c
 *
 * signal strength based on ip address from ieee/wavelan/wi device
 *	driver.
 *
 * acknowledgements:
 *
 * some code here taken from Bill Paul's wicontrol.c, at least
 * for ioctl access to wi device.  
 *
 * module exists to extract signal strength from wi0 device
 * and map it into 0.100 % range.
 *
 * how:
 *	use ioctl access mechanism to get entire cache.
 *	search cache for desired ip address.
 *	map quality (signal - noise) to 100%.  
 *	note this is *quality*.  
 *	if < 0, return 0
 *	if > 60, return 100
 *	else scale 0..60 in hw values to 0..100 in percent "goodness"
 */

#include <sys/types.h>
#include <sys/cdefs.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <sys/socket.h>

#include <net/if.h>
#include <net/if_var.h>
#include <net/ethernet.h>



#define WICACHE
#if defined(__FreeBSD__) && __FreeBSD_version < 470000
#include <machine/if_wavelan_ieee.h>
/*
#include <machine/if_wi_cache.h>
*/
#else
#include <dev/wi/if_wavelan_ieee.h>
#endif


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <err.h>

static void wi_getval		__P((char *, struct wi_req *));

static void wi_getval(iface, wreq)
	char			*iface;
	struct wi_req		*wreq;
{
	struct ifreq		ifr;
	int			s;

	bzero((char *)&ifr, sizeof(ifr));

	strcpy(ifr.ifr_name, iface);
	ifr.ifr_data = (caddr_t)wreq;

	s = socket(AF_INET, SOCK_DGRAM, 0);

	if (s == -1)
		err(1, "socket");

	if (ioctl(s, SIOCGWAVELAN, &ifr) == -1)
		err(1, "SIOCGWAVELAN");

	close(s);

	return;
}


static struct wi_sigcache *
wi_readcache(iface, items)
	char			*iface;
	int			*items;
{
	static struct wi_req	wreq;
	int 			*wi_sigitems;
	struct wi_sigcache 	*sc;
	char *			pt;
	int 			i;

	if (iface == NULL)
		return(0);

	bzero((char *)&wreq, sizeof(wreq));
	wreq.wi_len = WI_MAX_DATALEN;
	wreq.wi_type = WI_RID_READ_CACHE;

	wi_getval(iface, &wreq);

	wi_sigitems = (int *) &wreq.wi_val; 
	*items = *wi_sigitems;
	pt = ((char *) &wreq.wi_val);
	pt += sizeof(int);
	sc = (struct wi_sigcache *) pt;

	return(sc);
}


/*
 * given ip address, try and get signal strength out of cache
 *
 * returns:
 *	100 when in doubt, else scaled signal value
 *	0..100: signal strength
 */
int
wi_ip_to_signal(struct in_addr bip)
{
	int sigitems;
	int rc;
	int i;
	struct wi_sigcache *sc;

	sc = wi_readcache("wi0", &sigitems);
	if (sigitems == 0)
		return(100);

	for(i = 0; i < sigitems; i++, sc++) {
		if (bip.s_addr == sc->ipsrc) {
			if (sc->quality < 0) {
				return(0);
			}
			if (sc->quality > 60) {
				return(100);
			}
			return((sc->quality/60.0) * 100);
		}
	}
	return(100);
}

#ifdef TESTMAIN

/*
 * print out the cache
 *
 * to compile:
 *	cc -o wisignal -DTESTMAIN wisignal.c 
 */
int
main(argc, argv)
	int argc;
	char *argv[];
{
	int sigitems;
	int rc;
	int i;
	struct wi_sigcache *sc;

	sc = wi_readcache("wi0", &sigitems);
	if (sigitems == 0) {
		printf("no cache found\n");
		return;
	}
	for (i = 0; i < sigitems; i++, sc++) {
		printf("[%d/%d]:", i+1, sigitems);
		printf(" %02x:%02x:%02x:%02x:%02x:%02x,",
		  		    	sc->macsrc[0]&0xff,
		  		    	sc->macsrc[1]&0xff,
		   		    	sc->macsrc[2]&0xff,
		   			sc->macsrc[3]&0xff,
		   			sc->macsrc[4]&0xff,
		   			sc->macsrc[5]&0xff);
        	printf(" %d.%d.%d.%d,",((sc->ipsrc >> 0) & 0xff),
				        ((sc->ipsrc >> 8) & 0xff),
				        ((sc->ipsrc >> 16) & 0xff),
				        ((sc->ipsrc >> 24) & 0xff));
		printf(" sig: %d, noise: %d, qual: %d\n",
		   			sc->signal,
		   			sc->noise,
		   			sc->quality);
		sc++;
	}
}
#endif








