
/*
 * Copyright (c) 1996 Portland State University 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Portland State University.
 * 4. The name of Portland State University may not be used to endorse 
 *    or promote products derived from this software without specific 
 *    prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR/S ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR/S BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * wlsignal.c
 *
 * Module provides interface to wavelan driver signal strength
 * cache.  The cache allows a daemon or application to read
 * out signal strength associated with mac src and ip src
 * addresses.  
 *
 * The module uses the kernel virtual memory interface to
 * access the cache (array of structures) kept in the driver.
 *
 * interface routines:
 * int init_wlcache() - called to init the kvm descriptor 
 * int read_wlcache(int *sigitems, struct w_sigcache *wsc)
 *	- read out sigitems worth of cache structures into a local
 *	array.  sigitems is the current count 0..sigitems-1 which
 *	is <= the max # of items in the cache
 *
 * if compiled with -DTESTMAIN, we can produce a one-shot
 * application that will read and print out the cache.
 *
 * # wlsignal - utility to dump sigitems in the signal cache
 *
 * syntax: 
 *	wlsignal - will print entire cache one time and exit
 */ 
#include <sys/param.h>
#include <sys/file.h>
#include <sys/protosw.h>
#include <sys/socket.h>

#include <netinet/in.h>
#include <arpa/inet.h>
#include <ctype.h>
#include <errno.h>
#include <kvm.h>
#include <limits.h>
#include <netdb.h>
#include <nlist.h>
#include <paths.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <err.h>

/* the if_wl_wavelan.h include file includes the
 * cache structure definition
#include "/usr/src/sys/i386/include/if_wl_wavelan.h"  
 */
#include <machine/if_wl_wavelan.h>  

/*
 * w_sigitems - current number of signal cache entries
 * w_sigcache - the array of cache elements
 */
static
struct nlist nl[] = {
#define	N_WSIGITEMS	0
	{ "_w_sigitems" },
#define	N_WSIGCACHE	1
	{ "_w_sigcache" },
	"",
};

static kvm_t *kvmd;

static char *nlistf = NULL;
static char *memf = NULL;
static char buf[_POSIX2_LINE_MAX];

/* test main turns this on, else off for module use
*/
static int wldebug = 0;

/*
 * returns:
 *	0: ok
 *     -1: failed
 */
int
init_wlcache()
{
	kvmd = kvm_openfiles(nlistf, memf, NULL, O_RDONLY, buf);
	if (kvmd == NULL) {
		if (wldebug)
			printf("kvm_open failed: %s", buf);
		return(-1);
	}
	if (kvm_nlist(kvmd, nl) < 0) {
		if (wldebug)
			printf("kvm_nlist failed: %s\n", kvm_geterr(kvmd));
		return(-1);
	}
	if (nl[0].n_value == 0) {
		if (wldebug)
			printf("kvm_nlist failed: 0 for n_value\n");
		return(-1);
	}
	return(0);
}

/*
 * returns:
 *	0: ok
 *     -1: failed
 */
int 
read_wlcache(int *sigitems, struct w_sigcache *wsc)
{
	int rc;

	/* read actual count of valid items, which may be less
	 * than the max size of the array
	 */
	rc = kvm_read(kvmd, nl[N_WSIGITEMS].n_value, sigitems, 
		sizeof(int));
	if (rc < 0) {
		if (wldebug)
			printf("kvm_read of w_sigitems failed\n");
		return(-1);
	}
         
	/* now read that many items out of the cache
	*/
	rc = kvm_read(kvmd, nl[N_WSIGCACHE].n_value, wsc, 
	        	*sigitems * sizeof(struct w_sigcache));
	if (rc < 0) {
		if (wldebug)
			printf("kvm_read of cache failed\n");
		return(-1);
	}
	return(0);
}

/*
 * given ip address, try and get signal strength out of cache
 *
 * returns:
 *	-2: failed
 *	-1: didn't find it in cache (which is tough if it's working)
 *	0..100: signal strength
 */
int
wl_ip_to_signal(struct in_addr bip)
{
	static int cacheInit = 0;
	int sigitems;
	struct w_sigcache wsc[MAXCACHEITEMS];
	int rc;
	int i;

	if (cacheInit == 0) {
		rc = init_wlcache();
		if (rc < 0) {
			return(-2);
		}
		cacheInit = 1;
	}
	rc = read_wlcache(&sigitems, &wsc[0]);
	if (rc < 0) {
		return(-2);
	}

	for(i = 0; i < sigitems; i++) {
		if (bip.s_addr == wsc[i].ipsrc) {
			if ( wsc[i].signal > 40)
				return(100);
			return((wsc[i].signal/40.0) * 100);
		}
	}
	return(-1);
}

#ifdef TESTMAIN

/*
 * print out the cache
 *
 * to compile:
 *	cc -o wlsignal -DTESTMAIN wlsignal.c -lkvm
 */
int
main(argc, argv)
	int argc;
	char *argv[];
{
        int rc;
        int i;
	int sigitems;
	struct w_sigcache wsc[MAXCACHEITEMS];

	wldebug = 1;

	rc = init_wlcache();
	if (rc < 0)
		exit(1);

	rc = read_wlcache(&sigitems, &wsc[0]);
	if (rc < 0)
		exit(1);

	/*
	printf(" %d sigitems in the signal_cache\n", sigitems);
	*/
	printf("signal range 0..63: silence 0..63: quality 0..15\n");

	/* after you read it, loop through structure,i.e. wsc
         * print each item:
	 */
	for(i = 0; i < sigitems; i++) {
		printf("cache index [%d:%d]>\n", i+1, sigitems);
        	printf("\tip src: %d.%d.%d.%d,",((wsc[i].ipsrc >> 0) & 0xff),
				        ((wsc[i].ipsrc >> 8) & 0xff),
				        ((wsc[i].ipsrc >> 16) & 0xff),
				        ((wsc[i].ipsrc >> 24) & 0xff));
		printf(" mac: %02x:%02x:%02x:%02x:%02x:%02x\n",
		  		    	wsc[i].macsrc[0]&0xff,
		  		    	wsc[i].macsrc[1]&0xff,
		   		    	wsc[i].macsrc[2]&0xff,
		   			wsc[i].macsrc[3]&0xff,
		   			wsc[i].macsrc[4]&0xff,
		   			wsc[i].macsrc[5]&0xff);
		printf("\tsignal: %d, silence: %d, quality: %d, ",
		   			wsc[i].signal,
		   			wsc[i].silence,
		   			wsc[i].quality);
		printf("snr: %d\n", wsc[i].snr);
	}

}
#endif

