/**************************************************************************
 The GUI for displaying the mobile node state and statistics
 Programmer: Ginny Mak
 Location: Portland, Oregon.
 Date: Nov 20, 2001.
**************************************************************************/ 
#include<stdio.h>
#include<stdlib.h>
#include<unistd.h>
#include<string.h>
#include<FL/Fl.h>
#include<Fl/Fl_Window.h>
#include<FL/Fl_Box.h>
#include<FL/Enumerations.h>
#include<FL/fl_draw.h>
#include<FL/Fl_Output.h>
#include<FL/Fl_Button.h>
#include<FL/Fl_Menu.h>
#include<FL/Fl_Menu_Bar.h>
#include<FL/Fl_Menu_Item.h>
#include<FL/Fl_Scroll.h>
#include<FL/Fl_Counter.h>
#include<FL/Fl_Value_Slider.h>
#include "cMnStat.h"

#define FAILURE -1
#define SUCCESS 0

#define BEACON_FIELDS 9  // number of fields in beacon statistics
#define UDP_FIELDS 8     // number of fields in udp statistics
 
#define MAXNODES 20      // maximun number of nodes displayed
#define MAXIFS 40        
  
//------------------------------------------------------------

/**********************************
 * VARIABLES DECLARATIONS         *
 **********************************/
char beaconField[9][10];
char udpField[8][10];

char *iflist[MAXIFS];
char ifname[128];
int prism2 = 0;

int agentsCount = 0;
int nodesCount = 0;

char agentBeacon[MAXNODES][100];
int agentSignal[MAXNODES];
char agentLocation[MAXNODES][100];

char nodeBeacon[MAXNODES][100];
int nodeSignal[MAXNODES];
char nodeLocation[MAXNODES][100];

double pollInterval = 2.0;

// this object holds all operations for collecting
// the state and statistics data
cMnStat *mnstat;      
mnstate mnState;      // for storing state information

//---------------------------------------------------------------------

/***********************************
 * FUNCTIONS DECLARATION           *
 ***********************************/

void message( char *msg );
void ask( char *msg );
Fl_Color genColor( int level );
char *getFa( char *s );
int getState();
int getStat();
void updateDisplay( void* );
void startWindow();


//------------------------------------------------------------------

/******************************************************
 polling interval window declaration and definitions
******************************************************/
void pollInterval_cb(Fl_Widget *w, void  *)
{
  Fl_Counter * cw = (Fl_Counter *)w;
  pollInterval = cw->value();
}
  
 
class counterWindow:public Fl_Window {
public:
  counterWindow(int w, int h, const char *label=0 );
  ~counterWindow();
private:
  void initialize();
  Fl_Counter *pollIntervalCounter;
};

counterWindow::counterWindow(int w, int h, const char *label=0):Fl_Window(w,h,label)
{
  initialize();
}


counterWindow::~counterWindow()
{
  delete pollIntervalCounter;
}

void counterWindow::initialize()
{
  this->set_modal();
  this->color(255);
  pollIntervalCounter = new Fl_Counter(20, 20, 150,20,"Polling Interval (sec)");
  pollIntervalCounter->type(FL_SIMPLE_COUNTER);
  pollIntervalCounter->value(pollInterval);
  pollIntervalCounter->minimum(1.0);
  pollIntervalCounter->maximum(100.0);
  pollIntervalCounter->step(1.0);
  pollIntervalCounter->color(FL_YELLOW);
  pollIntervalCounter->labelcolor(FL_BLACK);
  pollIntervalCounter->selection_color(255);
  pollIntervalCounter->labelfont(FL_HELVETICA_BOLD);
  pollIntervalCounter->callback(pollInterval_cb);
}

counterWindow *myCounterWindow;

//------------------------------------------------------------------------

/********************************************************
 Class stateWindow declarations
********************************************************/

Fl_Menu_Item mnstatOptions[] = {
  {"Stat", 0, 0, 0, FL_SUBMENU },
  {"beacon"},
  {"mn Statistics"},
  {0},
  {"Nodes"},
  {"Agents"},
  {"Poll Int"},
  {0}
};

class stateWindow:public Fl_Window
{
public:
  stateWindow(int w, int h, char *label=0 );
  ~stateWindow();
  void setState( char *state );
  void setFa( char *fa );
  void setCoa( char *coa );
  void setLocation( char *location );
  void setMeter( double aValue );
  void draw();
  
 
private:
  int initialize();
  Fl_Menu_Bar *menuBar;
  Fl_Output *mstate;
  Fl_Output *mfa;
  Fl_Output *mcoa;
  Fl_Output *mlocation;
  Fl_Value_Slider *msignal;
};

stateWindow *myState;

// ----------------------------------------------------------------------
/********************************************
 * class agentWindow declarations
 ********************************************/
class agentWindow:public Fl_Window
{
public:
  agentWindow(int w, int h, char *label=0 );
  ~agentWindow();
  void setBeacon( int index, char *aValue );
  void setLocation( int index, char *aValue );
  void setMeter( int index, double value );
  void resetBeacon( int index );
  void resetLocation( int index );
  void resetMeter( int index );

private:
  int initialize();
  Fl_Scroll *scroll;
  Fl_Output *agents[MAXNODES][2];
  Fl_Output *dummy[MAXNODES];
  Fl_Value_Slider *meter[MAXNODES];
  
};

agentWindow *myAgent;

// ----------------------------------------------------------------------
/**************************************
 * class nodeWindow declarations
 **************************************/
class nodeWindow:public Fl_Window
{
public:
  nodeWindow(int w, int h, char *label=0 );
  ~nodeWindow();
  void setBeacon( int index, char *aValue );
  void setLocation( int index, char *aValue );
  void setMeter( int index, double value );
  void resetBeacon( int index );
  void resetLocation( int index );
  void resetMeter( int index );

private:
  int initialize();
  Fl_Scroll *scroll;
  Fl_Output *nodes[MAXNODES][2];
  Fl_Output *dummy[MAXNODES];
  Fl_Value_Slider *meter[MAXNODES];
};

nodeWindow *myNode;


//-------------------------------------------------------------------------
/*******************************************
 * class beaconWindow declarations
 *******************************************/
class beaconWindow:public Fl_Window
{
public:
  beaconWindow(int w, int h, char *label=0 );
  ~beaconWindow();
  void setBeaconField( int index, char *aValue);

private:
  int initialize();
  Fl_Output *beacons[9];
};

beaconWindow *myBeacon;


//----------------------------------------------------------------------------
/*********************************************
 * class udpStatWindow declarations
 *********************************************/
class udpStatWindow:public Fl_Window
{
public:
  udpStatWindow(int w, int h, char *label=0 );
  ~udpStatWindow();
  void setUdpField( int index, char *aValue );

private:
  int initialize();
  Fl_Output *udpStats[9];
};

udpStatWindow *myUdpStat;

//---------------------------------------------------------------

/*********************************************************
 * Class stateWindow : definition and callback functions
 *********************************************************/
void mnstat_cb( Fl_Widget*, void *data ) {
  ask( "Want to quit?" );
}

void statMenu_cb( Fl_Widget *w, void *data )
{
  Fl_Menu_ *mw = (Fl_Menu_*)w;
  const Fl_Menu_Item *m = mw->mvalue();
  if( strcmp( m->label(), "Poll Int" ) == 0 ) {
	myCounterWindow->show();
  }
  else if( strcmp( m->label(), "beacon" ) == 0 ) {
    myBeacon->show();
  }
  else if( strcmp( m->label(), "mn Statistics" ) == 0 ) {
    myUdpStat->show();
  }
  else if( strcmp( m->label(), "Nodes" ) == 0 ) {
    myNode->show();
  }
  else if( strcmp( m->label(), "Agents" ) == 0 ) {
    myAgent->show();
  }

}

stateWindow::stateWindow( int w, int h, char *label=0):Fl_Window( w, h, label )
{
  initialize();
}
  
stateWindow::~stateWindow()
{
  delete menuBar;
  delete mstate;
  delete mfa;
  delete mcoa;
  delete msignal;
  delete mlocation;
}

int stateWindow::initialize()
{
  double min = 0;
  double max = 100.0;
  double aStep = 0;
  
  this->callback( mnstat_cb );
  this->color(FL_BLACK);
  menuBar = new Fl_Menu_Bar( 0, 0, 200, 30, 0);
  menuBar->menu( mnstatOptions );
  menuBar->textfont( FL_HELVETICA_BOLD );
  menuBar->textsize( 10 );
  menuBar->textcolor( FL_BLACK );
  menuBar->color( 255 );
  menuBar->callback( statMenu_cb );

  mstate = new Fl_Output( 0, 30, 200, 20, 0 );
  mstate->color( 100 );
  mstate->textcolor( 255 );
  mstate->value( mnState.state );
  mstate->textsize( 12 );
  
  mfa = new Fl_Output( 0, 50, 200, 20, 0 );
  mfa->color( 255 );
  mfa->textcolor( FL_BLACK );
  mfa->value( mnState.fa );
  mfa->textsize( 12 );

  mcoa = new Fl_Output( 0, 70, 200, 20, 0 );
  mcoa->color( 100 );
  mcoa->textcolor( 255 );
  mcoa->value( mnState.coa );
  mcoa->textsize( 12 );

  mlocation = new Fl_Output( 0, 90, 200, 20, 0 );
  mlocation->color( 255 );
  mlocation->textcolor( FL_BLACK );
  mlocation->value( mnState.location );
  mlocation->textsize( 12 );

  msignal = new Fl_Value_Slider( 0, 112, 200, 20, 0 );
  msignal->type( FL_HOR_FILL_SLIDER );
  msignal->color( 255 );
  msignal->labelsize( 10 );
  msignal->labelcolor( 255 );
  msignal->range( max, min );
  msignal->step( aStep );
  msignal->minimum( min );
  msignal->maximum( max );
  double signal = mnState.signal;
  msignal->value( signal );
  msignal->selection_color( genColor( (int)signal ) );
  this->resizable();
  return SUCCESS;
}

inline void stateWindow::setState( char * aState )
{
  mstate->value( aState );
}

inline void stateWindow::setFa( char *aFa )
{
  mfa->value( aFa );
}

inline void stateWindow::setCoa( char *aCoa )
{
  mcoa->value( aCoa );
}

inline void stateWindow::setLocation( char *aLocation )
{
  mlocation->value( aLocation );
}

inline void stateWindow::setMeter( double aValue ) 
{
  msignal->value( aValue );
  msignal->selection_color( genColor( (int)aValue ) );
}

void stateWindow::draw()
{
  // draw the color coded graduation 
  // and graduation labels at the bottom of the window
  Fl_Window::draw();
  fl_color(FL_RED);
  #ifdef IPAQ
  fl_rectf(40,280,40,4);
  #else
  fl_rectf(40,296,40,4);
  #endif
  fl_color(FL_BLUE);
  #ifdef IPAQ
  fl_rectf(80,280,40,4);
  #else
  fl_rectf(80,296,40,4);
  #endif
  fl_color(FL_YELLOW);
  #ifdef IPAQ
  fl_rectf(120,280,40,4);
  #else
  fl_rectf(120,296,40,4);
  #endif 
  fl_color(FL_GREEN);
  #ifdef IPAQ
  fl_rectf(160,280,40,4);
  #else
  fl_rectf(160,296,40,4);
  #endif
  fl_color(255);
  fl_font(FL_HELVETICA,10);
  #ifdef IPAQ
  fl_draw("0",40,270);
  fl_draw("25", 75, 270);
  fl_draw("50", 110, 270);
  fl_draw("75", 148, 270);
  fl_draw("100", 172,270);
  #else
  fl_draw("0",40,290);
  fl_draw("25", 75, 290);
  fl_draw("50", 115, 290);
  fl_draw("75", 155, 290);
  fl_draw("100", 180,290);
  #endif
  
}
 
//---------------------------------------------------------------------------

/*************************************
 * class agentWindow: definition
 *************************************/

agentWindow::agentWindow( int w, int h, char *label=0 ):Fl_Window( w, h, label )
{
  initialize();
}

agentWindow::~agentWindow()
{
  for( int i = 0; i < MAXNODES; i++ ) {
    for( int j = 0; j < 2; j++ )
      delete agents[i][j];
    delete meter[i];
    delete dummy[i];
  }
  delete scroll;
}


int agentWindow::initialize()
{
  double min = 0;
  double max = 100.0;
  double aStep = 0;

  this->set_modal();
  this->color(FL_BLACK);
  this->box( FL_NO_BOX );
  scroll = new Fl_Scroll(0,0,200, 300);

  scroll->scrollbar.align(Fl_Scroll::BOTH+Fl_Scroll::BOTH_ALWAYS);
  scroll->box( FL_NO_BOX );
  scroll->align(FL_ALIGN_RIGHT);
  scroll->hscrollbar.align(FL_ALIGN_BOTTOM);
  scroll->color(FL_BLACK);

  int step = 20;
  for( int i = 0; i < MAXNODES; i++ ) {
    agents[i][0] = new Fl_Output( 0, 0+i*3*step, 300, 20, 0 );
    agents[i][1] = new Fl_Output( 0, 20+i*3*step, 300, 20, 0 );
    meter[i] = new Fl_Value_Slider( 0, 40+i*3*step, 200, 20, 0 );

    // dummy sits next to the meter so that the display
    // looks good when scrolled horizontally
    dummy[i] = new Fl_Output( 200, 40+i*3*step, 100, 20 , 0 );
    meter[i]->type( FL_HOR_FILL_SLIDER );
    meter[i]->color( 255 );
    meter[i]->labelsize( 10 );
    meter[i]->labelcolor( 255 );
    meter[i]->range( max, min );
    meter[i]->step( aStep );
    meter[i]->minimum( min );
    meter[i]->maximum( max );

	
    agents[i][0]->textsize(12);
    agents[i][1]->textsize(12);
    scroll->add( agents[i][0] );
    scroll->add( agents[i][1] );
     scroll->add( dummy[i] );
	
    agents[i][0]->color(100);
    agents[i][1]->color(100);
    dummy[i]->color(100);
    agents[i][0]->textcolor(255);
    agents[i][1]->textcolor(255);
  }
  scroll->end();
  this->resizable(scroll);
  this->end();
 
  return SUCCESS;
}

inline void agentWindow::setBeacon( int index, char *aValue )
{
  agents[index][0]->value( aValue );
}

inline void agentWindow::setLocation( int index, char *aValue )
{
  agents[index][1]->value( aValue );
}

inline void agentWindow::setMeter( int index, double aValue )
{
  meter[index]->value( aValue );
  meter[index]->selection_color( genColor( (int)aValue) );
}
  
inline void agentWindow::resetBeacon( int index )
{
  agents[index][0]->value( "" );
}

inline void agentWindow::resetLocation( int index )
{
  agents[index][1]->value( "" );
}

inline void agentWindow::resetMeter( int index )
{
  meter[index]->value( 0 );
}
  
//-------------------------------------------------------------------------------

/***********************************
 * class nodeWindow: definition
 ***********************************/
nodeWindow::nodeWindow( int w, int h, char *label=0 ):Fl_Window( w, h, label )
{
  initialize();
}

nodeWindow::~nodeWindow()
{
  for( int i = 0; i < MAXNODES; i++ ) {
    for( int j = 0; j < 2; j++ )
      delete nodes[i][j];
    delete meter[i];
    delete dummy[i];
  }
  delete scroll;
}


int nodeWindow::initialize()
{
  double min = 0;
  double max = 100.0;
  double aStep = 0;

  this->set_modal();
  this->color(255);
  this->box( FL_NO_BOX );
  scroll = new Fl_Scroll(0,0,200, 300);

  scroll->scrollbar.align(Fl_Scroll::BOTH+Fl_Scroll::BOTH_ALWAYS);
  scroll->box( FL_NO_BOX );
  scroll->align(FL_ALIGN_RIGHT);
  scroll->hscrollbar.align(FL_ALIGN_BOTTOM);
  scroll->color(FL_BLACK);

  int step = 20;
  int number = MAXNODES;
  for( int i = 0; i < number; i++ ) {
    nodes[i][0] = new Fl_Output( 0, 0+i*3*step, 300, 20, 0 );
    nodes[i][1] = new Fl_Output( 0, 20+i*3*step, 300, 20, 0 );
 
    // dummy sits next to the meter so that the display looks 
    // good when scrolled horizontally
    dummy[i] = new Fl_Output( 200, 40+i*3*step, 100, 20, 0 );
    meter[i] = new Fl_Value_Slider( 0, 40+i*3*step, 200, 20, 0 );
    meter[i]->type( FL_HOR_FILL_SLIDER );
    meter[i]->color( 255 );
    meter[i]->labelsize( 10 );
    meter[i]->labelcolor( 255 );
    meter[i]->range( max, min );
    meter[i]->step( aStep );
    meter[i]->minimum( min );
    meter[i]->maximum( max );
    nodes[i][0]->textsize(12);
    nodes[i][1]->textsize(12);
    scroll->add( nodes[i][0] );
    scroll->add( nodes[i][1] );
    scroll->add( dummy[i] );
    
    nodes[i][0]->color(253);
    nodes[i][1]->color(253);
    dummy[i]->color(253);
    nodes[i][0]->textcolor(FL_BLACK);
    nodes[i][1]->textcolor(FL_BLACK);
  }
  scroll->end();
  this->resizable(scroll);
  this->end();
  return SUCCESS;
}

inline void nodeWindow::setBeacon( int index, char *aValue )
{
  nodes[index][0]->value( aValue );
}


inline void nodeWindow::setLocation( int index, char *aValue )
{
  nodes[index][1]->value( aValue );
} 

inline void nodeWindow::setMeter( int index, double aValue )
{
  meter[index]->value( aValue );
  meter[index]->selection_color( genColor( (int)aValue) );
}

inline void nodeWindow::resetBeacon( int index )
{
  nodes[index][0]->value( "" );
}

inline void nodeWindow::resetLocation( int index )
{
  nodes[index][1]->value( "" );
}

inline void nodeWindow::resetMeter( int index )
{
  meter[index]->value( 0 );
}

//------------------------------------------------------------------------------

/**************************************
 * class beaconWindow: definition
 **************************************/
beaconWindow::beaconWindow( int w, int h, char *label=0 ):Fl_Window( w, h, label )
{
  initialize();
}

beaconWindow::~beaconWindow()
{
  for( int i = 0; i < 9; i++ ) {
    delete beacons[i];
  }
}


int beaconWindow::initialize()
{
  this->set_modal();
  this->color(100);
  
  Fl_Box *label1 = new Fl_Box( 0, 10, 130, 30, "Total beacons sent:" );
  label1->labelsize(8);
  label1->labelcolor(255);

  Fl_Box *label2 = new Fl_Box( 0, 40, 130, 30, "Total beacons recvd:" );
  label2->labelsize(8);
  label2->labelcolor( 255 );

  Fl_Box *label3 = new Fl_Box( 0, 70, 130, 30, "Total auth. beacon recvd:" );
  label3->labelsize(8);
  label3->labelcolor( 255 );

  Fl_Box *label4 = new Fl_Box( 0, 100, 130, 30, "Valid beacons recvd:" );
  label4->labelsize(8);
  label4->labelcolor( 255 );

  Fl_Box *label5 = new Fl_Box( 0, 130, 130, 30, "Valid non-auth. beacons recvd:" );
  label5->labelsize(8);
  label5->labelcolor( 255 );

  Fl_Box *label6 = new Fl_Box( 0, 160, 130, 30, "Valid auth. beacons recvd:" );
  label6->labelsize(8);
  label6->labelcolor( 255 );

  Fl_Box *label7 = new Fl_Box( 0, 190, 130, 30, "Total ICMP protocol errors:" );
  label7->labelsize(8);
  label7->labelcolor( 255 );

  Fl_Box *label8 = new Fl_Box( 0, 220, 130, 30, "Total authentication failure:" );
  label8->labelsize(8);
  label8->labelcolor( 255 );
  
  Fl_Box *label9 = new Fl_Box( 0, 250, 130, 30, "Total agents ignored\n(lack authentication):" );
  label9->labelsize(8);
  label9->labelcolor( 255 );

  int startY = 10;
  int step = 30;
  int number = 9;
  for( int i = 0; i < number; i++ ) {
    beacons[i]= new Fl_Output( 130, startY+i*step, 70, 30, 0 );
    beacons[i]->textsize(12);
  }

  return SUCCESS;
}


inline void beaconWindow::setBeaconField( int index, char *aValue )
{
  beacons[index]->value( aValue );
}
  

//------------------------------------------------------------------------------

/******************************************
 * class udpStatWindow definition
 ******************************************/

udpStatWindow::udpStatWindow (int w, int h, char *label=0 ):Fl_Window( w, h, label )
{
  initialize();
}

udpStatWindow::~udpStatWindow()
{
  for( int i = 0; i < 8; i++ ) {
    delete udpStats[i];
  }
}


int udpStatWindow::initialize()
{
  this->set_modal();
  this->color(FL_YELLOW);
  
  Fl_Box *label1 = new Fl_Box( 0, 20, 130, 30, "UDP registration sent:" );
  label1->color(255);
  label1->labelsize(8);

  Fl_Box *label2 = new Fl_Box( 0, 50, 130, 30, "UDP replies received:" );
  label2->color(255);
  label2->labelsize(8);

  Fl_Box *label3 = new Fl_Box( 0, 80, 130, 30, "UDP error: \nident didn't match:" );
  label3->color(255);
  label3->labelsize(8);


  Fl_Box *label4 = new Fl_Box( 0, 110, 130, 30, "UDP error: \ninvalid ha authentication:" );
  label4->color(255);
  label4->labelsize(8);

  Fl_Box *label5 = new Fl_Box( 0, 140, 130, 30, "UDP error: \ninvalid fa authentication:" );
  label5->color(255);
  label5->labelsize(8);

  Fl_Box *label6 = new Fl_Box( 0, 170, 130, 30, "UDP error: \nmissing ha authentication:" );
  label6->color(255);
  label6->labelsize(8);

  Fl_Box *label7 = new Fl_Box( 0, 200, 130, 30, "UDP error: \nmissing fa authentication:" );
  label7->color(255);
  label7->labelsize(8);

  Fl_Box *label8 = new Fl_Box( 0, 230, 130, 30, "UDP error: \nha made time update:" );
  label8->color(255);
  label8->labelsize(8);
  
  int startY = 20;
  int step = 30;
  int number = 8;
  for( int i = 0; i < number; i++ ) {
    
    udpStats[i]= new Fl_Output( 130, startY+i*step, 70, 30, 0 );
    udpStats[i]->textsize(12);
  }
  this->resizable();
  return SUCCESS;
}

inline void udpStatWindow::setUdpField( int index, char *aValue )
{
  udpStats[index]->value( aValue );
}
  

//=================================================================

/******************************************************************
 SOME UTILITY FUNCTIONS
******************************************************************/

//------------------------------------------------------------------

void msgok_cb(Fl_Widget *w, void *data)
{
	//Fl_Button *bw = (Fl_Widget *)w;
	//w->parent()->hide();
	delete w->parent();
}

// window for giving message passed as the function parameter 
void message(char *msg)
{	  
  Fl_Window*msgWindow = new Fl_Window(200, 100, "Notice");
  msgWindow->set_modal();
  msgWindow->color(FL_YELLOW);
  Fl_Box *msgBox = new Fl_Box( 5,5,150,50, msg);
  msgBox->color(255);
  msgBox->labelcolor(FL_BLUE);
  msgBox->labelfont(FL_HELVETICA_BOLD);
  Fl_Button *okButton = new Fl_Button( 140, 70, 50, 20, "OK");
  okButton->color(FL_BLUE);
  okButton->labelcolor(255);
  okButton->labelfont(FL_BOLD);
  okButton->callback(msgok_cb);
  msgWindow->end();
  msgWindow->show();
  Fl::run();
}

//----------------------------------------------------------------------
void yes_cb(Fl_Widget *w, void *data)
{
  system("rm -f /tmp/mnd.sock.client" );
  exit(0);
}

void no_cb(Fl_Widget *w, void *data)
{
  delete w->parent();
}

// window for confirming quitting program 
void ask(char *msg)
{	  
  Fl_Window *askWindow = new Fl_Window(200, 100, "Confirm quitting");
  askWindow->set_modal();
  askWindow->color(FL_YELLOW);
  Fl_Box *askBox = new Fl_Box( 5,5,150,50, msg);
  askBox->color(255);
  askBox->labelcolor(FL_BLUE);
  askBox->labelfont(FL_HELVETICA_BOLD);
  askBox->labelsize(20);
  Fl_Button *yesButton = new Fl_Button( 80, 70, 50, 20, "YES");
  yesButton->color(FL_BLUE);
  yesButton->labelcolor(255);
  yesButton->labelfont(FL_BOLD);
  yesButton->callback(yes_cb);
  Fl_Button *noButton = new Fl_Button( 140, 70, 50, 20, "NO");
  noButton->color(FL_BLUE);
  noButton->labelcolor(255);
  noButton->labelfont(FL_BOLD);
  noButton->callback(no_cb);
  askWindow->end();
  askWindow->show();
}

//--------------------------------------------------------------
Fl_Color genColor(int level)
{
  if ( level >=0 && level <= 25 )
	return FL_RED;
  else if( level > 25 && level <= 50 )
	return FL_BLUE;
  else if( level > 50 && level <= 75 )
	return FL_YELLOW;
  else
	return FL_GREEN;
}


//--------------------------------------------------------------
// This function is used to extract the IP address of a fa
char * getFa( char *s )
{
  int i, j;
  static char addr[16];
  // s is of the form fa: [ip address]
  // start position of ip in s is 4
  int start = 4;

  int len = strlen( s );
  for (i = start, j = 0; i < len; i++ , j++)
    addr[j] = s[i];
  addr[j] = '\0';
  return addr;
}

//---------------------------------------------------------------
// Gather all the statistics here

int getStat()
{
  mnstat->getState( &mnState );
  mnstat->getBeaconData( beaconField );
  mnstat->getUdpData( udpField );
  
  mnstat->getAgents( &agentsCount, 
		     agentBeacon, 
		     agentSignal,
		     agentLocation);

   // find signal and location for the state window
  char fa[16];
  char *temp;
 
  temp = getFa( mnState.fa) ;
  strcpy( fa, temp );
  int len = strlen( fa );
  // search among agent for this fa / coa address
  int start ; // where IP address in agentBeacon string starts
  for( int i = 0; i < agentsCount; i++ ) {
	start = 13;
	int k = 0;
	while( agentBeacon[i][start] == fa[k] && k < len ) {
	  ++start; ++k;
	}
	if( k  == len ) { // find a match
	  sprintf( mnState.location, "%s", agentLocation[i] );
	  sprintf( mnState.signalText, "Signal %d", agentSignal[i] );
	  mnState.signal = agentSignal[i];
	  break;
	}
  }
  

   mnstat->getNodes( &nodesCount, 
		    nodeBeacon,
		    nodeSignal,
		    nodeLocation );

  return SUCCESS;
}

//---------------------------------------------------------------------

void updateDisplay( void *)
{
  getStat();

  for( int i = 0; i < BEACON_FIELDS; i++ )
	myBeacon->setBeaconField( i, beaconField[i] );

  for( int i = 0; i < UDP_FIELDS; i++ )
	myUdpStat->setUdpField( i, udpField[i] );

  myState->setState( mnState.state );
  myState->setFa( mnState.fa );
  myState->setCoa( mnState.coa );
  myState->setLocation( mnState.location );
  myState->setMeter( mnState.signal );

  for( int i = 0; i < agentsCount; i++ ) {
	myAgent->setBeacon( i, agentBeacon[i] );
	myAgent->setMeter( i, agentSignal[i] );
	myAgent->setLocation( i, agentLocation[i] );
  }
  for( int i = agentsCount; i < MAXNODES; i++ ) {
	myAgent->resetBeacon( i );
	myAgent->resetMeter( i );
	myAgent->resetLocation( i );
  }
  for( int i = 0; i < nodesCount; i++ ) {
	myNode->setBeacon( i, nodeBeacon[i] );
	myNode->setMeter( i, nodeSignal[i] );
	myNode->setLocation( i, nodeLocation[i] );
  }
  for( int i = nodesCount; i < MAXNODES; i++ ) {
	myNode->resetBeacon( i );
	myNode->resetMeter( i );
	myNode->resetLocation( i );
  }
}
  
//---------------------------------------------------------------------------  
void startWindow( )
{
  mnstat = new cMnStat();
  myState = new stateWindow( 200, 300, "Mobile Node Status" );
  myState->end();
  myBeacon = new beaconWindow( 200, 300, "Beacon Statistics" );
  myBeacon->end();
  myUdpStat = new udpStatWindow( 200, 300, "UDP Registration Statistics" );
  myUdpStat->end();
  myAgent = new agentWindow( 200, 300, "Nearby Agents" );
  myAgent->end();
  myNode = new nodeWindow( 200, 300, "Nearby Nodes" );
  myNode->end();

  updateDisplay( 0 );

  myState->show();
  
  Fl::add_timeout(pollInterval, updateDisplay);
  for(;;) { 
    Fl::wait();
    updateDisplay(0);
  }
  Fl::run();
}


/**************************MAIN**************************************/

int main( int argc, char **argv )
{

  if( getuid()) {
    printf( "wmnstat: root only\n" );
    exit( FAILURE );
  }

  pollInterval = 2.0;
  myCounterWindow = new counterWindow( 190, 80, "Setting polling Interval");
  myCounterWindow->end();
  startWindow();
}

